// FILE: DvmLib.c
// COPYRIGHT: (c), Symmetric Research, 2009
//
// Source file for those parts of the USBxCH DVM application shared between
// the GUI (Graphical User Interface) and CMD (Command line) versions

#if defined( SROS_WINXP ) || defined( SROS_WIN2K )
//FIX - What is this used for ??? 
#ifndef LARGE_INTEGER
#include <windows.h>
#endif

#elif defined( SROS_LINUX )

#define _MAX_FNAME	256
#define _MAX_EXT	256

#endif // defined SROS_xxx


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#include "DvmLib.h"
#include "SrIniParse.h" // Parsing functions
#include "SrHelper.h"   // Helper funtions for keyboard and strings


MATCHTABLE AtodModelTable[] = {

        "USB1CH",  SRDAT_ATODMODEL_USB1CH,
        "USB4CH",  SRDAT_ATODMODEL_USB4CH,
        "USB8CH",  SRDAT_ATODMODEL_USB8CH,
        "DEFAULT", SRDAT_ATODMODEL_UNKNOWN,
         NULL,     0
        };

MATCHTABLE GpsModelTable[] = {

        "GARMIN",  SRDAT_GPSMODEL_GARMIN,
        "TRIMBLE", SRDAT_GPSMODEL_TRIMBLE,
        "ONCORE",  SRDAT_GPSMODEL_ONCORE,
        "PCTIME",  SRDAT_GPSMODEL_PCTIME,
        "DEFAULT", SRDAT_GPSMODEL_UNKNOWN,
         NULL,     0
        };

MATCHTABLE ChannelUnitsTypeTable[] = {

        "HexCounts", UNITS_HEX,
        "Counts",    UNITS_COUNTS,
        "Volts",     UNITS_FLOATS,
         NULL,       0
        };


// Global variables all collected up together in one structure

DVM_VARIABLES DvmInfo;
char          TempStr[MAXSTR];

char *OnOffString[] = { "Off", "On" };
char *TriggerString[] = { "Continuous", "Keypress", "Amplitude" };
char *OutFormatString[] = { "fw", "csv" };



// STRING HELPERS:
//
//
//


//FIX - Add to SrHelper instead ???
//------------------------------------------------------------------------------
// ROUTINE: Chop
// PURPOSE: Chop spaces ' ' off the end of a string and add '\n'.
//------------------------------------------------------------------------------
static char ChopBuffer[256];

char *Chop( char *s ) {

        char *p;


        // Chop spaces ' ' off the end of a string and add '\n'.

        SrStringCopy( ChopBuffer, s, 256 );

        p = ChopBuffer;

        while ( *p )

                p++;


        if ( *--p == '\n' ) {

                while ( *--p == ' ' )

                         ; // back up


                *++p = '\n';
                *++p = '\0';

                }


        return( ChopBuffer );

}

//------------------------------------------------------------------------------
// ROUTINE: DvmInit
// PURPOSE: Read in the dvm.ini file and initialize the DvmInfo structure.
//------------------------------------------------------------------------------
int DvmInit( DVMFNPTR NewShowLogMessageFunction,
             char *CmdLine, char *ProgramName ) {

        int    Success, Error, Half;
        char  *Suffix;
        double SpsTrue;

        // Set program name

        if ( ProgramName )
                DvmInfo.ProgramName = ProgramName;
        else
                DvmInfo.ProgramName = "Dvm";
                

        // Set up message functions

        if ( NewShowLogMessageFunction )
                DvmInfo.ShowLogMessage = NewShowLogMessageFunction;


        // Process command line for ini file name

        if ( !DvmIniCheck( CmdLine ) )
                return( 0 );


        // Initialize variables from ini file

        if ( !DvmIniParse( DvmInfo.IniFileName ) )
                return( 0 );

//FIX - Include ini verify checks to make sure things like AtodAnalogChannels
//      and AtodXchModel name are consistent with the selected AtodXchModel option


        // Set additional variables

        Success = SrUsbXchSpsRateValidate( (double)DVM_SPS, &SpsTrue, &Error );
//FIX - Do something with return value


        DvmInfo.IsRunning        = 0;
        DvmInfo.AtodLedState     = 0;
        DvmInfo.SampleNumber     = 0;
        DvmInfo.Nremain          = 0L;
        DvmInfo.LastMsg          = -1;
        DvmInfo.TriggerRequested = 0;
        DvmInfo.AtodSampleRate   = SpsTrue;
        DvmInfo.TimerInterval    = 1000;   // ms


        // Compute # of samples to read per timer interval making sure
        // the number is a little bit more so the program waits for the
        // last few points, assuring that the fifo is emptied.  Also,
        // ensure that the result is an even number since the firmware
        // always returns points two at a time.

        DvmInfo.Nread = (int)(SpsTrue * DvmInfo.TimerInterval / 1000.0L) +
                                                            DVM_READ_EXTRA;
        Half = DvmInfo.Nread / 2;
        if ( DvmInfo.Nread != Half*2 )
                DvmInfo.Nread++;


        // Prepare default output file name

        if ( DvmInfo.OutFileFormat == OUTPUT_FORMAT_CSV )
                Suffix = "csv";
        else
                Suffix = "asc";

        if ( DvmInfo.OutFileNaming == SRDAT_FILENAMETYPE_YMDHMS ) {
                sprintf( DvmInfo.OutFileNameShort, "YYYY-MM-DD-at-hh-mm-ss.%s", Suffix );
                sprintf( DvmInfo.OutFileName, "%s%c%s",
                         DVM_DEFAULT_DIR, SR_DIRSLASH, DvmInfo.OutFileNameShort );
                }
        else if ( DvmInfo.OutFileNaming == SRDAT_FILENAMETYPE_SINGLE ) {
                sprintf( DvmInfo.OutFileNameShort, "%s.%s", DVM_DEFAULT_OUT, Suffix );
                strncpy( DvmInfo.OutFileName, DvmInfo.OutFileNameShort, MAXSTR );
                }
        else { // if ( DvmInfo.OutFileNaming == SRDAT_FILENAMETYPE_NONE ) {
                strncpy( DvmInfo.OutFileNameShort, "None", MAXSTR );
                strncpy( DvmInfo.OutFileName, DvmInfo.OutFileNameShort, MAXSTR );
                }
        
        

//printf("DvmInit set Nread to %d\n", DvmInfo.Nread );

        // Initialize headers and data arrays

        DvmTitleInit( );
        DvmDataInit( );



        return( 1 );
}

//------------------------------------------------------------------------------
// ROUTINE: DvmAtodOpen
// PURPOSE: Open the driver and initialize the USBxCH A/D in preparation
//          for acquiring data.
//------------------------------------------------------------------------------
int DvmAtodOpen( void ) {

        int    Error;
        double SpsTrue;

//printf("DvmAtodOpen\n" );

        // Open driver and initialize board

        DvmInfo.AtodHandle = SrUsbXchOpen(
                                     DvmInfo.AtodDriverName,
                                     DvmInfo.AtodXchModel,
                                     DvmInfo.GpsModel,
                                     USERCFG_DEFAULT_BYTE,
                                     DvmInfo.AtodSampleRate,
                                     &SpsTrue,
                                     &Error
                                   );
        if ( !SrUsbXchIsOpenSuccess( Error ) ) {
                sprintf( TempStr,
                         "SrUsbXchOpen failed with error # %d: %s.",
                         Error,
                         USBXCH_ERROR_MSG[Error]
                       );
                DvmInfo.ShowLogMessage( MSGLEVEL_ERR, ERROR_NOT_FATAL, TempStr );
                return( 0 );
                }


        return( 1 );
}


//------------------------------------------------------------------------------
// ROUTINE: DvmAtodStart
// PURPOSE: Perform any actions needed to begin acquiring data.  Assumes that
//          driver is open and USBxCH board is initialized.
//------------------------------------------------------------------------------
int DvmAtodStart( void ) {

//printf("DvmAtodStart Enter\n" );

        // Start acquisition

        if ( !SrUsbXchStart( DvmInfo.AtodHandle, NULL ) )
                return( 0 );

        DvmInfo.AtodLedState = 1;
        SrUsbXchUserLed( DvmInfo.AtodHandle, DvmInfo.AtodLedState, DvmInfo.AtodLedState );


        // Set DVM specific variables to indicate we have just started a new run

        DvmInfo.IsRunning        = 1;
        DvmInfo.FilterInitNeeded = 1;
        DvmInfo.AtodLedState     = 0;
        DvmInfo.SampleNumber     = 0;
        DvmInfo.Nremain          = 0L;
        DvmInfo.LastMsg          = -1;


//printf("DvmAtodStart Exit\n" );
        return( 1 );
}

//------------------------------------------------------------------------------
// ROUTINE: DvmAtodStop
// PURPOSE: Perform any actions needed to stop acquiring data.  This should undo
//          anything done by DvmStart.
//------------------------------------------------------------------------------
int DvmAtodStop( ) {

//printf("DvmAtodStop Enter\n" );

        // Stop A/D execution
        
        if ( !SrUsbXchStop( DvmInfo.AtodHandle, NULL ) )
                return( 0 );

        DvmInfo.AtodLedState = 0;
        SrUsbXchUserLed( DvmInfo.AtodHandle, 
                         DvmInfo.AtodLedState, DvmInfo.AtodLedState );  // Turn off LED

        DvmInfo.IsRunning = 0;

//printf("DvmAtodStop Exit\n" );
        return( 1 );
}

//------------------------------------------------------------------------------
// ROUTINE: DvmAtodClose
// PURPOSE: Close the driver and undo any actions done by DvmAtodOpen.
//------------------------------------------------------------------------------
int DvmAtodClose( void ) {

//printf("DvmAtodClose\n" );

        // If running, stop before close

        if ( DvmIsRunning( ) > 0 )
                DvmAtodStop( );


        // Close the A/D driver

        SrUsbXchClose( DvmInfo.AtodHandle );
        DvmInfo.AtodHandle = BAD_DEVHANDLE;

        return( 1 );
}


//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
//FIX
void ShowCalReg( void ) {

        int           ichan, ireg, Nchan, Nreg;
        unsigned long ofc, fsc;
        char          RegData[ADREG_MAX*MAXCHANNEL];

        Nchan = 4;
        Nreg  = 6;

        DvmCalGet( RegData );


        // Show channel numbers

        printf( "            " );
        for ( ichan = 0 ; ichan < Nchan ; ichan++ )
                printf( "  ch%d", ichan );
        printf( "\n" );


        // Show register values

        for ( ireg = 0 ; ireg < Nreg ; ireg++ ) {
                printf( "Register %2d = ", ireg );
                for ( ichan = 0 ; ichan < Nchan ; ichan++ )
                        printf( "0x%02X ", (RegData[ireg*Nchan+ichan] & 0X00FF) );
                printf( "\n" );
                }


        // Show full OFC and FSC calibration numbers

        printf( "\n" );
        for ( ichan = 0 ; ichan < Nchan ; ichan++ ) {
                fsc = ( (RegData[5*Nchan+ichan]&0x00FF) << 16) +
                      ( (RegData[4*Nchan+ichan]&0x00FF) << 8) +
                        (RegData[3*Nchan+ichan]&0x00FF);

                ofc = ( (RegData[2*Nchan+ichan]&0x00FF) << 16) +
                      ( (RegData[1*Nchan+ichan]&0x00FF) << 8 ) +
                        (RegData[0*Nchan+ichan]&0x00FF);

                printf( "Chan %d:  OFC = 0x%06lX, FSC = 0x%06lX\n",
                          ichan, (ofc&0x00FFFFFF), (fsc&0x00FFFFFF) );
                }

}

//FIX

//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
double LgIntToDouble(  unsigned long HiPart, unsigned long LoPart ) {
        double Dbl;

        Dbl =  HiPart;
        Dbl *= 4294967296.0L;
        Dbl += LoPart;

        return( Dbl );
}

//------------------------------------------------------------------------------
// ROUTINE: DvmAtodRead
// PURPOSE: Read in data.  Three possible error codes. 1 success, -1 no data,
//          0 fatal read failure.
//------------------------------------------------------------------------------
int DvmAtodRead( int *Nsamples ) {

        int                    Result, ErrorCode, CheckSanity;
        SRUSBXCH_STATE         UsbState;
        SRDAT_USBXCH_SERIAL    SerialValues[5];
        SRDAT_USBXCH_EQUIPMENT EquipmentValues[5];
        unsigned long          NanalogRequested, NserialRequested, NequipmentRequested;
        unsigned long          NanalogRead, NserialRead, NequipmentRead, Nsamp;
        unsigned long   NpacketsRequested, NpacketsRead;
        unsigned long   NpacketsProcessed, NpacketsProcessedData, NpacketsProcessedStatus;


//printf("DvmAtodRead Enter\n" );

        // Read in data

        NanalogRequested    = (unsigned long)DvmInfo.Nread;
        NserialRequested    = 4;
        NequipmentRequested = 4;
        CheckSanity         = 1;


        NpacketsRequested = (unsigned long)DvmInfo.Nread/2 - DvmInfo.Nremain;

        Nsamp = SrUsbXchGetDataAsPackets(
                                         DvmInfo.AtodHandle,
                                        &DvmInfo.PacketData[DvmInfo.Nremain],
                                         NpacketsRequested,
                                        &NpacketsRead,
                                        &ErrorCode
                                        );
//FIX - Check error return

        Nsamp = SrUsbXchConvertPackets(
                                       DvmInfo.PacketData,
                                       NpacketsRead,
                                       CheckSanity,
                                      &DvmInfo.LastMsg,
                                       DvmInfo.DataRaw,
                                       NanalogRequested,
                                      &NanalogRead,
                                       SerialValues,
                                       NserialRequested,
                                      &NserialRead,
                                       EquipmentValues,
                                       NequipmentRequested,
                                      &NequipmentRead,
                                      &NpacketsProcessed,
                                      &NpacketsProcessedData,
                                      &NpacketsProcessedStatus,
                                      &ErrorCode
                                      );
//FIX - Check error return


        DvmInfo.Nremain = SrDatUsbPacketCleanUp(
                                                DvmInfo.PacketData,
                                                NpacketsRead,
                                                NpacketsProcessed,
                                                &ErrorCode
                                               );
//FIX - Check error return




        // Also get state information like power and temperature

        Result = SrUsbXchGetState( DvmInfo.AtodHandle, &UsbState );

        if ( Result ) {
                DvmInfo.AtodPowerInfo = UsbState.PowerInfo;
                DvmInfo.AtodDigitalIn = UsbState.DigitalIn;
                SrDatUsbTemperatureCombine( UsbState.TempHi, UsbState.TempLo,
                                            &DvmInfo.AtodTemperature );
                }
        else
                return( 0 );


        

//printf("DvmAtodRead calling time update\n" );
        DvmTimeUpdate( );
        DvmTitleUpdate( ); // to get new power/temp status


        // No data to process this time

        if ( Nsamp == 0L ) {
//printf("DvmAtodRead failed no data\n" );
//                DvmInfo.ShowLogMessage( MSGLEVEL_MSG, ERROR_NOT_FATAL, "No data read." );
                if ( Nsamples )  *Nsamples = 0;
//printf("DvmAtodRead Exit NO DATA\n" );
                return( -1 );
                }



        // Show data updated

//printf("DvmAtodRead toggle LED\n" );
        DvmInfo.AtodLedState ^= 1;
        SrUsbXchUserLed( DvmInfo.AtodHandle,
                         DvmInfo.AtodLedState, DvmInfo.AtodLedState );




        if (Nsamples)  *Nsamples = (int)Nsamp;

//printf("DvmAtodRead Exit\n" );
        return( 1 );
}

//------------------------------------------------------------------------------
// ROUTINE: DvmCalInit
// PURPOSE: Initialize calibration registers
//------------------------------------------------------------------------------
int DvmCalInit( void ) {

        int i;

        for ( i = 0 ; i < MAXCHANNEL ; i++ ) {
                DvmInfo.CalReg[i].OfcLow  = 0x00;
                DvmInfo.CalReg[i].OfcMid  = 0x00;
                DvmInfo.CalReg[i].OfcHigh = 0x00;
                DvmInfo.CalReg[i].FscLow  = 0x00;
                DvmInfo.CalReg[i].FscMid  = 0x00;
                DvmInfo.CalReg[i].FscHigh = 0x00;
                }

        return( 1 );
}

//------------------------------------------------------------------------------
// ROUTINE: DvmCalSet
// PURPOSE: Set calibration register values.
//------------------------------------------------------------------------------
int DvmCalSet( char *RegData ) {

        int i, Nchan, Nreg;

        Nchan = DvmInfo.AtodAnalogChannels;
        Nreg  = 6;

        for ( i = 0 ; i < Nchan ; i++ ) {
                DvmInfo.CalReg[i].OfcLow  = RegData[0*Nchan+i];
                DvmInfo.CalReg[i].OfcMid  = RegData[1*Nchan+i];
                DvmInfo.CalReg[i].OfcHigh = RegData[2*Nchan+i];
                DvmInfo.CalReg[i].FscLow  = RegData[3*Nchan+i];
                DvmInfo.CalReg[i].FscMid  = RegData[4*Nchan+i];
                DvmInfo.CalReg[i].FscHigh = RegData[5*Nchan+i];
                }

        return( 1 );
}

//------------------------------------------------------------------------------
// ROUTINE: DvmCalGet
// PURPOSE: Read calibration register values.
//------------------------------------------------------------------------------
int DvmCalGet( char *RegData ) {

        int i, Nchan, Nreg;

        Nchan = DvmInfo.AtodAnalogChannels;
        Nreg  = 6;

        for ( i = 0 ; i < Nchan ; i++ ) {
                RegData[0*Nchan+i] = DvmInfo.CalReg[i].OfcLow;
                RegData[1*Nchan+i] = DvmInfo.CalReg[i].OfcMid;
                RegData[2*Nchan+i] = DvmInfo.CalReg[i].OfcHigh;
                RegData[3*Nchan+i] = DvmInfo.CalReg[i].FscLow;
                RegData[4*Nchan+i] = DvmInfo.CalReg[i].FscMid;
                RegData[5*Nchan+i] = DvmInfo.CalReg[i].FscHigh;
                }

        return( 1 );
}

//------------------------------------------------------------------------------
// ROUTINE: DvmIsRunning
// PURPOSE: Indicate whether the USBXCH is currently acquiring data.  1 means
//          yes, 0 means no.
//------------------------------------------------------------------------------
int DvmIsRunning( void ) {
        return( DvmInfo.IsRunning );
}

//------------------------------------------------------------------------------
// ROUTINE: DvmTitleInit
// PURPOSE: Initialize the header title lines.  Currently there are three.  The
//          first has the name of the ini file, the second has the name of the
//          output file, and the third has the power and temperature status.
//------------------------------------------------------------------------------
int DvmTitleInit( void ) {

        int nt;

        // Set up Title lines

        nt = 0;
        while ( nt < TITLE_MAX ) {
                if ( nt == TITLE_ASCIIOUT ) {
                        if ( DvmInfo.OutFileNaming == SRDAT_FILENAMETYPE_NONE )
                                sprintf( DvmInfo.Title[nt], "Output file = %s", DvmInfo.OutFileName );
                        else
                                sprintf( DvmInfo.Title[nt], "Output file = .%c%s",
                                         SR_DIRSLASH, DvmInfo.OutFileName );
                        }
                else if ( nt == TITLE_STATUS )
                        sprintf( DvmInfo.Title[nt], "Power and Temperature unknown until running" );
                else
                        break;
                nt++;
                }

        DvmInfo.Ntitle = nt;

        return( DvmInfo.Ntitle );
}

//------------------------------------------------------------------------------
// ROUTINE: DvmTitleUpdate
// PURPOSE: Update the header title lines as the output file name and power
//          and temperature status may have changed.
//------------------------------------------------------------------------------
int DvmTitleUpdate( void ) {

        int    nt, PowerGoodH;
        char  *PowerString;
        double TempC, TempF;

        // Initialize values

        SrDatUsbTemperatureCompute( DvmInfo.AtodTemperature, &TempC, &TempF );
        PowerGoodH = SrUsbXchIsPowerBitGood( DvmInfo.AtodPowerInfo,
                                             USBXCH_POWER_HISTORY );
        if ( PowerGoodH == 1 )
                PowerString = "Good";
        else
                PowerString = "BAD";
        
        
        // Set up Title lines

        nt = 0;
        while ( nt < TITLE_MAX ) {
                if ( nt == TITLE_ASCIIOUT ) {
                        if ( DvmInfo.OutFileNaming == SRDAT_FILENAMETYPE_NONE )
                                sprintf( DvmInfo.Title[nt], "Output file = %s", DvmInfo.OutFileName );
                        else
                                sprintf( DvmInfo.Title[nt], "Output file = .%c%s",
                                         SR_DIRSLASH, DvmInfo.OutFileName );
                        }
                else if ( nt == TITLE_STATUS )
                        sprintf( DvmInfo.Title[nt], "Power is %s         Temp = %.1lfC  (%.1lfF)",
                                 PowerString, TempC, TempF );
                else
                        break;
                nt++;
                }

        DvmInfo.Ntitle = nt;

        return( DvmInfo.Ntitle );
}

//------------------------------------------------------------------------------
// ROUTINE: DvmDataInit
// PURPOSE: Reset the raw counts and scaled versions of the current data value
//          stored in the DvmInfo structure.
//------------------------------------------------------------------------------
void DvmDataInit( void ) {

    int i, k;


    // Initialize filter and display data arrays

    SrUsbXchClearData( DvmInfo.DataRaw, MAXSAMPLES );

    DvmInfo.FilterCoeff[0] = 0.98;       // previous influence
    DvmInfo.FilterCoeff[1] = 0.02;       // current x influence

    DvmInfo.FilterCoeff[0] = 0.00;       // previous influence
    DvmInfo.FilterCoeff[1] = 1.00;       // current x influence

    for ( i = 0 ; i < DvmInfo.AtodAnalogChannels ; i++ ) {
            DvmInfo.DataCounts[i] = 0L;
            DvmInfo.DataFloats[i] = 0L;
            for ( k = 0 ; k < MAXFILTER ; k++ ) {
                    DvmInfo.FilterTerms[k][i] = 0.0L;
                    }
            }

}

//------------------------------------------------------------------------------
// ROUTINE: DvmDataFilter
// PURPOSE: Process the data by scaling the measured count value to volts or
//          some other user selected units by linearly interpolating using
//          a slope and offset formula and then filtering with a two term
//          recursive filter.
//
//          The slope and offset formula is:
//             yk = slope * xk + offset
//          where slope and offset are values read in from the ini file
//          that were previously determined by calibrating the device using
//          the supplied calibration program, cal.
//
//          The filter equation is:
//              F(n) = C(n) * X + C(n-1) * F(n-1) + C(n-2) * F(n-2) + ...
//          where C is coeff, F is filter terms and X is current data value
//------------------------------------------------------------------------------
void DvmDataFilter( int Nsamples ) {

    int    i, j, k, nf;
    double Result, Round;


    // Error exit

    if ( Nsamples <= 0 )
            return;


//#define SHOW_FILTER_INFO

    DvmInfo.SampleNumber++;


    // Initialize filter if needed

    if ( DvmInfo.FilterInitNeeded == 1 ) {
        DvmInfo.FilterInitNeeded = 0;
        for ( i = 0 ; i < DvmInfo.AtodAnalogChannels ; i++ ) {
            for ( k = 0 ; k < MAXFILTER ; k++ ) {
                DvmInfo.FilterTerms[k][i] = DvmInfo.DataRaw[0].AnalogData[i];
                } // end for k
            } // end for i

#ifdef SHOW_FILTER_INFO
                for ( i = 0 ; i < DvmInfo.AtodAnalogChannels ; i++ ) {
                    printf("T   Raw    FilterTerm for Channel %d\n", i);
                    for ( k = 0 ; k < MAXFILTER ; k++ ) {

                        printf("%d 0x%lX %lf\n", k,
                               DvmInfo.DataRaw[0].AnalogData[i],
                               DvmInfo.FilterTerms[k][i] );
                        }
                    }
#endif // SHOW_FILTER_INFO


        } // end if FilterInitNeeded



    // Loop for all samples, updating filtered result for each channel
    //
    // Equation is: F(n) = C(n) * X + C(n-1) * F(n-1) + C(n-2) * F(n-2) + ...
    //
    // where C is coeff, F is filter terms and X is current data value
    //

#ifdef SHOW_FILTER_INFO

printf( "Filter Coeffs and Terms\n" );
printf( "term  Coeff      Ch 0          Ch 1          Ch 2          Ch 3\n" );
for ( k = 0 ; k < MAXFILTER ; k++ ) {
        printf( "%d  %1lf  ", k, DvmInfo.FilterCoeff[k] );

        for ( i = 0 ; i < DvmInfo.AtodAnalogChannels ; i++ ) {
                printf( "%12.6lf  ", DvmInfo.FilterTerms[k][i] );
                }
        printf( "\n" );
        }
#endif // SHOW_FILTER_INFO



    nf = MAXFILTER-1;

    for ( j = 0 ; j < Nsamples ; j++ ) {

#ifdef SHOW_FILTER_INFO
        printf( "%03d Raw:       ", j );
        for ( i = 0 ; i < DvmInfo.AtodAnalogChannels ; i++ ) {
                printf( "%-12ld  ", DvmInfo.DataRaw[j].AnalogData[i] );
                }

        printf( "\n" );


        printf( "%03d Results: ", j );
#endif // SHOW_FILTER_INFO

        for ( i = 0 ; i < DvmInfo.AtodAnalogChannels ; i++ ) {

            Result = DvmInfo.FilterCoeff[nf] * DvmInfo.DataRaw[j].AnalogData[i];

            for ( k = 0 ; k < nf ; k++ ) {
                DvmInfo.FilterTerms[k][i] = DvmInfo.FilterTerms[k+1][i];
                Result += ( DvmInfo.FilterCoeff[k] * DvmInfo.FilterTerms[k][i] );
                } // end for k

            DvmInfo.FilterTerms[nf][i] = Result;

#ifdef SHOW_FILTER_INFO
            printf( "%12.6lf  ", Result );
#endif // SHOW_FILTER_INFO

            } // end for i

#ifdef SHOW_FILTER_INFO
        printf( "\n" );
#endif // SHOW_FILTER_INFO

        } // end for j


    //  Compute display values for each channel

    for ( i = 0 ; i < DvmInfo.AtodAnalogChannels ; i++ ) {
        if (DvmInfo.FilterTerms[nf][i] < 0)
            Round = -0.5;
        else
            Round = 0.5;
        DvmInfo.DataCounts[i] = (long)(DvmInfo.FilterTerms[nf][i] + Round);
        DvmInfo.DataFloats[i] = (float)((DvmInfo.Channel[i].UnitsSlope *
                                         DvmInfo.FilterTerms[nf][i]) +
                                        (double)DvmInfo.Channel[i].UnitsOffset);
        }


// FIX FIX FIX *** just compute an average for now

#ifdef USE_AVERAGE_FILTER

    for ( i = 0 ; i < DvmInfo.AtodAnalogChannels ; i++ ) {
            Result = 0.0;
            for ( j = 0 ; j < Nsamples ; j++ ) {
                    Result += DvmInfo.DataRaw[j].AnalogData[i];
                    }
            Result  = Result / Nsamples;
            DvmInfo.DataCounts[i] = (long)(Result);
            DvmInfo.DataFloats[i] = (float)((DvmInfo.Channel[i].UnitsSlope * Result) +
                                            (double)DvmInfo.Channel[i].UnitsOffset);
            }

#endif // USE_AVERAGE_FILTER





}

//------------------------------------------------------------------------------
// ROUTINE: DvmDataFormat
// PURPOSE: Write the data to a string so it can be displayed or output.
//------------------------------------------------------------------------------
int DvmDataFormat( int Ichannel, char *DataString, int Nchar ) {

        int    i, uplace, udigit, decade;
        char   DataFmtStr[MAXSTR];
        double maxval;

        // Error check

        if ( Ichannel < 0                            ||
             Ichannel > DvmInfo.AtodAnalogChannels-1 ||
             DataString == NULL                      ||
             Nchar < 10                              )
                return( 0 );



//FIX - check this
        // Assume data already processed and scaled from counts to floats



        // Prepare format string


        if (DvmInfo.Channel[Ichannel].UnitsType == UNITS_HEX)
                sprintf( DataFmtStr, "0x%%06X" );

        else if (DvmInfo.Channel[Ichannel].UnitsType == UNITS_COUNTS)
                sprintf( DataFmtStr, "%%8ld" );

        else { // DvmInfo.Channel[Ichannel].UnitsType == UNITS_FLOATS
                uplace = DvmInfo.Channel[Ichannel].UnitsPlaces;
                udigit = DvmInfo.Channel[Ichannel].UnitsDigits;
                decade = uplace - udigit - 1;
                if (decade < 0)  decade = 0;
                maxval = 1.0L;
                for ( i = 0 ; i < decade ; i++ )
                        maxval *= 10.0L;
                if ( DvmInfo.DataFloats[Ichannel] > maxval )
                        sprintf( DataFmtStr, "# too big" );
                else if ( -DvmInfo.DataFloats[Ichannel] > maxval )
                        sprintf( DataFmtStr, "# too tiny" );
                else
                        sprintf( DataFmtStr, "%%+%d.%df", uplace, udigit );
                }


        // Fill provided string

        if ( DvmInfo.Channel[Ichannel].UnitsType == UNITS_HEX)
                sprintf( DataString, DataFmtStr, (DvmInfo.DataCounts[Ichannel] & 0x00FFFFFF) );

        else if (DvmInfo.Channel[Ichannel].UnitsType == UNITS_COUNTS)
                sprintf( DataString, DataFmtStr, DvmInfo.DataCounts[Ichannel] );

        else
                sprintf( DataString, DataFmtStr, DvmInfo.DataFloats[Ichannel] );

    return( 1 );
}


//------------------------------------------------------------------------------
// ROUTINE: DvmFileYmdHms
// PURPOSE: Create the YMDHMS data file name using the supplied suffix (eg asc 
//          or csv) and make the requested data directory if it doesn't
//          already exist.
//------------------------------------------------------------------------------
int DvmFileYmdHms( char *DataDirectory, char *FileSuffix ) {

            time_t UtcTimeInSecondsSince_1970;
        struct tm *LocTime;


        // Get the current time.

                       time( &UtcTimeInSecondsSince_1970 );
        LocTime = localtime( &UtcTimeInSecondsSince_1970 );



        // Error check

        if ( !DataDirectory || !FileSuffix )
                return( 0 );

        
        // Ensure the data directory exists

        if ( ! SrFileExist( DataDirectory ) ) {

                if ( SrMkDir( DataDirectory ) == -1 /* == failure */ )

                        return( 0 );
                }


        
        // Prepare the output file name: YYYY-MM-DD-at-HH-MM-SS.asc or .csv
        //
        //    YMDHMS = ( year, month, day, hours, minutes, seconds)
        //
        //

        sprintf( DvmInfo.OutFileNameShort, "%04d-%02d-%02d-at-%02d-%02d-%02d.%s",

                                        LocTime->tm_year+1900,
                                        LocTime->tm_mon+1,
                                        LocTime->tm_mday,
                                        LocTime->tm_hour,
                                        LocTime->tm_min,
                                        LocTime->tm_sec,
                                        FileSuffix
                                        );

        sprintf( DvmInfo.OutFileName, "%s%c%s",
                                        DataDirectory,
                                        SR_DIRSLASH,
                                        DvmInfo.OutFileNameShort
                                        );


        DvmTitleUpdate( ); // to get new YMDHMS file name
        
        return( 1 );
}

//------------------------------------------------------------------------------
// ROUTINE: DvmFileHeaderWrite
// PURPOSE: Open the DVM output file and write out the header info.
//------------------------------------------------------------------------------
int DvmFileHeaderWrite( void ) {

        int   i;
        char *Suffix;
        FILE *OutFp;

        // Quick exit if no output file desired

        if ( DvmInfo.OutFileOn == INISWITCH_OFF )
                return( 1 );


        // Prepare output file name and directory if needed

        if ( DvmInfo.OutFileFormat == OUTPUT_FORMAT_CSV )
                Suffix = "csv";
        else
                Suffix = "asc";
        
        if ( DvmInfo.OutFileNaming == SRDAT_FILENAMETYPE_YMDHMS ) {
                if ( !DvmFileYmdHms( DVM_DEFAULT_DIR, Suffix ) ) {
                        sprintf( TempStr,
                                 "Could not create directory %s for output data files.",
                                 DVM_DEFAULT_DIR );
                        DvmInfo.ShowLogMessage( MSGLEVEL_ERR, ERROR_FATAL, TempStr );
                        return( 0 );
                        }
                }

        else { // if ( DvmInfo.OutFileNaming == SRDAT_FILENAMETYPE_SINGLE ) {
                sprintf( DvmInfo.OutFileNameShort, "%s.%s", DVM_DEFAULT_OUT, Suffix );
                strncpy( DvmInfo.OutFileName, DvmInfo.OutFileNameShort, MAXSTR );
                }



        // Ensure the file does not exist

        if ( SrFileExist( DvmInfo.OutFileName ) ) {
                sprintf( TempStr,
                         "Output file %s already exists.  Please delete before running again.",
                         DvmInfo.OutFileName );
                DvmInfo.ShowLogMessage( MSGLEVEL_ERR, ERROR_FATAL, TempStr );
                return( 0 );
                }
        


        // Open the .out file for write.

        OutFp = fopen( DvmInfo.OutFileName, "w" );

        if ( !OutFp ) {
                sprintf( TempStr,
                         "Could not open out file %s to write header.",
                         DvmInfo.OutFileName );
                DvmInfo.ShowLogMessage( MSGLEVEL_ERR, ERROR_FATAL, TempStr );
                return( 0 );
                }


        // Only print header info if requested

        if ( DvmInfo.OutFileShowHeader == INISWITCH_ON ) {

                if ( DvmInfo.OutFileFormat == OUTPUT_FORMAT_FW ) {

                        // Output topmost part

                        fprintf( OutFp, "# %s\n# format=fw\n#\n", DvmInfo.ProgramName );


                        // Output general values

                        DvmIniWrite( OutFp, INIPART_GENERAL, 0, NULL, "# ", NULL );

                        // Output per channel values

                        for ( i = 0 ; i < DvmInfo.AtodAnalogChannels ; i++ )
                                DvmIniWrite( OutFp, INIPART_DISPLAY, i, NULL, "# ", NULL );



                        // Output channel titles

                        fprintf( OutFp, "#\n#   Pt#   " );

                        for ( i = 0 ; i < DvmInfo.AtodAnalogChannels ; i++ )
                                fprintf( OutFp, " %10s ", DvmInfo.Channel[i].Title );

                        fprintf( OutFp, "D0 D1 D2 D3 PwrH DegC  DegF   Time in Seconds     Time as YMD HMS    Frac Sec\n" );


                        // Output channel units

                        fprintf( OutFp, "#         " );

                        for ( i = 0 ; i < DvmInfo.AtodAnalogChannels ; i++ )
                                fprintf( OutFp, " %10s ", DvmInfo.Channel[i].UnitsName );

                        fprintf( OutFp, "\n" );
                        }
                
                else { // if ( DvmInfo.OutFileFormat == OUTPUT_FORMAT_CSV ) {
                        
                        // Output channel titles (units)

                        fprintf( OutFp, "Pt#," );

                        for ( i = 0 ; i < DvmInfo.AtodAnalogChannels ; i++ )
                                fprintf( OutFp, "\"%s (%s)\",",
                                         DvmInfo.Channel[i].Title,
                                         DvmInfo.Channel[i].UnitsName );

                        fprintf( OutFp, "D0,D1,D2,D3,PwrH,DegC,DegF,Seconds,Date(YMD),Time(HMS),FracSec\n" );
                        }

                } // end if OutFileShowHeader

        fclose( OutFp );

        return( 1 );
}

//------------------------------------------------------------------------------
// ROUTINE: DvmFileDataWrite
// PURPOSE: Open the DVM output file and write out the current data.
//------------------------------------------------------------------------------
int DvmFileDataWrite( void ) {

        int    i, PowerGoodH;
        double TempC, TempF;
        char   DataString[MAXSTR];
        FILE   *OutFp;


        // Quick exit if no output file desired

        if ( DvmInfo.OutFileOn == INISWITCH_OFF )
                return( 1 );


        // Open the .out file for append.

        OutFp = fopen( DvmInfo.OutFileName, "a" );

        if ( !OutFp ) {
                sprintf( TempStr,
                         "Could not open out file %s to write data.",
                         DvmInfo.OutFileName );
                DvmInfo.ShowLogMessage( MSGLEVEL_ERR, ERROR_FATAL, TempStr );
                return( 0 );
                }



        // Compute some status values

        PowerGoodH = SrUsbXchIsPowerBitGood( DvmInfo.AtodPowerInfo,
                                             USBXCH_POWER_HISTORY );
        
        SrDatUsbTemperatureCompute( DvmInfo.AtodTemperature, &TempC, &TempF );


        if ( DvmInfo.OutFileFormat == OUTPUT_FORMAT_FW ) {

                fprintf( OutFp, "%010ld", DvmInfo.SampleNumber );

                for ( i = 0 ; i < DvmInfo.AtodAnalogChannels ; i++ ) {
                        DvmDataFormat( i, DataString, MAXSTR-1 );
                        fprintf( OutFp, " %10s ", DataString );
                        }

                fprintf( OutFp, " %d  %d  %d  %d   ",
                                  (DvmInfo.AtodDigitalIn & 0x01),
                                 ((DvmInfo.AtodDigitalIn & 0x02)>>1),
                                 ((DvmInfo.AtodDigitalIn & 0x04)>>2),
                                 ((DvmInfo.AtodDigitalIn & 0x08)>>3) );

                fprintf( OutFp, "%1d %5.1lf %5.1lf  %17.6lf  %4d/%02d/%02d %02d:%02d:%02d  0.%06ld\n",
                                 PowerGoodH, TempC, TempF, DvmInfo.TimeSecSince1970,
                                 DvmInfo.TimeYear, DvmInfo.TimeMonth, DvmInfo.TimeDay,
                                 DvmInfo.TimeHour, DvmInfo.TimeMin,   DvmInfo.TimeSec,
                                 DvmInfo.TimeUsec );
                }

        else { // if ( DvmInfo.OutFileFormat == OUTPUT_FORMAT_CSV ) {

                fprintf( OutFp, "%ld,", DvmInfo.SampleNumber );

                for ( i = 0 ; i < DvmInfo.AtodAnalogChannels ; i++ ) {
                        DvmDataFormat( i, DataString, MAXSTR-1 );
                        fprintf( OutFp, "%s,", DataString );
                        }

                fprintf( OutFp, "%d,%d,%d,%d,",
                                  (DvmInfo.AtodDigitalIn & 0x01),
                                 ((DvmInfo.AtodDigitalIn & 0x02)>>1),
                                 ((DvmInfo.AtodDigitalIn & 0x04)>>2),
                                 ((DvmInfo.AtodDigitalIn & 0x08)>>3) );
                
                fprintf( OutFp, "%1d,%.1lf,%.1lf,%.6lf,%4d/%02d/%02d,%02d:%02d:%02d,0.%06d\n",
                                 PowerGoodH, TempC, TempF, DvmInfo.TimeSecSince1970,
                                 DvmInfo.TimeYear, DvmInfo.TimeMonth, DvmInfo.TimeDay,
                                 DvmInfo.TimeHour, DvmInfo.TimeMin,   DvmInfo.TimeSec,
                                 DvmInfo.TimeUsec );
                
                } // end else if OUTPUT_FORMAT_CSV

        fclose( OutFp );

        return( 1 );
}

//------------------------------------------------------------------------------
// ROUTINE: DvmTriggerCondition
// PURPOSE: To determine if a trigger condition has been met.
//------------------------------------------------------------------------------
int DvmTriggerCondition( void ) {

        // For keypress,

        if ( DvmInfo.TriggerType == TRIGGER_KEYPRESS )
                if ( DvmInfo.TriggerRequested == 1 ) {
                        DvmInfo.TriggerRequested = 0;
                        return( 1 );
                        }
                else
                        return( 0 );


        // For amplitude, trigger when threshold exceeded

        else if ( DvmInfo.TriggerType == TRIGGER_AMPLITUDE )
//FIX - not implemented yet
                return( 1 );


        // For continuous and default, always pass trigger check

        else // ( DvmInfo.TriggerType == TRIGGER_CONTINUOUS )
                return( 1 );
}

//------------------------------------------------------------------------------
// ROUTINE: DvmTriggerAction
// PURPOSE: Perform some action once the trigger condition has been met.
//          The default action is saving data to a file.
//------------------------------------------------------------------------------
int DvmTriggerAction( void ) {

        DvmFileDataWrite( );

        return( 1 );
}

//------------------------------------------------------------------------------
// ROUTINE: DvmTimeUpdate
// PURPOSE: Get the current PC time and update the DvmInfo structure.
//------------------------------------------------------------------------------
int DvmTimeUpdate( void ) {

        int Success;

        Success = SrGetPcTime( &DvmInfo.TimeSecSince1970,
                          &DvmInfo.TimeYear,
                          &DvmInfo.TimeMonth,
                          &DvmInfo.TimeDay,
                          &DvmInfo.TimeHour,
                          &DvmInfo.TimeMin,
                          &DvmInfo.TimeSec,
                          &DvmInfo.TimeUsec );

        return( Success );
}

//------------------------------------------------------------------------------
// ROUTINE: DvmIniSetDefaults
// PURPOSE: Setup default ini values to use in case the user either does not
//          specify an ini or does not include all values.
//------------------------------------------------------------------------------
int DvmIniSetDefaults( void ) {

        int i;

        // Set general parameters

        strcpy( DvmInfo.AtodDriverName,   "SrUsbXch0" );
        strcpy( DvmInfo.AtodXchModelName, "USB4CH" );
        strcpy( DvmInfo.GpsModelName,     "PCTIME" );
        strcpy( DvmInfo.OutFileName,      "YYYY-MM-DD-at-hh-mm-ss.asc" ); // YMDHMS name
        strcpy( DvmInfo.OutFileComment,   "Data acquired with USBxCH DVM" );
        strcpy( DvmInfo.EditorName,        DVM_DEFAULT_EDITOR );

        DvmCreateShortFileName( DvmInfo.IniFileName, DvmInfo.IniFileNameShort );
        DvmCreateShortFileName( DvmInfo.OutFileName, DvmInfo.OutFileNameShort );

        DvmInfo.AtodXchModel       = SRDAT_ATODMODEL_USB4CH;
        DvmInfo.GpsModel           = SRDAT_GPSMODEL_PCTIME;
        DvmInfo.AtodAnalogChannels = 4;
        DvmInfo.TriggerType        = TRIGGER_CONTINUOUS;
        DvmInfo.TriggerRequested   = 0;
        DvmInfo.OutFileOn          = INISWITCH_ON;
        DvmInfo.OutFileNaming      = SRDAT_FILENAMETYPE_YMDHMS;
        DvmInfo.OutFileFormat      = OUTPUT_FORMAT_FW;
        DvmInfo.OutFileShowHeader  = INISWITCH_ON;


        // Set per channel parameters

        for ( i = 0 ; i < MAXCHANNEL ; i++ ) {
                sprintf( DvmInfo.Channel[i].Title, "Channel %2d", i );
                sprintf( DvmInfo.Channel[i].UnitsName, "Volts" );
                DvmInfo.Channel[i].UnitsType   = UNITS_FLOATS;
                DvmInfo.Channel[i].UnitsPlaces = 7;
                DvmInfo.Channel[i].UnitsDigits = 5;
                DvmInfo.Channel[i].UnitsSlope  = (double)USBXCH_VOLTAGE_FULL_SCALE_RANGE /
                                                 (double)(0xFFFFFF); // volts per count
                DvmInfo.Channel[i].UnitsOffset = 0.0L;
                }

        return( 1 );
}

//------------------------------------------------------------------------------
// ROUTINE: DvmIniCheck
// PURPOSE: Check the ini file name given on the command line.
//          If no ini file was given, create a default ini, and return 2.
//          Else if the specified ini file exists, save its name and return 1.
//          Else it doesn't exist, so print an error message and exit.
//------------------------------------------------------------------------------
int DvmIniCheck( char *IniFilename ) {

        int iret, Ok;

        // If no filename given, create a default ini

        if ( IniFilename == NULL ) {
                strncpy( DvmInfo.IniFileName, DVM_DEFAULT_INI, MAXSTR );
                sprintf( TempStr, "%s\n\n%s\n%s\n",
                         "No ini file       ",
                         "  ... saving settings to default ini file",
                         "  ... running with defaults"
                       );
                DvmInfo.ShowLogMessage( MSGLEVEL_WARN, ERROR_NOT_FATAL, TempStr );
                DvmIniCreateDefault( DvmInfo.IniFileName );
                iret = 2;
                }


        // Else if the ini file does exist, save its name

        else if ( SrFileExist( IniFilename ) ) {
                strncpy( DvmInfo.IniFileName, IniFilename, MAXSTR );
                iret = 1;
                }


        // Otherwise print an message and exit

        else {
                sprintf( TempStr,
                         "ini file %s does not exist     \n\n   Try again or run with no ini file",
                         IniFilename );
                DvmInfo.ShowLogMessage( MSGLEVEL_ERR, ERROR_FATAL, TempStr );
                iret = 0;
                }


        // Create short version (with no path info) too

        Ok = DvmCreateShortFileName( DvmInfo.IniFileName, DvmInfo.IniFileNameShort );


        return( iret );
}


//------------------------------------------------------------------------------
// ROUTINE: DvmIniCreateDefault
// PURPOSE: Create a default ini file containing the current settings.  This
//          will overwrite the default ini if it already exists.
//------------------------------------------------------------------------------
int DvmIniCreateDefault( char *IniFilename ) {

        FILE *IniFp;
        int   i;


        // Set up the default values

        DvmIniSetDefaults( );


        // Open default .ini file for writing

        IniFp = fopen( IniFilename, "w" );

        if ( !IniFp ) {
                sprintf( TempStr,
                         "Could not open ini file %s to write settings",
                         IniFilename );
                DvmInfo.ShowLogMessage( MSGLEVEL_ERR, ERROR_FATAL, TempStr );
                return( 0 );
                }


        // Output overwrite warning notice

        fprintf( IniFp, "; *** This file is generated automatically and may be overwritten.\n" );
        fprintf( IniFp, "; *** Please copy to a new name before editing so edits won't be lost.\n\n" );
        strcpy( DvmInfo.OutFileComment, "Output from default DVM ini file" );


        // Output ini parameters, general and for each display

        DvmIniWrite( IniFp, INIPART_GENERAL, 0,
                     "; DVM default ini settings output by DvmGui or DvmCmd\n\n",
                     NULL, "\n" );

        for ( i = 0 ; i < MAXCHANNEL ; i++ ) {
                sprintf( TempStr, "; Parameters for Channel %d\n\n", i );
                DvmIniWrite( IniFp, INIPART_DISPLAY, i, TempStr, NULL, "\n" );
                }


        // Close file

        fclose( IniFp );

        return( 1 );
}

//------------------------------------------------------------------------------
// ROUTINE: DvmIniWrite
// PURPOSE: Print the ini information stored in the DvmInfo structure out to a
//          file.  This helper function is provided so anyone calling it will
//          be guaranteed to produce an ini file with all the required info.
//------------------------------------------------------------------------------
int DvmIniWrite( FILE *FileIni, int Part, int Ichannel,
                 char* HeaderLine, char *Indent, char* FooterLine ) {

        char *In, *NamingString;

        // Error check

        if ( !FileIni )
                return( 0 );


        // Guarantee we have a valid indent

        if (Indent)
                In = Indent;
        else
                In = "";



        if ( DvmInfo.OutFileNaming == SRDAT_FILENAMETYPE_NONE )
                NamingString = "None";
        else if ( DvmInfo.OutFileNaming == SRDAT_FILENAMETYPE_SINGLE )
                NamingString = "Single";
        else // if ( DvmInfo.OutFileNaming == SRDAT_FILENAMETYPE_YMDHMS )
                NamingString = "YmdHms";
        


        // Output general ini data

        if ( Part == INIPART_GENERAL ) {

                if (HeaderLine)  fprintf( FileIni, "%s", HeaderLine );

                fprintf( FileIni, "%sAtodDriverName       = %s\n",     In, DvmInfo.AtodDriverName );
                fprintf( FileIni, "%sAtodModelName        = %s\n",     In, DvmInfo.AtodXchModelName );
//              fprintf( FileIni, "%sGpsModelName         = %s\n",     In, DvmInfo.GpsModelName );
                fprintf( FileIni, "%sOutputFileNaming     = %s\n",     In, NamingString );
                fprintf( FileIni, "%sOutputFileComment    = \"%s\"\n", In, DvmInfo.OutFileComment );
                fprintf( FileIni, "%sOutputFileFormat     = %s\n",     In, OutFormatString[DvmInfo.OutFileFormat] );
                fprintf( FileIni, "%sOutputFileShowHeader = %s\n",     In, OnOffString[DvmInfo.OutFileShowHeader] );
                fprintf( FileIni, "%sTriggerType          = %s\n",     In, TriggerString[DvmInfo.TriggerType] );
                fprintf( FileIni, "%sEditorName           = \"%s\"\n", In, DvmInfo.EditorName );

                if (FooterLine)  fprintf( FileIni, "%s", FooterLine );
                }


        // Output ini data for the specified display

        else if ( Part == INIPART_DISPLAY ) {

                if (HeaderLine)  fprintf( FileIni, "%s", HeaderLine );

                fprintf( FileIni, "%sChannelTitle  %d      = \"%s\"\n", In, Ichannel, DvmInfo.Channel[Ichannel].Title );
                fprintf( FileIni, "%sChannelUnits  %d      = \"%s\"\n", In, Ichannel, DvmInfo.Channel[Ichannel].UnitsName );
                fprintf( FileIni, "%sChannelPlaces %d      = %d\n",     In, Ichannel, DvmInfo.Channel[Ichannel].UnitsPlaces );
                fprintf( FileIni, "%sChannelDigits %d      = %d\n",     In, Ichannel, DvmInfo.Channel[Ichannel].UnitsDigits );
                fprintf( FileIni, "%sChannelSlope  %d      = %lg\n",    In, Ichannel, DvmInfo.Channel[Ichannel].UnitsSlope );
                fprintf( FileIni, "%sChannelOffset %d      = %lg\n",    In, Ichannel, DvmInfo.Channel[Ichannel].UnitsOffset );

                if (FooterLine)  fprintf( FileIni, "%s", FooterLine );

                }

        else {
                return( 0 );
                }


        return( 1 );
}

//------------------------------------------------------------------------------
// ROUTINE: DvmIniParse
// PURPOSE: Parse the DVM ini file and save the results in the DvmInfo structure.
//------------------------------------------------------------------------------
int DvmIniParse( char *IniFile ) {

        int    Success, TokenType, Index, Channel, ChannelDigits, ChannelPlaces;
        double ChannelSlope, ChannelOffset;


        // Set up defaults to be used unless overridden by values from file

        DvmIniSetDefaults( );



        // Open the .ini file.

        Success = IniOpen( IniFile );

        if ( !Success ) {
                sprintf( TempStr, "Can't open ini file %s!", IniFile );
                DvmInfo.ShowLogMessage( MSGLEVEL_ERR, ERROR_NOT_FATAL, TempStr );
//FIX - check return value, was 1
                return( 0 );
                }

        DvmCreateShortFileName( DvmInfo.IniFileName, DvmInfo.IniFileNameShort );



        // Parse the ini file.

parse:
        TokenType = GetToken( );

        if ( TokenType == ERROR )
                return( 0 );


        if ( KeywordEquals( "AtodDriverName" ) ) {

                TokenType = GetToken( );
                if ( TokenType != STRING  &&  TokenType != SYMBOL ) {
                        ErrorIni( "bad AtoD Driver Name string" );
                        return( 0 );
                        }

                strncpy( DvmInfo.AtodDriverName, GetTokenPtr( ), MAXSTR );
                }



        else if ( KeywordEquals( "AtodModelName" ) ) {

                TokenType = GetToken( );
                if ( TokenType != STRING  &&  TokenType != SYMBOL ) {
                        ErrorIni( "bad AtodModelName string" );
                        return( 0 );
                        }

                if ( ! MatchTable( AtodModelTable, &DvmInfo.AtodXchModel, &Index ) ) {
                        ErrorIni( "AtodXchModelName must = {USB1CH,USB4CH,USB8CH,DEFAULT}" );
                        return( 0 );
                        }
                else
                        strncpy( DvmInfo.AtodXchModelName, AtodModelTable[Index].String, MAXSTR );
                }



        else if ( KeywordEquals( "GpsModelName" ) ) {

                TokenType = GetToken( );
                if ( TokenType != STRING  &&  TokenType != SYMBOL ) {
                        ErrorIni( "bad GpsModelName string" );
                        return( 0 );
                        }

                if ( ! MatchTable( GpsModelTable, &DvmInfo.GpsModel, &Index ) ) {
                        ErrorIni( "GpsModelName must = {GARMIN,TRIMBLE,ONCORE,PCTIME,DEFAULT}" );
                        return( 0 );
                        }
                else
                        strncpy( DvmInfo.GpsModelName, GpsModelTable[Index].String, MAXSTR );
                }



        else if ( KeywordEquals( "OutputFileNaming" ) ) {

                if ( GetToken() == ERROR )
                        return( 0 );

                if ( Match( "None" ) ) {
                        DvmInfo.OutFileNaming = SRDAT_FILENAMETYPE_NONE;
                        DvmInfo.OutFileOn     = INISWITCH_OFF;
                        }

                else if ( Match( "Single" ) ) {
                        DvmInfo.OutFileNaming = SRDAT_FILENAMETYPE_SINGLE;
                        DvmInfo.OutFileOn     = INISWITCH_ON;
                        }

                else if ( Match( "YmdHms" ) ) {
                        DvmInfo.OutFileNaming = SRDAT_FILENAMETYPE_YMDHMS;
                        DvmInfo.OutFileOn     = INISWITCH_ON;
                        }

                else {
                        ErrorIni( "OutputFileNaming must = {None,Single,YmdHms}" );
                        return( 0 );
                        }
                }

        
        else if ( KeywordEquals( "OutputFileName" ) ) {
                ErrorIni( "OutputFileName replaced by OutputFileNaming = {None,Single,YmdHms}" );
                return( 0 );
                }
        


        else if ( KeywordEquals( "OutputFileComment" ) ) {

                TokenType = GetToken( );
                if ( TokenType != STRING  &&  TokenType != SYMBOL ) {
                        ErrorIni( "bad comment string" );
                        return( 0 );
                        }

                strncpy( DvmInfo.OutFileComment, GetTokenPtr( ), MAXSTR );
                }


        else if ( KeywordEquals( "OutputFileFormat" ) ) {

                if ( GetToken( ) == ERROR )
                        return( 0 );

                if ( Match( OutFormatString[OUTPUT_FORMAT_FW] ) )
                        DvmInfo.OutFileFormat = OUTPUT_FORMAT_FW;

                else if ( Match( OutFormatString[OUTPUT_FORMAT_CSV] ) )
                        DvmInfo.OutFileFormat = OUTPUT_FORMAT_CSV;

                else {
                        sprintf( TempStr, "OutputFileFormat must = {%s, %s}",
                                 OutFormatString[OUTPUT_FORMAT_FW],
                                 OutFormatString[OUTPUT_FORMAT_CSV] );
                        ErrorIni( TempStr );
                        return( 0 );
                        }
                }



        else if ( KeywordEquals( "OutputFileShowHeader" ) ) {

                Success = GetOnOff( &DvmInfo.OutFileShowHeader,
                                    "OutputFileShowHeader must be ON or OFF" );
                if ( !Success )
                        return( 0 );
                }


        else if ( KeywordEquals( "OutputFileShowIndex" )   ||
                  KeywordEquals( "OutputFileShowTimeSec" ) ||
                  KeywordEquals( "OutputFileShowTimeYmd" ) ) {
                
                ErrorIni( "Option is always ON and no longer available" );
                return( 0 );
                }


        else if ( KeywordEquals( "TriggerType" ) ) {

                if ( GetToken( ) == ERROR )
                        return( 0 );

                if ( Match( "Continuous" ) )
                        DvmInfo.TriggerType = TRIGGER_CONTINUOUS;

                else if ( Match( "Keypress" ) )
                        DvmInfo.TriggerType = TRIGGER_KEYPRESS;

                else if ( Match( "Amplitude" ) )
                        DvmInfo.TriggerType = TRIGGER_AMPLITUDE;

                else {
                        ErrorIni( "TriggerType must = {Continuous, Keypress, Amplitude } " );
                        return( 0 );
                        }

                }


        else if ( KeywordEquals( "EditorName" ) ) {

                TokenType = GetToken( );
                if ( TokenType != STRING  &&  TokenType != SYMBOL ) {
                        ErrorIni( "bad ini editor program name string" );
                        return( 0 );
                        }

                strncpy( DvmInfo.EditorName, GetTokenPtr( ), MAXSTR );
                }



        else if ( ChannelKeywordEquals( "ChannelTitle", 0, MAXCHANNEL-1, &Channel ) ) {

                if ( GetToken( ) != STRING ) {
                        ErrorIni( "illegal CHANNEL title" );
                        return( 0 );
                        }

                strncpy( DvmInfo.Channel[Channel].Title, GetTokenPtr( ), MAXSTR );
                DvmInfo.Channel[Channel].Title[MAXSTR-1] = '\0';
                }




        else if ( ChannelKeywordEquals( "ChannelUnits", 0, MAXCHANNEL-1, &Channel ) ) {

                if ( GetToken( ) != STRING ) {
                        ErrorIni( "illegal CHANNEL units" );
                        return( 0 );
                        }

                strncpy( DvmInfo.Channel[Channel].UnitsName, GetTokenPtr( ), MAXSTR );
                DvmInfo.Channel[Channel].UnitsName[MAXSTR-1] = '\0';

                // Assume all types are UNITS_FLOAT, but override that with
                // UNITS_HEX or UNITS_COUNTS if the name matches HexCounts
                // or Counts respectively

                DvmInfo.Channel[Channel].UnitsType = UNITS_FLOATS;
                MatchTable( ChannelUnitsTypeTable,
                            &DvmInfo.Channel[Channel].UnitsType,
                            &Index );
                }



        else if ( ChannelKeywordEquals( "ChannelPlaces", 0, MAXCHANNEL-1, &Channel ) ) {

                Success = GetInt( &ChannelPlaces, "illegal CHANNEL places" );
                if ( !Success || ChannelPlaces < MINPLACES || MAXPLACES < ChannelPlaces ) {
                        sprintf( TempStr, "channel places must be %d to %d", MINPLACES, MAXPLACES );
                        ErrorIni( TempStr );
                        return( 0 );
                        }

                DvmInfo.Channel[Channel].UnitsPlaces = ChannelPlaces;
                }



        else if ( ChannelKeywordEquals( "ChannelDigits", 0, MAXCHANNEL-1, &Channel ) ) {

                Success = GetInt( &ChannelDigits, "illegal CHANNEL digits" );
                if ( !Success || ChannelDigits < MINDIGITS || MAXDIGITS < ChannelDigits ) {
                        sprintf( TempStr, "channel digits must be %d to %d", MINDIGITS, MAXDIGITS );
                        ErrorIni( TempStr );
                        return( 0 );
                        }

                DvmInfo.Channel[Channel].UnitsDigits = ChannelDigits;
                }
//FIX need to verify places and digits are consistent


        else if ( ChannelKeywordEquals( "ChannelSlope", 0, MAXCHANNEL-1, &Channel ) ) {

                Success = GetFloat( &ChannelSlope, "illegal CHANNEL slope" );
                if ( !Success )
                        return( 0 );
                DvmInfo.Channel[Channel].UnitsSlope = ChannelSlope;
                }



        else if ( ChannelKeywordEquals( "ChannelOffset", 0, MAXCHANNEL-1, &Channel ) ) {

                Success = GetFloat( &ChannelOffset, "illegal CHANNEL offset" );
                if ( !Success )
                        return( 0 );
                DvmInfo.Channel[Channel].UnitsOffset = ChannelOffset;
                }





        // Unused keywords
        //
        // These options are not used in DVM at this time, so just
        // pass over them here ...
        //
        //

        else if ( KeywordEquals( "Title" )                ||
                  KeywordEquals( "AtodPortMode" )         ||
                  KeywordEquals( "DisplayColor" ) ) {

                if ( GetToken( ) == ERROR )   // ignore them ...
                        return( 0 );
                }


        else if ( TokenType == EOF ) {

                Success = IniClose( );     // Close the ini file
                if ( !Success )
                        return( 0 );

                return( 1 );
                }



        else { // token unrecognized ...

                ErrorIni( "unknown .ini syntax" );
                return( 0 );
                }




        // Continue parsing.

        goto parse;

}

//------------------------------------------------------------------------------
// ROUTINE: DvmFormatIsValid
// PURPOSE: Verify that supplied places and digits values will combine to make
//          a valid display format.  Returns 1 for valid values and 0 if invalid.
//------------------------------------------------------------------------------
int DvmFormatIsValid( int Places, int Digits ) {

        int DecimalPt;

        // Prepare correction for case where decimal pt is not needed

        if ( Digits == 0 )
                DecimalPt = 0;
        else
                DecimalPt = 1;


        // Check validity

        if ( Places < MINPLACES    ||
             Places > MAXPLACES    ||
             Digits < MINDIGITS    ||
             Digits > MAXDIGITS    ||
             Digits > Places - 1 - DecimalPt )
                return( 0 );
        else
                return( 1 );
}

//------------------------------------------------------------------------------
// ROUTINE: DvmCreateShortFileName
// PURPOSE: Split off the directory path part of the specified file name.
//------------------------------------------------------------------------------
int DvmCreateShortFileName( char *LongName, char *ShortName ) {

#if defined( SROS_WINXP ) || defined( SROS_WIN2K )

        char fname[_MAX_FNAME], ext[_MAX_EXT];

        if (!LongName || !ShortName)
                return( 0 );

        if (strlen( LongName ) == 0)
                strcpy( ShortName, LongName );
        else {
                _splitpath( LongName, NULL, NULL, fname, ext );
                sprintf( ShortName, "%s%s", fname, ext );
                }
#else
//FIX	printf( "********* DvmCreateShortFileName not available for Linux yet ***************\n" );

	strcpy( ShortName, LongName );
	

#endif // SROS_xxx	
	

        return( 1 );
}


//------------------------------------------------------------------------------
// ROUTINE: DvmLogMessagePrep
// PURPOSE: Prepare a complete message string by adding the time and date
//          on to the supplied message.
//------------------------------------------------------------------------------
void DvmLogMessagePrep( int MsgLevel, char *ExtraMsg ) {

        char *datestr, dbuffer[16];
        char *timestr, tbuffer[16];
        char *levelstr;
        time_t     current_time;
        struct tm *local_time;

        // Get the date and time.

        time( &current_time );
        local_time = localtime( &current_time );

        strftime( dbuffer, 16, "%m/%d/%y", local_time );
        strftime( tbuffer, 16, "%H:%M:%S", local_time );

        datestr = dbuffer;
        timestr = tbuffer;


        // Prepare message string

        if (MsgLevel == MSGLEVEL_MSG)
                levelstr = "Message";
        else if (MsgLevel == MSGLEVEL_WARN)
                levelstr = "Warning";
        else if (MsgLevel == MSGLEVEL_ERR)
                levelstr = "Error";
        else
                levelstr = "Note";

        if (ExtraMsg)
                sprintf( DvmInfo.LogMessage, "%s occurred on %s at %s\n\n%s\n",
                                  levelstr, datestr, timestr, ExtraMsg );
        else
                sprintf( DvmInfo.LogMessage, "%s occurred on %s at %s\n\n",
                                  levelstr, datestr, timestr );
}
