// FILE: DvmLib.h
// COPYRIGHT: (c), Symmetric Research, 2009
//
// Header file for those parts of the USBxCH DVM application shared
// between the GUI (Graphical User Interface) and CMD (Command line)
// versions It is also used by the CAL calibration program.  DvmInfo, of
// type struct DvmVariables, is a container for all the global variables
// used in the DVM program.

#include "SrUsbXch.h"   // SrUsbXch library functions and defines


#define UNITS_HEX          0 // Hex counts
#define UNITS_COUNTS       1 // Decimal counts
#define UNITS_FLOATS       2 // Floating point units

#define MSGLEVEL_MSG       0 // Informational message
#define MSGLEVEL_WARN      1 // Warning message, something not right
#define MSGLEVEL_ERR       2 // Error message, there is trouble

#define ERROR_NOT_FATAL    0 // Program can continue after error
#define ERROR_FATAL        1 // Program must exit after error

#define OUTPUT_FORMAT_FW   0 // Output format is Fixed Width columns
#define OUTPUT_FORMAT_CSV  1 // Output format is Comma Separated Values

#define TRIGGER_CONTINUOUS 0
#define TRIGGER_KEYPRESS   1
#define TRIGGER_AMPLITUDE  2

#define CALIBRATE_NO       0
#define CALIBRATE_YES      1
#define CALIBRATE_SELF     2

//#define DVM_REV       2.01 // Version of DVM programs
//#define HDR_REV          2 // Version of header format

#define DVM_SPS          651
#define DVM_READ_EXTRA    40
#define DVM_TIMER_LESS    30
#define DVM_DEFAULT_INI    "DvmSetupDefault.ini"
#define DVM_DEFAULT_DIR    "DvmData"
#define DVM_DEFAULT_OUT    "DvmOutput"
#define DVM_DEFAULT_EDITOR "notepad.exe"

#define INIPART_GENERAL     0 // General part of ini file
#define INIPART_DISPLAY     1 // Display specific part of ini file

#define INISWITCH_OFF       0 // Set ini file on/off switch OFF
#define INISWITCH_ON        1 // Set ini file on/off switch ON

#define MINDIGITS           0 // Min number of characters after the decimal
#define MAXDIGITS           6 // Max number of characters after the decimal
#define MINPLACES           1 // Min number of total display characters
#define MAXPLACES          10 // Max number of total display characters

//FIX - remove when tested
//#define TITLE_INIFILE       0 // Title with input ini file name
#define TITLE_ASCIIOUT      0 // Title with output ascii file name
#define TITLE_STATUS        1 // Title with power and temperature info
#define TITLE_MAX           2 // Max number of constant titles

#define MAXSTR            256 // Maximum size of string
#define MAXCHANNEL          8 // Max number of analog channels

#define MAXFILTER           2
#define MAXSAMPLES          (DVM_SPS + DVM_READ_EXTRA + 1)
#define MAXPACKETS          (MAXSAMPLES/2)


// Structures, typedefs, and global variable declarations

// Type for pointer to log message function

typedef void (*DVMFNPTR)( int msglevel, int fatal, char *msg );

struct DvmChannel {

        char      Title[MAXSTR];
        char      UnitsName[MAXSTR];
        int       UnitsType;
        int       UnitsPlaces;
        int       UnitsDigits;
        double    UnitsSlope;
        double    UnitsOffset;
        };

typedef struct DvmChannel DVM_CHANNEL;

struct DvmCalibrationReg {
        unsigned char OfcLow;
        unsigned char OfcMid;
        unsigned char OfcHigh;
        unsigned char FscLow;
        unsigned char FscMid;
        unsigned char FscHigh;
        };

typedef struct DvmCalibrationReg DVM_CALREG;

        

struct DvmVariables {

        char       *ProgramName;              // DvmGui or DvmCmd

	char        IniFileName[MAXSTR];      // Ini input file name
	char        IniFileNameShort[MAXSTR]; // Ini input file name short form (no path)
        int         OutFileOn;                // Output file desired (or not)
        int         OutFileNaming;            // Output file naming style (none,single,ymdhms)
        char        OutFileName[MAXSTR];      // Output file name
	char        OutFileNameShort[MAXSTR]; // Output file name short form (no path)
        char        OutFileComment[MAXSTR];   // Output file user settable text
        int         OutFileFormat;            // Output file format (fw or csv)
        int         OutFileShowHeader;        // Output file with header info
        char        EditorName[MAXSTR];       // Name of program for editing ini
        
        
        DEVHANDLE      AtodHandle;
	char           AtodDriverName[MAXSTR];
	char           AtodXchModelName[MAXSTR];
	int            AtodXchModel;
	int            AtodAnalogChannels;
        double         AtodSampleRate;
        int            AtodLedState;
        int            AtodPowerInfo;
        int            AtodTemperature;
        int            AtodDigitalIn;

        DVM_CHANNEL Channel[MAXCHANNEL];
        DVM_CALREG  CalReg[MAXCHANNEL];
        
        int         TriggerType;
        int         TriggerRequested;

	int         Nread;
	int         TimerInterval;
        int         IsRunning;                // Indicates if A/D is running

        int         FilterInitNeeded;

        long        SampleNumber;

        int         TimeYear;
        int         TimeMonth;
        int         TimeDay;
        int         TimeHour;
        int         TimeMin;
        int         TimeSec;
        long        TimeUsec;
        double      TimeSecSince1970;

	int         GpsModel;
	char        GpsModelName[MAXSTR];
        
        SRUSBXCH_PACKET     PacketData[MAXPACKETS];
        SRDAT_USBXCH_ANALOG DataRaw[MAXSAMPLES];
	long          DataAverage[USBXCH_MAX_ANALOG_CHANNELS];
	long          DataCounts[USBXCH_MAX_ANALOG_CHANNELS];
	float         DataFloats[USBXCH_MAX_ANALOG_CHANNELS];
        double        FilterTerms[MAXFILTER][USBXCH_MAX_ANALOG_CHANNELS];
        double        FilterCoeff[MAXFILTER];
        unsigned long Nremain;
        int           LastMsg;

        char        Title[TITLE_MAX][MAXSTR]; // Title string text
        int         Ntitle;                   // Number of title strings

        char        LogMessage[MAXSTR];
        
        DVMFNPTR    ShowLogMessage;
        };

typedef struct DvmVariables DVM_VARIABLES;

extern DVM_VARIABLES DvmInfo;



// Function Prototypes

char *Chop( char *s );

int  DvmInit( DVMFNPTR NewShowLogMessageFunction,
              char *CmdLine,
              char *ProgramName );

int  DvmAtodOpen( void );
int  DvmAtodClose( void );
int  DvmAtodStart( void );
int  DvmAtodCalibrate( int Calibrate );
int  DvmAtodRead( int *Nsamples );
int  DvmAtodStop( void );

int  DvmCalInit( void );
int  DvmCalSet( char *RegData );
int  DvmCalGet( char *RegData );

int  DvmIsRunning( void );
int  DvmTitleInit( void );
int  DvmTitleUpdate( void );

int  DvmFileYmdHms( char *DataDirectory, char *FileSuffix );
int  DvmFileHeaderWrite( void );
int  DvmFileDataWrite( void );

void DvmDataInit( void );
void DvmDataFilter( int Nsamples );
int  DvmDataFormat( int Ichannel, char *DataString, int Nchar );

int  DvmTriggerCondition( void );
int  DvmTriggerAction( void );

int  DvmTimeUpdate( void );

int  DvmIniSetDefaults( void );
int  DvmIniCheck( char *IniFilename );
int  DvmIniCreateDefault( char *IniFilename );
int  DvmIniWrite( FILE *FileIni, int Part, int Ichannel,
                  char* HeaderLine, char *Indent, char* FooterLine );
int  DvmIniParse( char *IniFilename );

int  DvmFormatIsValid( int Places, int Digits );

int  DvmCreateShortFileName( char *LongName, char *ShortName );

void DvmLogMessagePrep( int MsgLevel, char *ExtraMsg );

