# FILE: Plot4.gp
# COPYRIGHT (c): Symmetric Research, 2010
#
# This GnuPlot file plots the same data as in Plot3.gp, but it adds several
# cosmetic improvements.  These include labeling the axes, titling the
# legend or key, titling the overall plot, setting the range on the axes
# and changing the line and point styles.
#
# While none of these changes are very complicated, they start adding up
# to alot of lines to type in at the GnuPlot command prompt.  So, you may
# want to put these commands into a .gp file like this one.  Then you can
# execute all the commands by just typing
#
#    load 'Plot4.gp'
#
# at the GnuPlot command prompt.
#
# It is nice to see this plot on the screen, but sometimes it would be
# better to generate a PDF file that could be used in a publication.
# Solar.gp along with MakePdf.bat and the free GhostScript program does this.
#



# REMOVE ANY PREVIOUS DEFAULT SETTINGS
#
#

reset


# DEFINE SOME HANDY CONSTANTS RELATED TO THE DATA FILE
#
#

DataFile = "DvmOutputSolar.asc"
ColSample = 1                   # sample number  is in column 1
ColChan0  = 2                   # channel 0 data is in column 2




# DEFINE LABEL FOR AXES, PLOT AND KEY AND CONTROL RANGE OF AXES
#
#

set xlabel "Sample"                     font "Helvetica Bold,14"
set ylabel "Light Level"                font "Helvetica Bold,14"
set title  "USBxCH Solar Panel Data"    font "Helvetica Bold,18"
KeyTitle = "Solar panel results"

set xtics nomirror                      font "Helvetica Bold,14"
set ytics nomirror                      font "Helvetica Bold,14"

set xrange [30:100]
set yrange [0:10]




# SET THE DATA PLOTTING STYLE
#
# The first set style says plot data with points connected by lines.
# The next set style defines linestyle 1 that can be referenced later
# by the plot command.  This linestyle has solid black thin lines
# connecting small filled circle points.
#

set style data linespoint
set style line 1 linetype -1 linewidth 1 pointtype 7 pointsize 1


# PLOT EXTRA LABELS AND ARROWS:
#
# Coordinates are given in x-axis and y-axis values
#

set label " Light very close" at 71,8 center
set arrow from  71,8.2 to 71,9.5

set label "Light off " at  95,2.7 right
set arrow from 95,2.7 to 99,1.3




# NOW DO THE PLOT
#
#

plot DataFile                   \
      every 2::30::100          \
      using ColSample:ColChan0  \
      linestyle 1               \
      title KeyTitle



# CLEAN UP SETTINGS
#
#

reset
