# FILE: Solar.gp
# COPYRIGHT (c): Symmetric Research, 2010
#
# This file contains GnuPlot commands for plotting the data from the USBxCH
# file DvmOutputSolar.asc into an intermediate postscript file that can be 
# turned into a PDF file.  The plot shows USB4CH data measuring the output 
# of an inexpensive solar panel in varying light conditions.
#
# Use the supplied batch file, MakePdf.bat, to run this GnuPlot file and 
# call the GhostScript program to go all the way to pdf.
#



# ENSURE WE START WITH A CLEAN SLATE:
#
#

reset



# SETUP THE POSTSCRIPT FILE OUTPUT:
#
# Process the Postscript output from GnuPlot through GhostScript or 
# Adobe Distiller to make a PDF.
#
# The GnuPlot set terminal postscript defaults are:
#
#    orientation = landscape
#         x size = 10" horizontal
#         y size =  7" vertical
#
#
#

set terminal postscript
set output 'Solar.ps'



# SET DATA AND LINE STYLES:
#
#

set style data linespoint              # include lines and points
set style line 1 lt -1 lw 1 pt 7 ps 1  # thin line with closed circles



# SET AXIS DATA LIKE TIC MARK STYLE, FORMAT, LABEL AND RANGE:
#
#

set xtics nomirror        font "Helvetica Bold,14"
set ytics nomirror        font "Helvetica Bold,14"

set xlabel "Sample"       font "Helvetica Bold,18"
set ylabel "Light Level"  font "Helvetica Bold,18"

set xrange [0:125]  # About 2 minutes of data
set yrange [0:10]   # Solar panel calibrated to light levels from 1 to 10




# SET PLOT AND LEGEND TITLES:
#
#

set title  "USBxCH Solar Panel Data" font "Helvetica Bold,18"
KeyTitle = "Solar panel results"




# PLOT EXTRA LABELS AND ARROWS:
#
#

set label " Light very close" at 70,7 center
set arrow from  70,7.2 to 70,9.5

set label "Light off " at  105,2.5 right
set arrow from 105,2.5 to 115,1.2




# PLOT SOLAR DATA:
#
#

plot "DvmOutputSolar.asc" using 1:2 linestyle 1 title KeyTitle




# CLEAN UP THINGS WE'VE SET
#
#

reset
