// FILE: SrIniParse.h
// COPYRIGHT: (c), Symmetric Research, 2009
//
// Header file for SR parsing functions used by several products and programs.
//

#include "SrDefines.h" // For function type, dev handle, 64 bit, + other defines


#define MAXTOKEN        256             // token info

#define ERROR           0
#define SYMBOL          1               // token types
#define EQUALS          2
#define HEX             3
#define INTEGER         4
#define FLOAT           5
#define EXPFLOAT        6
#define STRING          7
#define UNKNOWN         9

#define OFF             0
#define ON              1

typedef struct MatchTablePair {
        char *String;
        int   Value;
        } MATCHTABLE;

typedef void (*FNPTR)( char *msg ); // Pointer to error or warning function

extern FNPTR ErrorIni;
extern FNPTR WarnIni;


// Parsing helper function prototypes

void  DefaultErrorIni( char *msg );
void  DefaultWarnIni( char *msg );

FUNCTYPE( int )   IniOpen( char *IniFilename );
FUNCTYPE( int )   IniClose( void );
FUNCTYPE( void )  SetErrorIniFunction( FNPTR ErrorIniFunction );
FUNCTYPE( void )  SetWarnIniFunction( FNPTR WarnIniFunction );
FUNCTYPE( char* ) PrepareFullMsg( char *msg );
FUNCTYPE( int )   AddToToken( int c );
FUNCTYPE( int )   SkipSpace( int c );
FUNCTYPE( int )   GetToken( void );
FUNCTYPE( char* ) GetTokenPtr( void );
FUNCTYPE( FILE* ) GetFilePtr( void );
FUNCTYPE( int )   Match( char *s );
FUNCTYPE( int )   MatchTable( MATCHTABLE *table, int *value, int *index );
FUNCTYPE( int )   KeywordEquals( char *keyword );
FUNCTYPE( int )   ChannelKeywordEquals( char *keyword, int min, int max, int *chan );
FUNCTYPE( int )   GetOnOff( int *parameter, char *error );
FUNCTYPE( int )   GetInt( int *parameter, char *error );
FUNCTYPE( int )   GetUint( unsigned int *parameter, char *error );
FUNCTYPE( int )   GetUlong( unsigned long *parameter, char *error );
FUNCTYPE( int )   GetHex( int *parameter, char *error );
FUNCTYPE( int )   GetFloat( double *parameter, char *error );
