// FILE: ScopeLib.c
// COPYRIGHT: (c), Symmetric Research, 2008-2010
//
// Source file for shared portion of scope application common to both
// the GUI and CMD versions.
//


#if defined( SROS_WINXP ) || defined( SROS_WIN2K )
#include <windows.h>                               // for GetLocalTime
#define S_IRWXU         0
#define S_IRWXG         0
#define S_IRWXO         0
#elif defined( SROS_LINUX )
#define _MAX_FNAME      256
#define _MAX_EXT        256
#endif  // SROS_xxxxx

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <time.h>       // for time() and localtime() functions
#include <sys/stat.h>
#include <sys/types.h>


#include "ScopeLib.h"
#include "SrIniParse.h" // for PrepareFullMsg, GetOnOff, KeywordEquals etc
#include "SrHelper.h"   // for SrFileExist, SrGetKeyWait, SrStringCopy, etc


// Global variables

SCOPE_VARIABLES ScopeInfo;
SCOPE_VARIABLES ScopeInfoSave;
SRDAT_HDR       DatHdr;    // See SrUsbXch.h for layout

char msg[MAXSTR];


// Runtime error handling ...

extern void DisplayLogMessage( int MsgLevel, int DlgShow, int Fatal, char *ExtraMsg );



// REV MESSAGE:
//
// 2008/04/03 First version
//



//------------------------------------------------------------------------------
// ROUTINE: ScopeLibRev
// PURPOSE: Print the scope library revision info.
//------------------------------------------------------------------------------
void ScopeLibRev( void ) {

        // Print the rev info.

        sprintf( msg, "USBxCH ScopeLib [/?]: (c) SR 2008 : Rev {2010/08/09}\n" );
}

//------------------------------------------------------------------------------
// ROUTINE: DefaultShowLogMessage
// PURPOSE: Default function for showing a logging message.  It will exit the
//          program if the fatal flag is set.
//------------------------------------------------------------------------------
void DefaultShowLogMessage( int msglevel, int dlgshow, int fatal, char *msg ) {
        
        printf( "Level %d Message = %s\n", msglevel, msg );
        if ( fatal )
                exit( 1 );
        else
                return;
}

//------------------------------------------------------------------------------
// ROUTINE: ScopeToggleSwitch
// PURPOSE: Change the value of the provided switch variable.  It toggles
//          between ON and OFF settings.
//------------------------------------------------------------------------------
int ScopeToggleSwitch( int *Switch ) {

        if ( !Switch )
                return( 0 );

        if ( *Switch == INISWITCH_ON )
                *Switch = INISWITCH_OFF;

        else // ( *Switch == INISWITCH_OFF )
                *Switch = INISWITCH_ON;

        return( 1 );
}


// STRING HELPERS:
//
//
//


//FIX - Add to SrHelper instead ???
//------------------------------------------------------------------------------
// ROUTINE: Chop
// PURPOSE: Chop spaces ' ' off the end of a string and add '\n'.
//------------------------------------------------------------------------------
static char ChopBuffer[256];

char *Chop( char *s ) {

        char *p;


        // Chop spaces ' ' off the end of a string and add '\n'.

        SrStringCopy( ChopBuffer, s, 256 );

        p = ChopBuffer;

        while ( *p )

                p++;


        if ( *--p == '\n' ) {

                while ( *--p == ' ' )

                         ; // back up


                *++p = '\n';
                *++p = '\0';

                }


        return( ChopBuffer );

}

//------------------------------------------------------------------------------
// ROUTINE: ScopeDisplayInfo
// PURPOSE: Control the display of program information based on an ini switch.
//------------------------------------------------------------------------------
void ScopeDisplayInfo( char *info ) {

        if ( ScopeInfo.ProgramDisplayInfo == INISWITCH_ON )
                printf( info );
}

//------------------------------------------------------------------------------
// ROUTINE: ScopeDisplayInfoFlush
// PURPOSE: Control the display of program information based on an ini switch.
//          This version adds a flush to ensure the info is displayed even
//          if it does not include a trailing carriage return.
//------------------------------------------------------------------------------
void ScopeDisplayInfoFlush( char *info ) {

        if ( ScopeInfo.ProgramDisplayInfo == INISWITCH_ON ) {
                printf( info );
                SrFlushPrint();
                }

}


// These are the INI report helpers.

//------------------------------------------------------------------------------
// ROUTINE: ScopeStringMatch
// PURPOSE: Check for a match between s and t.
//------------------------------------------------------------------------------
int ScopeStringMatch( char *s, char *t ) {

        while ( *s && *t && toupper( *s ) == toupper( *t ) )

                { s++; t++; }


        return ( *s == '\0' && *t == '\0' ? 1 : 0 );
}

//------------------------------------------------------------------------------
// ROUTINE: ScopeMakeOnOffString
// PURPOSE: Convert the INISWITCH defines to the strings "On" and "Off".
//          Returns NULL on error;
//------------------------------------------------------------------------------
char* ScopeMakeOnOffString( int OnOffValue ) {

        char *s;

        switch ( OnOffValue ) {
                case INISWITCH_OFF:   s = "Off";    break;
                case INISWITCH_ON:    s = "On";     break;
                default:              s = NULL;
                }

        return( s );
}

//------------------------------------------------------------------------------
// ROUTINE: ScopeMakeOnOffValue
// PURPOSE: Convert the strings "On" and "Off" to the INISWITCH_ integers.
//          Returns -1 on error.
//------------------------------------------------------------------------------
int ScopeMakeOnOffValue( char *OnOffString ) {

        char *s;
        int   value;

        s = OnOffString;

        if ( !s )                                 value = -1;
        else if ( ScopeStringMatch( s, "Off" ) )  value = INISWITCH_OFF;
        else if ( ScopeStringMatch( s, "On" ) )   value = INISWITCH_ON;
        else                                      value = -1;

        return( value );
}

//------------------------------------------------------------------------------
// ROUTINE: ScopeMakeTriggerTypeValue
// PURPOSE: Converts a trigger type string into an integer representing one of
//          the allowed trigger types as defined by the SRDAT_TRIGGER constants in
//          SrUsbXch.h into a string.  It returns -1 if an invalid string is
//          given.
//------------------------------------------------------------------------------
int ScopeMakeTriggerTypeValue( char *TriggerTypeString ) {

        char *s;
        int   value;

        s = TriggerTypeString;

        if ( !s )                                         value = -1;
        else if ( ScopeStringMatch( s, "None" ) )         value = SRDAT_TRIGGERTYPE_NONE;
        else if ( ScopeStringMatch( s, "Threshold" ) )    value = SRDAT_TRIGGERTYPE_THRESHOLD;
        else if ( ScopeStringMatch( s, "DigitalLevel" ) ) value = SRDAT_TRIGGERTYPE_DIGLEVEL;
        else if ( ScopeStringMatch( s, "DigitalEdge" ) )  value = SRDAT_TRIGGERTYPE_DIGEDGE;
        else                                              value = -1;

        return( value );
}

//------------------------------------------------------------------------------
// ROUTINE: ScopeMakeTriggerActiveLevelValue
// PURPOSE: Convert the strings "High" and "Low" to the SRDAT_TRIGLEVEL_ defines.
//          Returns -1 on error.
//------------------------------------------------------------------------------
int ScopeMakeTriggerActiveLevelValue( char *TriggerActiveLevelString ) {

        char *s;
        int   value;

        s = TriggerActiveLevelString;

        if ( !s )                                  value = -1;
        else if ( ScopeStringMatch( s, "High" ) )  value = SRDAT_TRIGGERLEVEL_HIGH;
        else if ( ScopeStringMatch( s, "Low" ) )   value = SRDAT_TRIGGERLEVEL_LOW;
        else                                       value = -1;

        return( value );
}

//------------------------------------------------------------------------------
// ROUTINE: ScopeMakeTriggerActiveEdgeValue
// PURPOSE: Convert a "Rising" or "Falling" string to the matching SRDAT_TRIGEDGE_
//          define.  Returns -1 on error.
//------------------------------------------------------------------------------
int ScopeMakeTriggerActiveEdgeValue( char *TriggerActiveEdgeString ) {

        char *s;
        int   value;

        s = TriggerActiveEdgeString;

        if ( !s )                                    value = -1;
        else if ( ScopeStringMatch( s, "Rising" ) )  value = SRDAT_TRIGGEREDGE_RISING;
        else if ( ScopeStringMatch( s, "Falling" ) ) value = SRDAT_TRIGGEREDGE_FALLING;
        else                                         value = -1;

        return( value );
}

//------------------------------------------------------------------------------
// ROUTINE: ScopeMakeOutputFileNamingValue
// PURPOSE: Convert a FileNaming string to the matching SRDAT_FILENAMETYPE_
//          define.  Returns -1 on error.
//------------------------------------------------------------------------------
int ScopeMakeOutputFileNamingValue( char *OutputFileNamingString ) {

        char *s;
        int   value;

        s = OutputFileNamingString;

        if ( !s )                                       value = -1;
        else if ( ScopeStringMatch( s, "None" ) )       value = SRDAT_FILENAMETYPE_NONE;
        else if ( ScopeStringMatch( s, "Single" ) )     value = SRDAT_FILENAMETYPE_SINGLE;
        else if ( ScopeStringMatch( s, "YmdHms" ) )     value = SRDAT_FILENAMETYPE_YMDHMS;
        else                                            value = -1;

        return( value );
}

//------------------------------------------------------------------------------
// ROUTINE: ScopeMakeDirYmdHms
// PURPOSE: Create the YMDHMS data directory.
//------------------------------------------------------------------------------
int ScopeMakeDirYmdHms( void ) {

            time_t UtcTimeInSecondsSince_1970;
        struct tm *LocTime;


        // Get the current time.

                       time( &UtcTimeInSecondsSince_1970 );
        LocTime = localtime( &UtcTimeInSecondsSince_1970 );


        // Prepare the directory name: YYYY-MM-DD-at-HH-MM-SS
        //
        //    YMDHMS = ( year, month, day, hours, minuts, seconds)
        //
        //

        sprintf( ScopeInfo.OutputFileDirectory, "%04d-%02d-%02d-at-%02d-%02d-%02d",

                                        LocTime->tm_year+1900,
                                        LocTime->tm_mon+1,
                                        LocTime->tm_mday,
                                        LocTime->tm_hour,
                                        LocTime->tm_min,
                                        LocTime->tm_sec

                                        );

        // Make the directory.

        if ( SrMkDir( ScopeInfo.OutputFileDirectory ) == -1 /* == failure */ )

                return( 0 );
        else
                return( 1 );
}

//------------------------------------------------------------------------------
// ROUTINE: ScopeInit
// PURPOSE: Initialize Scope.
//------------------------------------------------------------------------------
int ScopeInit( const char *IniFileName ) {

        int Ok;

//FIX Do we need to init these ScopeInfo variables too, or just the IniParse ones?
//
//        Ok =  ScopeInitScopeInfo( int ScreenWidth, int ScreenHeight, char *CurrDir,
//                                  SRMSGFNPTR MessageFunc );


        // This function checks if the specified file exists.  If so, it
        // just returns.  Otherwise, it writes out a default ini file.  In
        // either case, the ScopeInfo.IniFileName variable is set, so that's
        // what we use as input to the ini parsing function.

        Ok = ScopeIniCheck( IniFileName );


        // Parse ini file and output a report

        if ( Ok ) {
                Ok = ScopeIniParse( ScopeInfo.IniFileName );
                ScopeIniReport();
                }

        return( Ok );
}

//------------------------------------------------------------------------------
// ROUTINE: ScopeInitScopeInfo
// PURPOSE: Initialize the global variables collected up together in the
//          ScopeInfo structure.
//------------------------------------------------------------------------------
int ScopeInitScopeInfo( int ScreenWidth, int ScreenHeight, char *CurrDir,
                        SRMSGFNPTR MessageFunc ) {

        double RateMin, RateMax;
        time_t CurrTime;
        long   CurrUsec;


        // Get current time

        GetCurrentSystemTime( &CurrTime, &CurrUsec );


        // Set time parameters

        ScopeInfo.TimeSource          = SRDAT_SOURCE_ACQ;
        ScopeInitTimes( CurrTime, CurrUsec );




        // Set max and min values

        SrUsbXchSpsRateTable( USBXCH_SPS_INDEX_MIN, &RateMin, NULL );
        SrUsbXchSpsRateTable( USBXCH_SPS_INDEX_MAX, &RateMax, NULL );

//FIX - We should select this based on sampling rate
//        ScopeInfo.MaxSample           = 1024; // 512; // max pts per chan per buffer
        ScopeInfo.MaxSample           = 512; // max pts per chan per buffer
// This is only used in doing a preliminary determination of bounds on
// first display sample selection.
        ScopeInfo.MaxSample           = 10000; // max pts per chan per buffer



        ScopeInfo.MinAmpShift         = 0;
        ScopeInfo.MaxAmpShift         = 24;
        ScopeInfo.MaxTriggerThreshold = MAX_TRIGGER_THRESHOLD;
        ScopeInfo.MaxTriggerDigioLine = MAX_TRIGGER_DIGIO_LINE;
        ScopeInfo.MinSPS              = (float)RateMin;
        ScopeInfo.MaxSPS              = (float)RateMax;


        // Set default values

        ScopeInfo.DefaultScreenWidth  = ScreenWidth;
        ScopeInfo.DefaultScreenHeight = ScreenHeight;
        ScopeInfo.DefaultColor        = CS_BLUE;
        ScopeInfo.DefaultThreshold    = ScopeInfo.MaxTriggerThreshold / 2;
        ScopeInfo.DefaultNdisplay     = USBXCH_MAX_ANAPLOT_CHANNELS;
        ScopeInfo.DefaultNsamples     = ScopeInfo.DefaultScreenWidth / 5;
        if (ScopeInfo.DefaultNsamples > ScopeInfo.MaxSample)
                ScopeInfo.DefaultNsamples = ScopeInfo.MaxSample / 2;

        ScopeInfo.DefaultNbuffers     = (ScopeInfo.DefaultScreenWidth - HTITLEMARGIN) /
                                         ScopeInfo.DefaultNsamples;
        if (ScopeInfo.DefaultNbuffers <= 0)
                ScopeInfo.DefaultNbuffers = 1;

        if (CurrDir && strlen( CurrDir ) != 0)
                SrStringCopy( ScopeInfo.DefaultDir, CurrDir, MAXSTR );
        else
                SrStringCopy( ScopeInfo.DefaultDir, ".", MAXSTR );


        // Set other parameters

        ScopeInfo.ChBufPts            = ScopeInfo.DefaultNsamples;
        ScopeInfo.Npackets            = ScopeInfo.ChBufPts/2 + 8; //Fix use good default here
        ScopeInfo.LastMsg             = -1;
        ScopeInfo.DataPtBytes         = sizeof(long);

        ScopeInfo.LedValueYellow      = 0;
        ScopeInfo.LedValueRed         = 0;

        ScopeInfo.NumGpsSat           = -1;
        ScopeInfo.NmeaMsgByte         = 0;

        ScopeInfo.UsbXchHandle        = BAD_DEVHANDLE;


        // Set current run parameters

        ScopeCurrentBufferInit();


        // Next and total point count parameters

        ScopePtCountersInitNext();
        ScopePtCountersInitTotal();



        // Set data buffer parameters

        ScopeInfo.UsbXchAnalogStruct    = NULL;
        ScopeInfo.UsbXchAnalogBuffer    = NULL;
        ScopeInfo.UsbXchAnalogBufferEnd = NULL;
        ScopeInfo.UsbXchSerialStruct    = NULL;
        ScopeInfo.UsbXchSerialBuffer    = NULL;
        ScopeInfo.UsbXchSerialBufferEnd = NULL;
        ScopeInfo.UsbXchEquipStruct     = NULL;


        // Set log message defaults

        SrStringCopy( ScopeInfo.LogMessage, "Scope Message\n", MAXSTR );
        if ( MessageFunc == NULL )
                ScopeInfo.ShowLogMessage = &DefaultShowLogMessage;
        else
                ScopeInfo.ShowLogMessage = MessageFunc;


        // Set additional defaults

        ScopeIniSetDefaults( );

        return( 1 );

}

//------------------------------------------------------------------------------
// ROUTINE: ScopeIniSetDefaults
// PURPOSE: Initialize to default values all the global ScopeInfo variables
//          that can be specified using the ini file.
//------------------------------------------------------------------------------
void ScopeIniSetDefaults( void ) {

        // When changing defaults, be sure to change in IniVerifyData too

        // Setup default INI parameters.

        // Program setting parameter defaults

        SrStringCopy( ScopeInfo.EditorName, SCOPE_DEFAULT_EDITOR, MAXSTR );

        ScopeInfo.ProgramAllowMultiple           = INISWITCH_OFF;
        ScopeInfo.ProgramKeypressToRun           = INISWITCH_ON;
        ScopeInfo.ProgramInteractive             = INISWITCH_ON;
        ScopeInfo.ProgramExitOnError             = INISWITCH_ON;
        ScopeInfo.ProgramShowStatusBar           = INISWITCH_ON;
        ScopeInfo.ProgramDisplayText             = INISWITCH_ON;
        ScopeInfo.ProgramDisplayInfo             = INISWITCH_ON;

        ScopeInfo.DisplaySerial       = INISWITCH_OFF;
        ScopeInfo.DisplayEquip        = INISWITCH_OFF;

        ScopeInfo.IniTitle[0] = '\0';


        // A/D parameter defaults

        ScopeInfo.AtodXtalSpeed = USBXCH_XTAL_SPEED;
        SrStringCopy( ScopeInfo.AtodDriverName, USBXCH_DEFAULT_DRIVER_NAME, MAXSTR );
//FIX   ScopeInfo.AtodXchModel = SrUsbXchGetDefaultModel( ScopeInfo.AtodDriverName );
        ScopeInfo.AtodXchModel  = SRDAT_ATODMODEL_USB4CH;
        ScopeSetChannelCountByModel(); // Assumes AtodXchModel is already set and
                                       // sets AtodXchModelName, UsbXchXxxChannels
        ScopeComputeActiveChannels();  // Set Channels, ChannelsPlot and  XxxIndex
        ScopeInitActiveChannels();     // Set ChannelType, ChannelTrigger,
                                       //     ChannelTitle and DigitalTitle


        // Gps parameter defaults

//FIX   ScopeInfo.GpsModel = SrUsbXchGetDefaultGpsModel( ScopeInfo.AtodDriverName );
        ScopeInfo.GpsModel               = SRDAT_GPSMODEL_PCTIME;
        ScopeUpdateGpsModel( 0 );        // Sets GpsModelName


        // Acquisition parameter defaults

        ScopeInfo.AcqToggleLed          = INISWITCH_ON;
        ScopeInfo.AcqSpsIndex           = USBXCH_SPS_INDEX_DEFAULT;
        ScopeSpsUpdateFromIndex(); // sets AcqSamplingRate, AcqSamplingRateTrue


        // Output file parameter defaults

        ScopeInfo.OutputFileNaming        = SRDAT_FILENAMETYPE_YMDHMS;
        ScopeInfo.OutputFileNbuffers      = ScopeInfo.DefaultNbuffers;
        sprintf( ScopeInfo.OutputFileDirectory, "." );


        // Plot display parameter defaults

        ScopeInfo.PlotNdisplay    = ScopeInfo.DefaultNdisplay; // reset below
        ScopeInfo.PlotFirstSample = 0;
        ScopeInfo.PlotFirstTrace  = 0;
        ScopeInfo.PlotColorScheme = ScopeInfo.DefaultColor;
        ScopeInfo.PlotShiftBy     = ScopeInfo.MaxAmpShift/2;
        ScopeInfo.PlotClipping    = INISWITCH_OFF;
        ScopeInfo.PlotAutoDC      = INISWITCH_ON;


        // Trigger parameter defaults

        ScopeInfo.TriggerType             = SRDAT_TRIGGERTYPE_NONE;
        ScopeInfo.TriggerThreshold        = ScopeInfo.DefaultThreshold;
        ScopeInfo.TriggerDigitalLevelLine = 0;
        ScopeInfo.TriggerDigitalEdgeLine  = 0;
        ScopeInfo.TriggerActiveLevel      = SRDAT_TRIGGERLEVEL_HIGH;
        ScopeInfo.TriggerActiveEdge       = SRDAT_TRIGGEREDGE_RISING;
        ScopeInfo.TriggerActionSave       = INISWITCH_ON;
        ScopeInfo.TriggerFilesToSave      = 2;
        ScopeInfo.TriggerActionSignal     = INISWITCH_OFF;
        ScopeInfo.TriggerLineToSignal     = 0;
        ScopeInfo.TriggerLatchLine        = INISWITCH_OFF;


        // XCR UserCfg configuration parameter defaults

        ScopeInfo.UserCfgNmeaTxPolarity  = USERCFG_NMEA_TX_IDLE_HIGH;
        ScopeInfo.UserCfgNmeaRxPolarity  = USERCFG_NMEA_RX_IDLE_HIGH;
        ScopeInfo.UserCfgPpsLed          = USERCFG_PPS_LED_TOGGLE;
        ScopeInfo.UserCfgPpsPolarity     = USERCFG_PPS_IDLE_LOW;
        ScopeInfo.UserCfgAdrdy           = USERCFG_DB25_ADRDY_DISABLED;
        ScopeInfo.UserCfgNotUsed1        = USERCFG_UNUSED;
        ScopeInfo.UserCfgTriggerPolarity = USERCFG_TRIGGER_IDLE_LOW;



        // Verify all is ok so far

        ScopeIniConsistencyCheck();



        // Set computed parameters
        // These include Channels, XxxIndex, ChannelType, XxxFileName, etc

        ScopeIniCompute();


        return;
}

//------------------------------------------------------------------------------
// ROUTINE: ScopePtCountersInitNext
// PURPOSE: Initialize some analog and serial data counters.
//------------------------------------------------------------------------------
void ScopePtCountersInitNext( void ) {

        ScopeInfo.NextPtAnalog  = 0L; // This is index of all data from one
                                      // time with one analog pt per channel
                                      // plus digital data, obc, etc
        ScopeInfo.NextPtSerial  = 0;
        ScopeInfo.NextPtEquip   = 0;
}

//------------------------------------------------------------------------------
// ROUTINE: ScopePtCountersInitTotal
// PURPOSE: Initialize more analog and serial data counters.
//------------------------------------------------------------------------------
void ScopePtCountersInitTotal( void ) {

        ScopeInfo.TotalSample   = 0;
        ScopeInfo.TotalPtSerial = 0;
        ScopeInfo.TotalPtEquip  = 0;
}

//------------------------------------------------------------------------------
// ROUTINE: ScopeInitTimes
// PURPOSE: Initialize ScopeInfo timing variables.
//------------------------------------------------------------------------------
void ScopeInitTimes( time_t StartTime, long StartUsec ) {

        ScopeInfo.StartTime           = StartTime;
        ScopeInfo.StartTimeUsec       = StartUsec;
        ScopeInfo.AcquireTime         = ScopeInfo.StartTime;
        ScopeInfo.AcquireTimeUsec     = ScopeInfo.StartTimeUsec;
        ScopeInfo.ReadyTime           = ScopeInfo.StartTime;
        ScopeInfo.ElapsedBufferTime   = 0.0L;
}

//------------------------------------------------------------------------------
// ROUTINE: ScopeSetChannelCountByModel
// PURPOSE: Initialize ScopeInfo channel variables so they are appropriate for
//          the selected model (eg USB4CH vs USB8CH).
//------------------------------------------------------------------------------
int ScopeSetChannelCountByModel( void ) {

        // Set USBxCH model specific channel counts

        // NOTE: Assumes ScopeInfo parameter AtodXchModel is already set

        switch( ScopeInfo.AtodXchModel ) {
                case SRDAT_ATODMODEL_USB4CH:
                        ScopeInfo.AtodMinTotalChannels= USB4CH_ANALOG_CHANNELS;
                        ScopeInfo.AtodAnalogChannels  = USB4CH_ANALOG_CHANNELS;
                        ScopeInfo.AtodDigitalChannels = USB4CH_DIGITAL_CHANNELS;
                        ScopeInfo.AtodGpsChannels     = USB4CH_GPS_CHANNELS;
                        ScopeInfo.AtodCounterChannels = USB4CH_COUNTER_CHANNELS;
                        ScopeInfo.AtodAnaPlotChannels = USB4CH_ANAPLOT_CHANNELS;
                        ScopeInfo.AtodDigPlotChannels = USB4CH_DIGPLOT_CHANNELS;
                        ScopeInfo.AtodGpsPlotChannels = USB4CH_GPSPLOT_CHANNELS;
                        ScopeInfo.AtodCntPlotChannels = USB4CH_CNTPLOT_CHANNELS;
                        SrStringCopy( ScopeInfo.AtodXchModelName, USB4CH_MODELNAME, MAXSTR );
                        break;

                case SRDAT_ATODMODEL_USB8CH:
                        ScopeInfo.AtodMinTotalChannels= USB8CH_ANALOG_CHANNELS;
                        ScopeInfo.AtodAnalogChannels  = USB8CH_ANALOG_CHANNELS;
                        ScopeInfo.AtodDigitalChannels = USB8CH_DIGITAL_CHANNELS;
                        ScopeInfo.AtodGpsChannels     = USB8CH_GPS_CHANNELS;
                        ScopeInfo.AtodCounterChannels = USB8CH_COUNTER_CHANNELS;
                        ScopeInfo.AtodAnaPlotChannels = USB8CH_ANAPLOT_CHANNELS;
                        ScopeInfo.AtodDigPlotChannels = USB8CH_DIGPLOT_CHANNELS;
                        ScopeInfo.AtodGpsPlotChannels = USB8CH_GPSPLOT_CHANNELS;
                        ScopeInfo.AtodCntPlotChannels = USB8CH_CNTPLOT_CHANNELS;
                        SrStringCopy( ScopeInfo.AtodXchModelName, USB8CH_MODELNAME, MAXSTR );
                        break;

                default:
                        ScopeInfo.AtodMinTotalChannels= USBXCH_MAX_ANALOG_CHANNELS;
                        ScopeInfo.AtodAnalogChannels  = USBXCH_MAX_ANALOG_CHANNELS;
                        ScopeInfo.AtodDigitalChannels = USBXCH_MAX_DIGITAL_CHANNELS;
                        ScopeInfo.AtodGpsChannels     = USBXCH_MAX_GPS_CHANNELS;
                        ScopeInfo.AtodCounterChannels = USBXCH_MAX_COUNTER_CHANNELS;
                        ScopeInfo.AtodAnaPlotChannels = USBXCH_MAX_ANAPLOT_CHANNELS;
                        ScopeInfo.AtodDigPlotChannels = USBXCH_MAX_DIGPLOT_CHANNELS;
                        ScopeInfo.AtodGpsPlotChannels = USBXCH_MAX_GPSPLOT_CHANNELS;
                        ScopeInfo.AtodCntPlotChannels = USBXCH_MAX_CNTPLOT_CHANNELS;
                        SrStringCopy( ScopeInfo.AtodXchModelName, USBXCH_MODELNAME, MAXSTR );
                }

        ScopeInfo.AtodTotalChannels = ScopeInfo.AtodAnalogChannels  +
                                      ScopeInfo.AtodDigitalChannels +
                                      ScopeInfo.AtodGpsChannels     +
                                      ScopeInfo.AtodCounterChannels;

        ScopeInfo.AtodTotalPlotChannels = ScopeInfo.AtodAnaPlotChannels +
                                          ScopeInfo.AtodDigPlotChannels +
                                          ScopeInfo.AtodGpsPlotChannels +
                                          ScopeInfo.AtodCntPlotChannels;

        return( 1 );
}

//------------------------------------------------------------------------------
// ROUTINE: ScopeUpdateGpsModel
// PURPOSE: Finalize GpsModel selection.  This controls things like the order
//          in which NMEA strings are expected.
//------------------------------------------------------------------------------
// FIX - should this be in SrUsbXch.c library
// FIX - what about default model?

int ScopeUpdateGpsModel( int AllowError ) {

        // Finalize GpsModel selection

        switch( ScopeInfo.GpsModel ) {
                case SRDAT_GPSMODEL_GARMIN:
                        SrStringCopy( ScopeInfo.GpsModelName, SRDAT_GPSNAME_GARMIN, MAXTITLE );
                        break;
                case SRDAT_GPSMODEL_TRIMBLE: // or SRDAT_GPSMODEL_PARGPS
                        SrStringCopy( ScopeInfo.GpsModelName, SRDAT_GPSNAME_TRIMBLE, MAXTITLE );
                        break;
                case SRDAT_GPSMODEL_ONCORE:
                        SrStringCopy( ScopeInfo.GpsModelName, SRDAT_GPSNAME_ONCORE, MAXTITLE );
                        break;
                case SRDAT_GPSMODEL_PCTIME:
                        SrStringCopy( ScopeInfo.GpsModelName, SRDAT_GPSNAME_PCTIME, MAXTITLE );
                        break;
                case SRDAT_GPSMODEL_NONE:
                        SrStringCopy( ScopeInfo.GpsModelName, SRDAT_GPSNAME_NONE, MAXTITLE );
                        break;
                default:
                        ScopeInfo.GpsModel = SRDAT_GPSMODEL_UNKNOWN;
                        SrStringCopy( ScopeInfo.GpsModelName, SRDAT_GPSNAME_UNKNOWN, MAXTITLE );
                        break;
                }

        if ( AllowError == 1                        &&
             (ScopeInfo.GpsModel < SRDAT_GPSMODEL_NONE    ||
              ScopeInfo.GpsModel >= SRDAT_GPSMODEL_MAX) ) {
                DisplayLogMessage( MSGLEVEL_ERR, MSGDLG_SHOW, ERROR_FATAL,
                                   "Invalid GPS model, please set GpsModelName correctly" );
                }


        return( 1 );
}

//------------------------------------------------------------------------------
// ROUTINE: ScopeIniConsistencyCheck
// PURPOSE: Verify there are no conflicting selections from the ini file.
//------------------------------------------------------------------------------
void ScopeIniConsistencyCheck( void ) {

        // Verify there are no conflicting INI selections.

        // Nothing to check right now ...

        return;
}

//------------------------------------------------------------------------------
// ROUTINE: ScopeIniCompute
// PURPOSE: Compute some additional convenient parameters based on the
//          user specified INI parameters.
//------------------------------------------------------------------------------
void ScopeIniCompute( void ) {

        // Convert requested sampling rate into index, code and true rate

        SrUsbXchSpsRateValidate( ScopeInfo.AcqSamplingRate,
                                 &ScopeInfo.AcqSamplingRateTrue,
                                 NULL );



        // Set USBxCH and GPS model dependent values

        ScopeSetChannelCountByModel( ); // Sets AtodXxxChannel and AtodXchModelName
        ScopeUpdateGpsModel( 1 );       // Sets GpsModelName



        // NOTE: Assumes ScopeInfo parameters AtodXxxChannel are already set

        ScopeComputeActiveChannels();  // Set Channels, ChannelsPlot and  XxxIndex

// don't re-init here, it wipes out channel title
//        ScopeInitActiveChannels();     // Set ChannelType, ChannelTrigger,
                                         //     ChannelTitle and DigitalTitle




        // Set title and file names

        // NOTE: Assumes ScopeInfo parameters OutputFileDirectory,
        //       OutputFileFormat and OutputFileNaming are already set

        ScopeInitFileNames( );
        ScopeCreateShortFileName( ScopeInfo.IniFileName, ScopeInfo.IniFileNameShort );
// don't re-init here, it wipes out scope title
//        SrStringCopy( ScopeInfo.IniTitle, ScopeInfo.IniFileNameShort, MAXSTR );


        // FIX - Don't allow any digital trigger channels
        // (amplitude triggering is on analog channels only)


        // This updates parameters that depend on buffer size which depends
        // on sample rate

        ScopeBuffersUpdateSize();
}

//------------------------------------------------------------------------------
// ROUTINE: ScopeIniCheck
// PURPOSE: Check the ini file name given on the command line.
//          If no ini file was given, create a default ini, and return 2.
//          Else if the specified ini file exists, save its name and return 1.
//          Else it doesn't exist, so print an error message and exit.
//------------------------------------------------------------------------------
int ScopeIniCheck( const char *IniFilename ) {

        int iret, Ok;

        // If no filename given, create a default ini

        if ( IniFilename == NULL || strlen( IniFilename ) == 0 ) {
                SrStringCopy( ScopeInfo.IniFileName, SCOPE_DEFAULT_INI, MAXSTR );
                sprintf( msg, "%s\n\n%s\n%s\n",
                         "No ini file       ",
                         "  ... saving settings to default ini file",
                         "  ... running with defaults"
                       );
                DisplayLogMessage( MSGLEVEL_WARN, MSGDLG_SHOW, ERROR_NOT_FATAL, msg );
                ScopeIniCreateDefault( ScopeInfo.IniFileName );
                iret = 2;
                }


        // Else if the ini file does exist, save its name

        else if ( SrFileExist( IniFilename ) ) {
                SrStringCopy( ScopeInfo.IniFileName, IniFilename, MAXSTR );
                iret = 1;
                }


        // Otherwise print an message and exit

        else {
                sprintf( msg,
                         "ini file %s does not exist     \n\n   Try again or run with no ini file",
                         IniFilename );
                DisplayLogMessage( MSGLEVEL_ERR, MSGDLG_SHOW, ERROR_FATAL, msg );
                iret = 0;
                }


        // Create short version (with no path info) too

        Ok = ScopeCreateShortFileName( ScopeInfo.IniFileName, ScopeInfo.IniFileNameShort );


        return( iret );
}


//------------------------------------------------------------------------------
// ROUTINE: ScopeIniCreateDefault
// PURPOSE: Create a default ini file containing the current settings.  This
//          will overwrite the default ini if it already exists.
//------------------------------------------------------------------------------
int ScopeIniCreateDefault( const char *IniFilename ) {

        FILE *IniFp;


        // Set up the default values

        ScopeIniSetDefaults( );


        // Open default .ini file for writing

        IniFp = fopen( IniFilename, "w" );

        if ( !IniFp ) {
                sprintf( msg,
                         "Could not open ini file %s to write settings",
                         IniFilename );
                DisplayLogMessage( MSGLEVEL_ERR, MSGDLG_SHOW, ERROR_FATAL, msg );
                return( 0 );
                }


        // Output overwrite warning notice

        fprintf( IniFp, "; FILE: %s\n;\n", IniFilename );
        fprintf( IniFp, "; *** This file is generated automatically and may be overwritten.\n" );
        fprintf( IniFp, "; *** Please copy to a new name before editing so edits won't be lost.\n\n" );


        // Output ini parameters

        ScopeIniWriteKeyword( IniFp );


        // Close file

        fclose( IniFp );

        return( 1 );
}

//------------------------------------------------------------------------------
// ROUTINE: ScopeIniWriteKeyword
// PURPOSE: Write a valid file using the current program settings.
//------------------------------------------------------------------------------
int ScopeIniWriteKeyword( FILE *OutFp ) {

        int   i;
        char *OutputFileString, *DisplayTextString;
        char *AutoDCString, *ClippingString, *StatusBarString, *ColorString;
        char *AutoRunString, *TriggerTypeString, *TriggerString;
        char *TriggerActionSignalString, *TriggerActionSaveString;
        char *DisplaySerialString, *DisplayEquipString, *ExitOnErrorString;
        char *TriggerLatchLineString, *TriggerLevelString, *TriggerEdgeString;
        char *GpsNameString, *AtodNameString, *ToggleLedString;
        char *NmeaTxPolarityString, *NmeaRxPolarityString;
        char *PpsLedString, *PpsPolarityString, *MultipleString;
        char *AdrdyString, *TriggerPolarityString;


        // Set up output strings

        ToggleLedString           = (  ScopeInfo.AcqToggleLed          ? "On" : "Off" );
        MultipleString            = (  ScopeInfo.ProgramAllowMultiple  ? "On" : "Off" );
        AutoRunString             = (  ScopeInfo.ProgramKeypressToRun  ? "On" : "Off" );
        ExitOnErrorString         = (  ScopeInfo.ProgramExitOnError    ? "On" : "Off" );
        StatusBarString           = (  ScopeInfo.ProgramShowStatusBar  ? "On" : "Off" );
        DisplayTextString         = (  ScopeInfo.ProgramDisplayText    ? "On" : "Off" );
        DisplaySerialString       = (  ScopeInfo.DisplaySerial         ? "On" : "Off" );
        DisplayEquipString        = (  ScopeInfo.DisplayEquip          ? "On" : "Off" );
        ClippingString            = (  ScopeInfo.PlotClipping          ? "On" : "Off" );
        AutoDCString              = (  ScopeInfo.PlotAutoDC            ? "On" : "Off" );
        TriggerActionSignalString = (  ScopeInfo.TriggerActionSignal   ? "On" : "Off" );
        TriggerActionSaveString   = (  ScopeInfo.TriggerActionSave     ? "On" : "Off" );
        TriggerLatchLineString    = (  ScopeInfo.TriggerLatchLine      ? "On" : "Off" );
        AdrdyString               = (  ScopeInfo.UserCfgAdrdy          ? "On" : "Off" );

        switch ( ScopeInfo.AtodXchModel ) {
                case SRDAT_ATODMODEL_USB1CH: AtodNameString = "USB1CH";   break;
                case SRDAT_ATODMODEL_USB4CH: AtodNameString = "USB4CH";   break;
                case SRDAT_ATODMODEL_USB8CH: AtodNameString = "USB8CH";   break;
                default:                     AtodNameString = "UNKNOWN";  break;
                }

        TriggerTypeString   = SrDatTriggerTypeString( ScopeInfo.TriggerType );
        TriggerLevelString  = SrDatTriggerLevelString( ScopeInfo.TriggerActiveLevel );
        TriggerEdgeString   = SrDatTriggerEdgeString( ScopeInfo.TriggerActiveEdge );

        OutputFileString    = SrDatFileNameTypeString( ScopeInfo.OutputFileNaming );

        switch ( ScopeInfo.PlotColorScheme ) {
                case CS_MONO:   ColorString = "Monochrome"; break;
                case CS_GREEN:  ColorString = "Green";      break;
                case CS_BLUE:   ColorString = "Blue";       break;
                case CS_BROWN:  ColorString = "Brown";      break;
                default:        ColorString = "Unknown";    break;
                }

        switch ( ScopeInfo.GpsModel ) {
                case SRDAT_GPSMODEL_PCTIME:  GpsNameString = "PcTime";   break;
                case SRDAT_GPSMODEL_GARMIN:  GpsNameString = "Garmin";   break;
                case SRDAT_GPSMODEL_TRIMBLE: GpsNameString = "Trimble";  break;
                case SRDAT_GPSMODEL_ONCORE:  GpsNameString = "Oncore";   break;
                case SRDAT_GPSMODEL_NONE:    GpsNameString = "None";     break;
                default:                     GpsNameString = "Unknown";  break;
                }

        switch ( ScopeInfo.UserCfgNmeaTxPolarity ) {
                case USERCFG_NMEA_TX_IDLE_HIGH: NmeaTxPolarityString = "IdleHigh";  break;
                case USERCFG_NMEA_TX_IDLE_LOW:  NmeaTxPolarityString = "IdleLow";   break;
                default:                        NmeaTxPolarityString = "Unknown";   break;
                }
        switch ( ScopeInfo.UserCfgNmeaRxPolarity ) {
                case USERCFG_NMEA_RX_IDLE_HIGH: NmeaRxPolarityString = "IdleHigh";  break;
                case USERCFG_NMEA_RX_IDLE_LOW:  NmeaRxPolarityString = "IdleLow";   break;
                default:                        NmeaRxPolarityString = "Unknown";   break;
                }
        switch ( ScopeInfo.UserCfgPpsLed ) {
                case USERCFG_PPS_LED_TOGGLE:     PpsLedString = "Toggle";      break;
                case USERCFG_PPS_LED_PROGRAMMED: PpsLedString = "Programmed";  break;
                default:                         PpsLedString = "Unknown";     break;
                }
        switch ( ScopeInfo.UserCfgPpsPolarity ) {
                case USERCFG_PPS_IDLE_LOW:  PpsPolarityString = "IdleLow";   break;
                case USERCFG_PPS_IDLE_HIGH: PpsPolarityString = "IdleHigh";  break;
                default:                    PpsPolarityString = "Unknown";   break;
                }
        switch ( ScopeInfo.UserCfgTriggerPolarity ) {
                case USERCFG_TRIGGER_IDLE_LOW:  TriggerPolarityString = "IdleLow";   break;
                case USERCFG_TRIGGER_IDLE_HIGH: TriggerPolarityString = "IdleHigh";  break;
                default:                        TriggerPolarityString = "Unknown";   break;
                }

        //********************************************************************************
        // Print out ini keywords in functional groups

        fprintf( OutFp, "                  Title = \"%s\"\n", ScopeInfo.IniTitle );
        fprintf( OutFp, "\n" );


        // Print driver and model keywords

        fprintf( OutFp, "         AtodDriverName = %s\n", ScopeInfo.AtodDriverName );
        fprintf( OutFp, "          AtodModelName = %s\n", AtodNameString );
        fprintf( OutFp, "           GpsModelName = %s\n", GpsNameString );
        fprintf( OutFp, "\n" );


        // Print acquisition keywords

        fprintf( OutFp, "           SamplingRate = %lf\n", ScopeInfo.AcqSamplingRateTrue );
        fprintf( OutFp, "              ToggleLed = %s\n",  ToggleLedString );
        fprintf( OutFp, "\n" );


        // Print channel keywords

        for ( i = 0 ; i < ScopeInfo.AtodTotalChannels ; i++ ) {
                if (ScopeInfo.ChannelTrigger[i] == INISWITCH_ON)
                        TriggerString = "On";
                else
                        TriggerString = "Off";

                fprintf(
                        OutFp,
                        "        ChannelTitle %2d = \"%s\"   Trigger %2d = %s\n",
                        i,
                        ScopeInfo.ChannelTitle[i],
                        i,
                        TriggerString
                      );
                }
        fprintf( OutFp, "\n" );


        // Print digital keywords

        for ( i = 0 ; i < ScopeInfo.AtodDigPlotChannels ; i++ ) {
                fprintf(
                        OutFp,
                        "        DigitalTitle %2d = \"%s\"\n",
                        i,
                        ScopeInfo.DigitalTitle[i]
                      );
                } // end for i

        fprintf( OutFp, "\n" );


        // Print output file keywords

        fprintf( OutFp, "       OutputFileNaming = %s\n", OutputFileString );
        fprintf( OutFp, "     OutputFileNbuffers = %d\n", ScopeInfo.OutputFileNbuffers );
        fprintf( OutFp, "\n" );


        // Print display keywords

        fprintf( OutFp, "            DisplayInfo = %s\n", StatusBarString );
        fprintf( OutFp, "        DisplayDataText = %s\n", DisplayTextString );
        fprintf( OutFp, "           DisplayColor = %s\n", ColorString );
        fprintf( OutFp, "           DisplayNbits = %d\n", ScopeInfo.MaxAmpShift - ScopeInfo.PlotShiftBy );
        fprintf( OutFp, "          DisplayAutoDC = %s\n", AutoDCString );
        fprintf( OutFp, "        DisplayClipping = %s\n", ClippingString );
        fprintf( OutFp, "     DisplayFirstSample = %d\n", ScopeInfo.PlotFirstSample );
        fprintf( OutFp, "      DisplayFirstTrace = %d\n", ScopeInfo.PlotFirstTrace );
        fprintf( OutFp, "         DisplayNtraces = %d\n", ScopeInfo.PlotNdisplay );
        fprintf( OutFp, "          DisplaySerial = Off\n" );
        fprintf( OutFp, "           DisplayEquip = Off\n" );
        fprintf( OutFp, "\n" );


        // Print trigger keywords

        fprintf( OutFp, "            TriggerType = %s\n",  TriggerTypeString );
        fprintf( OutFp, "       TriggerThreshold = %ld\n", ScopeInfo.TriggerThreshold );
        if (MAX_TRIGGER_DIGIO_LINE >= 0) {
        fprintf( OutFp, "TriggerDigitalLevelLine = %d\n",  ScopeInfo.TriggerDigitalLevelLine );
        fprintf( OutFp, "     TriggerActiveLevel = %s\n",  TriggerLevelString );
        fprintf( OutFp, " TriggerDigitalEdgeLine = %d\n",  ScopeInfo.TriggerDigitalEdgeLine );
        fprintf( OutFp, "      TriggerActiveEdge = %s\n",  TriggerEdgeString );
        }
        fprintf( OutFp, "      TriggerActionSave = %s\n",  TriggerActionSaveString );
        fprintf( OutFp, "     TriggerFilesToSave = %d\n",  ScopeInfo.TriggerFilesToSave );
        if (MAX_TRIGGER_DIGIO_LINE >= 0) {
        fprintf( OutFp, "    TriggerActionSignal = %s\n",  TriggerActionSignalString );
        fprintf( OutFp, "    TriggerLineToSignal = %d\n",  ScopeInfo.TriggerLineToSignal );
        fprintf( OutFp, "           TriggerLatch = %s\n",  TriggerLatchLineString );
        }
        fprintf( OutFp, "\n" );


        // Print XCR UserCfg configuration keywords

        fprintf( OutFp, "  UserCfgNmeaTxPolarity = %s\n", NmeaTxPolarityString );
        fprintf( OutFp, "  UserCfgNmeaRxPolarity = %s\n", NmeaRxPolarityString );
        fprintf( OutFp, "          UserCfgPpsLed = %s\n", PpsLedString );
        fprintf( OutFp, "     UserCfgPpsPolarity = %s\n", PpsPolarityString );
        fprintf( OutFp, "           UserCfgAdrdy = %s\n", AdrdyString );
        fprintf( OutFp, " UserCfgTriggerPolarity = %s\n", TriggerPolarityString );
        fprintf( OutFp, "\n" );



        // Print program control keywords

        fprintf( OutFp, "            ExitOnError = %s\n",     ExitOnErrorString );
        fprintf( OutFp, "          AllowMultiple = %s\n",     MultipleString );
        fprintf( OutFp, "          KeypressToRun = %s\n",     AutoRunString );
        fprintf( OutFp, "             EditorName = \"%s\"\n", ScopeInfo.EditorName );
        fprintf( OutFp, "\n" );



        return( 1 );
}

//------------------------------------------------------------------------------
// ROUTINE: ScopeIniParse
// PURPOSE: Parse specified ini file for keywords and update program settings
//          accordingly.
//------------------------------------------------------------------------------
int ScopeIniParse( const char *IniFilename ) {

        int Success, TokenType, itmp;
        int Channel, Digital, MaxTotalChannels;
        double SpsMin, SpsMax;
        static int LastChannel = -1;
        static int LastDigital = -1;
        static int IsChannel   = -1;

        // Set up defaults to be used unless overridden by values from file

        ScopeIniSetDefaults();


        // Allow running with defaults if no ini was specified

        if (strlen(IniFilename) == 0)
                return( 1 );


        // Open the .ini file.

        Success = IniOpen( (char *)IniFilename );

        if ( !Success ) {
                printf( "unable to open .ini file" );
                return( 0 );
                }

        else {
                SrStringCopy( ScopeInfo.IniFileName, IniFilename, MAXSTR );
                }


        MaxTotalChannels = USBXCH_MAX_ANALOG_CHANNELS  +
                           USBXCH_MAX_DIGITAL_CHANNELS +
                           USBXCH_MAX_GPS_CHANNELS     +
                           USBXCH_MAX_COUNTER_CHANNELS;



        // Parse the ini file.

parse:
        TokenType = GetToken();

        if ( TokenType == ERROR )
                return( 0 );


        // Check program setting parameters
        // IniTitle
        // ProgramAllowMultiple, ProgramKeypressToRun, ProgramExitOnError

        if ( KeywordEquals( "Title" ) ) {
                TokenType = GetToken();
                if ( TokenType != STRING ) {
                        ErrorIni( "illegal Title string");
                        return( 0 );
                        }
                else
                        SrStringCopy( ScopeInfo.IniTitle, GetTokenPtr(), MAXSTR );
                }



        else if ( KeywordEquals( "AllowMultiple" ) ) {

                Success = GetOnOff( &ScopeInfo.ProgramAllowMultiple, "AllowMultiple must = {ON,OFF}" );
                if ( !Success )
                        return( 0 );
                }


        else if ( KeywordEquals( "KeypressToRun" ) ) {

                Success = GetOnOff( &ScopeInfo.ProgramKeypressToRun, "KeypressToRun must = {ON,OFF}" );
                if ( !Success )
                        return( 0 );
                }


        else if ( KeywordEquals( "ExitOnError" ) ) {

                Success = GetOnOff( &ScopeInfo.ProgramExitOnError, "ExitOnError must = {ON,OFF}" );
                if ( !Success )
                        return( 0 );
                }


        else if ( KeywordEquals( "EditorName" ) ) {

                TokenType = GetToken( );
                if ( TokenType != STRING  &&  TokenType != SYMBOL ) {
                        ErrorIni( "bad ini editor program name string" );
                        return( 0 );
                        }

                strncpy( ScopeInfo.EditorName, GetTokenPtr( ), MAXSTR );
                }





        // Atod keywords
        // AtodDriverName, AtodXchModel
        // AtodXtalSpeed currently hardwired and not selectable

        else if ( KeywordEquals( "AtodDriverName" ) ) {

                TokenType = GetToken();
                if ( TokenType != STRING  &&  TokenType != SYMBOL ) {
                        ErrorIni( "bad AtoD Driver Name string" );
                        return( 0 );
                        }

                SrStringCopy( ScopeInfo.AtodDriverName, GetTokenPtr(), MAXSTR );
                }



        else if ( KeywordEquals( "AtodModelName" ) ) {

                TokenType = GetToken();
                if ( TokenType != STRING  &&  TokenType != SYMBOL ) {
                        ErrorIni( "bad AtodModelName string" );
                        return( 0 );
                        }

                if ( Match( SRDAT_ATODNAME_USB1CH ) )
                        ScopeInfo.AtodXchModel = SRDAT_ATODMODEL_USB1CH;

                else if ( Match( SRDAT_ATODNAME_USB4CH ) )
                        ScopeInfo.AtodXchModel = SRDAT_ATODMODEL_USB4CH;

                else if ( Match( SRDAT_ATODNAME_USB8CH ) )
                        ScopeInfo.AtodXchModel = SRDAT_ATODMODEL_USB8CH;

                else if ( Match( "DEFAULT" ) )
                        ScopeInfo.AtodXchModel = SRDAT_ATODMODEL_UNKNOWN;

                else {
                        ErrorIni( "AtodModelName must = {USB1CH,USB4CH,USB8CH,DEFAULT}" );
                        return( 0 );
                        }

                }




        // Gps keywords

        else if ( KeywordEquals( "GpsModelName" ) ) {

                if ( GetToken() == ERROR )
                        return( 0 );

//FIX - should this be a function in srusbxch.c or srdat.c library?

                if ( Match( "PCTIME" ) )
                        ScopeInfo.GpsModel = SRDAT_GPSMODEL_PCTIME;

                else if ( Match( "GARMIN" ) )
                        ScopeInfo.GpsModel = SRDAT_GPSMODEL_GARMIN;

                else if ( Match( "TRIMBLE" ) )
                        ScopeInfo.GpsModel = SRDAT_GPSMODEL_TRIMBLE;

                else if ( Match( "ONCORE" ) )
                        ScopeInfo.GpsModel = SRDAT_GPSMODEL_ONCORE;

                else if ( Match( "NONE" ) )
                        ScopeInfo.GpsModel = SRDAT_GPSMODEL_NONE;

                else {
                        ErrorIni( "GpsModelName must = {PCTIME,GARMIN,TRIMBLE,ONCORE,NONE}" );
                        return( 0 );
                        }

                }




        // Initialize acquisition constants:
        // AcqGain, AcqSPS

        else if ( KeywordEquals( "SamplingRate" ) ) {

                Success = GetFloat( &ScopeInfo.AcqSamplingRate,
                                    "SamplingRate must = float" );
                if ( !Success )
                        return( 0 );

                SrUsbXchSpsRateTable( USBXCH_SPS_INDEX_MIN, &SpsMin, NULL );
                SrUsbXchSpsRateTable( USBXCH_SPS_INDEX_MAX, &SpsMax, NULL );

                if ( ScopeInfo.AcqSamplingRate < SpsMin ||
                     SpsMax < ScopeInfo.AcqSamplingRate ) {

                        sprintf( msg,
                                 "SamplingRate must be %lf to %lf Hz",
                                 SpsMin, SpsMax );
                        ErrorIni( msg );
                        return( 0 );
                        }

//FIX                        SrUsbXchSpsRequest( dtmp, &ScopeInfo.AcqSpsIndex, NULL, NULL, NULL );
//FIX                        ScopeSpsUpdateFromIndex( );
                }




         else if ( KeywordEquals( "ToggleLed" ) ) {

                Success = GetOnOff( &ScopeInfo.AcqToggleLed, "ToggleLed must = {ON,OFF}" );
                if ( !Success )
                        return( 0 );
                }




        // Channel keywords

        else if ( Match( "Channel") ) {
                ErrorIni( "Channel keyword is obsolete.  Please use ChannelTitle instead.");
                return( 0 );
                }

        else if ( Match( "ChannelTitle") ) {

                Success = GetInt( &Channel, "illegal channel number");
                if ( !Success )
                        return( 0 );

                if ( GetToken() != EQUALS ) {
                        ErrorIni( "missing = sign on ChannelTitle");
                        return( 0 );
                        }

                if ( GetToken() != STRING ) {
                        ErrorIni( "illegal channel title");
                        return( 0 );
                        }


                // Make sure the channel is in the range and in order.

                if ( Channel < 0 || MaxTotalChannels <= Channel ) {
                        ErrorIni( "channel number out of range");
                        return( 0 );
                        }

                if ( Channel <= LastChannel ) {
                        sprintf( msg, "Channels MUST be listed in increasing order.\nChannel %d is out of order.",
                                 Channel );
                        ErrorIni( msg );
                        return( 0 );
                        }


                // Assign title to channel and null terminate.

                SrStringCopy( ScopeInfo.ChannelTitle[Channel], GetTokenPtr(), MAXTITLE );
                ScopeInfo.ChannelTitle[Channel][MAXTITLE-1] = '\0';


                LastChannel = Channel;
                IsChannel   = 1;
                }



        // Digital keywords

        else if ( Match( "Digital") ) {
                ErrorIni( "Digital keyword is obsolete.  Please use DigitalTitle instead.");
                return( 0 );
                }

        else if ( Match( "DigitalTitle") ) {

                Success = GetInt( &Digital, "illegal digital number");
                if ( !Success )
                        return( 0 );

                if ( GetToken() != EQUALS ) {
                        ErrorIni( "missing = sign on DigitalTitle");
                        return( 0 );
                        }

                if ( GetToken() != STRING ) {
                        ErrorIni( "illegal digital title");
                        return( 0 );
                        }


                // Make sure the Digital is in the range.

                if ( Digital < 0 || USBXCH_MAX_DIGPLOT_CHANNELS <= Digital ) {
                        sprintf( msg,
                                 "Digital channel number must be 0 to %d",
                                 USBXCH_MAX_DIGPLOT_CHANNELS-1
                                );
                        ErrorIni( "Digital number out of range");
                        return( 0 );
                        }

                if ( Digital <= LastDigital ) {
                        sprintf( msg, "Digital channels MUST be listed in increasing order.\nDigital %d is out of order.",
                                 Digital );
                        ErrorIni( msg );
                        return( 0 );
                        }

                // Assign title to Digital and null terminate.

                SrStringCopy( ScopeInfo.DigitalTitle[Digital], GetTokenPtr(), MAXTITLE );
                ScopeInfo.DigitalTitle[Digital][MAXTITLE-1] = '\0';

                LastDigital = Digital;
                IsChannel   = 0;
                }










        // Output keywords
        // Initialize output constants
        // OutputFileNaming, OutputFileNbuffers

        else if ( KeywordEquals( "OutputFileNaming" ) ) {

                if ( GetToken() == ERROR )
                        return( 0 );


                if ( Match( "None" ) )
                        ScopeInfo.OutputFileNaming = SRDAT_FILENAMETYPE_NONE;

                else if ( Match( "Single" ) )
                        ScopeInfo.OutputFileNaming = SRDAT_FILENAMETYPE_SINGLE;

                else if ( Match( "YmdHms" ) )
                        ScopeInfo.OutputFileNaming = SRDAT_FILENAMETYPE_YMDHMS;

                else {
                        ErrorIni( "OutputFileNaming must = {None,Single,YmdHms}" );
                        return( 0 );
                        }
                }


        else if ( KeywordEquals( "OutputFileNbuffers" ) ) {

                Success = GetUlong( &ScopeInfo.OutputFileNbuffers,
                          "illegal OutputFileNbuffers integer value" );
                if ( !Success )
                        return( 0 );
                }




        // Display keywords

        // Initialize display constants (uses MaxAmpShift,DefaultNdisplay,DefaultColor)
        // ProgramShowStatusBar, PltClipping, PltAutoDC
        // PltShiftBy, PltNdisplay, PltFirstSample,PltFirstTrace,PltColorScheme

        else if ( KeywordEquals( "DisplayDataText" ) ) {

                Success = GetOnOff( &ScopeInfo.ProgramDisplayText,
                                    "DisplayDataText must = {ON,OFF}" );
                if ( !Success )
                        return( 0 );
                }


        else if ( KeywordEquals( "DisplayInfo" ) ) {

                Success = GetOnOff( &ScopeInfo.ProgramDisplayInfo,
                                    "DisplayInfo must = {ON,OFF}" );
                if ( !Success )
                        return( 0 );
                }



        else if ( KeywordEquals( "DisplaySerial" ) ) {

                Success = GetOnOff( &ScopeInfo.DisplaySerial,
                          "DisplaySerial must = {On,Off}" );
                if ( !Success )
                        return( 0 );
                }

        else if ( KeywordEquals( "DisplayEquip" ) ) {

                Success = GetOnOff( &ScopeInfo.DisplayEquip,
                          "DisplayEquip must = {On,Off}" );
                if ( !Success )
                        return( 0 );
                }




        else if ( KeywordEquals( "DisplayColor" ) ) {

                TokenType = GetToken();
                if ( TokenType != STRING  &&  TokenType != SYMBOL ) {
                        ErrorIni( "bad DisplayColor string" );
                        return( 0 );
                        }

                if ( Match( "Monochrome" ) )
                        ScopeInfo.PlotColorScheme = CS_MONO;

                else if ( Match( "Green" ) )
                        ScopeInfo.PlotColorScheme = CS_GREEN;

                else if ( Match( "Blue" ) )
                        ScopeInfo.PlotColorScheme = CS_BLUE;

                else if ( Match( "Brown" ) )
                        ScopeInfo.PlotColorScheme = CS_BROWN;
                else
                    ErrorIni(
                        "DisplayColor must = {Monochrome,Green,Blue,Brown}" );
                }


        else if ( KeywordEquals( "DisplayClipping" ) ) {

                Success = GetOnOff( &ScopeInfo.PlotClipping,
                                    "DisplayClipping must = {ON,OFF}" );
                if ( !Success )
                        return( 0 );
                }


        else if ( KeywordEquals( "DisplayNbits" ) ) {

                Success = GetInt( &itmp, "illegal DisplayNbits integer value" );
                if ( !Success )
                        return( 0 );

                if ( itmp >= ScopeInfo.MinAmpShift && itmp <= ScopeInfo.MaxAmpShift )
                        ScopeInfo.PlotShiftBy = ScopeInfo.MaxAmpShift - itmp;
                else {
                        sprintf( msg, "DisplayNbits must be %d to %d",
                                 ScopeInfo.MinAmpShift, ScopeInfo.MaxAmpShift
                               );
                        ErrorIni( msg );
                        return( 0 );
                        }
                }


        else if ( KeywordEquals( "DisplayAutoDC" ) ) {

                Success = GetOnOff( &ScopeInfo.PlotAutoDC,
                                    "DisplayAutoDC must = {ON,OFF}" );
                if (!Success)
                        return( 0 );
                }


        else if ( KeywordEquals( "DisplayFirstSample" ) ) {

                Success = GetInt( &itmp, "illegal DisplayFirstSample integer value" );
                if ( !Success )
                        return( 0 );

                // True max is NsamplesPlot-TraceWidth, but that is not known yet
                if ( itmp >= 0 && itmp <= ScopeInfo.MaxSample )
                        ScopeInfo.PlotFirstSample = itmp;
                else {
                        sprintf( msg,
                                 "DisplayFirstSample must be 0 to %d samples",
                                 ScopeInfo.MaxSample );
                        ErrorIni( msg );
                        return( 0 );
                        }
                }


        else if ( KeywordEquals( "DisplayFirstTrace" ) ) {

                Success = GetInt( &itmp, "illegal DisplayFirstTrace integer value" );
                if ( !Success )
                        return( 0 );

                // True max is Nchannels-Ndisplay, but that is not known yet
                if ( itmp >= 0 && itmp <= ScopeInfo.AtodTotalPlotChannels )
                        ScopeInfo.PlotFirstTrace = itmp;
                else {
                        sprintf( msg,
                                 "DisplayFirstTrace must be 0 to %d - Ndisplay channels",
                                 ScopeInfo.AtodTotalPlotChannels );
                        ErrorIni( msg );
                        return( 0 );
                        }
                }


        else if ( KeywordEquals( "DisplayNtraces" ) ) {

                Success = GetInt( &itmp, "illegal DisplayNtraces integer value" );
                if ( !Success )
                        return( 0 );

                // True max is Nchannels, but that is not known yet
                if ( itmp >= 0 && itmp <= ScopeInfo.AtodTotalPlotChannels )
                        ScopeInfo.PlotNdisplay = itmp;
                else {
                        sprintf( msg,
                                 "DisplayNtraces must be 0 to %d channels",
                                 ScopeInfo.AtodTotalPlotChannels );
                        ErrorIni( msg );
                        return( 0 );
                        }
                }


        // Trigger keywords

        // Initialize trigger constants (uses ??,MaxTriggerThreshold,DefaultThreshold)
        // TriggerType, TriggerThreshold, TriggerDigitalLevelLine, TriggerDigitalEdgeLine,
        // TriggerActiveLevel, TriggerActiveEdge, TriggerActionSave, TriggerFilesToSave,
        // TriggerActionSignal, TriggerLineToSignal, TriggerLatchLine

        else if ( Match( "Trigger") ) {

                Success = GetInt( &Channel, "illegal Trigger channel number");
                if ( !Success )
                        return( 0 );

                if ( GetToken() != EQUALS ) {
                        ErrorIni( "missing = sign on TRIGGER");
                        return( 0 );
                        }

                Success = GetOnOff( &ScopeInfo.ChannelTrigger[LastChannel],
                                    "Trigger must = {ON,OFF}" );
                if ( !Success )
                        return( 0 );


                IsChannel   = -1;
                LastChannel = -1;
                LastDigital = -1;
                }



        else if ( KeywordEquals( "TriggerType" ) ) {

                if ( GetToken() == ERROR )
                        return( 0 );

                if ( Match( "None" ) )
                        ScopeInfo.TriggerType = SRDAT_TRIGGERTYPE_NONE;

                else if ( Match( "Threshold" ) )
                        ScopeInfo.TriggerType = SRDAT_TRIGGERTYPE_THRESHOLD;

                else if ( Match( "DigitalLevel" ) ||
                          Match( "DigitalSignal" ) ) {

                        if ( MAX_TRIGGER_DIGIO_LINE > 0 )
                                ScopeInfo.TriggerType = SRDAT_TRIGGERTYPE_DIGLEVEL;
                        else
                                WarnIni( "Triggering on digital line not supported" );
                        }

                else if ( Match( "DigitalEdge" ) ) {

                        if ( MAX_TRIGGER_DIGIO_LINE > 0 )
                                ScopeInfo.TriggerType = SRDAT_TRIGGERTYPE_DIGEDGE;
                        else
                                WarnIni( "Triggering on digital line not supported" );
                        }

                else {
                    ErrorIni( "TriggerType must = {None,Threshold,DigitalLevel,DigitalEdge}" );
                    return( 0 );
                        }

                }


        else if ( KeywordEquals( "TriggerThreshold" ) ) {

                Success = GetUlong( &ScopeInfo.TriggerThreshold,
                          "illegal TriggerThreshold integer value" );
                if ( !Success )
                        return( 0 );

                if ( (ScopeInfo.TriggerThreshold <= 0) ||
                     (ScopeInfo.TriggerThreshold > MAX_TRIGGER_THRESHOLD)  ) {

                        sprintf( msg,
                                 "TriggerThreshold must be > 0 and less than max amplitude %d (0x%X)",
                                 MAX_TRIGGER_THRESHOLD, MAX_TRIGGER_THRESHOLD );
                        ErrorIni( msg );
                        return( 0 );
                        }
                }


        else if ( KeywordEquals( "TriggerDigitalLevelLine" ) ) {

                Success = GetInt( &ScopeInfo.TriggerDigitalLevelLine,
                        "illegal TriggerDigitalLevelLine integer value" );
                if ( !Success )
                        return( 0 );

                if ( (ScopeInfo.TriggerDigitalLevelLine < 0) ||
                     (ScopeInfo.TriggerDigitalLevelLine > MAX_TRIGGER_DIGIO_LINE)  ) {

                        sprintf( msg,
                                 "TriggerDigitalLevelLine must be from 0 to %d",
                                 MAX_TRIGGER_DIGIO_LINE );
                        ErrorIni( msg );
                        return( 0 );
                        }
                }


        else if ( KeywordEquals( "TriggerDigitalEdgeLine" ) ) {

                Success = GetInt( &ScopeInfo.TriggerDigitalEdgeLine,
                        "illegal TriggerDigitalEdgeLine integer value" );
                if ( !Success )
                        return( 0 );

                if ( (ScopeInfo.TriggerDigitalEdgeLine < 0) ||
                     (ScopeInfo.TriggerDigitalEdgeLine > MAX_TRIGGER_DIGIO_LINE)  ) {

                        sprintf( msg,
                                 "TriggerDigitalEdgeLine must be from 0 to %d",
                                 MAX_TRIGGER_DIGIO_LINE );
                        ErrorIni( msg );
                        return( 0 );
                        }
                }



        else if ( KeywordEquals( "TriggerActiveLevel" ) ) {

                if ( GetToken() == ERROR )
                        return( 0 );

                if ( Match( "High" ) )
                        ScopeInfo.TriggerActiveLevel = SRDAT_TRIGGERLEVEL_HIGH;

                else if ( Match( "Low" ) )
                        ScopeInfo.TriggerActiveLevel = SRDAT_TRIGGERLEVEL_LOW;

                else {
                        ErrorIni( "TriggerActiveLevel must = {High,Low}" );
                        return( 0 );
                        }

                }


        else if ( KeywordEquals( "TriggerActiveEdge" ) ) {

                if ( GetToken() == ERROR )
                        return( 0 );

                if ( Match( "Rising" ) )
                        ScopeInfo.TriggerActiveEdge = SRDAT_TRIGGEREDGE_RISING;

                else if ( Match( "Falling" ) )
                        ScopeInfo.TriggerActiveEdge = SRDAT_TRIGGEREDGE_FALLING;

                else {
                        ErrorIni( "TriggerActiveEdge must = {Rising,Falling}" );
                        return( 0 );
                        }

                }


        else if ( KeywordEquals( "TriggerActionSave" ) ) {

                Success = GetOnOff( &ScopeInfo.TriggerActionSave, "TriggerActionSave must = {On,Off}" );
                if ( !Success )
                        return( 0 );
                }

        else if ( KeywordEquals( "TriggerFilesToSave" ) ) {

                Success = GetInt( &ScopeInfo.TriggerFilesToSave,
                          "illegal TriggerFilesToSave integer value" );
                if ( !Success )
                        return( 0 );

                if ( ScopeInfo.TriggerFilesToSave <= 0 ) {
                        ErrorIni( "TriggerFilesToSave must be > 0" );
                        return( 0 );
                        }
                }


        else if ( KeywordEquals( "TriggerActionSignal" ) ) {

                Success = GetOnOff( &ScopeInfo.TriggerActionSignal, "TriggerActionSignal must = {On,Off}" );
                if ( !Success )
                        return( 0 );
                }


        else if ( KeywordEquals( "TriggerLineToSignal" ) ) {

                Success = GetInt( &ScopeInfo.TriggerLineToSignal,
                          "illegal TriggerLineToSignal integer value" );
                if ( !Success )
                        return( 0 );

                if ( (ScopeInfo.TriggerLineToSignal < 0) ||
                     (ScopeInfo.TriggerLineToSignal > MAX_TRIGGER_DIGIO_LINE)  ) {

                        sprintf( msg,
                                 "TriggerLineToSignal must be 0 to %d",
                                 MAX_TRIGGER_DIGIO_LINE );
                        ErrorIni( msg );
                        return( 0 );
                        }
                }


        else if ( KeywordEquals( "TriggerLatch" ) ) {

                Success = GetOnOff( &ScopeInfo.TriggerLatchLine, "TriggerLatch must = {On,Off}" );
                if ( !Success )
                        return( 0 );
                }




        // UserCfg keywords

        else if ( KeywordEquals( "UserCfgNmeaTxPolarity" ) ) {

                if ( GetToken() == ERROR )
                        return( 0 );

                if ( Match( "IdleHigh" ) )
                        ScopeInfo.UserCfgNmeaTxPolarity = USERCFG_NMEA_TX_IDLE_HIGH;

                else if ( Match( "IdleLow" ) )
                        ScopeInfo.UserCfgNmeaTxPolarity = USERCFG_NMEA_TX_IDLE_LOW;

                else {
                        ErrorIni( "UserCfgNmeaTxPolarity must = {IdleHigh,IdleLow}" );
                        return( 0 );
                        }

                }

        else if ( KeywordEquals( "UserCfgNmeaRxPolarity" ) ) {

                if ( GetToken() == ERROR )
                        return( 0 );

                if ( Match( "IdleHigh" ) )
                        ScopeInfo.UserCfgNmeaRxPolarity = USERCFG_NMEA_RX_IDLE_HIGH;

                else if ( Match( "IdleLow" ) )
                        ScopeInfo.UserCfgNmeaRxPolarity = USERCFG_NMEA_RX_IDLE_LOW;

                else {
                        ErrorIni( "UserCfgNmeaRxPolarity must = {IdleHigh,IdleLow}" );
                        return( 0 );
                        }

                }


        else if ( KeywordEquals( "UserCfgPpsLed" ) ) {

                if ( GetToken() == ERROR )
                        return( 0 );

                if ( Match( "Toggle" ) )
                        ScopeInfo.UserCfgPpsLed = USERCFG_PPS_LED_TOGGLE;

                else if ( Match( "Programmed" ) )
                        ScopeInfo.UserCfgPpsLed = USERCFG_PPS_LED_PROGRAMMED;

                else {
                        ErrorIni( "UserCfgPpsLed must = {Toggle,Programmed}" );
                        return( 0 );
                        }

                }

        else if ( KeywordEquals( "UserCfgPpsPolarity" ) ) {

                if ( GetToken() == ERROR )
                        return( 0 );

                if ( Match( "IdleHigh" ) )
                        ScopeInfo.UserCfgPpsPolarity = USERCFG_PPS_IDLE_HIGH;

                else if ( Match( "IdleLow" ) )
                        ScopeInfo.UserCfgPpsPolarity = USERCFG_PPS_IDLE_LOW;

                else {
                        ErrorIni( "UserCfgPpsPolarity must = {IdleHigh,IdleLow}" );
                        return( 0 );
                        }

                }

        else if ( KeywordEquals( "UserCfgAdrdy" ) ) {

                Success = GetOnOff( &ScopeInfo.UserCfgAdrdy, "UserCfgAdrdy must = {On,Off}" );
                if ( !Success )
                        return( 0 );
                }

        else if ( KeywordEquals( "UserCfgTriggerPolarity" ) ) {

                if ( GetToken() == ERROR )
                        return( 0 );

                if ( Match( "IdleHigh" ) )
                        ScopeInfo.UserCfgTriggerPolarity = USERCFG_TRIGGER_IDLE_HIGH;

                else if ( Match( "IdleLow" ) )
                        ScopeInfo.UserCfgTriggerPolarity = USERCFG_TRIGGER_IDLE_LOW;

                else {
                        ErrorIni( "UserCfgTriggerPolarity must = {IdleHigh,IdleLow}" );
                        return( 0 );
                        }

                }






        // Unused keywords
        //
        // These options are not used at this time so just pass over
        // them here ...
        //
        //

        else if (
                  KeywordEquals( "ObsoleteOptions" )  ||
                  KeywordEquals( "ReallyOldOptions" ) ) {

                if ( GetToken() == ERROR )  // ignore them ...
                        return( 0 );

                }


        else if ( TokenType == EOF ) {

                Success = IniClose();     // Close the ini file
                if ( !Success )
                        return( 0 );

                ScopeIniVerifyData();
                ScopeIniCompute();

                return( 1 );
                }



        else { // token unrecognized ...

                ErrorIni( "unknown .ini syntax" );
                return( 0 );
                }




        // Continue parsing.

        goto parse;
}

//------------------------------------------------------------------------------
// ROUTINE: ScopeIniVerifyData
// PURPOSE: Ensure that all ScopeInfo variables that may have been updated from
//          the ini file are all still valid and in range.
//------------------------------------------------------------------------------
int ScopeIniVerifyData( void ) {

        // When changing defaults, be sure to change in ScopeIniSetDefaults too

        int i, maxValue;

        // Since the channel counts depend on which USBxCH device
        // we are using, we check all the hardware parameters first.

        // Hardware parameter verification

        if (strlen(ScopeInfo.AtodDriverName) == 0)
                SrStringCopy(ScopeInfo.AtodDriverName, USBXCH_DEFAULT_DRIVER_NAME, MAXSTR );

        if ( (ScopeInfo.AtodXchModel != SRDAT_ATODMODEL_USB1CH) &&
             (ScopeInfo.AtodXchModel != SRDAT_ATODMODEL_USB4CH) &&
             (ScopeInfo.AtodXchModel != SRDAT_ATODMODEL_USB8CH)
           )
                ScopeInfo.AtodXchModel = SRDAT_ATODMODEL_UNKNOWN;



        // Device id unknown or default requested, so use default

        if (ScopeInfo.AtodXchModel == SRDAT_ATODMODEL_UNKNOWN)
                ScopeInfo.AtodXchModel = SrUsbXchGetDefaultModel(
                                                ScopeInfo.AtodDriverName );

        // Set channel counts based on selected model

        ScopeSetChannelCountByModel( );

        if ( ScopeInfo.AtodXtalSpeed != USBXCH_XTAL_SPEED )
                ScopeInfo.AtodXtalSpeed = USBXCH_XTAL_SPEED;




        // Gps parameter verification

        if ( (ScopeInfo.GpsModel < SRDAT_GPSMODEL_NONE) ||
             (ScopeInfo.GpsModel >= SRDAT_GPSMODEL_MAX) )
                ScopeInfo.GpsModel = SRDAT_GPSMODEL_PCTIME;



        // Program setting parameter verification

        if ((ScopeInfo.ProgramAllowMultiple != INISWITCH_ON) &&
            (ScopeInfo.ProgramAllowMultiple != INISWITCH_OFF))
                ScopeInfo.ProgramAllowMultiple = INISWITCH_OFF;

        if ((ScopeInfo.ProgramKeypressToRun != INISWITCH_ON) &&
            (ScopeInfo.ProgramKeypressToRun != INISWITCH_OFF))
                ScopeInfo.ProgramKeypressToRun = INISWITCH_ON;

        if ((ScopeInfo.ProgramExitOnError != INISWITCH_ON) &&
            (ScopeInfo.ProgramExitOnError != INISWITCH_OFF))
                ScopeInfo.ProgramExitOnError = INISWITCH_ON;

        if ((ScopeInfo.ProgramInteractive != INISWITCH_ON) &&
            (ScopeInfo.ProgramInteractive != INISWITCH_OFF))
                ScopeInfo.ProgramInteractive = INISWITCH_ON;

        if ((ScopeInfo.ProgramShowStatusBar != INISWITCH_ON) &&
            (ScopeInfo.ProgramShowStatusBar != INISWITCH_OFF))
                ScopeInfo.ProgramShowStatusBar = INISWITCH_ON;

        if ((ScopeInfo.ProgramDisplayText != INISWITCH_ON) &&
            (ScopeInfo.ProgramDisplayText != INISWITCH_OFF))
                ScopeInfo.ProgramDisplayText = INISWITCH_ON;

        if (strlen(ScopeInfo.EditorName) == 0)
                SrStringCopy(ScopeInfo.EditorName, SCOPE_DEFAULT_EDITOR, MAXSTR );



        // Acquisition parameter verification

        if ((ScopeInfo.AcqSpsIndex < USBXCH_SPS_INDEX_MIN) ||
            (ScopeInfo.AcqSpsIndex > USBXCH_SPS_INDEX_MAX) ) {
                ScopeInfo.AcqSpsIndex = USBXCH_SPS_INDEX_DEFAULT;
                ScopeSpsUpdateFromIndex();
                }

        if ((ScopeInfo.AcqToggleLed != INISWITCH_ON) &&
            (ScopeInfo.AcqToggleLed != INISWITCH_OFF))
                ScopeInfo.AcqToggleLed = INISWITCH_ON;




        // Channel parameter verification

        ScopeComputeActiveChannels();   // NOTE: Assumes ScopeInfo parameters and
                                        //       AtodXxxChannel are already set



        for ( i = 0 ; i < ScopeInfo.AtodTotalChannels ; i++ ) {

//FIX - Do NOT let ChannelTitle ever get read in too big

                if (strlen(ScopeInfo.ChannelTitle[i]) > MAXTITLE)
                        ScopeInfo.ChannelTitle[i][MAXTITLE-1] = '\0';
                }


        for ( i = 0 ; i < ScopeInfo.AtodDigPlotChannels ; i++ ) {

                if (strlen(ScopeInfo.DigitalTitle[i]) > MAXTITLE)
                        ScopeInfo.DigitalTitle[i][MAXTITLE-1] = '\0';
                }

        if ((ScopeInfo.ChBufPts<=0) ||
            ((ScopeInfo.MaxSample!=0) && (ScopeInfo.ChBufPts>ScopeInfo.MaxSample))) {
                ScopeInfo.ChBufPts = ScopeInfo.DefaultNsamples;
                ScopeInfo.Npackets = ScopeInfo.ChBufPts/2 + 8; //Fix use good default here
                }

        if ( ScopeInfo.Npackets < 0 )
                ScopeInfo.Npackets = ScopeInfo.ChBufPts/2 + 8; //Fix use good default here




        // Output file parameter verification

        if ((ScopeInfo.OutputFileNaming != SRDAT_FILENAMETYPE_NONE)   &&
            (ScopeInfo.OutputFileNaming != SRDAT_FILENAMETYPE_SINGLE) &&
            (ScopeInfo.OutputFileNaming != SRDAT_FILENAMETYPE_YMDHMS) )
                ScopeInfo.OutputFileNaming = SRDAT_FILENAMETYPE_YMDHMS;

        if (ScopeInfo.OutputFileNbuffers <= 0)
                ScopeInfo.OutputFileNbuffers = 1; // FIX was Nplotbuffers;

        if ( ScopeInfo.OutputFileNaming == SRDAT_FILENAMETYPE_YMDHMS ) {
                ScopeInfo.OutputFileSeqStart = 0;
                }




        // Display parameter verification

        if ((ScopeInfo.PlotClipping != INISWITCH_ON) &&
            (ScopeInfo.PlotClipping != INISWITCH_OFF))
                ScopeInfo.PlotClipping = INISWITCH_OFF;

        if ((ScopeInfo.PlotAutoDC != INISWITCH_ON) &&
            (ScopeInfo.PlotAutoDC != INISWITCH_OFF))
                ScopeInfo.PlotAutoDC = INISWITCH_ON;

        if ((ScopeInfo.PlotShiftBy < 0) || (ScopeInfo.PlotShiftBy > ScopeInfo.MaxAmpShift-1))
                ScopeInfo.PlotShiftBy = ScopeInfo.MaxAmpShift/2;

//FIX - check NchannelsPlot vs DefaultNdisplay see ScopeComputeActiveChannels above
        if ((ScopeInfo.PlotNdisplay <= 0) || (ScopeInfo.PlotNdisplay > ScopeInfo.ChannelsPlot))
                ScopeInfo.PlotNdisplay = ScopeInfo.DefaultNdisplay;

        maxValue = ScopeInfo.ChannelsPlot - ScopeInfo.PlotNdisplay;
        if ((ScopeInfo.PlotFirstTrace < 0) || (ScopeInfo.PlotFirstTrace > maxValue))
                ScopeInfo.PlotFirstTrace = maxValue;

//FIX        maxValue = NsamplesPlot - GetSystemMetrics(SM_CXFULLSCREEN) - 1;
        maxValue = ScopeInfo.DefaultScreenWidth - 1;
        if (maxValue < 0)  maxValue = 0;
        if ((ScopeInfo.PlotFirstSample < 0) || (ScopeInfo.PlotFirstSample > maxValue))
                ScopeInfo.PlotFirstSample = 0;

        if ((ScopeInfo.PlotColorScheme < CS_MONO) || (ScopeInfo.PlotColorScheme > CS_BROWN))
                ScopeInfo.PlotColorScheme = ScopeInfo.DefaultColor;



        // Trigger parameter verification

        if ((ScopeInfo.TriggerType < SRDAT_TRIGGERTYPE_NONE) ||
            (ScopeInfo.TriggerType > SRDAT_TRIGGERTYPE_DIGEDGE))
                ScopeInfo.TriggerType = SRDAT_TRIGGERTYPE_NONE;

        if ((ScopeInfo.TriggerThreshold < 0) ||
            (ScopeInfo.TriggerThreshold > MAX_TRIGGER_THRESHOLD))
                ScopeInfo.TriggerThreshold = ScopeInfo.DefaultThreshold;

        if ((ScopeInfo.TriggerDigitalLevelLine < 0) ||
            (ScopeInfo.TriggerDigitalLevelLine > MAX_TRIGGER_DIGIO_LINE))
                ScopeInfo.TriggerDigitalLevelLine = 0;

        if ((ScopeInfo.TriggerDigitalEdgeLine < 0) ||
            (ScopeInfo.TriggerDigitalEdgeLine > MAX_TRIGGER_DIGIO_LINE))
                ScopeInfo.TriggerDigitalEdgeLine = 0;

        if ((ScopeInfo.TriggerActiveLevel != SRDAT_TRIGGERLEVEL_LOW) &&
            (ScopeInfo.TriggerActiveLevel != SRDAT_TRIGGERLEVEL_HIGH)  )
                ScopeInfo.TriggerActiveLevel = SRDAT_TRIGGERLEVEL_HIGH;

        if ((ScopeInfo.TriggerActiveEdge != SRDAT_TRIGGEREDGE_FALLING) &&
            (ScopeInfo.TriggerActiveEdge != SRDAT_TRIGGEREDGE_RISING)  )
                ScopeInfo.TriggerActiveEdge = SRDAT_TRIGGEREDGE_RISING;

        if ((ScopeInfo.TriggerActionSave != INISWITCH_ON) &&
            (ScopeInfo.TriggerActionSave != INISWITCH_OFF))
                ScopeInfo.TriggerActionSave = INISWITCH_ON;

        if (ScopeInfo.TriggerFilesToSave <= 0)
                ScopeInfo.TriggerFilesToSave = 2;

        if ((ScopeInfo.TriggerActionSignal != INISWITCH_ON) &&
            (ScopeInfo.TriggerActionSignal != INISWITCH_OFF))
                ScopeInfo.TriggerActionSignal = INISWITCH_OFF;

        if ((ScopeInfo.TriggerLineToSignal < 0) ||
            (ScopeInfo.TriggerLineToSignal > MAX_TRIGGER_DIGIO_LINE))
                ScopeInfo.TriggerLineToSignal = 0;

        if ((ScopeInfo.TriggerLatchLine != INISWITCH_ON) &&
            (ScopeInfo.TriggerLatchLine != INISWITCH_OFF))
                ScopeInfo.TriggerLatchLine = INISWITCH_OFF;

        for ( i = 0 ;  i < ScopeInfo.AtodTotalChannels ; i++ ) {
                if ((ScopeInfo.ChannelTrigger[i] != INISWITCH_ON) &&
                    (ScopeInfo.ChannelTrigger[i] != INISWITCH_OFF))
                        ScopeInfo.ChannelTrigger[i] = INISWITCH_OFF;
                }



        // XCR UserCfg configuration parameter verification

        if ( (ScopeInfo.UserCfgNmeaTxPolarity != USERCFG_NMEA_TX_IDLE_HIGH) &&
             (ScopeInfo.UserCfgNmeaTxPolarity != USERCFG_NMEA_TX_IDLE_LOW) )
                ScopeInfo.UserCfgNmeaTxPolarity = USERCFG_NMEA_TX_IDLE_HIGH;

        if ( (ScopeInfo.UserCfgNmeaRxPolarity != USERCFG_NMEA_RX_IDLE_HIGH) &&
             (ScopeInfo.UserCfgNmeaRxPolarity != USERCFG_NMEA_RX_IDLE_LOW) )
                ScopeInfo.UserCfgNmeaRxPolarity = USERCFG_NMEA_RX_IDLE_HIGH;

        if ( (ScopeInfo.UserCfgPpsLed != USERCFG_PPS_LED_TOGGLE) &&
             (ScopeInfo.UserCfgPpsLed != USERCFG_PPS_LED_PROGRAMMED) )
                ScopeInfo.UserCfgPpsLed = USERCFG_PPS_LED_TOGGLE;

        if ( (ScopeInfo.UserCfgPpsPolarity != USERCFG_PPS_IDLE_LOW) &&
             (ScopeInfo.UserCfgPpsPolarity != USERCFG_PPS_IDLE_HIGH) )
                ScopeInfo.UserCfgPpsPolarity = USERCFG_PPS_IDLE_LOW;

        if ( (ScopeInfo.UserCfgAdrdy != USERCFG_DB25_ADRDY_DISABLED) &&
             (ScopeInfo.UserCfgAdrdy != USERCFG_DB25_ADRDY_ENABLED) )
                ScopeInfo.UserCfgAdrdy = USERCFG_DB25_ADRDY_DISABLED;

        if (ScopeInfo.UserCfgNotUsed1 != USERCFG_UNUSED )
                ScopeInfo.UserCfgNotUsed1 = USERCFG_UNUSED;

        if ( (ScopeInfo.UserCfgTriggerPolarity != USERCFG_TRIGGER_IDLE_LOW) &&
             (ScopeInfo.UserCfgTriggerPolarity != USERCFG_TRIGGER_IDLE_HIGH) )
                ScopeInfo.UserCfgTriggerPolarity = USERCFG_TRIGGER_IDLE_LOW;


        return( 1 );
}


//------------------------------------------------------------------------------
// ROUTINE: ScopeIniReport
// PURPOSE: Generate an output report file named scope.rpt using the following
//          format.
//------------------------------------------------------------------------------
char *ScopeIniReportFmt[] = {

        "; USBxCH INI DATA from [%s]:                            \n",
        "                                                        \n",
        "          AtodDriverName = %s                           \n",
        "           AtodModelName = %s                           \n",
        "            GpsModelName = %s                           \n",
        "                                                        \n",
        "            SamplingRate = %lf                          \n",
        "               ToggleLed = %s                           \n",
        "                                                        \n",
        "         ChannelTitle %2d = \"%s\"   Trigger %2d = %s       \n",
        "                                                        \n",
        "         DigitalTitle %2d = \"%s\"                         \n",
        "                                                        \n",
        "             DisplayInfo = %s                           \n",
        "         DisplayDataText = %s                           \n",
        "           DisplaySerial = %s                           \n",
        "            DisplayEquip = %s                           \n",
        "                                                        \n",
        "        OutputFileNaming = %s                           \n",
        "      OutputFileNbuffers = %lu                          \n",
        "                                                        \n",
        "             TriggerType = %s                           \n",
        "        TriggerThreshold = %lu                          \n",
        " TriggerDigitalLevelLine = %d                           \n",
        "  TriggerDigitalEdgeLine = %d                           \n",
        "      TriggerActiveLevel = %s                           \n",
        "       TriggerActiveEdge = %s                           \n",
        "       TriggerActionSave = %s                           \n",
        "      TriggerFilesToSave = %d                           \n",
        "     TriggerActionSignal = %s                           \n",
        "     TriggerLineToSignal = %d                           \n",
        "            TriggerLatch = %s                           \n",
        "                                                        \n",
        "   UserCfgNmeaTxPolarity = %d                           \n",
        "   UserCfgNmeaRxPolarity = %d                           \n",
        "           UserCfgPpsLed = %d                           \n",
        "      UserCfgPpsPolarity = %d                           \n",
        "            UserCfgAdrdy = %d                           \n",
        "  UserCfgTriggerPolarity = %d                           \n",
        "                                                        \n",
        "           AllowMultiple = %s                           \n",
        "           KeypressToRun = %s                           \n",
        "             ExitOnError = %s                           \n",
        "                                                        \n",
        "                                                        \n",
        "; Additional items implied by these parameters:         \n",
        ";                                                       \n",
        ";               Channels = %d                           \n",
        ";               Npackets = %d                           \n",
        ";               ChBufPts = %d                           \n",
        ";                Nvalues = %u                           \n",
        ";             NgpsValues = %u                           \n",
        ";            AcqBufBytes = %u bytes                     \n",
        ";            SerBufBytes = %u bytes                     \n",
        ";         OutputFileSize = %lu bytes                    \n",
        ";      TriggerTimeToSave = %d seconds                   \n",
        ";                                                       \n",
        ";       TrueSamplingRate = %lf Hz                       \n",
        ";             BufferTime = %lf sec                      \n",
        ";                                                       \n",

        NULL,

        };

void ScopeIniReport( void ) {

        FILE *RptFp;
        int  i;
        char **fmt, *TriggerTypeStr, *TriggerActiveLevelStr, *TriggerActiveEdgeStr;
        char *OutputFileNamingStr;

#define OnOff ScopeMakeOnOffString

        TriggerTypeStr        = SrDatTriggerTypeString(  ScopeInfo.TriggerType        );
        TriggerActiveLevelStr = SrDatTriggerLevelString( ScopeInfo.TriggerActiveLevel );
        TriggerActiveEdgeStr  = SrDatTriggerEdgeString(  ScopeInfo.TriggerActiveEdge  );
        OutputFileNamingStr   = SrDatFileNameTypeString( ScopeInfo.OutputFileNaming   );



        // Generate the ASCII INI report file, Scope.rpt.

        RptFp = fopen( ScopeInfo.RptFileName, "w" );

        if ( !RptFp )
                DisplayLogMessage( MSGLEVEL_ERR, MSGDLG_SHOW, ERROR_FATAL,
                                   "unable to open Scope.rpt report file" );



        ScopeDisplayInfo( "Generating Scope.rpt ... \n" );



        fmt = ScopeIniReportFmt;

        fprintf( RptFp, Chop(*fmt++),                  ScopeInfo.IniFileName          );
        fprintf( RptFp, Chop(*fmt++)                                                  );

        fprintf( RptFp, Chop(*fmt++),                  ScopeInfo.AtodDriverName       );
        fprintf( RptFp, Chop(*fmt++),                  ScopeInfo.AtodXchModelName     );
        fprintf( RptFp, Chop(*fmt++),                  ScopeInfo.GpsModelName         );
        fprintf( RptFp, Chop(*fmt++)                                                  );
        fprintf( RptFp, Chop(*fmt++),                  ScopeInfo.AcqSamplingRate      );
        fprintf( RptFp, Chop(*fmt++),           OnOff( ScopeInfo.AcqToggleLed       ) );
        fprintf( RptFp, Chop(*fmt++)                                                  );


        for ( i = 0 ; i < ScopeInfo.Channels ; i++ )
                fprintf( RptFp, Chop( *fmt ), i, ScopeInfo.ChannelTitle[i], i,
                                              OnOff( ScopeInfo.ChannelTrigger[i] ) );
        fmt++;

        fprintf( RptFp, Chop(*fmt++)                                         );
        for ( i = 0 ; i < ScopeInfo.AtodDigPlotChannels ; i++ )
                fprintf( RptFp, Chop( *fmt ), i, ScopeInfo.DigitalTitle[i] );
        fmt++;

        fprintf( RptFp, Chop(*fmt++)                                            );
        fprintf( RptFp, Chop(*fmt++), OnOff( ScopeInfo.ProgramDisplayInfo )     );
        fprintf( RptFp, Chop(*fmt++), OnOff( ScopeInfo.ProgramDisplayText )     );
        fprintf( RptFp, Chop(*fmt++), OnOff( ScopeInfo.DisplaySerial )          );
        fprintf( RptFp, Chop(*fmt++), OnOff( ScopeInfo.DisplayEquip )           );
        fprintf( RptFp, Chop(*fmt++)                                            );
        fprintf( RptFp, Chop(*fmt++),        OutputFileNamingStr                );
        fprintf( RptFp, Chop(*fmt++),        ScopeInfo.OutputFileNbuffers       );
        fprintf( RptFp, Chop(*fmt++)                                            );
        fprintf( RptFp, Chop(*fmt++),        TriggerTypeStr                     );
        fprintf( RptFp, Chop(*fmt++),        ScopeInfo.TriggerThreshold         );
        fprintf( RptFp, Chop(*fmt++),        ScopeInfo.TriggerDigitalLevelLine  );
        fprintf( RptFp, Chop(*fmt++),        ScopeInfo.TriggerDigitalEdgeLine   );
        fprintf( RptFp, Chop(*fmt++),        TriggerActiveLevelStr              );
        fprintf( RptFp, Chop(*fmt++),        TriggerActiveEdgeStr               );
        fprintf( RptFp, Chop(*fmt++), OnOff( ScopeInfo.TriggerActionSave  )     );
        fprintf( RptFp, Chop(*fmt++),        ScopeInfo.TriggerFilesToSave       );
        fprintf( RptFp, Chop(*fmt++), OnOff( ScopeInfo.TriggerActionSignal )    );
        fprintf( RptFp, Chop(*fmt++),        ScopeInfo.TriggerLineToSignal      );
        fprintf( RptFp, Chop(*fmt++), OnOff( ScopeInfo.TriggerLatchLine )       );
        fprintf( RptFp, Chop(*fmt++)                                            );
//FIX - Convert these items into strings
        fprintf( RptFp, Chop(*fmt++),        ScopeInfo.UserCfgNmeaTxPolarity    );
        fprintf( RptFp, Chop(*fmt++),        ScopeInfo.UserCfgNmeaRxPolarity    );
        fprintf( RptFp, Chop(*fmt++),        ScopeInfo.UserCfgPpsLed            );
        fprintf( RptFp, Chop(*fmt++),        ScopeInfo.UserCfgPpsPolarity       );
        fprintf( RptFp, Chop(*fmt++),        ScopeInfo.UserCfgAdrdy             );
        fprintf( RptFp, Chop(*fmt++),        ScopeInfo.UserCfgTriggerPolarity   );
        fprintf( RptFp, Chop(*fmt++)                                            );
        fprintf( RptFp, Chop(*fmt++), OnOff( ScopeInfo.ProgramAllowMultiple )   );
        fprintf( RptFp, Chop(*fmt++), OnOff( ScopeInfo.ProgramKeypressToRun )   );
        fprintf( RptFp, Chop(*fmt++), OnOff( ScopeInfo.ProgramExitOnError )     );
        fprintf( RptFp, Chop(*fmt++)                                            );
        fprintf( RptFp, Chop(*fmt++)                                            );

        fprintf( RptFp, Chop(*fmt++)                                            );
        fprintf( RptFp, Chop(*fmt++)                                            );
        fprintf( RptFp, Chop(*fmt++),        ScopeInfo.Channels                 );
        fprintf( RptFp, Chop(*fmt++),        ScopeInfo.Npackets                 );
        fprintf( RptFp, Chop(*fmt++),        ScopeInfo.ChBufPts                 );
        fprintf( RptFp, Chop(*fmt++),        ScopeInfo.Nvalues                  );
        fprintf( RptFp, Chop(*fmt++),        ScopeInfo.NgpsValues               );
        fprintf( RptFp, Chop(*fmt++),        ScopeInfo.AcqBufBytes              );
        fprintf( RptFp, Chop(*fmt++),        ScopeInfo.SerBufBytes              );
        fprintf( RptFp, Chop(*fmt++),        ScopeInfo.OutputFileSize           );
        fprintf( RptFp, Chop(*fmt++),        ScopeInfo.TriggerTimeToSave        );
        fprintf( RptFp, Chop(*fmt++)                                            );
        fprintf( RptFp, Chop(*fmt++),        ScopeInfo.AcqSamplingRateTrue      );
        fprintf( RptFp, Chop(*fmt++),        ScopeInfo.BufferTime               );
        fprintf( RptFp, Chop(*fmt++)                                            );

        fclose( RptFp );


}



// HEADER DATA AND REPORTING:
//
// Two different header formats currently in use.  One, defined in
// ScopeCmd.h, is for the .HDR file used with .OUT data files.  The
// other, defined in SrUsbXch.h, is for the 4096 byte header that starts
// each .DAT file.
//


//------------------------------------------------------------------------------
// ROUTINE: ScopeHdrDefaults
// PURPOSE: Fill the specified header type with appropriate default values.
//------------------------------------------------------------------------------
void ScopeHdrDefaults( void ) {

        int  i;

        switch ( ScopeInfo.OutputFileNaming ) {

          case SRDAT_FILENAMETYPE_SINGLE:
          case SRDAT_FILENAMETYPE_YMDHMS:

            // Initialize .DAT file 4096 byte header.
            // Fill with generic defaults, then with USBxCH specific values.

            SrDatHdrInit( &DatHdr );

            SrStringCopy( DatHdr.DatProduct, HDR_PRODUCT_STRING, SRDAT_DIM_STRING );
            SrStringCopy( DatHdr.DatApplication, "SR SCOPE", SRDAT_DIM_STRING );

            DatHdr.DatGpsModel  = ScopeInfo.GpsModel;
            DatHdr.DatAtodModel = ScopeInfo.AtodXchModel;

            DatHdr.FileNameType = ScopeInfo.OutputFileNaming;
            if ( ScopeInfo.OutputFileNaming == SRDAT_FILENAMETYPE_YMDHMS )
                    DatHdr.FileNameInc = 1;
            else
                    DatHdr.FileNameInc = 0;

            DatHdr.Channels       = ScopeInfo.Channels;
            DatHdr.ChannelsAnalog = ScopeInfo.AtodAnalogChannels;

            if ( ScopeInfo.AtodDigitalChannels > 0 ) {
                    DatHdr.ChannelsDigital  = ScopeInfo.AtodDigitalChannels;
                    DatHdr.DigitalPerChan   = ScopeInfo.AtodDigPlotChannels;
                    DatHdr.ChannelDigIndex  = ScopeInfo.DigitalIndex;
                    }
            else {
                    DatHdr.ChannelsDigital  = 0;
                    DatHdr.DigitalPerChan   = 0;
                    DatHdr.ChannelDigIndex  = SRDAT_INDEXCHANNEL_NONE;
                    }

            if ( ScopeInfo.AtodGpsChannels > 0) {
                    DatHdr.ChannelsGps      = ScopeInfo.AtodGpsChannels;
                    DatHdr.ChannelMarkIndex = ScopeInfo.GpsMarkIndex;
                    }
            else {
                    DatHdr.ChannelsGps      = 0;
                    DatHdr.ChannelMarkIndex = SRDAT_INDEXCHANNEL_NONE;
                    }

            if ( ScopeInfo.AtodCounterChannels > 0) {
                    DatHdr.ChannelsCounter  = ScopeInfo.AtodCounterChannels;
                    DatHdr.ChannelObcIndex  = ScopeInfo.CounterIndex;
                    }
            else {
                    DatHdr.ChannelsCounter  = 0;
                    DatHdr.ChannelObcIndex  = SRDAT_INDEXCHANNEL_NONE;
                    }



            DatHdr.ChannelPtsPerRecord = ScopeInfo.ChBufPts;
            DatHdr.ChannelPtsPerFile   = ScopeInfo.ChBufPts * ScopeInfo.OutputFileNbuffers;
            DatHdr.AtodWordSize        = ScopeInfo.DataPtBytes;
            DatHdr.AtodDataCoding      = SRDAT_DATACODING_SIGNED;
            DatHdr.AtodSamplingRate    = ScopeInfo.AcqSamplingRateTrue;
            DatHdr.AtodGain            = 1.0;
            DatHdr.AtodFilterCoeff     = 0.0;
            DatHdr.AtodFilterScale     = 1.0;


            for ( i = 0 ; i < DatHdr.Channels ; i++ ) {
                    DatHdr.ChannelList[i].Channel = i;
//FIX make sure SRDAT_DIM_TITLE <= MAXTITLE
                    SrStringCopy(
                                 DatHdr.ChannelList[i].Title,
                                 ScopeInfo.ChannelTitle[i],
                                 SRDAT_DIM_TITLE
                                );
                    }

            for ( i = 0 ; i < DatHdr.DigitalPerChan ; i++ ) {
                    DatHdr.DigitalList[i].Channel = i;
                    SrStringCopy(
                                 DatHdr.DigitalList[i].Title,
                                 ScopeInfo.DigitalTitle[i],
                                 SRDAT_DIM_TITLE );
                    } // end for i


            DatHdr.TriggerType = ScopeInfo.TriggerType;
            if (ScopeInfo.TriggerType == SRDAT_TRIGGERTYPE_THRESHOLD) {
                    DatHdr.TriggerParm1 = ScopeInfo.TriggerThreshold;
                    DatHdr.TriggerParm2 = ScopeInfo.ChannelTrigger[ScopeInfo.Channels-1];
                    for ( i = 2 ; i < ScopeInfo.Channels ; i++ ) {
                            DatHdr.TriggerParm2 *= 2; // shift left 1 bit
                            DatHdr.TriggerParm2 |= ScopeInfo.ChannelTrigger[ScopeInfo.Channels-i];
                            }
                    }
            else if (ScopeInfo.TriggerType == SRDAT_TRIGGERTYPE_DIGLEVEL) {
                    DatHdr.TriggerParm1 = ScopeInfo.TriggerDigitalLevelLine;
                    DatHdr.TriggerParm2 = ScopeInfo.TriggerActiveLevel;
                    }
            else if (ScopeInfo.TriggerType == SRDAT_TRIGGERTYPE_DIGEDGE) {
                    DatHdr.TriggerParm1 = ScopeInfo.TriggerDigitalEdgeLine;
                    DatHdr.TriggerParm2 = ScopeInfo.TriggerActiveEdge;
                    }


            // Set GPS counter frequency in the Dat Hdr.

            DatHdr.CounterFreqHi  = 0L;
            DatHdr.CounterFreq    = 0L;
            DatHdr.Counter2FreqHi = 0L;
            DatHdr.Counter2Freq   = 0L;

            break;



          default:

            // No header requested

            break;


          } // end switch


}

//------------------------------------------------------------------------------
// ROUTINE: ScopeHdrUpdateWrite
// PURPOSE: Update a .DAT header and write it out.
//------------------------------------------------------------------------------
int ScopeHdrUpdateWrite( FILE *fp ) {

        struct tm *utc_time;
        long       modtime;
        int        nstr;
//FIX        LARGE_INTEGER tmp64;



        // No output desired

        if ( ScopeInfo.OutputFileNaming == SRDAT_FILENAMETYPE_NONE )
                return( 1 );


        // Update the DAT header structure with current information.

        nstr = (int)strlen( ScopeInfo.DataFileName );
        if (nstr < SRDAT_DIM_STRING)
                SrStringCopy( DatHdr.FileName, ScopeInfo.DataFileName, SRDAT_DIM_STRING );
        else {
                nstr = nstr - SRDAT_DIM_STRING + 1;
                SrStringCopy( DatHdr.FileName, &ScopeInfo.DataFileName[nstr], SRDAT_DIM_STRING );
                }

        // Get file "create" time.  The st_ctime value doesn't seem
        // to be properly updated when re-creating a file, so use the
        // "modify" time instead since we've just opened it for creation.

        modtime = SrFileModTime( fp );
        if ( modtime != -1 ) {
                DatHdr.FileCreateTimeHi    = 0L;
                DatHdr.FileCreateTime      = modtime;
                }
// WCT: This section will need to be added to support times greater than year 2038
//        LARGE_INTEGER tmp64;
//        if ( fstat64( fh, &filestats ) != -1 ) {
//                        tmp64.QuadPart = (LONGLONG)filestats.st_mtime;
//                        DatHdr.FileCreateTimeHi    = tmp64.HighPart;
//                        DatHdr.FileCreateTime      = tmp64.LowPart;
//                }

        DatHdr.FileSeqNum                  = ScopeInfo.OutputFileSeqStart;
        DatHdr.FileStartPtNum              = ScopeInfo.CurrPointsFromStart;
        DatHdr.TimePoint[0].Type           = SRDAT_TIME_TYPE_FIRSTPT;
        DatHdr.TimePoint[0].SampleFromRun  = ScopeInfo.CurrPointsFromStart;
        DatHdr.TimePoint[0].SampleFromFile = 0;
        DatHdr.TimePoint[0].GpsMarkIndex   = 0;

//FIX - check gmt vs local behaves correctly here, we think Acquire is in gmt

        utc_time = gmtime( &ScopeInfo.AcquireTime );  // Split time into YMD and HMS

        DatHdr.TimePoint[0].Year           = utc_time->tm_year + 1900;
        DatHdr.TimePoint[0].Month          = utc_time->tm_mon + 1;
        DatHdr.TimePoint[0].Day            = utc_time->tm_mday;
        DatHdr.TimePoint[0].YMDSource      = ScopeInfo.TimeSource;
        DatHdr.TimePoint[0].Hour           = utc_time->tm_hour;
        DatHdr.TimePoint[0].Minute         = utc_time->tm_min;
        DatHdr.TimePoint[0].Second         = utc_time->tm_sec;
        DatHdr.TimePoint[0].Microsecond    = ScopeInfo.AcquireTimeUsec;
        DatHdr.TimePoint[0].HMSSource      = ScopeInfo.TimeSource;



        // Write out the updated header

        if ( !SrDatHdrWrite( fp, &DatHdr ) ) {
                DisplayLogMessage( MSGLEVEL_WARN, MSGDLG_SHOW, ERROR_MAYBE_FATAL,
                                   "Could not write updated .DAT header" );
                return( 0 );
                }

        return( 1 );
}

//------------------------------------------------------------------------------
// ROUTINE: ScopeTmpFileComputeFinalName
// PURPOSE: Compute file name to use for output data depending on the selected
//          data format.
//------------------------------------------------------------------------------
char *ScopeTmpFileComputeFinalName( void ) {

        char *Suffix;
        static char ShortName[MAXSTR];
        static char FinalName[MAXSTR];


        // Set file suffix

        Suffix = "dat";


        // Compute the final file name using the requested naming mode.

        if ( ScopeInfo.OutputFileNaming == SRDAT_FILENAMETYPE_SINGLE ) {
                sprintf( ShortName, "Scope.%s", Suffix );
                sprintf( FinalName, "%s/%s", ScopeInfo.OutputFileDirectory, ShortName );
                }


        else if ( ScopeInfo.OutputFileNaming == SRDAT_FILENAMETYPE_YMDHMS ) {
                sprintf( ShortName, "%08ld.%s", ScopeInfo.OutputFileSeqStart, Suffix );
                sprintf( FinalName, "%s/%s", ScopeInfo.OutputFileDirectory, ShortName );
                }


        else  { // ScopeInfo.OutputFileNaming == SRDAT_FILENAMETYPE_NONE

                sprintf( ShortName, "" );
                sprintf( FinalName, "" );
                }


        SrStringCopy( ScopeInfo.DataFileName,      FinalName, MAXSTR );
        SrStringCopy( ScopeInfo.DataFileNameShort, ShortName, MAXSTR );

        return( FinalName );

}

//------------------------------------------------------------------------------
// ROUTINE: ScopeTmpFileOpen
// PURPOSE: Open the temp file for adding more data.
//------------------------------------------------------------------------------
FILE *ScopeTmpFileOpen( void ) {

        FILE *fp;
        char *oflag;

        // Open the tmp file.

        // If this is the first block, it must be created when opening.

        if ( ScopeInfo.CurrBufferInFile == 0 )
                oflag = "w+b"; // was OFLAG | (O_CREAT|O_TRUNC);
        else
                oflag = "a+b"; // was OFLAG | O_APPEND;


        fp = fopen( ScopeInfo.TmpFileName, oflag );

        if (!fp)
                DisplayLogMessage( MSGLEVEL_WARN, MSGDLG_SHOW, ERROR_MAYBE_FATAL,
                                   "unable to open Scope.tmp" );


        //FIX - need seektoend ???

        return( fp );
}

//------------------------------------------------------------------------------
// ROUTINE: ScopeTmpFileClose
// PURPOSE: Close the temp file after adding more data.  This ensures that at
//          least some data will be saved in the event of a system crash.
//------------------------------------------------------------------------------
int ScopeTmpFileClose( FILE *fp ) {

        // Close tmp file.

        //FIX - need fflush ???

        fclose( fp );

        return( 1 );
}

//------------------------------------------------------------------------------
// ROUTINE: ScopeTmpFileRename
// PURPOSE: Rename the temp file to its final name after all requested data
//          has been added.
//------------------------------------------------------------------------------
int ScopeTmpFileRename( char *FinalName ) {


        // Verify that new name is available.

        if ( SrFileExist( FinalName ) ) {

                sprintf( msg, "Rename failed.  File %s already exists.",
                         FinalName );
                DisplayLogMessage( MSGLEVEL_WARN, MSGDLG_SHOW, ERROR_MAYBE_FATAL, msg );
                return( 0 );
                }



        // Rename the .tmp file.

        if ( rename( ScopeInfo.TmpFileName, FinalName ) == 0 ) {

                sprintf( msg, "Scope.tmp renamed to %s ... \n", FinalName );
                ScopeDisplayInfo( msg );
                }

        else {
                DisplayLogMessage( MSGLEVEL_WARN, MSGDLG_SHOW, ERROR_MAYBE_FATAL,
                                   "Scope.tmp rename failed" );
                return( 0 );
                }


        return( 1 );
}

//------------------------------------------------------------------------------
// ROUTINE: ScopeTmpFileDelete
// PURPOSE: Delete the temp file.
//------------------------------------------------------------------------------
int ScopeTmpFileDelete( void ) {

#ifdef DELETE_FUNCTION_IMPLEMENTED
        // If sending data to standard out, there is nothing to delete.

        if ( ScopeInfo.OutputFileNaming == SRDAT_FILENAMETYPE_STDOUT )
                return( 1 );


        // Verify that specified name is available.

        if ( SrFileExist( ScopeInfo.TmpFileName ) ) {

                sprintf( msg, "Delete failed.  File %s does not exist.",
                         ScopeInfo.TmpFileName );
                DisplayLogMessage( MSGLEVEL_WARN, MSGDLG_SHOW, ERROR_MAYBE_FATAL, msg );
                return( 0 );
                }



        // Delete the .tmp file.

        sprintf( FullFileName, "%s/Scope.tmp", ScopeInfo.OutputFileDirectory );

        if ( remove( ScopeInfo.TmpFileName ) == 0 ) {

                sprintf( msg, "Deleted file %s ... \n", ScopeInfo.TmpFileName );
                ScopeDisplayInfo( msg );
                }

        else {
                DisplayLogMessage( MSGLEVEL_WARN, MSGDLG_SHOW, ERROR_MAYBE_FATAL,
                                   "Scope.tmp delete failed" );
                return( 0 );
                }
#else
        DisplayLogMessage( MSGLEVEL_WARN, MSGDLG_SHOW, ERROR_MAYBE_FATAL,
                           "Scope.tmp delete not implemented yet" );

#endif // DELETE_FUNCTION_IMPLEMENTED

        return( 1 );
}


//------------------------------------------------------------------------------
// ROUTINE: ScopeWriteData
// PURPOSE: Write out the acquired analog and serial data to the .tmp file
//          using the selected file format.
//------------------------------------------------------------------------------
int ScopeWriteData( void ) {

        FILE *fp;


        // Don't save anything if no file was requested.
        // But track buffer and file count for triggering.

        if ( ScopeInfo.OutputFileNaming == SRDAT_FILENAMETYPE_NONE ) {

                sprintf( msg, "Buffer[%08lX] processed\n", ScopeInfo.CurrBuffer );
                ScopeDisplayInfo( msg );

                ScopeInfo.CurrBufferInFile++;
                if (ScopeInfo.CurrBufferInFile >= ScopeInfo.OutputFileNbuffers) {
                        ScopeInfo.CurrBufferInFile = 0;
                        if ( ScopeInfo.CurrTriggerFile > 0 )
                                ScopeInfo.CurrTriggerFile--;
                        }

                return( 1 );
                }



        // Open tmp file and get file handle.

        fp = ScopeTmpFileOpen( );
        if ( !fp )
                return( 0 );



        // Before first buffer, update and write the header (if .DAT
        // files requested).  The final file name is now computed just
        // after the previous file was written.

        if ( ScopeInfo.CurrBufferInFile == 0 )
                ScopeHdrUpdateWrite( fp );



        // Save data in DAT format.

#ifdef SCOPE_WRITE_PACKETS
        if ( !SrDatRecordWrite( fp,
                          SRDAT_TAGID_USBPACKET,
                          ScopeInfo.UsbXchPacketArray,
                          ScopeInfo.CurrPacketsProcessed * sizeof(SRUSBXCH_PACKET) ) ) {
                DisplayLogMessage( MSGLEVEL_WARN, MSGDLG_SHOW, ERROR_MAYBE_FATAL,
                                   "Could not write USB packet data to .DAT" );
                return( 0 );
                }
#endif // SCOPE_WRITE_PACKETS


        if ( !SrDatRecordWrite( fp,
                          SRDAT_TAGID_USBANALOG,
                          ScopeInfo.UsbXchAnalogStruct,
                          ScopeInfo.ChBufPts * sizeof(SRDAT_USBXCH_ANALOG) ) ) {
                DisplayLogMessage( MSGLEVEL_WARN, MSGDLG_SHOW, ERROR_MAYBE_FATAL,
                                   "Could not write analog data to .DAT" );
                return( 0 );
                }


        if ( ScopeInfo.NextPtSerial > 0 ) {
                if ( !SrDatRecordWrite( fp,
                                  SRDAT_TAGID_USBSERIAL,
                                  ScopeInfo.UsbXchSerialStruct,
                                  ScopeInfo.NextPtSerial * sizeof(SRDAT_USBXCH_SERIAL) ) ) {
                        DisplayLogMessage( MSGLEVEL_WARN, MSGDLG_SHOW, ERROR_MAYBE_FATAL,
                                           "Could not write serial data to .DAT" );
                        return( 0 );
                        }
                }

        if ( ScopeInfo.NextPtEquip > 0 ) {
                if ( !SrDatRecordWrite( fp,
                                  SRDAT_TAGID_USBEQUIP,
                                  ScopeInfo.UsbXchEquipStruct,
                                  ScopeInfo.NextPtEquip * sizeof(SRDAT_USBXCH_EQUIPMENT) ) ) {
                        DisplayLogMessage( MSGLEVEL_WARN, MSGDLG_SHOW, ERROR_MAYBE_FATAL,
                                           "Could not write equipment data to .DAT" );
                        return( 0 );
                        }
                }


        if (ScopeInfo.CurrBufferInFile >= ScopeInfo.OutputFileNbuffers-1) // ScopeInfo.CurrBufferInFile not incremented yet
                if ( !SrDatRecordWrite( fp, SRDAT_TAGID_EOF, NULL, 0 ) ) {
                        DisplayLogMessage( MSGLEVEL_WARN, MSGDLG_SHOW, ERROR_MAYBE_FATAL,
                                "Could not write EOF to .DAT" );
                        return( 0 );
                        }



        // Close tmp file

        ScopeTmpFileClose( fp );

        sprintf( msg, "Buffer[%010lu], Nbuffer %02ld -> Scope.tmp\n",
                 ScopeInfo.CurrBuffer, ScopeInfo.CurrBufferInFile );
        ScopeDisplayInfo( msg );



        // Increment the number of blocks processed.

        ScopeInfo.CurrBufferInFile++;

        if (ScopeInfo.CurrBufferInFile >= ScopeInfo.OutputFileNbuffers) {
                if (!ScopeTmpFileRename( ScopeInfo.DataFileName ) ) {

                        // If scope can't rename the tmp file, it was
                        // deleted as part of cleaning up.  This was
                        // changed on 03/25/2005 so the user will have a
                        // chance to recover the data manually when scope
                        // stops due to the error.
                        //
                        // WCT: There may be some situations when
                        // running automatically with DoExitOnError off
                        // where this is not the best behavior.  This code
                        // should be changed again if those situtations can
                        // be clearly identified.
                        //
                        // int SaveFileError
                        // SaveFileError = FileError;
                        // ScopeTmpFileDelete();
                        // FileError = SaveFileError;

                        ScopeInfo.OutputFileSeqStart++; // Next output file name number
                        ScopeInfo.CurrBufferInFile = 0;
                        return( 0 );
                        }

                ScopeInfo.CurrBufferInFile = 0;
                if ( ScopeInfo.CurrTriggerFile > 0)
                        ScopeInfo.CurrTriggerFile--;
                ScopeInfo.OutputFileSeqStart++;         // Next output file name number

                // Prepare the final filename for the NEXT file, using the
                // ReadyTime of the last buffer in the file just written
                // as an approximation of the start time of the first point
                // in the next file.  This is a change from versions of the
                // code before 6/30/03 that computed the file name using
                // the time at the END of the first buffer in the next file.

                ScopeTmpFileComputeFinalName( );
                }

        return( 1 );
}



//------------------------------------------------------------------------------
// ROUTINE: ScopeTriggerThresholdSetRange
// PURPOSE: Compute the min and max values for the trigger threshold range
//          relative to the DC offset for the specified channel.
//------------------------------------------------------------------------------
int ScopeTriggerThresholdSetRange( long Channel, long *ThMin, long *ThMax ) {

        double TMPoffset;
        long   DCoffset, j, k;

        // Find channel offset

        k         = Channel;
        TMPoffset = 0;

        for ( j = 0 ; j < ScopeInfo.ChBufPts ; j++ ) {
                TMPoffset += ScopeInfo.UsbXchAnalogBuffer[k];
                k         += ScopeInfo.Channels;
                }
        DCoffset = (long)(TMPoffset / (double)ScopeInfo.ChBufPts);


        // Compute threshold range from offset

        *ThMin = DCoffset - ScopeInfo.TriggerThreshold;
        if (*ThMin < 0L)
                *ThMin = 0L;

        *ThMax = DCoffset + ScopeInfo.TriggerThreshold;
        if (*ThMax > 0x00FFFFFFL)
                *ThMax = 0x00FFFFFFL;

        return( 1 );
}

//------------------------------------------------------------------------------
// ROUTINE: ScopeTriggerCheckAnalog
// PURPOSE: Check if the analog data for the specified channel exceeds the
//          trigger threshold.
//------------------------------------------------------------------------------
int ScopeTriggerCheckAnalog( int Channel ) {

        int j, pt;
        long ThresholdMin, ThresholdMax;

        ScopeTriggerThresholdSetRange( Channel, &ThresholdMin, &ThresholdMax );

        pt = Channel;
        for ( j = 0 ; j < ScopeInfo.ChBufPts ; j++ ) {

            if ( (ScopeInfo.UsbXchAnalogBuffer[pt] < ThresholdMin) ||
                 (ScopeInfo.UsbXchAnalogBuffer[pt] > ThresholdMax) ) {

                    DisplayLogMessage( MSGLEVEL_MSG, MSGDLG_SHOW, ERROR_NOT_FATAL,
                                       "Amplitude exceeds threshold\n" );

                    return( 1 );
                    }

                pt += ScopeInfo.Channels;

            } // end for j

        return( 0 );
}

//------------------------------------------------------------------------------
// ROUTINE: ScopeTriggerDigitalCompare
// PURPOSE: This function compares the data point with the trigger mask and
//          returns 1 if any of the corresponding bits are set.  Since
//          each digital channel is packed into a single bit in a single
//          long, the mask should have a 1 in the bit position for each
//          digital channel that is participating in the trigger check.
//------------------------------------------------------------------------------
int ScopeTriggerDigitalCompare( int Point, long TriggerMask ) {

        if (ScopeInfo.UsbXchAnalogBuffer[Point] & TriggerMask)
                return( 1 );
        else
                return( 0 );
}

//------------------------------------------------------------------------------
// ROUTINE: ScopeTriggerThresholdCompare
// PURPOSE: This function converts the data point pt to offset format and
//          then compares it with the supplied threshold.  It returns 1
//          if the amplitude of the point exceeds the DCoffset corrected
//          threshold in either direction (min or max).
//------------------------------------------------------------------------------
int ScopeTriggerThresholdCompare( int Point, long ThMin, long ThMax ) {

        long OffsetPt;

        OffsetPt = (ScopeInfo.UsbXchAnalogBuffer[Point] & CLEARTOPBYTE)
                   ^ CONVERTTOOFFSET;

        if ( OffsetPt >= ThMax )
                return( 1 );
        else if ( OffsetPt <= ThMin )
                return( 1 );
        else
                return( 0 );
}

//------------------------------------------------------------------------------
// ROUTINE: ScopeTriggerDigioCompare
// PURPOSE: This function compares the signal on the specified digital
//          I/O line and returns 1 if the signal matches the specified
//          state.
//------------------------------------------------------------------------------
int ScopeTriggerDigioCompare( int Line, int ActiveState ) {

        int SignalState, DigitalIoNibble;

        SrUsbXchUserIoRd( ScopeInfo.UsbXchHandle, &DigitalIoNibble );

        SignalState = (DigitalIoNibble >> Line) & 0x1;

        if ( SignalState == ActiveState )
                return( 1 );
        else
                return( 0 );
}

//------------------------------------------------------------------------------
// ROUTINE: ScopeTriggerCondition
// PURPOSE: Check to see if the trigger condition has been met.  If so, return
//          1, otherwise return 0.  What is checked, depends on which of
//          several different trigger modes has been selected.
//------------------------------------------------------------------------------
int ScopeTriggerCondition( void ) {

        int i;
        int DigActive, trig;


        // Continuous recording

        if ( ScopeInfo.TriggerType == SRDAT_TRIGGERTYPE_NONE )
                return( 1 );


        // Amplitude threshold

        else if (ScopeInfo.TriggerType == SRDAT_TRIGGERTYPE_THRESHOLD) {

            // Update in progress state based on number of files saved

              if ( (ScopeInfo.CurrTriggerState == TRIG_STATE_INPROGRESS) &&
                   ( ScopeInfo.CurrTriggerFile <= 0 )        )
                   ScopeInfo.CurrTriggerState = TRIG_STATE_ARMED;

            // Compute the amplitude threshold trigger condition

            for ( i = 0 ; i < ScopeInfo.Channels ; i++ ) {

              // Only check channels selected for triggering

                if (ScopeInfo.ChannelTrigger[i] == INISWITCH_ON) {

                    if (i == ScopeInfo.DigitalIndex)
                        trig = 0;
                    else
                        trig = ScopeTriggerCheckAnalog( i );


                    if (trig) {

                        // If already saving, don't count current partial file

                        if (ScopeInfo.CurrTriggerFile > 0)
                            ScopeInfo.CurrTriggerFile = ScopeInfo.TriggerFilesToSave + 1;
                        else
                            ScopeInfo.CurrTriggerFile = ScopeInfo.TriggerFilesToSave;

                        ScopeInfo.CurrTriggerState = TRIG_STATE_INPROGRESS;
//FIX - is this return correct???
                        return( 1 );

                        } // end if trig

                    } // end if INISWITCH_ON

                } // end for i

             } // end if SRDAT_TRIGGERTYPE_THRESHOLD



        // Digital signal on I/O line level triggered

        else if (ScopeInfo.TriggerType == SRDAT_TRIGGERTYPE_DIGLEVEL) {

                // Immediate return if this feature not available

                if (ScopeInfo.TriggerDigitalLevelLine < 0 ||
                    ScopeInfo.TriggerDigitalLevelLine > MAX_TRIGGER_DIGIO_LINE )
                        return( 0 );

                // Update in progress state based on number of files saved

                  if ( (ScopeInfo.CurrTriggerState == TRIG_STATE_INPROGRESS) &&
                       (ScopeInfo.CurrTriggerFile <= 0)                 )
                       ScopeInfo.CurrTriggerState = TRIG_STATE_ARMED;


                // Test digital I/O line

                if ( ScopeTriggerDigioCompare( ScopeInfo.TriggerDigitalLevelLine,
                                               ScopeInfo.TriggerActiveLevel) ) {

                        sprintf( msg, "Digital level detected\n" );
                        ScopeDisplayInfo( msg );

                        // If already saving, don't count current partial file

                        if (ScopeInfo.CurrTriggerFile > 0)
                                ScopeInfo.CurrTriggerFile = ScopeInfo.TriggerFilesToSave + 1;
                        else
                                ScopeInfo.CurrTriggerFile = ScopeInfo.TriggerFilesToSave;

                        ScopeInfo.CurrTriggerState = TRIG_STATE_INPROGRESS;
//FIX - return here?
                        return( 1 );

                        } // end Digio Compare

                } // end if SRDAT_TRIGGERTYPE_DIGLEVEL


        // Digital signal on EDGE triggered I/O line (rising or falling)

        else if (ScopeInfo.TriggerType == SRDAT_TRIGGERTYPE_DIGEDGE) {

                // Immediate return if this feature not available

                if (ScopeInfo.TriggerDigitalEdgeLine < 0 ||
                    ScopeInfo.TriggerDigitalEdgeLine > MAX_TRIGGER_DIGIO_LINE)
                        return( 0 );

                // Update in progress state based on number of files saved

                if ( (ScopeInfo.CurrTriggerState == TRIG_STATE_INPROGRESS) &&
                     (ScopeInfo.CurrTriggerFile <= 0)                 ) {
                     ScopeInfo.CurrTriggerState = TRIG_STATE_WAITING;
                        }


                // Test digital I/O line for matching the active trigger
                // state (ie rising edge is active when high, falling edge
                // is active when low)

                DigActive = ScopeTriggerDigioCompare( ScopeInfo.TriggerDigitalEdgeLine,
                                                      ScopeInfo.TriggerActiveEdge );

                // Now process the match results based on the current
                // trigger state ...


                // We are armed and looking for trigger.  If we find one,
                // setup the number of files to save.

                if (ScopeInfo.CurrTriggerState == TRIG_STATE_ARMED) {
                        if (DigActive) { // triggered
                                sprintf( msg, "Digital edge detected\n" );
                                ScopeDisplayInfo( msg );
                                ScopeInfo.CurrTriggerState = TRIG_STATE_INPROGRESS;
                                ScopeInfo.CurrTriggerFile = ScopeInfo.TriggerFilesToSave;
                                return( 1 );
                                }
                        else
                                return( 0 );
                        }


                // We are currently processing a trigger.  If we are
                // almost done, we check to see if we've been re-armed.
                // This lets us skip the waiting state so we're ready to
                // trigger again immediately after the current trigger
                // finishes.  In any case, we indicate the current
                // trigger continues.

                else if (ScopeInfo.CurrTriggerState == TRIG_STATE_INPROGRESS) {
                        if ((ScopeInfo.CurrTriggerFile == 1)                               && // if last file +
                            (ScopeInfo.CurrBufferInFile == ScopeInfo.OutputFileNbuffers-1) && // last buffer +
                            !DigActive                                           ) {          // re-armed then
                                ScopeInfo.CurrTriggerState = TRIG_STATE_ARMED;                // skip waiting
                                }
                        return( 1 );

                        }


                // We are waiting to be re-armed by seeing the digital
                // line go to the inactive or non-triggered state.  With
                // rising edge triggering, a high value causes a trigger
                // and a low value re-arms.

                else { // (ScopeInfo.CurrTriggerState == TRIG_STATE_WAITING)
                        if (!DigActive) { // re-armed
                                ScopeInfo.CurrTriggerState = TRIG_STATE_ARMED;
                                }

                        return( 0 );     // Never return a trigger when waiting
                        }

                } // end if SRDAT_TRIGGERTYPE_DIGEDGE


        // Unknown trigger type

        else {
                DisplayLogMessage( MSGLEVEL_WARN, MSGDLG_SHOW, ERROR_MAYBE_FATAL,
                                   "Unknown trigger type" );
                return( 0 );
                }



        // A positive trigger file count may come from either
        //   1) An amplitude exceeding the threshold in the current buffer
        //   2) More files still remaining to be written after a
        //      previous high amplitude was found
        //
        // In other words, if we haven't already returned from somewhere
        // above because of finding a curent trigger, return a result
        // based on whether or not we still have more "after trigger"
        // files to save.

        if (ScopeInfo.CurrTriggerFile > 0)
                return( 1 );
        else
                return( 0 );
}

//------------------------------------------------------------------------------
// ROUTINE: ScopeTriggerAction
// PURPOSE: Action to be performed once the trigger condition has been met.
//          The typical action is to save files.
//------------------------------------------------------------------------------
int ScopeTriggerAction( void ) {

        // If this action chosen for true triggering,
        // turn on digital output bit

        // WCT: Consider keeping a global UserIo value and or-ing in signal
        //      so we only change the one signal line

        if ( ScopeInfo.TriggerType != SRDAT_TRIGGERTYPE_NONE  &&
             ScopeInfo.TriggerActionSignal == INISWITCH_ON )
                SrUsbXchUserIoWr( ScopeInfo.UsbXchHandle,
                                ( 0x01 << ScopeInfo.TriggerLineToSignal ) );



        // The standard trigger action is to save files

        if ( ScopeInfo.TriggerActionSave == INISWITCH_ON ) {
                return( ScopeWriteData( ) );
                }

        // Otherwise, just count down files that would have been written
        // so trigger will finish as usual

        else {
                ScopeInfo.CurrBufferInFile++;
                if (ScopeInfo.CurrBufferInFile >= ScopeInfo.OutputFileNbuffers) {
                        ScopeInfo.CurrBufferInFile = 0;
                        if (ScopeInfo.CurrTriggerFile > 0)
                                ScopeInfo.CurrTriggerFile--;
                        }
                sprintf( msg, "Buffer[%08lX] discarded\n", ScopeInfo.CurrBuffer );
                DisplayLogMessage( MSGLEVEL_MSG, MSGDLG_SHOW, ERROR_NOT_FATAL, msg );
                }
                return( 1 );
}

//------------------------------------------------------------------------------
// ROUTINE: ScopeTriggerDone
// PURPOSE: Action to be performed once a trigger condition has expired.
//          The typical action is turn off a digital output bit that was turned
//          on when the trigger condition was first met.
//------------------------------------------------------------------------------
int ScopeTriggerDone( void ) {

        // Turn off digital output bit if it was turned on
        // and the trigger state is not being latched

        if ( ScopeInfo.TriggerType != SRDAT_TRIGGERTYPE_NONE       && // triggering
             ScopeInfo.TriggerActionSignal == INISWITCH_ON && // digital output action
             ScopeInfo.TriggerLatchLine == INISWITCH_OFF   ) // not latching
                ScopeTriggerClearLatch( );

        return( 1 );

}

//------------------------------------------------------------------------------
// ROUTINE: ScopeTriggerClearLatch
// PURPOSE: Allow the user to turn off a digital output bit that was turned
//          on when the trigger condition was first met and stayed on even after
//          the trigger condition expired (ie it was latched).
//------------------------------------------------------------------------------
int ScopeTriggerClearLatch( void ) {

        // Turn off digital output bit to clear

        // WCT: Consider keeping a global UserIo value and and-ing in signal
        //      so we only change the one signal line

        SrUsbXchUserIoWr( ScopeInfo.UsbXchHandle, (0x00 << ScopeInfo.TriggerLineToSignal) );

        return( 1 );
}



//------------------------------------------------------------------------------
// ROUTINE: ScopeLogMessagePrep
// PURPOSE: Prepare a complete message string by adding the time and date
//          on to the supplied message.
//------------------------------------------------------------------------------
void ScopeLogMessagePrep( int MsgLevel, const char *ExtraMsg ) {

        int   msglen;
        char *datestr, dbuffer[16];
        char *timestr, tbuffer[16];
        char *levelstr;
        time_t     current_time;
        struct tm *local_time;
        FILE      *LogFp;


        // Get the date and time.

        time( &current_time );
        local_time = localtime( &current_time );

        strftime( dbuffer, 16, "%Y/%m/%d", local_time );
        strftime( tbuffer, 16, "%H:%M:%S", local_time );

        datestr = dbuffer;
        timestr = tbuffer;


        // Prepare message string

        if ( MsgLevel == MSGLEVEL_MSG )
                levelstr = "Message";
        else if ( MsgLevel == MSGLEVEL_WARN )
                levelstr = "Warning";
        else if ( MsgLevel == MSGLEVEL_ERR )
                levelstr = "Error";
        else
                levelstr = "Note";

        if ( ExtraMsg ) {
                msglen = strlen( ExtraMsg );
                if ( msglen > 210 ) // MAXSTR of 256-(16+1+16+1+7+6+2)
                        sprintf( ScopeInfo.LogMessage, "%s %s %s = ...%s\n",
                                 datestr, timestr, levelstr, &ExtraMsg[msglen-210] );
                else
                        sprintf( ScopeInfo.LogMessage, "%s %s %s = %s\n",
                                 datestr, timestr, levelstr, ExtraMsg );
                }
        else
                sprintf( ScopeInfo.LogMessage, "%s %s %s\n",
                                  datestr, timestr, levelstr );



        // Append the error to the log file Scope.log.

        if ( LogFp = fopen( ScopeInfo.LogFileName, "a" ) ) {

                fprintf( LogFp, "%s", ScopeInfo.LogMessage );

                   fclose( LogFp );

                }


}


// A/D functions

//------------------------------------------------------------------------------
// ROUTINE: ScopeAtodInit
// PURPOSE: Open and initialize the USBxCH A/D.
//------------------------------------------------------------------------------
int ScopeAtodInit( void ) {

        int    ErrorCode, UserCfgByte;
        char   ExtraMsg[MAXSTR];


        // Initialize to bad values

        ScopeInfo.UsbXchHandle = BAD_DEVHANDLE;


        // Prepare user mode byte and sampling rate

        SrUsbXchFillUserCfgByte( ScopeInfo.UserCfgNmeaTxPolarity,
                                 ScopeInfo.UserCfgNmeaRxPolarity,
                                 ScopeInfo.UserCfgPpsLed,
                                 ScopeInfo.UserCfgPpsPolarity,
                                 ScopeInfo.UserCfgAdrdy,
                                 ScopeInfo.UserCfgNotUsed1,
                                 ScopeInfo.UserCfgTriggerPolarity,
                                 &UserCfgByte );



        // Open USBxCH driver

        ScopeInfo.UsbXchHandle = SrUsbXchOpen(
                                ScopeInfo.AtodDriverName,
                                ScopeInfo.AtodXchModel,
                                ScopeInfo.GpsModel,
                                UserCfgByte,
                                ScopeInfo.AcqSamplingRate,
                                &ScopeInfo.AcqSamplingRateTrue,
                                &ErrorCode
                               );


        // Set specific error information

        if ( ErrorCode == USBXCH_ERROR_OPEN_SAMPLE_RATE )

                strcpy( ExtraMsg, "Invalid sample rate." );

        else if ( ErrorCode == USBXCH_ERROR_DRIVER_NOT_OPEN )

                strcpy( ExtraMsg, "Check for active drivers by running DevMan showall or SetDid showall." );

        else if ( !SrUsbXchIsOpenSuccess( ErrorCode ) )

                strcpy( ExtraMsg, "Check power and cable.  Run DevMan showall to check driver." );

        else
                strcpy( ExtraMsg, "" );



        // Set general error info, close if needed, and return failure

        if ( ScopeInfo.UsbXchHandle == BAD_DEVHANDLE ) {
                sprintf( msg, "Could not open and initialize the %s.  (%s)\n%s",
                                  ScopeInfo.AtodXchModelName,
                                  USBXCH_ERROR_MSG[ErrorCode],
                                  ExtraMsg );
                DisplayLogMessage( MSGLEVEL_WARN, MSGDLG_SHOW, ERROR_MAYBE_FATAL, msg );
                return( 0 );
                }
        if ( !SrUsbXchIsOpenSuccess( ErrorCode ) ) {
                sprintf( msg, "Could not initialize the %s.  (%s)\n%s",
                                  ScopeInfo.AtodXchModelName,
                                  USBXCH_ERROR_MSG[ErrorCode],
                                  ExtraMsg );
                DisplayLogMessage( MSGLEVEL_WARN, MSGDLG_SHOW, ERROR_MAYBE_FATAL, msg );
                SrUsbXchClose( ScopeInfo.UsbXchHandle );
                ScopeInfo.UsbXchHandle = BAD_DEVHANDLE;
                return( 0 );
                }



        // Initialize the USBxCH Yellow + Red LEDs to off.

        ScopeInfo.LedValueYellow = 0;
        ScopeInfo.LedValueRed = 0;

        SrUsbXchUserLed( ScopeInfo.UsbXchHandle,
                       ScopeInfo.LedValueYellow,
                       ScopeInfo.LedValueRed );
        SrUsbXchUserIoWr( ScopeInfo.UsbXchHandle, (ScopeInfo.LedValueYellow<<3) );



        // Return success

        return( 1 );
}

//------------------------------------------------------------------------------
// ROUTINE: ScopeAtodStart
// PURPOSE: Begin acquiring data with the USBxCH A/D.
//------------------------------------------------------------------------------
int ScopeAtodStart( void ) {

        int Result, Error;

        // Initialize current position counter.  Used for gathering a
        // whole buffer of ready data before indicating ready.

        ScopePtCountersInitNext( );
        ScopePtCountersInitTotal();


        // Prepare YMDHMS directory if requested

        if ( ScopeInfo.OutputFileNaming == SRDAT_FILENAMETYPE_YMDHMS ) {
                ScopeInfo.OutputFileSeqStart = 0;
                ScopeMakeDirYmdHms();
                ScopeInitFileNames();
                }


        // Start the A/D

        Result = SrUsbXchStart( ScopeInfo.UsbXchHandle, &Error );



        // Write start message to log, but don't show dialog

        if ( Result )  {
                if ( strcmp( ScopeInfo.OutputFileDirectory, ".") == 0 )
                        sprintf( msg, "Starting acquisition in current working directory." );
                else
                        sprintf( msg, "Starting acquisition in %s.", ScopeInfo.OutputFileDirectory );
                }
        else
                sprintf( msg, "Failed to start acquisition.  Error %s.",
                         USBXCH_ERROR_MSG[Error] );
        

        DisplayLogMessage( MSGLEVEL_MSG, MSGDLG_NO_SHOW, ERROR_NOT_FATAL, msg );

        return( Result );
}

//------------------------------------------------------------------------------
// ROUTINE: ScopeAtodStop
// PURPOSE: Stop acquiring data with the USBxCH A/D.
//------------------------------------------------------------------------------
int ScopeAtodStop( void ) {

        int Result, Error;

        Result = SrUsbXchStop( ScopeInfo.UsbXchHandle, &Error );  // Stop A/D execution
        SrUsbXchClose( ScopeInfo.UsbXchHandle );                  // Close the A/D driver
        ScopeInfo.UsbXchHandle = BAD_DEVHANDLE;


        // Write stop message to log, but don't show dialog

        if ( !Result ) 
                sprintf( msg, "Failed to stop acquisition.  Error %s.",
                         USBXCH_ERROR_MSG[Error] );
        else if ( strcmp( ScopeInfo.OutputFileDirectory, ".") == 0 )
                sprintf( msg, "Stopping acquisition in current working directory." );
        else
                sprintf( msg, "Stopping acquisition in %s.", ScopeInfo.OutputFileDirectory );

        // NOTE: To avoid any possibility of circular calls, it is good to use
        //       NO_SHOW on the MSGDLG since ScopeAtodStop can be called from
        //       DisplayLogMessage via ShowGuiLogMessage in some circumstances.
        
        DisplayLogMessage( MSGLEVEL_MSG, MSGDLG_NO_SHOW, ERROR_NOT_FATAL, msg );

        return( Result );
}

//------------------------------------------------------------------------------
// ROUTINE: ScopeAtodGetReadyData
// PURPOSE: Read acquired data from the USBxCH A/D.  This includes both analog,
//          serial, and equipment data.  Only get ready data.  Returns 1 if a
//          buffer worth of data is ready, 0 if more data is needed to fill
//          the buffer, and -1 if an error has occurred.
//------------------------------------------------------------------------------
int ScopeAtodGetReadyData( void ) {

        int           ErrorCode, CheckSanity;
        unsigned long NanalogRead, NanalogRequest, Nsamples, Npackets;
        unsigned long NserialRead, NserialRequest, NequipmentRequest, NequipmentRead;
        unsigned long NpacketRequest, NpacketRead, Nremain;
        unsigned long Nprocessed, NprocessedData, NprocessedStatus;

//FIX - Check for negative NanalogRequest and NpacketRequest, maybe done below?
//
//        if ( ScopeInfo.ChBufPts <= ScopeInfo.NextPtAnalog)
//                return( 0 );



        // Read data.  If none is ready, Nsamples will be 0.

        NanalogRequest    = ScopeInfo.ChBufPts - ScopeInfo.NextPtAnalog;
        NserialRequest    = ScopeInfo.NgpsValues;
        NequipmentRequest = ScopeInfo.NgpsValues;
        NpacketRequest    = ScopeInfo.Npackets - ScopeInfo.CurrPacketsRead;
        CheckSanity       = 1;

        if ( NpacketRequest == 0L ) // No packets requested, so quick exit
                return( 0 );

        Npackets = SrUsbXchGetDataAsPackets(
                                        ScopeInfo.UsbXchHandle,
                                        &ScopeInfo.UsbXchPacketArray[ScopeInfo.CurrPacketsRead],
                                        NpacketRequest,
                                        &NpacketRead,
                                        &ErrorCode
                                       );


        if ( NpacketRead == 0L ) // No packet read, so quick exit
                return( 0 );

        ScopeInfo.CurrPacketsRead += NpacketRead;               // previous + newly read

        if ( ScopeAtodCheckError( ErrorCode, "SrUsbXchGetDataAsPackets" ) == -1 )
                return( -1 );



        // Computed number of unprocessed packets

        Nremain = ScopeInfo.CurrPacketsRead - ScopeInfo.CurrPacketsProcessed;


        // Process data from Usb Packets into UsbXch Structures for Analog,Serial,Equip

        Nsamples = SrUsbXchConvertPackets(
                      &ScopeInfo.UsbXchPacketArray[ScopeInfo.CurrPacketsProcessed],
                      Nremain,
                      CheckSanity,
                      &ScopeInfo.LastMsg,
                      &ScopeInfo.UsbXchAnalogStruct[ScopeInfo.NextPtAnalog],
                      NanalogRequest,
                      &NanalogRead,
                      &ScopeInfo.UsbXchSerialStruct[ScopeInfo.NextPtSerial],
                      NserialRequest,
                      &NserialRead,
                      &ScopeInfo.UsbXchEquipStruct[ScopeInfo.NextPtEquip],
                      NequipmentRequest,
                      &NequipmentRead,
                      &Nprocessed,
                      &NprocessedData,
                      &NprocessedStatus,
                      &ErrorCode
                      );

        ScopeInfo.CurrPacketsProcessed += Nprocessed;

        Nremain = ScopeInfo.CurrPacketsRead - ScopeInfo.CurrPacketsProcessed;

        if ( ScopeAtodCheckError( ErrorCode, "SrUsbXchConvertPackets" ) == -1 )
                return( -1 );

//FIX - remove after debugging

        if (Nsamples != NanalogRead) {
                printf( "WARNING: ScopeAtodGetReadyData Nsamples %lu != NanalogRead %lu\n",
                          Nsamples, NanalogRead );
                }



        // Process analog, serial, and equipment data structures into
        // format required by Scope program

        ScopeProcessAnalog( NanalogRead );
        ScopeProcessSerial( (int)NserialRead );
        ScopeProcessEquip( (int)NequipmentRead );


        // Error check

        if ( ScopeInfo.ChBufPts < ScopeInfo.NextPtAnalog ) {
                printf( "WARNING: ChBufPts %d < NextPtAnalog %d\n",
                          ScopeInfo.ChBufPts, ScopeInfo.NextPtAnalog );
                if ( ScopeAtodCheckError( USBXCH_ERROR_READ_TOO_LARGE,
                                          "ScopeAtodGetReadyData" ) == -1 )
                        return( -1 );
                }




        if ( Nremain == 0 && (ScopeInfo.ChBufPts != ScopeInfo.NextPtAnalog) ) {

//                printf("WARNING: Processed all available packets, but haven't filled analog buffer!!!\n" );
//                printf("WARNING: Read requested number of packets, but haven't filled analog buffer!!!\n" );
//                if ( ScopeAtodCheckError( USBXCH_ERROR_READ_NOT_ENOUGH_DATA,
//                                          "ScopeAtodGetReadyData" ) == -1 )
//                        return( -1 );
                }


        // If analog buffer is full

        if (ScopeInfo.ChBufPts == ScopeInfo.NextPtAnalog) {

                // Save updated sample totals

                ScopeInfo.TotalSample   += ScopeInfo.NextPtAnalog;
                ScopeInfo.TotalPtSerial += ScopeInfo.NextPtSerial;
                ScopeInfo.TotalPtEquip  += ScopeInfo.NextPtEquip;


                // Twiddle yellow LED to show we are working and a buffer is ready

                ScopeTwiddleYellowLed( );

                return( 1 );  // Indicate buffer ready
                }
        else {
                return( 0 );  // Buffer not ready yet
                }
}

//------------------------------------------------------------------------------
// ROUTINE: ScopeAtodDataPrepareToRead
// PURPOSE: Prepare to read new data by removing processed packets from the
//          array of packet data previously read from the USBxCH.
//------------------------------------------------------------------------------
int ScopeAtodDataPrepareToRead( void ) {

        int           Error;
        unsigned long FirstPacket;

        // Clean up processed packets and send back the number of unprocessed
        // packets that remain in the array

        FirstPacket = SrDatUsbPacketCleanUp(
                                            ScopeInfo.UsbXchPacketArray,
                                            ScopeInfo.CurrPacketsRead,
                                            ScopeInfo.CurrPacketsProcessed,
                                            &Error
                                           );


        // Reset current number of packets read and processed

        ScopeInfo.CurrPacketsRead      = 0L + FirstPacket;
        ScopeInfo.CurrPacketsProcessed = 0L;


        if ( Error != USBXCH_ERROR_NONE )
                return( 0 );
        else
                return( 1 );
}

//------------------------------------------------------------------------------
// ROUTINE: ScopeAtodCheckOverflow
// PURPOSE: If the USBxCH onboard DRAM FIFO fills up, an overflow is declared
//          and this function returns 1.  Otherwise it returns 0.  Once an
//          overflow has occurred, the data in the FIFO is valid and can be
//          read.  But you must stop and restart acquisition before any data
//          after the overflow can be acquired.
//------------------------------------------------------------------------------
int ScopeAtodCheckOverflow( void ) {

        return( SrUsbXchOverflow( ScopeInfo.UsbXchHandle ) );
}


//------------------------------------------------------------------------------
// ROUTINE: ScopeAtodCheckError
// PURPOSE: Check for an error during data read.
//------------------------------------------------------------------------------
//FIX - Original version did this, do we ?
// Trap abnormal overflow where no points are coming in
// so we avoid getting stuck in an infinite polling loop
//        if (ErrorCode == USBXCH_ERROR_OVERFLOW && n <= 0) {

int ScopeAtodCheckError( int errorcode, char *str ) {


        if (errorcode != USBXCH_ERROR_NONE) {
                sprintf( msg,
                                 "%s failed with error # %d: %s.",
                                 str,
                                 errorcode,
                                 USBXCH_ERROR_MSG[errorcode]
                                );
                DisplayLogMessage( MSGLEVEL_ERR, MSGDLG_SHOW, ERROR_FATAL, msg );
                return( -1 );
                }

        return( 1 );
}

//------------------------------------------------------------------------------
// ROUTINE: ScopeProcessAnalog
// PURPOSE: Transfer analog data from the UsbXch data structure into the simple
//          array use by scope when plotting the data.
//------------------------------------------------------------------------------
int ScopeProcessAnalog( unsigned long Nsamples ) {

        unsigned long i, j, ku, kd;
        int           SampleCount;


        // Transfer data from UsbXchAnalogStruct into simpler scope arrays


        for ( i = 0 ; i < Nsamples ; i++ ) {

                SampleCount = ScopeInfo.TotalSample + i;

                for ( j = 0 ; j < (unsigned long)ScopeInfo.Channels ; j++ ) {

                        ku = ScopeInfo.NextPtAnalog+i;
                        kd = (ku*ScopeInfo.Channels)+j;

                        if ( j == ScopeInfo.DigitalIndex ) {
                                ScopeInfo.UsbXchAnalogBuffer[kd] =
                                        (ScopeInfo.UsbXchAnalogStruct[ku].DigitalIn & 0x000F);
//FIX test this                         (ScopeInfo.UsbXchAnalogStruct[ku].PpsToggle & 0x000F);
//printf("DigitaIndex i = %d, j = %d, ku = %d, kd = %d DGTL = 0x%X, DigIn = 0x%02X\n",
//                                i, j, ku, kd,
//                                ScopeInfo.UsbXchAnalogBuffer[kd],
//                                (ScopeInfo.UsbXchAnalogStruct[ku].DigitalIn & 0x000F)
//      );
                                }

                        else if ( j == ScopeInfo.GpsMarkIndex ) {
                                if (SampleCount == 0) {
//printf("Changed OBC[ku=%ld] from 0x%X to 0\n", ku, ScopeInfo.UsbXchAnalogStruct[ku].OnBoardCount);
                                        ScopeInfo.UsbXchAnalogStruct[ku].OnBoardCount = 0;
                                        }
                                if (ScopeInfo.UsbXchAnalogStruct[ku].OnBoardCount == 0)
                                        ScopeInfo.UsbXchAnalogBuffer[kd] = 0;
                                else
                                        ScopeInfo.UsbXchAnalogBuffer[kd] = 1;
                                }

                        else if ( j == ScopeInfo.CounterIndex ) {
                                ScopeInfo.UsbXchAnalogBuffer[kd] =
                                        (ScopeInfo.UsbXchAnalogStruct[ku].OnBoardCount);
//printf("CounterIndex i = %d, j = %d, ku = %d, kd = %d CNTR = 0x%X, OBC = 0x%X\n",
//                                i, j, ku, kd,
//                                ScopeInfo.UsbXchAnalogBuffer[kd],
//                                (ScopeInfo.UsbXchAnalogStruct[ku].OnBoardCount)
//      );
                                }

                        else {
                                ScopeInfo.UsbXchAnalogBuffer[kd] =
                                        ScopeInfo.UsbXchAnalogStruct[ku].AnalogData[j];
                                }

                        } // end for j

                } // end for i


        // Update next starting sample

        ScopeInfo.NextPtAnalog += ((int)Nsamples);


        return( 1 );
}


//------------------------------------------------------------------------------
// ROUTINE: ScopeProcessSerial
// PURPOSE: Perform any processing on the serial data held in a SRDAT_USBXCH_SERIAL
//          structure into any formats needed by scope for displaying it.
//          Currently, the data is displayed and the only processing is to
//          update the next serial point variable.
//------------------------------------------------------------------------------
int ScopeProcessSerial( int Nserial ) {

        int i, j, jstart;

//FIX - what should this routine really do

        // Display NMEA Data

        if ( ScopeInfo.DisplaySerial == INISWITCH_ON ) {

                for ( i = ScopeInfo.NextPtSerial ; i < Nserial ; i++ ) {

                        // Display them

                        printf( "\nUsbXchSerialStruct[%d]:\n", i );
                        printf( " PpsEventNum = %d\n",   ScopeInfo.UsbXchSerialStruct[i].PpsEventNum );
                        printf( " NmeaCount   = %d\n",   ScopeInfo.UsbXchSerialStruct[i].NmeaCount  );
                        printf( " ValidFields = 0x%X\n", ScopeInfo.UsbXchSerialStruct[i].ValidFields  );
                        for ( j = 0 ; j < ScopeInfo.UsbXchSerialStruct[i].NmeaCount ; j++ ) {
                                jstart = j * SRDAT_NMEA_MAX_SIZE;
                                printf( " Msg[%d]= %s", j, &ScopeInfo.UsbXchSerialStruct[i].NmeaMsg[jstart] );
                                }
                        printf( "\n" );

                        }
                }

        ScopeInfo.NextPtSerial += Nserial;

        return( ScopeInfo.NextPtSerial );
}

//------------------------------------------------------------------------------
// ROUTINE: ScopeProcessEquip
// PURPOSE: Perform any processing on the equipment data held in an SRDAT_USBXCH_EQUIPMENT
//          structure into any formats needed by scope for displaying it.
//          Currently, the data is displayed and the only processing is to
//          update the next equipment point variable.
//------------------------------------------------------------------------------
int ScopeProcessEquip( int Nequip ) {

        int    i, Empty, PartFull, Overflow;
        double TempC, TempF;

//FIX - what should this routine really do

        // Display NMEA Data

        if ( ScopeInfo.DisplayEquip == INISWITCH_ON ) {

                for ( i = ScopeInfo.NextPtEquip ; i < Nequip ; i++ ) {

                        SrDatUsbTemperatureCompute( ScopeInfo.UsbXchEquipStruct[i].Temperature,
                                                    &TempC, &TempF );
                        SrUsbXchDramFlagsSplit( ScopeInfo.UsbXchEquipStruct[i].DramFlags,
                                                &Empty, &PartFull, &Overflow );


                        // Display them

                        printf( "UsbXchEquipStruct[%d]:\n", i );
                        printf( "   PpsEventNum  = %d\n",   ScopeInfo.UsbXchEquipStruct[i].PpsEventNum );
                        printf( "   Voltage Good = %d\n",   ScopeInfo.UsbXchEquipStruct[i].VoltageGood );
                        printf( "   Raw Temp     = 0x%X\n", ScopeInfo.UsbXchEquipStruct[i].Temperature );
                        printf( "   Degrees C    = %5.1lf\n", TempC );
                        printf( "   Degrees F    = %3.0lf\n", (TempF+0.5L) );
                        printf( "   DramFlags    = 0x%02X\n",  ScopeInfo.UsbXchEquipStruct[i].DramFlags );
                        printf( "   EM/PF/OV     = %2d/%2d/%2d\n", Empty, PartFull, Overflow );
                        printf( "\n" );

                        }
                }

        ScopeInfo.NextPtEquip += Nequip;

        return( ScopeInfo.NextPtEquip );
}

//------------------------------------------------------------------------------
// ROUTINE: ScopeBuffersUpdateSize
// PURPOSE: Compute various parameters that are related to buffer size.
//------------------------------------------------------------------------------
int ScopeBuffersUpdateSize( void ) {

        long GpsBytesPerSec, NsecPerFile;


        // Compute buffer size so it is about 1/2 second long

        ScopeInfo.ChBufPts    = 2*((int)(ScopeInfo.AcqSamplingRateTrue/4.0));

        ScopeInfo.Nvalues     = ScopeInfo.ChBufPts * ScopeInfo.Channels;
        ScopeInfo.AcqBufBytes = ScopeInfo.Nvalues  * ScopeInfo.DataPtBytes;

        if ( ScopeInfo.AcqSamplingRateTrue > 0.0001 )
                ScopeInfo.BufferTime = ScopeInfo.ChBufPts / ScopeInfo.AcqSamplingRateTrue;
        else
                ScopeInfo.BufferTime = 0.0L;


// FIX - 2 pts/analog packet + 2*x status packets/second
//       we allocate enough space for 1 sets of GPS in case there are NMEA strings
//       at the beginning and end of the approximately 1 second of data buffer
//       but when using PCTIME the strings may be for often so add an extra
//       factor of 2

        if ( ScopeInfo.GpsModel == SRDAT_GPSMODEL_NONE )
                ScopeInfo.Npackets = ScopeInfo.ChBufPts/2;          // no NMEA strings
        else if ( ScopeInfo.GpsModel == SRDAT_GPSMODEL_PCTIME )
                ScopeInfo.Npackets = ScopeInfo.ChBufPts/2 + 1*2*1;  // short ZDA only
        else if ( ScopeInfo.GpsModel == SRDAT_GPSMODEL_TRIMBLE )
                ScopeInfo.Npackets = ScopeInfo.ChBufPts/2 + 1*3;    // short ZDA and long GGA
        else // ( ScopeInfo.GpsModel == SRDAT_GPSMODEL_GARMIN )
                ScopeInfo.Npackets = ScopeInfo.ChBufPts/2 + 1*4;    // long RMC and long GGA


        // Compute some GPS parameters
        //
        // GPS data comes once per second.  So, to determine the number
        // of points needed in the GPS data arrays, we compute the
        // length of time for one buffer and add 2 for good measure.
        //
        // ScopeInfo.NgpsValues  = (int)(ScopeInfo.BufferTime) + 2;
        //
        // But when using PCTIME, we may send ZDA strings more often.
        // So lets round up the buffer time and assume we may have two
        // strings per second.

        ScopeInfo.NgpsValues  = (int)(ScopeInfo.BufferTime+1)*2;

        ScopeInfo.SerBufBytes = ScopeInfo.NgpsValues * sizeof( SRDAT_USBXCH_SERIAL );

        GpsBytesPerSec        = sizeof( SRDAT_USBXCH_SERIAL );
        NsecPerFile           = (long)(((ScopeInfo.ChBufPts * ScopeInfo.OutputFileNbuffers) /
                                         ScopeInfo.AcqSamplingRateTrue) + 0.5);


        // Compute some trigger info

        ScopeInfo.TriggerTimeToSave = (int)((float)(ScopeInfo.TriggerFilesToSave *
                                 ScopeInfo.OutputFileNbuffers * ScopeInfo.ChBufPts) /
                                 ScopeInfo.AcqSamplingRateTrue );


        // The sleepms variable controls how many milliseconds to wait
        // between each attempt to read new data from the USBxCH.  A
        // value of BufferTime * 200 means wait about 1/5th the amount
        // of time needed to acquire one buffer worth of data.
        //
        //   SleepMs = (BufferTime sec / 5) * 1000 ms/sec
        //
//FIX   // When changing this value, be sure to change TimeCount limit

        ScopeInfo.SleepMs        = (int)(ScopeInfo.BufferTime * 200.0L);


        // Compute basic output file size, then extend to include DAT
        // headers and tags and GPS parameters.

        ScopeInfo.OutputFileSize = ScopeInfo.AcqBufBytes *
                                   ScopeInfo.OutputFileNbuffers;

        ScopeInfo.OutputFileSize += (SRDAT_HDR_TOTALSIZE +
                                     SRDAT_TAGSIZE * (ScopeInfo.OutputFileNbuffers+1));

        ScopeInfo.OutputFileSize += (GpsBytesPerSec * NsecPerFile);


        return( 1 );
}

//------------------------------------------------------------------------------
// ROUTINE: ScopeBuffersAllocate
// PURPOSE: Allocate USBxCH and PARGPS buffer memory:
//          The UsbXchAnalogBuffer array in PC memory is used to store analog
//          data on its way out from the USBxCH to the disk and screen.
//          The UsbXchSerialBuffer array is for GPS timing data.
//------------------------------------------------------------------------------
int ScopeBuffersAllocate(

                SRUSBXCH_PACKET **UsbXchPacketArray,
                SRDAT_USBXCH_ANALOG **UsbXchAnalogStruct,
                long **UsbXchAnalogBuffer,
                long **UsbXchAnalogBufferEnd,
                SRDAT_USBXCH_SERIAL **UsbXchSerialStruct,
                char **UsbXchSerialBuffer,
                char **UsbXchSerialBufferEnd,
                SRDAT_USBXCH_EQUIPMENT **UsbXchEquipStruct

               ) {

        int i;


        // Allocate USBxCH packet array.


        *UsbXchPacketArray   = (SRUSBXCH_PACKET *)malloc( ScopeInfo.Npackets *
                                                      sizeof(SRUSBXCH_PACKET) );
        if ( !*UsbXchPacketArray ) {
                sprintf( msg, "Unable to malloc %s usb packet array",
                         ScopeInfo.AtodXchModelName );
                DisplayLogMessage( MSGLEVEL_ERR, MSGDLG_SHOW, ERROR_FATAL, msg );
                return( 0 );
                }

//FIX - do we need to zero out CurrPacket variables here too?



        // Allocate USBxCH data buffers.
        // The DataStruct contains data in the exact format output by the USBxCH.
        // The DataBuffer contains data in the format expected by scope.

        *UsbXchAnalogStruct    = (SRDAT_USBXCH_ANALOG *)malloc( ScopeInfo.ChBufPts *
                                                      sizeof(SRDAT_USBXCH_ANALOG) );
        *UsbXchAnalogBuffer    = (long *)malloc( ScopeInfo.AcqBufBytes );
        *UsbXchAnalogBufferEnd = *UsbXchAnalogBuffer + ScopeInfo.Nvalues;

        if ( !*UsbXchAnalogStruct || !*UsbXchAnalogBuffer ) {
                sprintf( msg, "Unable to malloc %s data structure and buffer",
                         ScopeInfo.AtodXchModelName );
                DisplayLogMessage( MSGLEVEL_ERR, MSGDLG_SHOW, ERROR_FATAL, msg );
                return( 0 );
                }


        // Allocate USBxCH serial data buffers.
        // The SerialStruct contains data in the exact format used in DAT files.
        // The SerialBuffer contains data in character format as provided by USBxCH.

        *UsbXchSerialStruct    = (SRDAT_USBXCH_SERIAL *)malloc( ScopeInfo.SerBufBytes );
        *UsbXchSerialBuffer    = (char*)malloc( ScopeInfo.AcqBufBytes );
        *UsbXchSerialBufferEnd = *UsbXchSerialBuffer + ScopeInfo.AcqBufBytes;

        if ( !*UsbXchSerialStruct || !*UsbXchSerialBuffer ) {
                DisplayLogMessage( MSGLEVEL_ERR, MSGDLG_SHOW, ERROR_FATAL,
                                   "Unable to malloc GPS data structure and buffer" );
                return( 0 );
                }


        // Allocate USBxCH equipment data buffers.
        // The EquipStruct contains data in the exact format used in DAT files.

        *UsbXchEquipStruct    = (SRDAT_USBXCH_EQUIPMENT *)malloc( ScopeInfo.SerBufBytes );

        if ( !*UsbXchEquipStruct ) {
                DisplayLogMessage( MSGLEVEL_ERR, MSGDLG_SHOW, ERROR_FATAL,
                                   "Unable to malloc equipment data structure" );
                return( 0 );
                }



        // Initialize SerialData and EquipData structures

        for ( i = 0 ; i < (int)ScopeInfo.NgpsValues ; i++ ) {

                SrDatUsbSerialInit( &ScopeInfo.UsbXchSerialStruct[ScopeInfo.NextPtSerial] );
                SrDatUsbEquipInit( &ScopeInfo.UsbXchEquipStruct[ScopeInfo.NextPtEquip] );
                }
        ScopeInfo.LastMsg = -1;



        return( 1 );
}

//------------------------------------------------------------------------------
// ROUTINE: ScopeBuffersFree
// PURPOSE: Free the USBxCH and PARGPS buffer memory allocated in AllocateBuffers.
//------------------------------------------------------------------------------
int ScopeBuffersFree(

                SRUSBXCH_PACKET **UsbXchPacketArray,
                SRDAT_USBXCH_ANALOG **UsbXchAnalogStruct,
                long **UsbXchAnalogBuffer,
                long **UsbXchAnalogBufferEnd,
                SRDAT_USBXCH_SERIAL **UsbXchSerialStruct,
                char **UsbXchSerialBuffer,
                char **UsbXchSerialBufferEnd,
                SRDAT_USBXCH_EQUIPMENT **UsbXchEquipStruct

               ) {



        // Free USBxCH data structure and related buffer.

//FIX - do we need to zero out CurrPacket variables here too?

        if (*UsbXchPacketArray)
               { free( *UsbXchPacketArray ); *UsbXchPacketArray = NULL; }

        if (*UsbXchAnalogStruct)
               { free( *UsbXchAnalogStruct ); *UsbXchAnalogStruct = NULL; }

        if (*UsbXchAnalogBuffer)
               { free( *UsbXchAnalogBuffer ); *UsbXchAnalogBuffer = NULL; }

        *UsbXchAnalogBufferEnd = NULL;



        // Free USBxCH GPS serial data structure and related buffer.

        if (*UsbXchSerialStruct)
                { free( *UsbXchSerialStruct ); *UsbXchSerialStruct = NULL; }

        if (*UsbXchSerialBuffer)
                { free( *UsbXchSerialBuffer ); *UsbXchSerialBuffer = NULL; }

        *UsbXchSerialBufferEnd    = NULL;



        // Free USBxCH equipment structure.

        if (*UsbXchEquipStruct)
               { free( *UsbXchEquipStruct ); *UsbXchEquipStruct = NULL; }


        return( 1 );
}

//------------------------------------------------------------------------------
// ROUTINE: ScopeTwiddleYellowLed
// PURPOSE: If toggling was requested, twiddle the yellow led.
//------------------------------------------------------------------------------
int ScopeTwiddleYellowLed( void ) {

        if ( ScopeInfo.AcqToggleLed == INISWITCH_ON ) {
                ScopeInfo.LedValueYellow ^= 1;
                SrUsbXchUserLed( ScopeInfo.UsbXchHandle,
                               ScopeInfo.LedValueYellow,
                               ScopeInfo.LedValueRed );
                }

        return( 1 );
}

//------------------------------------------------------------------------------
// ROUTINE: ScopeInitFileNames
// PURPOSE: Compute full file names that include the directory path for several
//          of the files used by scope such as temporary, error, header, and
//          report files.
//------------------------------------------------------------------------------
int ScopeInitFileNames( void ) {

        // NOTE: Assumes ScopeInfo parameters OutputFileDirectory,
        //       OutputFileNaming are already set


        // Note: ini file name gotten from command line or menu item
        //       and passed to ScopeIniParse which fills in ScopeInfo.IniFileName

        // Fill in rpt, tmp, and err

        sprintf( ScopeInfo.RptFileName, "%s/Scope.rpt", ScopeInfo.OutputFileDirectory );
        sprintf( ScopeInfo.TmpFileName, "%s/Scope.tmp", ScopeInfo.OutputFileDirectory );
        sprintf( ScopeInfo.LogFileName, "Scope.log" ); // Always in current working directory

        // Fill in Final and Short data file names

        ScopeTmpFileComputeFinalName();


        return( 1 );
}

//------------------------------------------------------------------------------
// ROUTINE: ScopeSpsUpdateFromIndex
// PURPOSE: Determine the actual sampling rate based on the index number for
//          one of the rates allowed by the A/D chips on the USBxCH.
//------------------------------------------------------------------------------
int ScopeSpsUpdateFromIndex( void ) {

        int Err;

        SrUsbXchSpsRateTable( ScopeInfo.AcqSpsIndex,
                              &ScopeInfo.AcqSamplingRate,
                              &Err );

        ScopeInfo.AcqSamplingRateTrue = ScopeInfo.AcqSamplingRate;

        return( Err );
}

//------------------------------------------------------------------------------
// ROUTINE: ScopeSpsGetString
// PURPOSE: Write the sampling rate corresponding to the specified index into
//          the provided string.  Returns 1 for success, 0 for invalid parameter.
//------------------------------------------------------------------------------
int ScopeSpsGetString( int SpsIndex, char *SpsString, int Nchar ) {

        double SpsRate;

        if  ( (SpsIndex < USBXCH_SPS_INDEX_MIN) ||
              (SpsIndex > USBXCH_SPS_INDEX_MAX) ||
              (!SpsString)                      ||
              (Nchar < 11)                      )
                return( 0 );


        SrUsbXchSpsRateTable( SpsIndex, &SpsRate, NULL );

        sprintf( SpsString, "%-11.5lf", SpsRate );

        return( 1 );
}

//------------------------------------------------------------------------------
// ROUTINE: ScopeCreateShortFileName
// PURPOSE: Split off the directory path part of the specified file name.
//------------------------------------------------------------------------------
int ScopeCreateShortFileName( char *LongName, char *ShortName ) {

#if defined( SROS_WINXP ) || defined( SROS_WIN2K )
        char fname[_MAX_FNAME], ext[_MAX_EXT];

        if (!LongName || !ShortName)
                return( 0 );

        if (strlen( LongName ) == 0)
                strcpy( ShortName, LongName );
        else {
                _splitpath( LongName, NULL, NULL, fname, ext );
                sprintf( ShortName, "%s%s", fname, ext );
                }

#else
//FIX   printf( "********* ScopeCreateShortFileName not available for Linux yet *************\n" );
        strcpy( ShortName, LongName );
#endif // SROS_xxx

        return( 1 );
}

//------------------------------------------------------------------------------
// ROUTINE: ScopeComputeActiveChannels
// PURPOSE: Compute number of active channels and index to special channels.
//          Assumes ScopeInfo parameters AtodXxxChannel are already set.
//------------------------------------------------------------------------------
int ScopeComputeActiveChannels( void ) {


        // Set defaults to no special channels

        ScopeInfo.DigitalIndex = SRDAT_INDEXCHANNEL_NONE;
        ScopeInfo.GpsMarkIndex = SRDAT_INDEXCHANNEL_NONE;
        ScopeInfo.CounterIndex = SRDAT_INDEXCHANNEL_NONE;



        // Analog channels always included

        ScopeInfo.Channels     = ScopeInfo.AtodAnalogChannels;
        ScopeInfo.ChannelsPlot = ScopeInfo.AtodAnaPlotChannels;


        // Digital channels

        ScopeInfo.DigitalIndex  = ScopeInfo.Channels;
        ScopeInfo.Channels     += ScopeInfo.AtodDigitalChannels;
        ScopeInfo.ChannelsPlot += ScopeInfo.AtodDigPlotChannels;


        // Gps channels

        ScopeInfo.GpsMarkIndex  = ScopeInfo.Channels;
        ScopeInfo.Channels     += ScopeInfo.AtodGpsChannels;
        ScopeInfo.ChannelsPlot += ScopeInfo.AtodGpsPlotChannels;


        // Counter channels

        ScopeInfo.CounterIndex  = ScopeInfo.Channels;
        ScopeInfo.Channels     += ScopeInfo.AtodCounterChannels;
        ScopeInfo.ChannelsPlot += ScopeInfo.AtodCntPlotChannels;


// FIX - what to do if there were NO channels

// FIX - check NchannelsPlot vs DefaultNdisplay

        // Ensure valid display number

        if (ScopeInfo.DefaultNdisplay > ScopeInfo.ChannelsPlot )
                ScopeInfo.DefaultNdisplay = ScopeInfo.ChannelsPlot;

        if (ScopeInfo.PlotNdisplay > ScopeInfo.ChannelsPlot ) {
                ScopeInfo.PlotNdisplay   = ScopeInfo.ChannelsPlot ;
                ScopeInfo.PlotFirstTrace = 0;
                }

        return( 1 );
}

//------------------------------------------------------------------------------
// ROUTINE: ScopeInitActiveChannels
// PURPOSE: Sets type, trigger, and title for active channels.
//          Assumes ScopeInfo parameters XxxIndex, and AtodXxxChannel
//          are already set.
//------------------------------------------------------------------------------
int ScopeInitActiveChannels( void ) {

        int i;
        char SigString[MAXSTR];

        //

        // NOTE: Assumes ScopeInfo parameters XxxIndex, and
        //       AtodXxxChannel are already set


        // Analog channels

        for ( i = 0 ; i < ScopeInfo.AtodAnalogChannels ; i++ ) {
                sprintf( SigString, "CH%02d", i );
                SrStringCopy( ScopeInfo.ChannelTitle[i], SigString, MAXTITLE );
                ScopeInfo.ChannelTrigger[i] = INISWITCH_ON;
                ScopeInfo.ChannelType[i]    = CH_TYPE_ANALOG;
                }


        // Digital channels

        i = ScopeInfo.DigitalIndex;
        SrStringCopy( ScopeInfo.ChannelTitle[i], "DGTL", MAXTITLE );
        ScopeInfo.ChannelTrigger[i] = INISWITCH_OFF;
        ScopeInfo.ChannelType[i]    = CH_TYPE_DIGITAL;

        for ( i = 0 ; i < ScopeInfo.AtodDigPlotChannels ; i++ ) {
                sprintf( SigString, "DIG%1d", i );
                SrStringCopy( ScopeInfo.DigitalTitle[i], SigString, MAXTITLE );
                }


        // Gps channels

        i = ScopeInfo.GpsMarkIndex;
        SrStringCopy( ScopeInfo.ChannelTitle[i], "PPS", MAXTITLE );
        ScopeInfo.ChannelTrigger[i] = INISWITCH_OFF;
        ScopeInfo.ChannelType[i]    = CH_TYPE_GPSMARK;


        // Counter channels

        i = ScopeInfo.CounterIndex;
        SrStringCopy( ScopeInfo.ChannelTitle[i], "CNTR", MAXTITLE );
        ScopeInfo.ChannelTrigger[i] = INISWITCH_OFF;
        ScopeInfo.ChannelType[i]    = CH_TYPE_GPSCOUNTER;


        return( 1 );
}

//------------------------------------------------------------------------------
// ROUTINE: ScopeOverflowRecovery
// PURPOSE: Let scope recover after an overflow occurs.
//          After an overflow has been detected, the USBxCH must be stopped
//          and restarted.  Some users might want to read out the valid data
//          data currently saved in the fifo first.  If so, put that code
//          here ...
//------------------------------------------------------------------------------
void ScopeOverflowRecovery( void ) {

        // Perform orderly shutdown and restart after overflow.

        ScopeAtodStop( );
        ScopeAtodInit( );
        ScopeAtodStart( );
}

//------------------------------------------------------------------------------
// ROUTINE: ScopeAtodPcTimeStamp
// PURPOSE: Send a timestamp to the A/D hardware if the user requested PCTIME as
//          the GPS model.  The timestamp is made by creating a pseudo NMEA
//          string from the current PC time.
//------------------------------------------------------------------------------
int ScopeAtodPcTimeStamp( void ) {

        int Error;

        if ( ScopeInfo.GpsModel == SRDAT_GPSMODEL_PCTIME )
                SrUsbXchSendTimeStamp( ScopeInfo.UsbXchHandle, &Error );
        else
                Error = USBXCH_ERROR_NONE;

        return( Error );
}

//------------------------------------------------------------------------------
// ROUTINE: ScopeCurrentBufferInit
// PURPOSE: Initialize buffer related variables such as counters for number of
//          buffers in tmp file, trigger files and number of points.
//------------------------------------------------------------------------------
int ScopeCurrentBufferInit( void ) {

        // Set initial values

        ScopeInfo.CurrPacketsRead      = 0L;
        ScopeInfo.CurrPacketsProcessed = 0L;

        ScopeInfo.CurrBuffer           = 0L; // total number of buffers
        ScopeInfo.CurrPointsFromStart  = 0L;
        ScopeInfo.CurrBufferInFile     = 0;  // number of buffers in tmp file
        ScopeInfo.CurrTriggerFile      = 0;

        if (ScopeInfo.TriggerType == SRDAT_TRIGGERTYPE_DIGEDGE)
                ScopeInfo.CurrTriggerState = TRIG_STATE_WAITING;
        else
                ScopeInfo.CurrTriggerState = TRIG_STATE_ARMED;

        return( 1 );
}

//------------------------------------------------------------------------------
// ROUTINE: ScopeCurrentBufferIncrement
// PURPOSE: Update buffer related variables to account for another buffer
//          processed.
//------------------------------------------------------------------------------
int ScopeCurrentBufferIncrement( void ) {

        time_t     EbTime;  // this is a long
        double     MicroTime;


        // Increment the number of buffers and points processed.

        ScopeInfo.CurrBuffer++;
        ScopeInfo.CurrPointsFromStart += ScopeInfo.ChBufPts;


        // Roughly update the next buffer time.  For a more accurate
        // value, use the GPS data to compute time.

        ScopeInfo.ElapsedBufferTime   += ScopeInfo.BufferTime;

        EbTime                         = (time_t)ScopeInfo.ElapsedBufferTime;
        MicroTime                      = (ScopeInfo.ElapsedBufferTime - EbTime) *
                                          SR_USPERSEC;
        ScopeInfo.AcquireTimeUsec      = ScopeInfo.StartTimeUsec + (long)MicroTime;

        while (ScopeInfo.AcquireTimeUsec > SR_USPERSEC) {
                ScopeInfo.AcquireTimeUsec -= SR_USPERSEC;
                EbTime++;
                }

        ScopeInfo.AcquireTime          = ScopeInfo.StartTime + EbTime;


        return( 1 );
}


//------------------------------------------------------------------------------
// ROUTINE: GetCurrentSystemTime
// PURPOSE: Get the current system time while protecting scope from OS
//          dependencies.
//------------------------------------------------------------------------------
// FIX - Add this function to SrHelper library ???

#if defined SROS_LINUX
int GetCurrentSystemTime( time_t *CurrTime, long *CurrUsec ) {

        time_t StartTime;

        time( &StartTime );


        // Fill in returns with time_t format and long

        if (CurrTime)  *CurrTime = StartTime;
        if (CurrUsec)  *CurrUsec = 0;

        return( 1 );
}
#else // defined (SROS_WINXP || SROS_WIN2K)
int GetCurrentSystemTime( time_t *CurrTime, long *CurrUsec ) {

        SYSTEMTIME SysTime;
        struct tm  TmTime;
        time_t     UtcSeconds;

        // Get system time in local time (not UTC)

        GetLocalTime( &SysTime );


        // Convert SYSTEMTIME structure to struct tm format

        TmTime.tm_year = SysTime.wYear - 1900;
        TmTime.tm_mon  = SysTime.wMonth - 1;   // systime month from 1, tmtime from 0
        TmTime.tm_mday = SysTime.wDay;
        TmTime.tm_hour = SysTime.wHour;
        TmTime.tm_min  = SysTime.wMinute;
        TmTime.tm_sec  = SysTime.wSecond;



        // Combine time by applying local correction to get UTC epoch seconds

        TmTime.tm_isdst = -1; // Force DST unknown
        UtcSeconds = mktime( &TmTime );


        // Fill in returns with time_t format and long

        if (CurrTime)  *CurrTime = UtcSeconds;
        if (CurrUsec)  *CurrUsec = SysTime.wMilliseconds * 1000;

        return( 1 );
}
#endif
