// FILE: ScopeLib.h
// COPYRIGHT: (c), Symmetric Research, 2009-2010
//
// Include file for scope defines and prototypes common to both the GUI and CMD
// versions for scope.
//


   
#include "SrDefines.h"  // for function type, devhandle, and other common defines
#include "SrUsbXch.h"   // for USBXCH_TOTAL_CHANNELS define, dat header format, etc


#define SCOPE_DEFAULT_EDITOR    "notepad.exe"
#define SCOPE_DEFAULT_INI       "ScopeSetupDefault.ini"

#define INISWITCH_OFF           0
#define INISWITCH_ON            1

#define CONVERT_STR2VAL         0
#define CONVERT_VAL2STR         1

#define HTITLEMARGIN            128 // FIX - duplicated from scopegui.cpp for now

#define CS_MONO                 0
#define CS_GREEN                1
#define CS_BLUE                 2
#define CS_BROWN                3

#define TRIG_STATE_NONE         0       // TriggerState values ... 
#define TRIG_STATE_ARMED        1
#define TRIG_STATE_INPROGRESS   2
#define TRIG_STATE_WAITING      3

#define CH_TYPE_NONE            0
#define CH_TYPE_ANALOG          1
#define CH_TYPE_DIGITAL         2
#define CH_TYPE_GPSMARK         3
#define CH_TYPE_GPSCOUNTER      4

#define MSGLEVEL_MSG            0
#define MSGLEVEL_WARN           1
#define MSGLEVEL_ERR            2

#define MSGDLG_NO_SHOW          0
#define MSGDLG_SHOW             1

#define ERROR_NOT_FATAL         0
#define ERROR_FATAL             1
#define ERROR_AUTO_FATAL        2
#define ERROR_MAYBE_FATAL       3


#define CONVERTTOOFFSET         0x00800000 // Assumes data is USBXCH_DF_SIGNED
#define CLEARTOPBIT             0x7FFFFFFF
#define CLEARTOPBYTE            0x00FFFFFF



#define MAXTITLE                12
#define MAXSTR                  256
#define MAXDIGITAL              4                 // USBXCH_MAX_DIGPLOT_CHANNELS

#define MAX_TRIGGER_THRESHOLD   8388608L          // half of full scale 24 bits
#define MAX_TRIGGER_DIGIO_LINE  3                 // I/O line 3



#define INIMAXDIGITAL          USBXCH_MAX_DIGPLOT_CHANNELS
#define INIMAXCHANNEL         (USBXCH_MAX_ANALOG_CHANNELS  +  \
                               USBXCH_MAX_DIGITAL_CHANNELS +  \
                               USBXCH_MAX_GPS_CHANNELS +      \
                               USBXCH_MAX_COUNTER_CHANNELS)


#define HDR_TOTAL_CHANNELS    11 //FIX - compute this
#define HDR_PRODUCT_STRING   "SR USBxCH"





// Type for pointer to log message function

typedef void (*SRMSGFNPTR)( int msglevel, int dlgshow, int fatal, char *msg );


typedef struct ScopeVariables {

        // Max and mins

        int     MaxSample;
        int     MinAmpShift;
        int     MaxAmpShift;
        long    MaxTriggerThreshold;
        int     MaxTriggerDigioLine;
        float   MinSPS;
        float   MaxSPS;


        // Defaults

        int     DefaultScreenWidth;
        int     DefaultScreenHeight;
        int     DefaultColor;
        int     DefaultThreshold;
        int     DefaultNdisplay;
        int     DefaultNsamples;
        int     DefaultNbuffers;
        char    DefaultDir[MAXSTR];


        // Program parameters

        int     ProgramAllowMultiple;
        int     ProgramKeypressToRun;
        int     ProgramInteractive;
        int     ProgramExitOnError;
        int     ProgramShowStatusBar;
        int     ProgramDisplayTrace;
        int     ProgramDisplayText;
        int     ProgramDisplayInfo;
        char    ProgramName[MAXSTR];

// FIX - should these be ProgramXxx or PlotXxx or DisplayXxx variables
        
        int     DisplaySerial;
        int     DisplayEquip;


        // Atod hardware parameters

        double  AtodXtalSpeed;
        char    AtodDriverName[MAXSTR];
        char    AtodXchModelName[MAXSTR];
        int     AtodXchModel;
        int     AtodAnalogChannels;
        int     AtodDigitalChannels;
        int     AtodGpsChannels;
        int     AtodCounterChannels;
        int     AtodTotalChannels;
        int     AtodAnaPlotChannels;
        int     AtodDigPlotChannels;
        int     AtodGpsPlotChannels;
        int     AtodCntPlotChannels;
        int     AtodTotalPlotChannels;
        int     AtodMinTotalChannels;


        // Acquisition parameters

        int     AcqToggleLed;
//FIX is AcqSpsIndex still needed or useful?
        int     AcqSpsIndex;
        double  AcqSamplingRate;
        double  AcqSamplingRateTrue;


        // Gps parameters

        int     GpsModel;
       char     GpsModelName[MAXTITLE];

        
        // Channel and digital parameters

        int     Channels;
        int     ChannelsPlot;
        
        int     ChannelType[INIMAXCHANNEL];
        int     ChannelTrigger[INIMAXCHANNEL];
        char    ChannelTitle[INIMAXCHANNEL][MAXTITLE];

        char    DigitalTitle[MAXDIGITAL][MAXTITLE];
        
        int     GpsMarkIndex;
        int     DigitalIndex;
        int     CounterIndex;


        // Output file parameters

         int    OutputFileNaming;
        long    OutputFileNbuffers;
        char    OutputFileDirectory[MAXSTR];
         int    OutputFileSeqStart;


        // Plot parameters

        int    PlotNdisplay;
        int    PlotFirstSample;
        int    PlotFirstTrace;
        int    PlotColorScheme;
        int    PlotClipping;
        int    PlotShiftBy;
        int    PlotAutoDC;


        // Trigger parameters

        int     TriggerType;
        long    TriggerThreshold;
        int     TriggerDigitalLevelLine;
        int     TriggerDigitalEdgeLine;
        int     TriggerActiveLevel;
        int     TriggerActiveEdge;
        int     TriggerActionSave;
        int     TriggerFilesToSave;
        int     TriggerActionSignal;
        int     TriggerLineToSignal;
        int     TriggerLatchLine;


        // XCR UserCfg user configuration parameters

        int UserCfgNmeaTxPolarity;
        int UserCfgNmeaRxPolarity;
        int UserCfgPpsLed;
        int UserCfgPpsPolarity;
        int UserCfgAdrdy;
        int UserCfgNotUsed1;
        int UserCfgTriggerPolarity;
        

        // Other parameters

        int       Npackets;
        int       FirstPacket;
        int       LastMsg;
        
        int       ChBufPts;
        int       DataPtBytes;

        int       LedValueYellow;
        int       LedValueRed;
        
        int       NumGpsSat;

        DEVHANDLE UsbXchHandle;


        
        // Current run parameters

        unsigned long CurrPacketsRead;
        unsigned long CurrPacketsProcessed;
        unsigned long CurrBuffer;
        unsigned long CurrPointsFromStart;
        int           CurrBufferInFile;
        int           CurrTriggerFile;
        int           CurrTriggerState;

        
        // Time parameters

        int     TimeSource;
        time_t  StartTime;
        long    StartTimeUsec;
        time_t  AcquireTime;
        long    AcquireTimeUsec;
        time_t  ReadyTime;
        double  ElapsedBufferTime;


        // Computed parameters
        
       unsigned int  Nvalues;
       int           AcqBufBytes;
       double        BufferTime;
       int           SleepMs;
       int           TriggerTimeToSave;
       unsigned long OutputFileSize;
       unsigned int  NgpsValues;
       int           SerBufBytes;



       
        // File names

        char IniFileName[MAXSTR];
        char RptFileName[MAXSTR];
        char LogFileName[MAXSTR];
        char TmpFileName[MAXSTR];
        char DataFileName[MAXSTR];

        char IniFileNameShort[MAXSTR];
        char DataFileNameShort[MAXSTR];

        
        // Other strings
        
        char IniTitle[MAXSTR];
        char EditorName[MAXSTR];

        
        // Data buffer parameters

        SRDAT_USBXCH_ANALOG *UsbXchAnalogStruct;
        long *UsbXchAnalogBuffer;
        long *UsbXchAnalogBufferEnd;
        SRDAT_USBXCH_SERIAL *UsbXchSerialStruct;
        char *UsbXchSerialBuffer;
        char *UsbXchSerialBufferEnd;
        SRDAT_USBXCH_EQUIPMENT *UsbXchEquipStruct;
        SRUSBXCH_PACKET *UsbXchPacketArray;


        
        // Next and total point count parameters
        
//FIX - sample vs point ???        
        // Sample is index for all info acquired at one instant in time
        // Pt is used for index of each a/d value at a channel in that sample

        int     NextPtAnalog;  // was NextSample
        int     NextPtSerial;
        int     NextPtEquip;
        
        int     TotalSample;
        int     TotalPtEquip;
        int     TotalPtSerial;       // Number of GPS data read while waiting
                                     // for 1 analog buffer to fill

        int     NmeaMsgByte;
        char    SerialBytes[65];

        char       LogMessage[MAXSTR];
        SRMSGFNPTR ShowLogMessage;

        } SCOPE_VARIABLES;




// Global variables in library

extern SCOPE_VARIABLES ScopeInfo;
extern SCOPE_VARIABLES ScopeInfoSave;
extern SRDAT_HDR       DatHdr;    // See SrUsbXch.h for layout




// Function prototypes

char *Chop( char *s );

void  ScopeLibRev( void );

void  ScopeDisplayInfo( char *info );
void  ScopeDisplayInfoFlush( char *info );

int   ScopeToggleSwitch( int *Switch );
int   ScopeStringMatch( char *s, char *t );

char *ScopeMakeOutputFileNamingString( int OutputFileNamingValue );
char *ScopeMakeOnOffString( int  OnOffValue );
int   ScopeMakeOnOffValue( char *OnOffString );
int   ScopeMakeTriggerActiveLevelValue( char *TriggerActiveLevelString );
int   ScopeMakeTriggerActiveEdgeValue( char *TriggerActiveEdgeString );
int   ScopeMakeOutputFileNamingValue( char *OutputFileNamingString );
int   ScopeMakeTriggerTypeValue( char *TriggerTypeString );

int   ScopeMakeDirYmdHms( void );

int   ScopeInitScopeInfo( int ScreenWidth, int ScreenHeight, char *CurrDir,
                          SRMSGFNPTR MessageFunc );
void  ScopeInitTimes( time_t StartTime, long StartUsec );
int   ScopeInitFileNames( void );
int   ScopeInitActiveChannels( void );

//FIX
int   ScopeInit( const char *IniFilename );

void  ScopeIniSetDefaults( void );
int   ScopeIniVerifyData( void );
void  ScopeIniConsistencyCheck( void );
void  ScopeIniCompute( void );
int   ScopeIniCheck( const char *IniFilename );
int   ScopeIniCreateDefault( const char *IniFilename );
int   ScopeIniWriteKeyword( FILE *OutFp );
int   ScopeIniParse( const char *IniFilename );
void  ScopeIniReport( void );

void  ScopeHdrDefaults( void );
int   ScopeHdrWriteFile( void );
int   ScopeHdrUpdateWrite(  FILE *fp );

char *ScopeTmpFileComputeFinalName( void );
FILE *ScopeTmpFileOpen( void );
int   ScopeTmpFileClose( FILE *fp );
int   ScopeTmpFileRename( char *FinalName );
int   ScopeTmpFileDelete( void );

int   ScopeWriteData( void );
int   ScopeSrOutWrite( FILE *fp, long *buffer, unsigned int nbytes );

int   ScopeTriggerThresholdSetRange( long Channel, long *ThMin, long *ThMax );
int   ScopeTriggerCheckAnalog( int Channel );
int   ScopeTriggerDigitalCompare( int Point, long TriggerMask );
int   ScopeTriggerThresholdCompare( int Point, long ThMin, long ThMax );
int   ScopeTriggerDigioCompare( int Line, int ActiveState );
int   ScopeTriggerCondition( void );
int   ScopeTriggerAction( void );
int   ScopeTriggerDone( void );
int   ScopeTriggerClearLatch( void );

int   ScopeAtodInit( void );
int   ScopeAtodStart( void );
int   ScopeAtodStop( void );
int   ScopeAtodGetReadyData( void );
int   ScopeAtodDataPrepareToRead( void );
int   ScopeAtodCheckOverflow( void );
int   ScopeAtodCheckError( int ErrorCode, char *FuncName );

int   ScopeProcessAnalog( unsigned long Nsamples );
int   ScopeProcessSerial( int Nserial );
int   ScopeProcessEquip( int Nequip );

int   ScopeBuffersUpdateSize( void );

int   ScopeBuffersAllocate(
                SRUSBXCH_PACKET **UsbXchPacketArray,
                SRDAT_USBXCH_ANALOG **UsbXchAnalogStruct,
                long **UsbXchAnalogBuffer,
                long **UsbXchAnalogBufferEnd,
                SRDAT_USBXCH_SERIAL **UsbXchSerialStruct,
                char **UsbXchSerialBuffer,
                char **UsbXchSerialBufferEnd,
                SRDAT_USBXCH_EQUIPMENT **UsbXchEquipStruct
               );
int   ScopeBuffersFree(
                SRUSBXCH_PACKET **UsbXchPacketArray,
                SRDAT_USBXCH_ANALOG **UsbXchAnalogStruct,
                long **UsbXchAnalogBuffer,
                long **UsbXchAnalogBufferEnd,
                SRDAT_USBXCH_SERIAL **UsbXchSerialStruct,
                char **UsbXchSerialBuffer,
                char **UsbXchSerialBufferEnd,
                SRDAT_USBXCH_EQUIPMENT **UsbXchEquipStruct
               );


int   ScopeSetChannelCountByModel( void );
int   ScopeUpdateGpsModel( int AllowError );


int   ScopeTwiddleYellowLed( void );
int   ScopeSpsUpdateFromIndex( void );
int   ScopeSpsGetString( int SpsIndex, char *SpsString, int Nchar );
int   ScopeCreateShortFileName( char *LongName, char *ShortName );
int   ScopeComputeActiveChannels( void );
void  ScopeOverflowRecovery( void );
int   ScopeAtodPcTimeStamp( void );

int   ScopeCurrentBufferInit( void );
int   ScopeCurrentBufferIncrement( void );
void  ScopePtCountersInitNext( void );
void  ScopePtCountersInitTotal( void );

void  ScopeLogMessagePrep( int MsgLevel, const char *ExtraMsg );

int   GetCurrentSystemTime( time_t *CurrTime, long *CurrUsec );
