// FILE: DevMan.c
// COPYRIGHT: (c), Symmetric Research, 2009-2010
//
// This utility program controls kernel mode device driver for the USBxCH.
//


#include <stdio.h>
#include <stdlib.h>
#include <string.h>      // for memset

#include "SrUsbXchDrv.h"

// Defines and globals

#define CMD_NONE        0
#define CMD_SHOW        1
#define CMD_MAX         9

#define VALUE_NO        0
#define VALUE_YES       1
#define VALUE_UNKNOWN   2

#define MAXSTR          256

#define NAME_UNKNOWN    "Unknown"

char ErrorMessage[MAXSTR];



// DevMan RUN TIME PARAMETERS:
//
// Both command line and fixed hardwired parameters are located here.
//    ... access this structure with statments like: DP.Command ...
//
//

struct DevManParametersLayout {

        int Command;
        int DevIndex;
        int LastError;
        int Debug;
        int DevId;
        char DevInstanceId[MAXSTR];

        } DevManParameters = {


                // Argument values:

                CMD_NONE,           // Command, default = none
                -1,                 // DevIndex, default = -1 (no selection)
                0,                  // LastError, default = success
                0,                  // Debug, default = 0 (no)
                -1,                 // DevId, default = -1 (unknown value)


                // character strings area ...
                
                };


#define DP  DevManParameters  // < DP = shorthand for DevManParameters ...


#define TYPE_REG_UNKNOWN 0
#define TYPE_REG_DWORD   1
#define TYPE_REG_SZ      2



// USAGE MESSAGE ARRAY:

char *UsageMsg[] = {
                
    "Controls the USBxCH kernel mode device driver                                 ",
    "                                                                              ",
    "USAGE: DevMan [showall] [/?]                                                  ",
    "                                                                              ",
    "    showall = displays info about known USBxCH devices                        ",
    "         /? = show help screen and quit                                       ",
    "                                                                              ",
    " Examples:                                                                    ",
    "                                                                              ",
    "    DevMan /?           < show help screen and quit                           ",
    "    DevMan showall      < show device info and status for USBxCH driver       ",
    "                                                                              ",
   NULL
   };

//Protos
  int  DoCmdShow( void );
  int  GetYorN( char *Prompt );
  void ErrorCmd( char *msg );
  void ErrorExec( char *msg );
  void ProcessCommandLine( int argc, char *argv[] );
  void PrintMessage( char *Message[] );
  void Usage( void );
   int StrEqu( char *s1, char *s2 );



// Main program

int main( int argc, char** argv ) {

	int Ok;


        // Print banner

        printf( "USBxCH DevMan : (c) SR 2010 : Rev {2010/09/05}\n\n" );


        // Process command line


        ProcessCommandLine( argc, argv );


        
        // Perform requested command

        if ( DP.Command == CMD_SHOW )
                Ok = DoCmdShow();

	else
		printf( "This command not supported under Linux\n" );



        exit( 0 );
}

//------------------------------------------------------------------------------
// ROUTINE: DoCmdShow
// PURPOSE: Show info about the SrInstrumentation/USB driver and devices.
//------------------------------------------------------------------------------
int DoCmdShow( void ) {

	int   i, c, ns, Nworking, SysDeviceNum[MAX_DEVICES];
	char  DriverName[MAXSTR], DeviceNameBuffer[MAXSTR],DeviceNumBuffer[MAXSTR];
	char *SysDeviceMap, *InstallStr;
	FILE *SysFp;

        printf( "\nSrUsbXch Devices Listed in SysFS:\n\n" );
        printf( "  S#  Port Id   Driver Name (DID)  Device Name\n" );


	// Initialize device map variables

	memset( DeviceNameBuffer, 0, MAXSTR );
	memset(	DeviceNumBuffer, 0, MAXSTR );

	for ( i = 0 ; i < MAX_DEVICES ; i++ )
		SysDeviceNum[i] = -1;


	// Open sysfs devicemap file prepared by the driver.

	SysDeviceMap = "/sys/bus/usb/drivers/SrUsbXch/devicemap";

	SysFp = fopen( SysDeviceMap, "r" );
	if ( SysFp ) {

		InstallStr = "YES";

		// Read devicemap contents from file into buffer

		i = 0;
		c = fgetc( SysFp );
		while ( c != EOF  &&  c != '\n' ) {
			DeviceNumBuffer[i++] = c;
			c = fgetc( SysFp );
			}

		fclose( SysFp );

		
		// Scanf buffer to get matching device numbers

		//                              0  1  2  3  4  5  6  7  8  9 
		ns = sscanf( DeviceNumBuffer, "%d %d %d %d %d %d %d %d %d %d",
			     &SysDeviceNum[0],
			     &SysDeviceNum[1],
			     &SysDeviceNum[2],
			     &SysDeviceNum[3],
			     &SysDeviceNum[4],
			     &SysDeviceNum[5],
			     &SysDeviceNum[6],
			     &SysDeviceNum[7],
			     &SysDeviceNum[8],
			     &SysDeviceNum[9]
			   );

		if ( ns != 10 )
			printf("Only read %d device numbers\n", ns );
		}
	else {
		InstallStr = "NO";
		}




	// Display info about existing devices

	Nworking = 0;
	
	for ( i = 0 ; i < MAX_DEVICES ; i++ ) {

		if ( SysDeviceNum[i] > -1 ) {

			Nworking++;

			printf( "  %d   ???????   %s%d     %d    /dev/%s%d\n",
				  Nworking,
				  SRUSBXCH_DRIVER_NAME, i, i,
				  SRUSBXCH_DEVICE_NAME, SysDeviceNum[i] );
			}
		} // end for i

        

        // Show device summary

        printf( "\n\nSrUsbXch Device Summary:\n" );
        printf( "   Working    =  %d\n",  Nworking );



        // Show driver summary info


	printf( "\n\nSrUsbXch Driver Summary:\n" );
	printf( "   Installed  = %s\n", InstallStr );

	printf( "\n" );
        

        return( 1 );
}


//------------------------------------------------------------------------------
// ROUTINE: GetYorN
// PURPOSE: Read a single character and determine if it is a yes or no response.
//          Returns 0 for no, 1 for yes, and 2 for unknown response.
//------------------------------------------------------------------------------
int GetYorN( char *Prompt ) {

        int c, c2, Answer;

        printf( "%s (y or n) ?\n", Prompt );

	// Get answer character
	
        c = getchar();
        if (c == 0xD) printf("\n");  // add newline if char was ENTER
        if (c == 0x00 || c == 0xE0 ) // deal with extended key (fn,->)
                c = getchar();        // c has scan code


	// If found a character, flush any remaining characters (eg ENTER)

	c2 = c;
	while ( c2 != EOF )
		c2 = getchar();
	

	// Determine what answer was

        if ( c == 'n' || c == 'N' )
                Answer = VALUE_NO;
        else if ( c == 'y' || c == 'Y' )
                Answer = VALUE_YES;
        else
                Answer = VALUE_UNKNOWN;

        return( Answer );
}

//------------------------------------------------------------------------------
// ROUTINE: ErrorCmd
// PURPOSE: Print a command line error message and quit.
//------------------------------------------------------------------------------
void ErrorCmd( char *msg ) {

        // Print a command line error message and quit.

        printf(

               "ERROR: Command line usage : %s \n\n ... DevMan /? for help\n",
               msg

               );

        exit( 1 );  // < quit at errorlevel 1.

}

//------------------------------------------------------------------------------
// ROUTINE: ErrorExec
// PURPOSE: An execution error occured.
//------------------------------------------------------------------------------
void ErrorExec( char *msg ) {

        // A fatal execution error occured.

        printf( "\nERROR: %s \n\n", msg );

        exit( 2 );  // < quit at errorlevel 2.

}

//------------------------------------------------------------------------------
// ROUTINE: ProcessCommandLine
// PURPOSE: Gather the command line arguments, filling in PP.
//------------------------------------------------------------------------------
void ProcessCommandLine( int argc, char *argv[] ) {

        int i, drvnum;


        // No command line arguments gives the usage message.

        if ( argc == 1 )

                Usage( );


        // Check for too many arguments.

        if ( argc > 4 )

                ErrorCmd( "too many command line arguments" );


        // Scan for valid arguments.

        for ( i = 1 ; i < argc ; i++ ) {


                // Check for usage message request.

                if ( StrEqu( argv[i], "/?") )

                        Usage( );


                // Check for the show command

                else if ( StrEqu( argv[i], "showall" ) ) {

                        if ( DP.Command == CMD_NONE )
                                DP.Command = CMD_SHOW;

                        else
                                ErrorCmd( "only one command is allowed" );

                        }


                // Check for driver index selection.

                else if ( sscanf( argv[i], "%d", &drvnum ) == 1 ) {

                        if ( DP.DevIndex == -1 )
                                DP.DevIndex = drvnum;

                        else
                                ErrorCmd( "only one driver index selection is allowed" );
                        }


                // If not one of the above, it is unknown

                else {
                        ErrorCmd( "unknown argument" );
                        }

                }

        if ( DP.Command == CMD_NONE )
                ErrorCmd( "at least one command must be given" );
        
}

//------------------------------------------------------------------------------
// ROUTINE: PrintMessage
// PURPOSE: Display all lines in the specified message.  Message must be
//          terminated with a NULL line.
//------------------------------------------------------------------------------
void PrintMessage( char *Message[] ) {

        int i;

        // Error check

        if ( !Message )
                return;


        // Print the specified message.

        for ( i = 0 ; Message[i] ; i++ )

                printf( "%s\n", Message[i] );

        return;
}

//------------------------------------------------------------------------------
// ROUTINE: Usage
// PURPOSE: Display the usage message and exit.
//------------------------------------------------------------------------------
void Usage( void ) {

        // Print the usage message.

        PrintMessage( UsageMsg );

        exit( 0 );  // < quit at errorlevel 0.
}

//------------------------------------------------------------------------------
// ROUTINE: StrEqu
// PURPOSE: Compare for exact equality between two strings.
//------------------------------------------------------------------------------
int StrEqu( char *s1, char *s2 ) {

        while ( *s1 && *s2 && (*s1 == *s2) )

                { s1++; s2++; }

        return ( (*s1 == '\0') && (*s2 == '\0') ? 1 : 0 );

}

