// FILE: Diag.c
// COPYRIGHT: (c), Symmetric Research, 2009-2010
//
// This file is for performing USB diagnostics.  It will start simple
// and include more tests as they are developed.
//
// To Do:
//  - Update TestData so it handles serial and equip data properly
//
//

#if defined( SROS_WIN2K ) || defined( SROS_WINXP )
#include <windows.h>     // windef.h -> winnt.h -> defines GENERIC_READ etc
#include <winioctl.h>    // device driver I/O control macros
#include <io.h>          // access function
#include <cfgmgr32.h>    // configuration manager functions for plug n play
#endif

#include <stdio.h>
#include <stdlib.h>
#include <time.h>        // time function
#include <string.h>      // strupr, strcpy, etc

#include "SrDefines.h"   // for function type, dev handle, 64 bit, + other defines
#include "SrHelper.h"    // for SrStringCopy, SrSleep, SrGetKeyWait, SrFlushPrint, etc
#include "SrUsbXch.h"    // for SrUsbXch library functions like SrUsbXchConvertPackets, etc
#include "SrUsbXchDrv.h" // for SrUsbXchDrv defines and structures like USB_DEVICE_DESCRIPTOR

#if defined( SROS_WIN2K ) || defined( SROS_WINXP )

#include <PSHPACK1.H> // required for close packing of USB structures
struct _USB_DEVICE_DESCRIPTOR {
        unsigned char bLength;
        unsigned char bDescriptorType;
        unsigned short bcdUSB;
        unsigned char bDeviceClass;
        unsigned char bDeviceSubClass;
        unsigned char bDeviceProtocol;
        unsigned char bMaxPacketSize0;
        unsigned short idVendor;
        unsigned short idProduct;
        unsigned short bcdDevice;
        unsigned char iManufacturer;
        unsigned char iProduct;
        unsigned char iSerialNumber;
        unsigned char bNumConfigurations;
};
#include <POPPACK.H> // Undoes PSHPACK1.H

typedef struct _USB_DEVICE_DESCRIPTOR         USB_DEVICE_DESCRIPTOR;

#define USB_DEVICE_DESCRIPTOR_TYPE                0x01
#define USB_CONFIGURATION_DESCRIPTOR_TYPE         0x02
#define USB_STRING_DESCRIPTOR_TYPE                0x03
#define USB_INTERFACE_DESCRIPTOR_TYPE             0x04
#define USB_ENDPOINT_DESCRIPTOR_TYPE              0x05

#else // if defined( SROS_LINUX )
#include <linux/types.h>        /* __u8 etc */

struct usb_device_descriptor {
        __u8  bLength;
        __u8  bDescriptorType;

        __le16 bcdUSB;
        __u8  bDeviceClass;
        __u8  bDeviceSubClass;
        __u8  bDeviceProtocol;
        __u8  bMaxPacketSize0;
        __le16 idVendor;
        __le16 idProduct;
        __le16 bcdDevice;
        __u8  iManufacturer;
        __u8  iProduct;
        __u8  iSerialNumber;
        __u8  bNumConfigurations;
} __attribute__ ((packed));

typedef struct usb_device_descriptor         USB_DEVICE_DESCRIPTOR;

#define USB_DEVICE_DESCRIPTOR_TYPE           0x01 // USB_DT_DEVICE
#define USB_CONFIGURATION_DESCRIPTOR_TYPE    0x02 // USB_DT_CONFIG
#define USB_STRING_DESCRIPTOR_TYPE           0x03 // USB_DT_STRING
#define USB_INTERFACE_DESCRIPTOR_TYPE        0x04 // USB_DT_INTERFACE
#define USB_ENDPOINT_DESCRIPTOR_TYPE         0x05 // USB_DT_ENDPOINT

#endif // SROS_xxx


#if defined( SROS_WIN2K ) || defined( SROS_WINXP )
#define SRDIAGDIR       "\\SR\\USBXCH\\Utilities\\Diag\\"
#else // if defined( SROS_LINUX )
#define SRDIAGDIR       "/usr/local/SR/USBXCH/Utilities/Diag/"
#endif // SROS_xxx


//#include "SrUsbXchDiag.hw"    // for USBXCH encoded hardware info
extern char *SrUsbXch_HwData[];


// Program copyright and screen title:

char *Copyright =

          "\n\nDiag.exe : Copyright (c), Symmetric Research, 2009-2010\n\n";

char *ScreenTitle =

      "USBxCH Diagnostic Program diag [/?]: (c) SR 2009 : %s\n"; // DIAG_REV


// Defines

#define DIAG_REV       "2010/08/25"


// Maximum FILENAME and generic string lengths:
//
// A filename is comprised of a path, basename, and extension.  The
// path and extension are optional.  In the string:
//
//   c:\directory\foo.exe
//
// the leading c:\directory\ is the path, foo is the basename, and exe
// is the extension.  In Windows, the last '\' denotes the end of the
// path, and the last '.' denotes the beginning of the extension.
//
// A complete filename must be less than MAXFILENAME characters long.
// Even if buffers for the path, basename, and ext are all individually
// dimensioned at MAXFILENAME, the total filename constructed from their
// parts must still be less than MAXFILENAME.
//
//

#define MAXFILENAME             1024
#define MAXSTR                  256

#define MASKVAL        0x00FFFFFF
#define NCHANNELS      4
#define NTIMEPTS       130                        // 1 second at 130 sps (130/2 packets)
#define MAXSERIAL      4                          // 2 str/set * 2 pack/str
#define MAX_PACKET_REQ (NTIMEPTS+(3*MAXSERIAL))   // enough packets for 2 sec
                                                  // of analog + 3 sets of nmea
#define MAXANALOG      MAX_PACKET_REQ*2           // more than we need, but lets us be
                                                  // sure to process all packets at 2 analog/packet

#define MAX_DIGVAL       10                       // For TestDigIn and TestDigOut
#define MAXTESTNMEABYTES (SRDAT_USBPACKET_SIZE+1) // For TestSerial

#define DRAM_TEST_SIZE1 0x40                      // For TestDramSingleWR only
#define DRAM_TEST_SIZE  0x4000                    // For all other Dram tests
#define DRAM_TEST_MAX   (DRAM_TEST_SIZE+1)

// Simple version for initial tests
//#define DRAM_NFILL         (400-1)
//#define DRAM_ARRAY_SIZE    (20*4)

// Full version
#define DRAM_NFILL       4194303 // 2^22 - 1
#define DRAM_ARRAY_SIZE  (65536*4)



#define PARTFULL_TEST_SIZE 80
#define EMPTY_TEST_SIZE 6

#define MAX_RESET_OPEN_TRIES    50                // For TestReset

#define IGNORE_EMPTY -1
#define    NOT_EMPTY  0
#define     IS_EMPTY  1

#define IGNORE_OVERFLOW -1
#define    NOT_OVERFLOW  0
#define     IS_OVERFLOW  1

#define IGNORE_PARTFULL -1
#define    NOT_PARTFULL  0
#define     IS_PARTFULL  1

#define READALL_REQUEST_NONE    0x00
#define READALL_REQUEST_ANALOG  0x01
#define READALL_REQUEST_SERIAL  0x02
#define READALL_REQUEST_EQUIP   0x04
#define READALL_REQUEST_ALL     0x07


#define BEEP_COUNT     1


// Sample rate index defines

#define USBXCH_SPS_INDEX_39000  15
#define USBXCH_SPS_INDEX_9765   13
#define USBXCH_SPS_INDEX_4883   12
#define USBXCH_SPS_INDEX_2600   11
#define USBXCH_SPS_INDEX_1300   10
#define USBXCH_SPS_INDEX_650     9
#define USBXCH_SPS_INDEX_130     8

// FIX - This should be some USBXCH_PIPE define in include perhaps?

#define USBXCH_PIPE_MIN          0
#define USBXCH_PIPE_MAX          5


// These MUST match SIE values and those in SrUsbXch.c
#define  USB8051_CPU_RUN    0
#define  USB8051_CPU_HOLD   1

// Defines for display helper function

#define PS_NONE     0x00
#define PS_DATE     0x01
#define PS_FLUSH    0x02
#define PS_LOG      0x10
#define PS_REPORT   0x20
#define PS_DISPLAY  0x40

#define PrintStr( msg )           PrintString( msg, PS_LOG | PS_REPORT | PS_DISPLAY )
#define PrintStrDate( msg )       PrintString( msg, PS_LOG | PS_REPORT | PS_DISPLAY | PS_DATE )
#define PrintStrDateFlush( msg )  PrintString( msg, PS_LOG | PS_REPORT | PS_DISPLAY | PS_DATE | PS_FLUSH )
#define PrintStrNoDisplay( msg )  PrintString( msg, PS_LOG )
#define PrintStrRpt( msg )        PrintString( msg, PS_REPORT | PS_DISPLAY )
#define PrintStrRptFlush( msg )   PrintString( msg, PS_REPORT | PS_DISPLAY | PS_FLUSH )

#define LogFileBaseExt  "300 Diag History.log"
#define RptFileBaseExt  "301 Diag Last Run.rpt"



// DIAG RUN TIME PARAMETERS:
//
// Both the user specified command line parameters and fixed hardwired
// parameters are allocated here.  Access this structure with statments
// like: DP.ShowHidden.
//

struct DiagParametersLayout {

        struct TimeTableRecord *pTimeSourceSelected;
        struct DiagTestOption  *pDiagTestSelected;
        
        DEVHANDLE UsbXchHandle;
          int LibRev;
          int ShowHidden;
          int UserCfgByte;
          int AtodGpsModel;
        char *AtodGpsModelName;
          int AtodXchModel;
        char *AtodXchModelName;

        FILE *RptFp;
          
         char AtodDriverName[MAXSTR];
         char CmdLine[MAXSTR];
         char Msg[MAXSTR];
         char ReadData[DRAM_ARRAY_SIZE];  // only used in TestOverflow
         char WriteData[DRAM_ARRAY_SIZE]; // only used in TestOverflow
SRDAT_USBXCH_ANALOG    UsbxchAnalogData[MAXANALOG];
SRDAT_USBXCH_SERIAL    UsbxchSerialData[MAXSERIAL];
SRDAT_USBXCH_EQUIPMENT UsbxchEquipData[MAXSERIAL]; // equip data included with serial

        char FilePathOnly[MAXFILENAME];
        char ExeFullPathName[MAXFILENAME];      // generic full path work area ...
        char LogFullPathName[MAXFILENAME];
        char RptFullPathName[MAXFILENAME];

        char *pFullPathNameBase;                // -> base+ext part of FullPathName[] ...

        } DiagParameters = {


                // default command line parameter values:
                //
                //      ShowHidden = 0      -> do not show special debugging options
                //     UserCfgByte = 0x00   -> standard GPS config
                //              gn = g2     -> Garmin
                //
                //   ( see SrUsbXch.h for xxxMODEL defines )
                //

                NULL,                    // optional "gn" time table record
                NULL,                    // optional diag test option
                BAD_DEVHANDLE,           // default DEVHANDLE
                0,                       // default LibRev
                0,                       // default ShowHidden
                USERCFG_DEFAULT_BYTE,    // default UserCfgByte
                SRDAT_GPSMODEL_PCTIME,   // default "gn" AtodGpsModel
                SRDAT_GPSNAME_PCTIME,    // AtodGpsModelName;

                SRDAT_ATODMODEL_USB4CH,  // fixed AtodXchModel
                SRDAT_ATODNAME_USB4CH,   // AtodXchModelName;

                NULL,                    // RptFp


                // all trailing array and filename areas
                // initialized to 0 by compiler ...

                };

#define DP  DiagParameters  // DP = shorthand for DiagTime parameter struct



// Globals


char *TrueFalse[] = { "FALSE", "TRUE "};




// Usage message array

char *UsageMsg[] = {

    "Run diagnostic tests to verify USBxCH operation                            ",
    "                                                                           ",
    "USAGE: Diag test [gn] [driver] [/?]                                        ",
    "                                                                           ",
    "        test = test to run (see table below)                               ",
    "          gn = time source (see table below)                               ",
    "      driver = SrUsbXch# where # is 0 to 9    << default = SrUsbXch0       ",
    "          /? = show help screen and quit                                   ",
    "                                                                           ",
    "  Test table: test selects the action ...                                  ",
    "                                                                           ",
    NULL,
//  "   Test ->  DESC                                                           ",
    "  %-9s = %s\n",
    "                                                                           ",
    "  Time source table: gn selects the time source ...                        ",
    "                                                                           ",
    NULL,
//  "    GG   ->  DESC                                                          ",
    "    %s   ->  %s\n",
    "                                                                           ",
    "  Log and report files ->                                                  ",
    "                                                                           ",
    "  \"300 Diag History.log\" lists all diag runs                               ",
    "  \"301 Diag Last Run.rpt\" saves screen text from the last run              ",
    "                                                                           ",
    "                                                                           ",
    "Examples:                                                                  ",
    "                                                                           ",
    "    Diag /?             < show help screen and quit                        ",
    "    Diag install        < run recommended tests to check installation      ",
    "    Diag nmea g2        < check NMEA strings from Garmin GPS               ",
    "    Diag led SrUsbXch1  < test yellow LED on USBxCH unit 1                 ",
    "                                                                           ",

    NULL
    };



// TIME SOURCE TABLE:
//
// These are the possible time sources for Diag: PC Time, GPS Time ...
//
//

struct TimeTableRecord {

        char  *GnString;        // < allowed command line "gn" string
        int   AtodGpsModel;     // < GPS model, NONE, PCTIME, GARMIN etc
        char  *UsageDesc;       // < usage message description

        } TimeTable[] = {

          { "g1",  SRDAT_GPSMODEL_PCTIME  , "PCTIME   ( NMEA = ZDA )      << default" },
          { "g2",  SRDAT_GPSMODEL_GARMIN  , "GARMIN   ( NMEA = RMC, GGA )           " },
          { "g3",  SRDAT_GPSMODEL_TRIMBLE , "TRIMBLE  ( NMEA = ZDA, GGA )           " },

          // end of table ...

          { NULL, 0, NULL },

          };


#define TR  struct TimeTableRecord  // TR = shorthand for time source record


// TEST OPTIONS STRUCTURE AND GLOBAL VARIABLES:
//
//
//

typedef int (TESTFN)( int );    // test function type

int  TestInstallation( int OpenClose );
int  TestDebug( int OpenClose );
int  TestDevice( int OpenClose );
int  TestPower( int OpenClose );
int  TestLed( int OpenClose );
int  TestAnalog( int OpenClose );
int  TestNmea( int OpenClose );
int  TestEquip( int OpenClose );
int  TestDigIn( int OpenClose );
int  TestDigOut( int OpenClose );
int  TestDram( int OpenClose );
int  TestDramSlam( int OpenClose );
int  TestDramSlide( int OpenClose );
int  TestDramInc( int OpenClose );
int  TestDramVariableWR( int OpenClose );
int  TestDramSingleWR( int OpenClose );
int  TestDramAllWR( int OpenClose );
int  TestEmpty( int OpenClose );
int  TestPartFull( int OpenClose );
int  TestOverflow( int OpenClose );
int  TestFirmRev( int OpenClose );
int  TestReset( int OpenClose );
int  TestRescan( int OpenClose );
int  TestPacket( int OpenClose );
int  TestSanity( int OpenClose );
int  TestSerial( int OpenClose );
int  TestGetFrame( int OpenClose );
int  TestPipeReset( int OpenClose );
int  TestPipeAbort( int OpenClose );
int  TestPortReset( int OpenClose );
int  TestPortCycle( int OpenClose );

struct DiagTestOption {
        int hidden;
        char *name;
        char *text;
        TESTFN *testfunction;
        } DiagTestTable[] = {

   0,   "install",    "Installation test suite (dev,power,led,analog,nmea)",             TestInstallation,
   0,   "debug",      "Debug tests for customer support",                                TestDebug,
   0,   "dev",        "Show device descriptor",                                          TestDevice,
   0,   "power",      "Check if power is currently good",                                TestPower,
   0,   "led",        "Toggle the yellow LED",                                           TestLed,
   0,   "analog",     "Start A/D's and display analog data",                             TestAnalog,
   0,   "nmea",       "Start A/D's and display NMEA strings",                            TestNmea,
   0,   "equip",      "Start A/D's and display equipment info like temp",                TestEquip,
   0,   "digin",      "Read digital input",                                              TestDigIn,
   0,   "digout",     "Send digital output",                                             TestDigOut,
   0,   "dram",       "Run several DRAM FIFO tests",                                     TestDram,
#ifdef OPTIONS_NOT_READY_YET
   0,   "empty",      "Check the DRAM FIFO empty flag",                                  TestEmpty,
#endif // OPTIONS_NOT_READY_YET
   0,   "pfull",      "Check the DRAM FIFO part full flag",                              TestPartFull,
   0,   "rev",        "Get firmware revision number",                                    TestFirmRev,
   0,   "portreset",  "Software reset PC motherboard USB hub",                           TestPortReset,
   0,   "portcycle",  "Hardware reset PC motherboard USB hub",                           TestPortCycle,
   0,   "reset",      "Hardware reset USBxCH board with full power cycle",               TestReset,
   0,   "rescan",     "Run Plug 'n' Play scan for hardware changes",                     TestRescan,
   1,   "packet",     "Start A/D's and display raw USB packets",                         TestPacket,
   1,   "sanity",     "Start A/D's and display analog data (+ sanity)",                  TestSanity,
   1,   "serial",     "Start A/D's and read 8051 serial memory area",                    TestSerial,
   1,   "frame",      "Get USB 1 ms frame number",                                       TestGetFrame,
   1,   "dslam",      "Run DRAM slam test",                                              TestDramSlam,
   1,   "dslide",     "Run DRAM slide test",                                             TestDramSlide,
   1,   "dinc",       "Run DRAM inc test",                                               TestDramInc,
   1,   "dvarwr",     "Run DRAM variable write/read test",                               TestDramVariableWR,
   1,   "dsglwr",     "Run DRAM single write/read test",                                 TestDramSingleWR,
   1,   "dallwr",     "Run DRAM all write/read test",                                    TestDramAllWR,
   1,   "overflow",   "Write the DRAM FIFO until overflow (VERY SLOW)",                  TestOverflow,
#ifdef OPTIONS_NOT_READY_YET
   1,   "reset",      "Reset stall condition on all pipes",                              TestPipeReset,
   1,   "abort",      "Abort all pending I/O for all pipes",                             TestPipeAbort,
#endif // OPTIONS_NOT_READY_YET
   1,    NULL,         NULL,                                                              NULL
   };

#define DO  struct DiagTestOption  // DO = shorthand for diag test option




// Prototypes

int  IsDescriptorGood( USB_DEVICE_DESCRIPTOR *Target );
int  ReadAll( int OpenClose, int DataRequested,
              unsigned long *NanalogRequested,
              unsigned long *NanalogReturned,
              unsigned long *NserialReturned,
              unsigned long *NequipReturned );
void ShowTime( void );
void ShowDramData( char *WriteData, char *ReadData );
int  CheckDramResults( char *WriteData, char *ReadData,
                       int Nvalues, int Nwrote, int Nread );
void ShowDramFlags( char *Msg, int *Empty, int *Overflow, int *PartFull );
int  CheckDramFlags( char *Msg, int ExpectedEmpty, int ExpectedOverflow,
                     int ExpectedPartFull );
int  CheckReadValue( char ExpectedValue, char ReadValue, int Nread, int Error );
void DisplayPackets( SRUSBXCH_PACKET *PacketArray,
                     unsigned long NpacketsRead, int MaxPrint, int StartCount );
void DisplayAnalogData( int NanalogRead, SRDAT_USBXCH_ANALOG *UsbxchAnalogData,
                        int Nchan, int MaxPrint, int StartCount, int ShowExtras );
void DisplaySerialData( int NserialRead, SRDAT_USBXCH_SERIAL *UsbxchSerialData );
void DisplayEquipData( int NequipRead, SRDAT_USBXCH_EQUIPMENT *UsbxchEquipData );
void OpenUsbXchFull( char *UsbxchDriverName, int GpsModel );
void OpenUsbXchMinimal( char *UsbxchDriverName );
void OpenUsbXchHardware( char *UsbxchDriverName );
int  OpenUsbXchVerify( char *UsbxchDriverName );
void CloseUsbXch( void );
void PrintRunTimeSummary( void );
void ProcessCommandLine( int argc, char *argv[] );
TR  *TimeTableLookUp( char *s );
DO  *DiagTestLookUp( char *testname );
void Usage( void );
void ErrorCmd( char *msg );
void PrintString( char *msg, int mode );
int  StrEqu( char *s1, char *s2 );
int  PrepareFileNames( void );
int  CleanExit( int ErrLevel );


int  GetDeviceDescriptor( DEVHANDLE UsbXchHandle,
                          USB_DEVICE_DESCRIPTOR *pDeviceDescriptor );
int  DescriptorTypeString( int DescType, char *DescTypeString, size_t MaxLen );


// Prototypes from SrUsbXch.c

int SrUsbXchHelperSpsInit( DEVHANDLE UsbXchHandle, int SpsIndex, int *Error );
int SrUsbXchHelperDownload( DEVHANDLE UsbXchHandle, char **HwData );
int SrUsbXchHelper8051RunHold( DEVHANDLE UsbXchHandle, int RunHold );
int SrUsbXchOsDriverIoctl( DEVHANDLE UsbXchHandle,
                           unsigned long IoCtlCode,
                           void *pValueIn, unsigned long InSize,
                           void *pValueOut, unsigned long  OutSize,
                           unsigned long *ReturnSize );




//------------------------------------------------------------------------------
// ROUTINE: Main
// PURPOSE: Organize and perform all the USBXCH diagnostic tests requested on
//          the command line by the user.
//------------------------------------------------------------------------------
int main( int argc, char *argv[] ) {

        int Result;


        // Set keyboard behavior to allow for non blocked key checks.

        SrKeyboardMode( KEYMODE_NONBLOCK );


        // Generate log and report file names.

        PrepareFileNames();


        // Open last run report file

        DP.RptFp = fopen( DP.RptFullPathName, "w" );
        if ( !DP.RptFp ) {
                sprintf( DP.Msg, "Could not open last run report file %s\n", DP.RptFullPathName );
                PrintStrDate( DP.Msg );
                return( 1 );
                }


        // Announce program version.

        sprintf( DP.Msg, ScreenTitle, DIAG_REV );
        PrintStrDate( DP.Msg );
        PrintStrRpt( "\n" );



        // Select desired tests from command line.

        ProcessCommandLine( argc, argv );


        // Get additional version and model name info

        SrUsbXchGetRev( &DP.LibRev );


        // Display starting info

        PrintRunTimeSummary( );



        // RUN SELECTED TEST:
        //
        //
        //

        if ( DP.pDiagTestSelected )
                Result = DP.pDiagTestSelected->testfunction( 1 ); // 1 -> request open

        else {
                PrintStrDate( "No test specified ... \n" );
                Result = 0;
                }



        // Close report file

        if ( DP.RptFp ) {
                fclose( DP.RptFp );
                DP.RptFp = NULL;
                }


        if ( Result ) {
                PrintStrRpt( "\nSUCCESS on diag test!\n" );
                CleanExit( 0 );
                }
        else {
                PrintStrRpt( "\nFAILURE on diag test.\n" );
                CleanExit( 1 );
                }
}



// Test functions
//
//
//

//------------------------------------------------------------------------------
// ROUTINE: TestInstallation
// PURPOSE: 
//------------------------------------------------------------------------------
int TestInstallation( int OpenClose ) {

        int Result, NumTest;

        // Run a set of standard tests to verify hardware installation


        PrintStrDate( "Starting standard installation tests ... \n" );

        if ( OpenClose )  OpenUsbXchFull( DP.AtodDriverName, DP.AtodGpsModel );

        Result = 0;

        PrintStr( "\n" );
        PrintStrRpt( "\n" );
        PrintStrDate( "Test 1: Device descriptor ... \n" );
        PrintStrRpt( "\n" );
        Result += TestDevice( 0 );

        PrintStr( "\n" );
        PrintStrRpt( "\n" );
        PrintStrDate( "Test 2: Power ... \n" );
        PrintStrRpt( "\n" );
        Result += TestPower( 0 );

        PrintStr( "\n" );
        PrintStrRpt( "\n" );
        PrintStrDate( "Test 3: Yellow LED ... \n" );
        PrintStrRpt( "\n" );
        Result += TestLed( 0 );

        PrintStr( "\n" );
        PrintStrRpt( "\n" );
        PrintStrDate( "Test 4: Analog data ... \n" );
        PrintStrRpt( "\n" );
        Result += TestAnalog( 0 );

        PrintStr( "\n" );
        PrintStrRpt( "\n" );
        PrintStrDate( "Test 5: NMEA data ... \n" );
        PrintStrRpt( "\n" );
        Result += TestNmea( 0 );

        NumTest = 5;

        if ( OpenClose )  CloseUsbXch( );


        if ( Result == NumTest ) {
                PrintStr( "\n" );
                PrintStrDate( "SUCCESS: Standard install tests all completed successfully.\n" );
                return( 1 );
                }
        else {
                sprintf( DP.Msg, "FAILURE: Standard install tests completed: %d with SUCCESS, %d with FAILURE.\n",
                         Result, (NumTest-Result) );
                PrintStr( "\n" );
                PrintStrDate( DP.Msg );
                return( 0 );
                }
}

//------------------------------------------------------------------------------
// ROUTINE: TestDebug
// PURPOSE: 
//------------------------------------------------------------------------------
int TestDebug( int OpenClose ) {

        int Result, NumTest;

        PrintStrDate( "Starting debugging tests ... \n" );

        if ( OpenClose )  OpenUsbXchFull( DP.AtodDriverName, DP.AtodGpsModel );


        // Run tests

        Result = 0;

        PrintStr( "\n" );
        PrintStrRpt( "\n" );
        PrintStrDate( "Test 1: Device descriptor ... \n" );
        PrintStrRpt( "\n" );
        Result += TestDevice( 0 );

        PrintStr( "\n" );
        PrintStrRpt( "\n" );
        PrintStrDate( "Test 2: Firmware revision ... \n" );
        PrintStrRpt( "\n" );
        Result += TestFirmRev( 0 );

        PrintStr( "\n" );
        PrintStrRpt( "\n" );
        PrintStrDate( "Test 3: Power ... \n" );
        PrintStrRpt( "\n" );
        Result += TestPower( 0 );

        PrintStr( "\n" );
        PrintStrRpt( "\n" );
        PrintStrDate( "Test 4: DRAM ... \n" );
        PrintStrRpt( "\n" );
        Result += TestDram( 0 );

//        PrintStr( "\n" );
//        PrintStrRpt( "\n" );
//        PrintStrDate( "Test 5: Empty flag ... \n" );
//        PrintStrRpt( "\n" );
//        Result += TestEmpty( 0 );

        PrintStr( "\n" );
        PrintStrRpt( "\n" );
        PrintStrDate( "Test 5: Raw packet data ... \n" );
        PrintStrRpt( "\n" );
        Result += TestPacket( 0 );

        PrintStr( "\n" );
        PrintStrRpt( "\n" );
        PrintStrDate( "Test 6: Analog/sanity data ... \n" );
        PrintStrRpt( "\n" );
        Result += TestSanity( 0 );

        PrintStr( "\n" );
        PrintStrRpt( "\n" );
        PrintStrDate( "Test 7: NMEA data ... \n" );
        PrintStrRpt( "\n" );
        Result += TestNmea( 0 );

        PrintStr( "\n" );
        PrintStrRpt( "\n" );
        PrintStrDate( "Test 8: Equipment data ... \n" );
        PrintStrRpt( "\n" );
        Result += TestEquip( 0 );

        NumTest = 8 + 5; // Dram has 5 extra tests

        if ( OpenClose )  CloseUsbXch( );


        if ( Result == NumTest ) {
                PrintStr( "\n" );
                PrintStrDate( "SUCCESS: Debugging tests all completed successfully.\n" );
                return( 1 );
                }
        else {
                sprintf( DP.Msg, "FAILURE: Debugging tests completed: %d with SUCCESS, %d with FAILURE.\n",
                         Result, (NumTest-Result) );
                PrintStr( "\n" );
                PrintStrDate( DP.Msg );
                sprintf( DP.Msg, "Please email file %s to info@symres.com for support.\n", DP.RptFullPathName );
                PrintStrDate( DP.Msg );
                return( 0 );
                }
}

//------------------------------------------------------------------------------
// ROUTINE: TestDevice
// PURPOSE: 
//------------------------------------------------------------------------------
int TestDevice( int OpenClose ) {

        USB_DEVICE_DESCRIPTOR DeviceDescriptor, *pDD;
        char                  VendorString[MAXSTR], ProductString[MAXSTR], DescTypeString[MAXSTR];
        int                   Result;


        PrintStrDate( "Device descriptor running ... \n" );



        // Get device descriptor

        if ( OpenClose )  OpenUsbXchHardware( DP.AtodDriverName );

        Result = GetDeviceDescriptor( DP.UsbXchHandle, &DeviceDescriptor );

        if ( OpenClose )  CloseUsbXch( );




        if ( Result ) {

                // Normally, we'd just call SrUsbXchShowDeviceDescriptor
                // but that does simple printf's and we want to call our
                // special PrintStrRpt instead.  So, we just duplicate
                // the function here.

                pDD = &DeviceDescriptor;

                // Set descriptor type string

                DescriptorTypeString( pDD->bDescriptorType, DescTypeString, MAXSTR-1 );


                // Set vendor and product strings

                strcpy( VendorString, " " );
                strcpy( ProductString, " " );

                if (pDD->idVendor == 0x15D3) {
                        strcpy( VendorString, "Symmetric Research" );
                        if (pDD->idProduct == 0x5504)
                                strcpy( ProductString, "USB4CH" );
                        else if (pDD->idProduct == 0x2235)
                                strcpy( ProductString, "USBXCH" );
                        }
                else if (pDD->idVendor == 0x0547) {
                        strcpy( VendorString, "Anchor Chips / Cypress" );
                        if (pDD->idProduct == 0x2235)
                                strcpy( ProductString, "USBXCH" );
                        }


                // Display Device descriptor data

                sprintf( DP.Msg, "\nDEVICE DESCRIPTOR =\n" );                                                                         PrintStrRpt( DP.Msg );
                sprintf( DP.Msg, "Device Descriptor: at 0x%p\n",       pDD );                                                         PrintStrRpt( DP.Msg );
                sprintf( DP.Msg, "Length             = 0x%X (%d)\n",    pDD->bLength, pDD->bLength );                                 PrintStrRpt( DP.Msg );
                sprintf( DP.Msg, "DescriptorType     = 0x%X (%d) %s\n", pDD->bDescriptorType, pDD->bDescriptorType, DescTypeString ); PrintStrRpt( DP.Msg );
                sprintf( DP.Msg, "USB                = 0x%X (%d)\n",    pDD->bcdUSB, pDD->bcdUSB );                                   PrintStrRpt( DP.Msg );
                sprintf( DP.Msg, "DeviceClass        = 0x%X\n",         pDD->bDeviceClass );                                          PrintStrRpt( DP.Msg );
                sprintf( DP.Msg, "DeviceSubClass     = 0x%X\n",         pDD->bDeviceSubClass );                                       PrintStrRpt( DP.Msg );
                sprintf( DP.Msg, "DeviceProtocol     = 0x%X\n",         pDD->bDeviceProtocol );                                       PrintStrRpt( DP.Msg );
                sprintf( DP.Msg, "MaxPacketSize0     = 0x%X (%d)\n",    pDD->bMaxPacketSize0, pDD->bMaxPacketSize0 );                 PrintStrRpt( DP.Msg );
                sprintf( DP.Msg, "idVendor           = 0x%X %s\n",      pDD->idVendor, VendorString );                                PrintStrRpt( DP.Msg );
                sprintf( DP.Msg, "idProduct          = 0x%X %s\n",      pDD->idProduct, ProductString );                              PrintStrRpt( DP.Msg );
                sprintf( DP.Msg, "Device             = 0x%X\n",         pDD->bcdDevice );                                             PrintStrRpt( DP.Msg );
                sprintf( DP.Msg, "iManufacturer      = 0x%X\n",         pDD->iManufacturer );                                         PrintStrRpt( DP.Msg );
                sprintf( DP.Msg, "iProduct           = 0x%X\n",         pDD->iProduct );                                              PrintStrRpt( DP.Msg );
                sprintf( DP.Msg, "iSerialNumber      = 0x%X\n",         pDD->iSerialNumber );                                         PrintStrRpt( DP.Msg );
                sprintf( DP.Msg, "NumConfigurations  = 0x%X\n",         pDD->bNumConfigurations );                                    PrintStrRpt( DP.Msg );
                sprintf( DP.Msg, "==================================\n\n" );                                                          PrintStrRpt( DP.Msg );


                if ( ! IsDescriptorGood( &DeviceDescriptor ) )
                        Result = 0;
                }

        if ( Result )
                PrintStrDate( "SUCCESS: Device descriptor completed\n" );
        else
                PrintStrDate( "FAILURE: Device descriptor completed\n" );

        return( Result );
}

//------------------------------------------------------------------------------
// ROUTINE: TestPower
// PURPOSE: 
//------------------------------------------------------------------------------
int TestPower( int OpenClose ) {

        int Result;


        PrintStrDate( "Power status running ... \n" );


        // Check the power status

        if ( OpenClose )  OpenUsbXchHardware( DP.AtodDriverName );

        Result = SrUsbXchIsPowerGood( DP.UsbXchHandle );

        if ( OpenClose )  CloseUsbXch( );


        if ( Result ) {
                PrintStrRpt( "Power is GOOD\n" );
                PrintStrDate( "SUCCESS: Power status completed\n" );
                }
        else {
                PrintStrRpt( "Power is BAD\n" );
                PrintStrDate( "FAILURE: Power status completed\n" );
                }

        return( Result );
}

//------------------------------------------------------------------------------
// ROUTINE: TestLed
// PURPOSE: 
//------------------------------------------------------------------------------
int TestLed( int OpenClose ) {

        int Ans, SleepMs, LedValue;

        PrintStrDate( "Toggle yellow LED running ... \n" );



        // Turn on yellow LED in a pattern (Red LED behavior controlled by PPS)

        if ( OpenClose ) OpenUsbXchHardware( DP.AtodDriverName );

        SrUsbXchUserIoWr( DP.UsbXchHandle, 0x03 ); // Start with digital out bits 0+1 on


        LedValue = LED_ON;

        while ( 1 ) {

                sprintf( DP.Msg, "Press ENTER to toggle yellow led or %s to quit\n", SR_QENTER );
                PrintStrRpt( DP.Msg );
                Ans = SrGetKeyWait( );
                if ( Ans == 'q' || Ans == 'Q' )
                        break;

                SrUsbXchUserLed( DP.UsbXchHandle, LedValue,  LED_OFF );

                if ( LedValue == LED_ON )
                        LedValue = LED_OFF;
                else
                        LedValue = LED_ON;
                }

        SrUsbXchUserLed( DP.UsbXchHandle, LED_ON,  LED_ON );

        if ( OpenClose )  CloseUsbXch( );


        PrintStrRpt( "\n" );
        PrintStrDate( "DONE: Toggle yellow LED complete\n" );
        return( 1 );


//FIX - This version not currently accessable

        SleepMs = 500;

#ifdef COMPILE_AUTOMATIC_VERSION

        SrUsbXchUserLed( DP.UsbXchHandle, LED_ON,  LED_OFF );  SrSleep( SleepMs );  // Yellow ON
        SrUsbXchUserLed( DP.UsbXchHandle, LED_OFF, LED_ON );   SrSleep( SleepMs );  // Red ON
        SrUsbXchUserLed( DP.UsbXchHandle, LED_ON,  LED_OFF );  SrSleep( SleepMs );  // Yellow ON
        SrUsbXchUserLed( DP.UsbXchHandle, LED_OFF, LED_ON );   SrSleep( SleepMs );  // Red ON
        SrUsbXchUserLed( DP.UsbXchHandle, LED_ON,  LED_ON );   SrSleep( SleepMs );  // Both ON

        SrUsbXchUserLed( DP.UsbXchHandle, LED_ON,  LED_OFF );  SrSleep( SleepMs );  // Yellow ON
        SrUsbXchUserLed( DP.UsbXchHandle, LED_OFF, LED_ON );   SrSleep( SleepMs );  // Red ON
        SrUsbXchUserLed( DP.UsbXchHandle, LED_ON,  LED_OFF );  SrSleep( SleepMs );  // Yellow ON
        SrUsbXchUserLed( DP.UsbXchHandle, LED_OFF, LED_ON );   SrSleep( SleepMs );  // Red ON
        SrUsbXchUserLed( DP.UsbXchHandle, LED_ON,  LED_ON );   SrSleep( SleepMs );  // Both ON
        return( 1 );

#endif // COMPILE_AUTOMATIC_VERSION
}

//------------------------------------------------------------------------------
// ROUTINE: TestDigIn
// PURPOSE: 
//------------------------------------------------------------------------------
int TestDigIn( int OpenClose ) {

        int i, Ans, SleepMs, DigValue;

        PrintStrDate( "Read digital inputs running ... \n" );

        if ( OpenClose )  OpenUsbXchHardware( DP.AtodDriverName );


        while ( 1 ) {

                sprintf( DP.Msg, "Press enter to read digital input or %s to quit\n", SR_QENTER );
                PrintStrRpt( DP.Msg );
                Ans = SrGetKeyWait( );
                if ( Ans == 'q' )
                        break;


                SrUsbXchUserIoRd( DP.UsbXchHandle, &DigValue );

                sprintf( DP.Msg, "Digital input is 0x%X\n", DigValue ); PrintStrRpt( DP.Msg );

                } // end while

        if ( OpenClose )  CloseUsbXch( );


        PrintStrRpt( "\n" );
        PrintStrDate( "DONE: Read digital inputs complete\n" );
        return( 1 );


//FIX - This version not currently accessable

        i       = 0;
        SleepMs = 500;

#ifdef COMPILE_AUTOMATIC_VERSION

        for ( i = 0 ; i < MAX_DIGVAL ; i++ ) {
                SrUsbXchUserIoRd( DP.UsbXchHandle, &DigValue );
                sprintf( DP.Msg, "Digital input %d is 0x%X\n", i, DigValue ); PrintStrRpt( DP.Msg );
                SrSleep( SleepMs );
                }

        return( 1 );

#endif // COMPILE_AUTOMATIC_VERSION
}


//------------------------------------------------------------------------------
// ROUTINE: TestDigOut
// PURPOSE: 
//------------------------------------------------------------------------------
int TestDigOut( int OpenClose ) {

        int i, Ans, SleepMs, Values[MAX_DIGVAL];

        PrintStrDate( "Send digital outputs running ... \n" );

        // Send digital output in a pattern

        Values[0] = 0x01;  // DIG_1RED
        Values[1] = 0x02;  // DIG_2YELLOW
        Values[2] = 0x04;  // DIG_3GREEN
        Values[3] = 0x08;  // DIG_4YELLOW
        Values[4] = 0x05;  // DIG_3GREEN  | DIG_1RED
        Values[5] = 0x0A;  // DIG_4YELLOW | DIG_2YELLOW
        Values[6] = 0x05;  // DIG_3GREEN  | DIG_1RED
        Values[7] = 0x0A;  // DIG_4YELLOW | DIG_2YELLOW
        Values[8] = 0x0F;  // DIG_ALL
        Values[9] = 0x00;  // DIG_NONE
        //    10  = MAX_DIGVAL


        if ( OpenClose )  OpenUsbXchHardware( DP.AtodDriverName );

        SrUsbXchUserLed( DP.UsbXchHandle, LED_OFF, LED_ON );        // Turn RED led on

        sprintf( DP.Msg, "\nSend %d digital outputs\n\n", MAX_DIGVAL );
        PrintStrRpt( DP.Msg );

        for ( i = 0 ; i < MAX_DIGVAL ; i++ ) {
                sprintf( DP.Msg, "Press enter to send dig value[%d] = 0x%X or %s to quit\n",
                         i, Values[i], SR_QENTER );
                PrintStrRpt( DP.Msg );
                Ans = SrGetKeyWait( );
                if ( Ans == 'q' )
                        break;
                else
                        SrUsbXchUserIoWr( DP.UsbXchHandle, Values[i] );
                }

        if ( OpenClose )  CloseUsbXch( );

        PrintStrRpt( "\n" );
        PrintStrDate( "DONE: Send digital outputs complete\n" );
        return( 1 );


//FIX - This version not currently accessable

        SleepMs = 500;

#ifdef COMPILE_AUTOMATIC_VERSION

        for ( i = 0 ; i < MAX_DIGVAL ; i++ ) {
                SrUsbXchUserIoWr( DP.UsbXchHandle, Values[i] );
                SrSleep( SleepMs );
                }

        return( 1 );

#endif // COMPILE_AUTOMATIC_VERSION
}

//------------------------------------------------------------------------------
// ROUTINE: TestAnalog
// PURPOSE: 
//------------------------------------------------------------------------------
int TestAnalog( int OpenClose ) {

        int           i, j, Nread, Result;
        unsigned long NanalogReturned, NanalogRequested;

        PrintStrDate( "Read analog data running ... \n" );


        Result = ReadAll( OpenClose, READALL_REQUEST_ANALOG,
                          &NanalogRequested, &NanalogReturned, NULL, NULL );

        if ( Result ) {

                // Error check that we got all expected analog data

                if ( NanalogReturned != NanalogRequested ) {
                        sprintf( DP.Msg, "ERROR: Analog data NanalogReturned %lu, NanalogRequested %lu\n",
                                 NanalogReturned, NanalogRequested );
                        PrintStrRpt( DP.Msg );
                        }


                // Error check that the data is not all zero

                Nread = (int)NanalogReturned;

                Result = 0;
                for ( i = 0 ; i < Nread ; i++ ) {
                        for ( j = 0 ; j < NCHANNELS ; j++ ) {
                                if ( DP.UsbxchAnalogData[i].AnalogData[j] != 0x00000000 ) {
                                        Result = 1;
                                        break;
                                        }
                                }
                        if ( Result == 1 )
                                break;
                        }

                if ( Result == 0 ) {
                        sprintf( DP.Msg, "ERROR: All analog data values are exactly zero.\n" );
                        PrintStrRpt( DP.Msg );
                        }


                // Display the data

                DisplayAnalogData( NanalogReturned, DP.UsbxchAnalogData, NCHANNELS, 20, 0, 0 );

                }

        if ( Result )
                PrintStrDate( "SUCCESS: Read analog data complete\n" );
        else
                PrintStrDate( "FAILURE: Read analog data complete\n" );

        return( Result );
}

//------------------------------------------------------------------------------
// ROUTINE: TestNmea
// PURPOSE: 
//------------------------------------------------------------------------------
int TestNmea( int OpenClose ) {

        int           Result;
        unsigned long NserialReturned;

        sprintf( DP.Msg, "Testing NMEA data with GpsModel = %s\n", DP.AtodGpsModelName );
        PrintStrDate( DP.Msg );
        PrintStrRpt( "\n" );
        PrintStrDate( "Read NMEA serial data running ... \n" );


        Result = ReadAll( OpenClose, READALL_REQUEST_SERIAL,
                          NULL, NULL, &NserialReturned, NULL );

        if ( Result ) {

                // Check that some serial data has been acquired

                if ( NserialReturned == 0L )
                        Result = 0;

                // Display the data

                DisplaySerialData( NserialReturned, DP.UsbxchSerialData );
                }


        if ( Result )
                PrintStrDate( "SUCCESS: Read NMEA serial data complete\n" );
        else
                PrintStrDate( "FAILURE: Read NMEA serial data complete\n" );

        if ( DP.AtodGpsModel == SRDAT_GPSMODEL_PCTIME ) {
                PrintStrRpt( "\nNOTE: To see NMEA strings provided by a Garmin GPS\n" );
                PrintStrRpt( "hook up your GPS antenna and run 'diag nmea g2'\n" );
                }

        return( Result );
}

//------------------------------------------------------------------------------
// ROUTINE: TestEquip
// PURPOSE: 
//------------------------------------------------------------------------------
int TestEquip( int OpenClose ) {

        int           Result;
        unsigned long NequipReturned;

        PrintStrDate( "Read equipment data running ... \n" );


        Result = ReadAll( OpenClose, READALL_REQUEST_EQUIP,
                          NULL, NULL, NULL, &NequipReturned );


        if ( Result ) {

                // Check that some equipment data has been acquired

                if ( NequipReturned == 0L )
                        Result = 0;

                // Display the data

                DisplayEquipData( NequipReturned, DP.UsbxchEquipData );
                }

        if ( Result )
                PrintStrDate( "SUCCESS: Read equipment data complete\n" );
        else
                PrintStrDate( "FAILURE: Read equipment data complete\n" );

        return( Result );
}

//------------------------------------------------------------------------------
// ROUTINE: TestDram
// PURPOSE: 
//------------------------------------------------------------------------------
int TestDram( int OpenClose ) {

        int Result, NumTest;

        PrintStrDate( "DRAM tests running ... \n" );


        if ( OpenClose )  OpenUsbXchHardware( DP.AtodDriverName );

        Result = 0;
        Result += TestDramSlam( 0 );        // 1
        Result += TestDramSlide( 0 );       // 2
        Result += TestDramInc( 0 );         // 3
        Result += TestDramVariableWR( 0 );  // 4
        Result += TestDramSingleWR( 0 );    // 5
        Result += TestDramAllWR( 0 );       // 6

        NumTest = 6;

        if ( OpenClose )  CloseUsbXch( );


        if ( Result == NumTest )
                PrintStrDate( "SUCCESS: DRAM tests complete\n" );
        else {
                sprintf( DP.Msg, "FAILURE: %d DRAM tests (SUCCESS %d tests).\n",
                         (NumTest-Result), Result );
                PrintStrRpt( "\n" );
                PrintStrDate( DP.Msg );
                }

        return( Result );
}

//------------------------------------------------------------------------------
// ROUTINE: TestDramSlam
// PURPOSE: 
//------------------------------------------------------------------------------
int TestDramSlam( int OpenClose ) {

        int  i, j, StartValue, SlamValue, Nstart;
        int  Nwrote, Nread, Nvalues, Error, Result;
        char *TestName;
        char ReadData[DRAM_TEST_MAX], WriteData[DRAM_TEST_MAX];

        Nvalues    = DRAM_TEST_SIZE;
        TestName   = "Slam ";
        StartValue = 0x00;
        Nstart     = 2;
        Result     = 1;


        if ( OpenClose )  OpenUsbXchHardware( DP.AtodDriverName );


        for ( i = 0 ; i < Nstart ; i++ ) {

                sprintf( DP.Msg, "DRAM %s 0x%02X running ... \n", TestName, StartValue );
                PrintStrDate( DP.Msg );


                // Initialize data

                for ( j = 0 ; j < DRAM_TEST_SIZE ; j++ )
                        ReadData[j] = 0x0;

                SlamValue = StartValue;
                for ( j = 0 ; j < DRAM_TEST_SIZE ; j++ ) {
                        WriteData[j] = (SlamValue & 0x0F);
                        SlamValue ^= 0xFF;
                        }

                SrUsbXchDramReset( DP.UsbXchHandle );


                // Write then read DRAM

                Nwrote = SrUsbXchDramWrite( DP.UsbXchHandle, WriteData,
                                            Nvalues, &Error );

                if ( Error != USBXCH_ERROR_NONE ) {
                        sprintf( DP.Msg, "ERROR: Failed Dram write with error %d\n", Error );
                        PrintStrRpt( DP.Msg );
                        sprintf( DP.Msg, "FAILURE: DRAM %s 0x%02X complete\n", TestName, StartValue );
                        PrintStrDate( DP.Msg );
                        return( 0 );
                        }


                Nread = SrUsbXchDramRead( DP.UsbXchHandle, ReadData,
                                          Nvalues, &Error );

                if ( Error != USBXCH_ERROR_NONE ) {
                        sprintf( DP.Msg, "ERROR: Failed Dram read with error %d\n", Error );
                        PrintStrRpt( DP.Msg );
                        sprintf( DP.Msg, "FAILURE: DRAM %s 0x%02X complete\n", TestName, StartValue );
                        PrintStrDate( DP.Msg );
                        return( 0 );
                        }



                // Check results

                Result = CheckDramResults( WriteData, ReadData, Nvalues, Nwrote, Nread );

                if ( Result ) {
                        sprintf( DP.Msg, "SUCCESS: DRAM %s 0x%02X complete\n", TestName, StartValue );
                        PrintStrDate( DP.Msg );
                        }
                else {
                        sprintf( DP.Msg, "FAILURE: DRAM %s 0x%02X complete\n", TestName, StartValue );
                        PrintStrDate( DP.Msg );
                        return( 0 );
                        }



                // Prepare start value for next loop

                if ( StartValue == 0xFF )
                        StartValue = 0x00;
                else
                        StartValue = 0xFF;


                } // end for i < Nstart


        // Reset FIFO

        SrUsbXchDramReset( DP.UsbXchHandle );

        if ( OpenClose )  CloseUsbXch( );

        return( 1 );
}

//------------------------------------------------------------------------------
// ROUTINE: TestDramSlide
// PURPOSE: 
//------------------------------------------------------------------------------
int TestDramSlide( int OpenClose ) {

        int  i, j, StartValue, SlideValue, Nstart;
        int  Nwrote, Nread, Nvalues, Error, Result;
        char *TestName;
        char ReadData[DRAM_TEST_MAX], WriteData[DRAM_TEST_MAX];

        Nvalues    = DRAM_TEST_SIZE;
        TestName   = "Slide";
        StartValue = 0x01;
        Nstart     = 4;
        Result     = 1;

        if ( OpenClose )  OpenUsbXchHardware( DP.AtodDriverName );


        for ( i = 0 ; i < Nstart ; i++ ) {

                sprintf( DP.Msg, "DRAM %s 0x%02X running ... \n", TestName, StartValue );
                PrintStrDate( DP.Msg );


                // Initialize data

                for ( j = 0 ; j < DRAM_TEST_SIZE ; j++ )
                        ReadData[j] = 0x0;

                SlideValue = StartValue;
                for ( j = 0 ; j < DRAM_TEST_SIZE ; j++ ) {
                        WriteData[j] = SlideValue;
                        SlideValue <<= 1;
                        SlideValue &= 0x0F;
                        if (SlideValue == 0)
                                SlideValue = 0x01;
                        }

                SrUsbXchDramReset( DP.UsbXchHandle );


                // Write then read DRAM

                Nwrote = SrUsbXchDramWrite( DP.UsbXchHandle, WriteData,
                                            Nvalues, &Error );

                if ( Error != USBXCH_ERROR_NONE ) {
                        sprintf( DP.Msg, "ERROR: Failed Dram write with error %d\n", Error );
                        PrintStrRpt( DP.Msg );
                        sprintf( DP.Msg, "FAILURE: DRAM %s 0x%02X complete\n", TestName, StartValue );
                        PrintStrDate( DP.Msg );
                        return( 0 );
                        }


                Nread = SrUsbXchDramRead( DP.UsbXchHandle, ReadData,
                                          Nvalues, &Error );

                if ( Error != USBXCH_ERROR_NONE ) {
                        sprintf( DP.Msg, "ERROR: Failed Dram read with error %d\n", Error );
                        PrintStrRpt( DP.Msg );
                        sprintf( DP.Msg, "FAILURE: DRAM %s 0x%02X complete\n", TestName, StartValue );
                        PrintStrDate( DP.Msg );
                        return( 0 );
                        }



                // Check results

                Result = CheckDramResults( WriteData, ReadData, Nvalues, Nwrote, Nread );

                if ( Result ) {
                        sprintf( DP.Msg, "SUCCESS: DRAM %s 0x%02X complete\n", TestName, StartValue );
                        PrintStrDate( DP.Msg );
                        }
                else {
                        sprintf( DP.Msg, "FAILURE: DRAM %s 0x%02X complete\n", TestName, StartValue );
                        PrintStrDate( DP.Msg );
                        return( 0 );
                        }


                // Prepare start value for next loop

                StartValue <<= 1;
                StartValue &= 0x0F;
                if (StartValue == 0)
                        StartValue = 0x01;


                } // end for i < Nstart


        // Reset FIFO

        SrUsbXchDramReset( DP.UsbXchHandle );

        if ( OpenClose )  CloseUsbXch( );

        return( 1 );
}

//------------------------------------------------------------------------------
// ROUTINE: TestDramInc
// PURPOSE: 
//------------------------------------------------------------------------------
int TestDramInc( int OpenClose ) {

        int  i, j, StartValue, Nstart;
        int  Nwrote, Nread, Nvalues, Error, Result;
        char *TestName;
        char ReadData[DRAM_TEST_MAX], WriteData[DRAM_TEST_MAX];

        Nvalues    = DRAM_TEST_SIZE;
        TestName   = "Inc  ";
        StartValue = 1;
        Nstart     = 8;
        Result     = 1;


        if ( OpenClose )  OpenUsbXchHardware( DP.AtodDriverName );

        for ( i = 0 ; i < Nstart ; i++ ) {

                sprintf( DP.Msg, "DRAM %s 0x%02X running ... \n", TestName, StartValue );
                PrintStrDate( DP.Msg );


                // Initialize data

                for ( j = 0 ; j < DRAM_TEST_SIZE ; j++ )
                        ReadData[j] = 0x0;

                for ( j = 0 ; j < DRAM_TEST_SIZE ; j++ ) {
                        WriteData[j] = ((StartValue+j)&0x0F);
                        }

                SrUsbXchDramReset( DP.UsbXchHandle );


                // Write then read DRAM

                Nwrote = SrUsbXchDramWrite( DP.UsbXchHandle, WriteData,
                                            Nvalues, &Error );

                if ( Error != USBXCH_ERROR_NONE ) {
                        sprintf( DP.Msg, "ERROR: Failed Dram write with error %d\n", Error );
                        PrintStrRpt( DP.Msg );
                        sprintf( DP.Msg, "FAILURE: DRAM %s 0x%02X complete\n", TestName, StartValue );
                        PrintStrDate( DP.Msg );
                        return( 0 );
                        }


                Nread = SrUsbXchDramRead( DP.UsbXchHandle, ReadData,
                                          Nvalues, &Error );

                if ( Error != USBXCH_ERROR_NONE ) {
                        sprintf( DP.Msg, "ERROR: Failed Dram read with error %d\n", Error );
                        PrintStrRpt( DP.Msg );
                        sprintf( DP.Msg, "FAILURE: DRAM %s 0x%02X complete\n", TestName, StartValue );
                        PrintStrDate( DP.Msg );
                        return( 0 );
                        }



                // Check results

                Result = CheckDramResults( WriteData, ReadData, Nvalues, Nwrote, Nread );

                if ( Result ) {
                        sprintf( DP.Msg, "SUCCESS: DRAM %s 0x%02X complete\n", TestName, StartValue );
                        PrintStrDate( DP.Msg );
                        }
                else {
                        sprintf( DP.Msg, "FAILURE: DRAM %s 0x%02X complete\n", TestName, StartValue );
                        PrintStrDate( DP.Msg );
                        return( 0 );
                        }


                // Prepare start value for next loop

                StartValue++;
                StartValue &= 0x0F;

                } // end for i < Nstart


        // Reset FIFO

        SrUsbXchDramReset( DP.UsbXchHandle );

        if ( OpenClose )  CloseUsbXch( );

        return( 1 );
}

//------------------------------------------------------------------------------
// ROUTINE: TestDramVariableWR
// PURPOSE: 
//------------------------------------------------------------------------------
int TestDramVariableWR( int OpenClose ) {

        int  i, Error;
        int  WriteIndex, ReadIndex, NtoWrite, NtoRead, Nread;
        char ReadData[DRAM_TEST_MAX], WriteData[DRAM_TEST_MAX];


        // Perform variable interspersed write/read test

        PrintStrDate( "DRAM variable write/read running ... \n" );



        // Initialize data

        for ( i = 0 ; i < DRAM_TEST_SIZE ; i++ )
                WriteData[i] = (char)( (i+1) & 0x0F );

        for ( i = 0 ; i < DRAM_TEST_SIZE ; i++ )
                ReadData[i] = 0x00;


        WriteIndex = 0;
        ReadIndex  = 0;


        if ( OpenClose )  OpenUsbXchHardware( DP.AtodDriverName );

        SrUsbXchDramReset( DP.UsbXchHandle );



        // First write/read pair

        NtoWrite = 4;
        NtoRead  = 3;

        SrUsbXchDramWrite( DP.UsbXchHandle, &WriteData[WriteIndex], NtoWrite, &Error );
        WriteIndex += NtoWrite;

        Nread = SrUsbXchDramRead( DP.UsbXchHandle, &ReadData[ReadIndex], NtoRead, &Error  );
        ReadIndex  += Nread;
        if (Nread != NtoRead) {
                sprintf( DP.Msg, "ERROR: Dram asked to read %d, but actually read %d\n", NtoRead, Nread );
                PrintStrRpt( DP.Msg );
                PrintStrDate( "FAILURE: DRAM variable write/read complete\n" );
                return( 0 );
                }



        // Second write/read pair

        NtoWrite = 4;
        NtoRead  = 2;

        SrUsbXchDramWrite( DP.UsbXchHandle, &WriteData[WriteIndex], NtoWrite, &Error );
        WriteIndex += NtoWrite;

        Nread = SrUsbXchDramRead( DP.UsbXchHandle, &ReadData[ReadIndex], NtoRead, &Error  );
        ReadIndex  += Nread;
        if (Nread != NtoRead){
                sprintf( DP.Msg, "ERROR: Dram asked to read %d, but actually read %d\n", NtoRead, Nread );
                PrintStrRpt( DP.Msg );
                PrintStrDate( "FAILURE: DRAM variable write/read complete\n" );
                return( 0 );
                }


        // Third write/read pair

        NtoWrite = 1;
        NtoRead  = 4;

        SrUsbXchDramWrite( DP.UsbXchHandle, &WriteData[WriteIndex], NtoWrite, &Error );
        WriteIndex += NtoWrite;

        Nread = SrUsbXchDramRead( DP.UsbXchHandle, &ReadData[ReadIndex], NtoRead, &Error  );
        ReadIndex  += Nread;
        if (Nread != NtoRead) {
                sprintf( DP.Msg, "ERROR: Dram asked to read %d, but actually read %d\n", NtoRead, Nread );
                PrintStrRpt( DP.Msg );
                PrintStrDate( "FAILURE: DRAM variable write/read complete\n" );
                return( 0 );
                }


        // Display results

//FIX        ShowDramData( WriteData, ReadData );


        PrintStrDate( "SUCCESS: DRAM variable write/read complete\n" );


        // Reset FIFO

        SrUsbXchDramReset( DP.UsbXchHandle );

        if ( OpenClose )  CloseUsbXch( );

        return( 1 );
}

//------------------------------------------------------------------------------
// ROUTINE: TestDramSingleWR
// PURPOSE: 
//------------------------------------------------------------------------------
int TestDramSingleWR( int OpenClose ) {

        int  i, Error;
        int  WriteIndex, ReadIndex, NtoWrite, NtoRead, Nread;
        char ReadData[DRAM_TEST_MAX], WriteData[DRAM_TEST_MAX];

        // Perform single interspersed write/read test

        PrintStrDate( "DRAM single write/read running  ... \n" );



        // Initialize data

        for ( i = 0 ; i < DRAM_TEST_SIZE1 ; i++ )
                WriteData[i] = (char)( (i+1) & 0x0F );

        for ( i = 0 ; i < DRAM_TEST_SIZE1 ; i++ )
                ReadData[i] = 0x00;


        WriteIndex = 0;
        ReadIndex  = 0;


        if ( OpenClose )  OpenUsbXchFull( DP.AtodDriverName, DP.AtodGpsModel );

        SrUsbXchDramReset( DP.UsbXchHandle );



        // Single write/read pairs

        NtoWrite   = 1;
        NtoRead    = 1;

        for ( i = 0 ; i < DRAM_TEST_SIZE1 ; i++ ) {

                SrUsbXchDramWrite( DP.UsbXchHandle, &WriteData[WriteIndex], NtoWrite, &Error );
                WriteIndex += NtoWrite;

                Nread = SrUsbXchDramRead( DP.UsbXchHandle, &ReadData[ReadIndex], NtoRead, &Error  );
                ReadIndex  += Nread;
                if (Nread != NtoRead) {
                        sprintf( DP.Msg, "ERROR: Dram asked to read %d, but actually read %d\n", NtoRead, Nread );
                        PrintStrRpt( DP.Msg );
                        PrintStrDate( "FAILURE: DRAM single write/read complete\n" );
                        return( 0 );
                        }

                } // end for i < DRAM_TEST_SIZE1


        // Display results

//FIX        ShowDramData( WriteData, ReadData );


        PrintStrDate( "SUCCESS: DRAM single write/read complete\n" );


        // Reset FIFO

        SrUsbXchDramReset( DP.UsbXchHandle );

        if ( OpenClose )  CloseUsbXch( );

        return( 1 );
}

//------------------------------------------------------------------------------
// ROUTINE: TestDramAllWR
// PURPOSE: 
//------------------------------------------------------------------------------
int TestDramAllWR( int OpenClose ) {

        int  i, Error;
        int  WriteIndex, ReadIndex, NtoWrite, NtoRead, Nread;
        char ReadData[DRAM_TEST_MAX], WriteData[DRAM_TEST_MAX];


        PrintStrDate( "DRAM all write then read running ... \n" );


        // Initialize data

        for ( i = 0 ; i < DRAM_TEST_SIZE ; i++ )
                WriteData[i] = (char)( (i+1) & 0x0F );

        for ( i = 0 ; i < DRAM_TEST_SIZE ; i++ )
                ReadData[i] = 0x00;


        WriteIndex = 0;
        ReadIndex  = 0;


        if ( OpenClose )  OpenUsbXchHardware( DP.AtodDriverName );

        SrUsbXchDramReset( DP.UsbXchHandle );




        // All writes/reads

        NtoWrite = 9;
        NtoRead  = 9;

        SrUsbXchDramWrite( DP.UsbXchHandle, &WriteData[WriteIndex], NtoWrite, &Error );
        WriteIndex += NtoWrite;

        Nread = SrUsbXchDramRead( DP.UsbXchHandle, &ReadData[ReadIndex], NtoRead, &Error  );
        ReadIndex  += Nread;
        if (Nread != NtoRead) {
                sprintf( DP.Msg, "ERROR: Dram asked to read %d, but actually read %d\n", NtoRead, Nread );
                PrintStrRpt( DP.Msg );
                PrintStrDate( "FAILURE: DRAM all write then read complete\n" );
                return( 0 );
                }


        // Display results

//FIX        ShowDramData( WriteData, ReadData );


        PrintStrDate( "SUCCESS: DRAM all write then read complete\n" );


        // Reset FIFO

        SrUsbXchDramReset( DP.UsbXchHandle );

        if ( OpenClose )  CloseUsbXch( );

        return( 1 );
}

//------------------------------------------------------------------------------
// ROUTINE: TestEmpty
// PURPOSE: 
//------------------------------------------------------------------------------
int TestEmpty( int OpenClose ) {

        int  i, Error, Nread, Result;
        char WriteData[EMPTY_TEST_SIZE], ReadData;


        PrintStrDate( "DRAM FIFO empty flag check running ...\n" );


        // Initialize data to write

        for ( i = 0 ; i < EMPTY_TEST_SIZE ; i++ )
                WriteData[i] = i & 0x0F;



        if ( OpenClose )  OpenUsbXchHardware( DP.AtodDriverName );


        // This while loop is only to run once, it just provides
        // an easy way (break) to exit on error

        while ( 1 ) {

                // Reset FIFO

                SrUsbXchDramReset( DP.UsbXchHandle );
                Result = CheckDramFlags( "Fifo starts with ", IS_EMPTY, NOT_OVERFLOW, IGNORE_PARTFULL );
                if ( !Result )
                        break;


                // Write 1

                SrUsbXchDramWrite( DP.UsbXchHandle, &WriteData[1], 1, &Error );
                Result = CheckDramFlags( "After writing 1  ", NOT_EMPTY, NOT_OVERFLOW, IGNORE_PARTFULL );
                if ( !Result )
                        break;


                // Write 2

                SrUsbXchDramWrite( DP.UsbXchHandle, &WriteData[2], 1, &Error );
                Result = CheckDramFlags( "After writing 2  ", NOT_EMPTY, NOT_OVERFLOW, IGNORE_PARTFULL );
                if ( !Result )
                        break;


                // Read 1

                ReadData  = 0x00;
                Nread = SrUsbXchDramRead( DP.UsbXchHandle, &ReadData, 1, &Error  );
                Result = CheckDramFlags( "After reading 1  ", NOT_EMPTY, NOT_OVERFLOW, IGNORE_PARTFULL );
                if ( !Result )
                        break;
                Result = CheckReadValue( WriteData[1], ReadData, Nread, Error );
                if ( !Result )
                        break;


                // Read 2

                ReadData  = 0x00;
                Nread = SrUsbXchDramRead( DP.UsbXchHandle, &ReadData, 1, &Error  );
                Result = CheckDramFlags( "After reading 2  ", IS_EMPTY, NOT_OVERFLOW, IGNORE_PARTFULL );
                if ( !Result )
                        break;
                Result = CheckReadValue( WriteData[2], ReadData, Nread, Error );
                if ( !Result )
                        break;


                // Write 3

                SrUsbXchDramWrite( DP.UsbXchHandle, &WriteData[3], 1, &Error );
                Result = CheckDramFlags( "After writing 3  ", NOT_EMPTY, NOT_OVERFLOW, IGNORE_PARTFULL );
                if ( !Result )
                        break;


                // Read 3

                ReadData = 0x00;
                Nread = SrUsbXchDramRead( DP.UsbXchHandle, &ReadData, 1, &Error  );
                Result = CheckDramFlags( "After reading 3  ", IS_EMPTY, NOT_OVERFLOW, IGNORE_PARTFULL );
                if ( !Result )
                        break;
                Result = CheckReadValue( WriteData[3], ReadData, Nread, Error );
                if ( !Result )
                        break;


                // Read X

                ReadData = 0x00;
                Nread = SrUsbXchDramRead( DP.UsbXchHandle, &ReadData, 1, &Error  );
                Result = CheckDramFlags( "After reading X  ", IS_EMPTY, NOT_OVERFLOW, IGNORE_PARTFULL );
                if ( !Result )
                        break;
                if ( Nread == 1 && Error == USBXCH_ERROR_NONE) {
                        sprintf( DP.Msg, "ERROR: Read of %d, but should have failed\n", ReadData );
                        PrintStrRpt( DP.Msg );
                        Result = 0;
                        break;
                        }


                // Write 4

                SrUsbXchDramWrite( DP.UsbXchHandle, &WriteData[4], 1, &Error );
                Result = CheckDramFlags( "After writing 4  ", NOT_EMPTY, NOT_OVERFLOW, IGNORE_PARTFULL );
                if ( !Result )
                        break;


                // Write 5

                SrUsbXchDramWrite( DP.UsbXchHandle, &WriteData[5], 1, &Error );
                Result = CheckDramFlags( "After writing 5  ", NOT_EMPTY, NOT_OVERFLOW, IGNORE_PARTFULL );
                if ( !Result )
                        break;


                // Read 4

                ReadData = 0x00;
                Nread = SrUsbXchDramRead( DP.UsbXchHandle, &ReadData, 1, &Error  );
                Result = CheckDramFlags( "After reading 4  ", NOT_EMPTY, NOT_OVERFLOW, IGNORE_PARTFULL );
                if ( !Result )
                        break;
                Result = CheckReadValue( WriteData[4], ReadData, Nread, Error );
                if ( !Result )
                        break;


                // Read 5

                ReadData = 0x00;
                Nread = SrUsbXchDramRead( DP.UsbXchHandle, &ReadData, 1, &Error  );
                Result = CheckDramFlags( "After reading 5  ", IS_EMPTY, NOT_OVERFLOW, IGNORE_PARTFULL );
                if ( !Result )
                        break;
                Result = CheckReadValue( WriteData[5], ReadData, Nread, Error );
                if ( !Result )
                        break;


                // End

                Result = CheckDramFlags( "Fifo ends   with ", IS_EMPTY, NOT_OVERFLOW, IGNORE_PARTFULL );
                if ( !Result )
                        break;


                // Force exit from dummy while loop
                break;

                } // end while


        // Reset FIFO

        SrUsbXchDramReset( DP.UsbXchHandle );

        if ( OpenClose )  CloseUsbXch( );


        if ( Result )
                PrintStrDate( "SUCCESS: DRAM FIFO empty flag check complete\n" );
        else
                PrintStrDate( "FAILURE: DRAM FIFO empty flag check complete\n" );


        return( Result );
}

//------------------------------------------------------------------------------
// ROUTINE: TestPartFull
// PURPOSE: 
//------------------------------------------------------------------------------
int TestPartFull( int OpenClose ) {

        int  i, Error, Nread, Nwrote, ExpectEmpty, ExpectPFull, Result;
        char WriteData[PARTFULL_TEST_SIZE];
        char ReadData[PARTFULL_TEST_SIZE], ReadStr[256];


        PrintStrDate( "DRAM FIFO part full flag check running ... \n" );


        // Initialize data to write

        for ( i = 0 ; i < PARTFULL_TEST_SIZE ; i++ )
                WriteData[i] = i & 0x0F;



        if ( OpenClose )  OpenUsbXchHardware( DP.AtodDriverName );


        // This while loop is only to run once, it just provides
        // an easy way (break) to exit on error

        while ( 1 ) {

                // Reset FIFO

                SrUsbXchDramReset( DP.UsbXchHandle );
                Result = CheckDramFlags( "Fifo starts with ", IS_EMPTY, NOT_OVERFLOW, NOT_PARTFULL );
                if ( !Result )
                        break;


                // Write 64

                SrUsbXchDramWrite( DP.UsbXchHandle, &WriteData[1], 64, &Error );
                Result = CheckDramFlags( "After writing 64 ", NOT_EMPTY, NOT_OVERFLOW, NOT_PARTFULL );
                if ( !Result )
                        break;
                Nwrote = 64;


                // Write 65th

                SrUsbXchDramWrite( DP.UsbXchHandle, &WriteData[65], 1, &Error );
                Result = CheckDramFlags( "After writing 65 ", NOT_EMPTY, NOT_OVERFLOW, IS_PARTFULL );
                if ( !Result )
                        break;
                Nwrote += 1;


                // Write 66th

                SrUsbXchDramWrite( DP.UsbXchHandle, &WriteData[66], 1, &Error );
                Result = CheckDramFlags( "After writing 66 ", NOT_EMPTY, NOT_OVERFLOW, IS_PARTFULL );
                if ( !Result )
                        break;
                Nwrote += 1;


                // Write 67th

                SrUsbXchDramWrite( DP.UsbXchHandle, &WriteData[67], 1, &Error );
                Result = CheckDramFlags( "After writing 67 ", NOT_EMPTY, NOT_OVERFLOW, IS_PARTFULL );
                if ( !Result )
                        break;
                Nwrote += 1;


                // Write 68th

                SrUsbXchDramWrite( DP.UsbXchHandle, &WriteData[68], 1, &Error );
                Result = CheckDramFlags( "After writing 68 ", NOT_EMPTY, NOT_OVERFLOW, IS_PARTFULL );
                if ( !Result )
                        break;
                Nwrote += 1;


                // Write 69th

                SrUsbXchDramWrite( DP.UsbXchHandle, &WriteData[69], 1, &Error );
                Result = CheckDramFlags( "After writing 69 ", NOT_EMPTY, NOT_OVERFLOW, IS_PARTFULL );
                if ( !Result )
                        break;
                Nwrote += 1;


                // Write 70th

                SrUsbXchDramWrite( DP.UsbXchHandle, &WriteData[70], 1, &Error );
                Result = CheckDramFlags( "After writing 70 ", NOT_EMPTY, NOT_OVERFLOW, IS_PARTFULL );
                if ( !Result )
                        break;
                Nwrote += 1;


                // Force exit from dummy while 1 loop
                break;

                } // end while 1

        PrintStrRpt( "\n" );



        // Do readback tests

        // This while loop is only to run once and only if the previous
        // loop was successful, it just provides an easy way (break) to
        // exit on error

        while ( Result ) {

                ExpectPFull = IS_PARTFULL;
                ExpectEmpty = NOT_EMPTY;

                for ( i = 0 ; i < Nwrote ; i++ ) {

                        ReadData[i+1] = 0x00;
                        sprintf( ReadStr, "After reading %2d ", Nwrote - i );

                        if ( i == Nwrote - 1 )
                                ExpectEmpty = IS_EMPTY;
                        if ( i == Nwrote - 64 )
                                ExpectPFull = NOT_PARTFULL;

                        Nread = SrUsbXchDramRead( DP.UsbXchHandle, &ReadData[i+1], 1, &Error  );
                        Result = CheckDramFlags( ReadStr, ExpectEmpty, NOT_OVERFLOW, ExpectPFull );
                        if ( !Result )
                                break;
                        Result = CheckReadValue( WriteData[i+1], ReadData[i+1], Nread, Error );
                        if ( !Result )
                                break;
                        }

                // Force exit from dummy while Result loop
                break;

                } // end while Result



        // Reset FIFO

        SrUsbXchDramReset( DP.UsbXchHandle );

        if ( OpenClose )  CloseUsbXch( );


        if ( Result )
                PrintStrDate( "SUCCESS: DRAM FIFO part full flag check complete\n" );
        else
                PrintStrDate( "FAILURE: DRAM FIFO part full flag check complete\n" );


        return( 1 );
}

//------------------------------------------------------------------------------
// ROUTINE: TestOverflow
// PURPOSE: 
//------------------------------------------------------------------------------
int TestOverflow( int OpenClose ) {

        int  Error, Nread, ShowCount, Nvalues, Result;
        long ii, jj, kk, StartValue, Nloops;
        char WriteData1, ReadData1;

//FIX - remove once empty flag checking restored

        PrintStrDate( "DRAM FIFO overflow flag check running ... \n" );

        PrintStrRpt( "\n *** NOTE: *** Empty flag errors not currently reported\n");


        if ( OpenClose )  OpenUsbXchHardware( DP.AtodDriverName );

        ShowTime();


        // This while loop is only to run once, it just provides
        // an easy way (break) to exit on error

        while ( 1 ) {

                // Reset FIFO

                SrUsbXchDramReset( DP.UsbXchHandle );
                Result = CheckDramFlags( "Fifo starts out ", IS_EMPTY, NOT_OVERFLOW, IGNORE_PARTFULL );
                if ( !Result )
                        break;


                ShowCount  = 0;

                Nloops = (long)(DRAM_NFILL+1) / (long)DRAM_ARRAY_SIZE;
                sprintf( DP.Msg, "DRAM_NFILL      = %ld\n", DRAM_NFILL );             PrintStrRpt( DP.Msg );
                sprintf( DP.Msg, "DRAM_ARRAY_SIZE = %ld\n", DRAM_ARRAY_SIZE );        PrintStrRpt( DP.Msg );
                sprintf( DP.Msg, "Nloops          = %ld\n", Nloops );                 PrintStrRpt( DP.Msg );
                sprintf( DP.Msg, "Computed Nfill  = %ld\n", Nloops*DRAM_ARRAY_SIZE ); PrintStrRpt( DP.Msg );

                if (Nloops*DRAM_ARRAY_SIZE != DRAM_NFILL+1) {
                        sprintf( DP.Msg, "Uneven number loops %ld != %ld\n",
                              (Nloops*DRAM_ARRAY_SIZE),DRAM_NFILL+1);
                        PrintStrRpt( DP.Msg );
                        Result = 0;
                        break;
                        }


                PrintStrRpt( "\nBeginning write section, please wait ...\n" );

                StartValue = 0x0L;

                for ( kk = 0 ; kk < Nloops ; kk++ ) {


                        // Initialize some values for write

                        for ( ii = 0 ; ii < DRAM_ARRAY_SIZE ; ii += 4 ) {
                                for ( jj = 0 ; jj < 4 ; jj++ ) {
                                        DP.ReadData[ii+jj]  = 0x0;
                                        DP.WriteData[ii+jj] =
                                                (char)((StartValue >> (jj*4)) & 0xF);
                                        }
                                StartValue++;
                                }


        //#define SHOW_DBG_DATA
        #ifdef SHOW_DBG_DATA
                        // Display values

                        jj = 3;
                        for ( ii = 0 ; ii < DRAM_ARRAY_SIZE ; ii++ ) {
                                jj++;
                                if (jj == 4) {
                                        jj = 0;
                                        sprintf( DP.Msg, "\n %2d: 0x ", ii );
                                        PrintStrRpt( DP.Msg );
                                        }
                                sprintf( DP.Msg, "%02X ", WriteData[ii] );
                                PrintStrRpt( DP.Msg );
                                }
                        PrintStrRpt(  "\n" );
        #endif // SHOW_DBG_DATA


                        // Write out the majority of values

                        if ( kk == Nloops-1 )
                                Nvalues = DRAM_ARRAY_SIZE - 4;
                        else
                                Nvalues = DRAM_ARRAY_SIZE;

                        SrUsbXchDramWrite( DP.UsbXchHandle, DP.WriteData, Nvalues, &Error );
        //                        sprintf( DP.Msg, "Loop %3ld at %7ld writing %ld points ", kk, StartValue*4, Nvalues );
                        sprintf( DP.Msg, "Loop %3ld: ", kk );
                        Result = CheckDramFlags( DP.Msg, NOT_EMPTY, NOT_OVERFLOW, IGNORE_PARTFULL );
                        if ( !Result )
                                break;

                        } // end for k


                // Write the last few values approaching overflow

                PrintStrRpt( "\nWriting last few points ...\n" );

                WriteData1 = 1;
                SrUsbXchDramWrite( DP.UsbXchHandle, &WriteData1, 1, &Error );
                Result = CheckDramFlags( "After writing 1 ", NOT_EMPTY, NOT_OVERFLOW, IGNORE_PARTFULL );
                if ( !Result )
                        break;

                WriteData1 = 2;
                SrUsbXchDramWrite( DP.UsbXchHandle, &WriteData1, 1, &Error );
                Result = CheckDramFlags( "After writing 2 ", NOT_EMPTY, NOT_OVERFLOW, IGNORE_PARTFULL );
                if ( !Result )
                        break;

                WriteData1 = 3;
                SrUsbXchDramWrite( DP.UsbXchHandle, &WriteData1, 1, &Error );
                Result = CheckDramFlags( "After writing 3 ", NOT_EMPTY, IS_OVERFLOW, IGNORE_PARTFULL );
                if ( !Result )
                        break;

                WriteData1 = 4;
                SrUsbXchDramWrite( DP.UsbXchHandle, &WriteData1, 1, &Error );
                Result = CheckDramFlags( "After writing 4 ", NOT_EMPTY, IS_OVERFLOW, IGNORE_PARTFULL );
                if ( !Result )
                        break;

                // Force exit from dummy while 1 loop
                break;

                } // end while 1



        ShowTime();


        // Initialize some values for read


        // This while loop is only to run once and only if the previous
        // loop was successful, it just provides an easy way (break) to
        // exit on error

        while ( Result ) {

                StartValue = 0x0L;

                PrintStrRpt( "\nBeginning read section, please wait ...\n" );

                for ( kk = 0 ; kk < Nloops ; kk++ ) {

                        for ( ii = 0 ; ii < DRAM_ARRAY_SIZE ; ii += 4 ) {
                                for ( jj = 0 ; jj < 4 ; jj++ ) {
                                        DP.ReadData[ii+jj]  = 0x0;
                                        DP.WriteData[ii+jj] =
                                                (char)((StartValue >> (jj*4)) & 0xF);
                                        }
                                StartValue++;
                                }


                        // Read the majority of values

                        if ( kk == Nloops-1 )
                                Nvalues = DRAM_ARRAY_SIZE - 4;
                        else
                                Nvalues = DRAM_ARRAY_SIZE;

                        Nread = SrUsbXchDramRead( DP.UsbXchHandle, DP.ReadData, Nvalues, &Error  );
        //                        sprintf( DP.Msg, "Loop %3ld at %7ld reading %ld points ", kk, StartValue*4, Nvalues );
                        sprintf( DP.Msg, "Loop %3ld: ", kk );
                        Result = CheckDramFlags( DP.Msg, NOT_EMPTY, IS_OVERFLOW, IGNORE_PARTFULL );
                        if ( !Result )
                                break;


        #ifdef SHOW_DBG_DATA
                        // Display values

                        jj = 3;
                        for ( ii = 0 ; ii < Nvalues ; ii++ ) {
                                jj++;
                                if (jj == 4) {
                                        jj = 0;
                                        sprintf( DP.Msg, "\n %2d: 0x ", ii );
                                        PrintStrRpt( DP.Msg );
                                        }
                                sprintf( DP.Msg, "%02X ", DP.ReadData[ii] );
                                PrintStrRpt( DP.Msg );
                                }
                        PrintStrRpt( "\n" );
        #endif // SHOW_DBG_DATA

                        for ( ii = 0 ; ii < Nvalues ; ii++ ) {
                                if ( DP.ReadData[ii] != DP.WriteData[ii] ) {
                                        if ( ShowCount < 20 ) {
                                                ShowCount++;
                                                sprintf( DP.Msg, "ERROR: Pt %7ld: Expected 0x%X but Read 0x%X\n",
                                                          ii, DP.WriteData[ii], DP.ReadData[ii] );
                                                PrintStrRpt( DP.Msg );
                                                Result = 0;
                                                break;
                                                }
                                        }
                                if ( !Result )
                                        break;
                                }

                        } // end for k


                // Read the last few values approaching empty

                PrintStrRpt( "\nReading last few points ...\n" );

                WriteData1 = 1;
                ReadData1  = 0;
                Nread = SrUsbXchDramRead( DP.UsbXchHandle, &ReadData1, 1, &Error );
                Result = CheckDramFlags( "After reading 1 ", NOT_EMPTY, IS_OVERFLOW, IGNORE_PARTFULL );
                if ( !Result )
                        break;
                Result = CheckReadValue( WriteData1, ReadData1, Nread, Error );
                if ( !Result )
                        break;

                WriteData1 = 2;
                ReadData1  = 0;
                Nread = SrUsbXchDramRead( DP.UsbXchHandle, &ReadData1, 1, &Error );
                Result = CheckDramFlags( "After reading 2 ", NOT_EMPTY, IS_OVERFLOW, IGNORE_PARTFULL );
                if ( !Result )
                        break;
                Result = CheckReadValue( WriteData1, ReadData1, Nread, Error );
                if ( !Result )
                        break;

                WriteData1 = 3;
                ReadData1  = 0;
                Nread = SrUsbXchDramRead( DP.UsbXchHandle, &ReadData1, 1, &Error );
                Result = CheckDramFlags( "After reading 3 ", IS_EMPTY, IS_OVERFLOW, IGNORE_PARTFULL );
                if ( !Result )
                        break;
                Result = CheckReadValue( WriteData1, ReadData1, Nread, Error );
                if ( !Result )
                        break;

                WriteData1 = 4;
                ReadData1  = 0;
                Nread = SrUsbXchDramRead( DP.UsbXchHandle, &ReadData1, 1, &Error );
                Result = CheckDramFlags( "After reading 4 ", IS_EMPTY, IS_OVERFLOW, IGNORE_PARTFULL );
                if ( !Result )
                        break;
                Result = CheckReadValue( WriteData1, ReadData1, Nread, Error );
                if ( !Result )
                        break;



                // Check final status (expecting empty and overflow)

                SrUsbXchDramReset( DP.UsbXchHandle );
                Result = CheckDramFlags( "Fifo ends with  ", IS_EMPTY, NOT_OVERFLOW, IGNORE_PARTFULL );
                if ( !Result )
                        break;

                // Force exit from dummy while Result loop
                break;

                } // end while Result


        ShowTime();

        if ( OpenClose )  CloseUsbXch( );


        if ( Result )
                PrintStrDate( "SUCCESS: DRAM FIFO overflow flag check complete\n" );
        else
                PrintStrDate( "FAILURE: DRAM FIFO overflow flag check complete\n" );

        return( Result );
}

//------------------------------------------------------------------------------
// ROUTINE: TestReset
// PURPOSE: 
//------------------------------------------------------------------------------
int TestReset( int OpenClose ) {

        int Result, Trys;

        PrintStrDate( "Reset running ... \n" );


        // Ensure that we start with driver closed

        if ( !OpenClose )  CloseUsbXch( );


        // First do a minimal open and try cycling the power to the port

        OpenUsbXchMinimal( DP.AtodDriverName );

        Result = SrUsbXchPortCycle( DP.UsbXchHandle );

        CloseUsbXch( );

        if ( Result ) {
                PrintStrRpt( "Port power cycle SUCCESS\n" );
                }
        else {
                PrintStrRpt( "Port power cycle FAILURE\n" );
                }



        // Wait while power cycle finished

        PrintStrRpt( "\nWaiting a few seconds for power cycle to finish ('q' to abort)\n" );

        Trys = 0;

        while ( !SrUsbXchIsResetComplete( DP.AtodDriverName ) ) {

                // Allow 'q' to abort

                if ( SrGetKeyCheck() == 'q' ) {
                        PrintStrRpt( "\n" );
                        PrintStr( "FAILURE: Reset stopped by keypress\n" );
                        return( 0 );
                        }

                // Cause abort after too many tries

                else if ( Trys > MAX_RESET_OPEN_TRIES ) {
                        sprintf( DP.Msg, "FAILURE: Reset stopped after %d tries to hardware open driver\n", Trys );
                        PrintStrRpt( "\n" );
                        PrintStr( DP.Msg );
                        return( 0 );
                        }

                // Otherwise wait a little longer

                else {
                        PrintStrRptFlush( "." );
                        SrSleep( 500 );       // half second
                        Trys++;
                        }

                } // end while

        PrintStrRpt( "\n" );



        // Now reopen the board and do the on board power cycle which closes the driver

        OpenUsbXchHardware( DP.AtodDriverName );

        DP.UsbXchHandle = SrUsbXchPowerReset( DP.UsbXchHandle );



        // Wait for reset to finish and driver is left closed

        PrintStrRpt( "\nWaiting a few seconds for driver reload ('q' to abort)\n" );

        Trys = 0;

        while ( !SrUsbXchIsResetComplete( DP.AtodDriverName ) ) {

                // Allow 'q' to abort

                if ( SrGetKeyCheck() == 'q' ) {
                        PrintStrRpt( "\n" );
                        PrintStr( "FAILURE: Reset stopped by keypress\n" );
                        return( 0 );
                        }

                // Cause abort after too many tries

                else if ( Trys > MAX_RESET_OPEN_TRIES ) {
                        sprintf( DP.Msg, "FAILURE: Reset stopped after %d tries to full open driver\n", Trys );
                        PrintStrRpt( "\n" );
                        PrintStr( DP.Msg );
                        return( 0 );
                        }

                // Otherwise wait a little longer

                else {
                        PrintStrRptFlush( "." );
                        SrSleep( 500 );       // half second
                        Trys++;
                        }

                } // end while

        PrintStrRpt( "\n" );


        // Check power

        OpenUsbXchHardware( DP.AtodDriverName );

        Result = SrUsbXchIsPowerGood( DP.UsbXchHandle );

        CloseUsbXch( );


        if ( Result ) {
                PrintStrRpt( "\nPower is GOOD\n" );
                PrintStrDate( "SUCCESS: Reset completed\n" );
                }
        else {
                PrintStrRpt( "\nPower is BAD\n" );
                PrintStrDate( "FAILURE: Reset completed\n" );
                }

        return( Result );
}

//------------------------------------------------------------------------------
// ROUTINE: TestRescan
// PURPOSE: 
//------------------------------------------------------------------------------
int TestRescan( int OpenClose ) {

        int      Result;

#if defined ( SROS_WIN2K ) || defined ( SROS_WINXP )        

        HMACHINE MachineHandle;
        DEVINST  DevRoot;
        DWORD    CmResult;


        PrintStrDate( "Rescan running ... \n" );


        // Ensure that we start with driver closed

        if ( !OpenClose )  CloseUsbXch( );



        // Initialize variables

        MachineHandle = NULL; // working with local machine only
        Result        = 0;


        // Get root device

        CmResult = CM_Locate_DevNode_Ex( &DevRoot,
                                         NULL,                     // request root device
                                         CM_LOCATE_DEVNODE_NORMAL, // get present device
                                         MachineHandle
                                       );
        
        // Re-enumerate all child devices

        if ( CmResult == CR_SUCCESS ) {

                PrintStrRpt( "\n  Located root device, please wait while rescan proceeds ... \n" );
                PrintStrRpt(   "  It may take a few minutes ...\n\n" );

                // WCT consider which is best
                // CM_REENUMERATE_NORMAL             - typical version
                // CM_REENUMERATE_RETRY_INSTALLATION - stronger version
                
                CmResult = CM_Reenumerate_DevNode_Ex( DevRoot,
                                                      CM_REENUMERATE_NORMAL,
                                                      MachineHandle );
                if( CmResult == CR_SUCCESS )
                        Result = 1;
                }

#else // SROS_LINUX

        Result = 0;
        PrintStrRpt( "Rescan not supported for Linux\n" );
        
#endif // SROS_xxx
        
        if ( Result )
                PrintStrDate( "SUCCESS: Rescan completed\n" );
        else 
                PrintStrDate( "FAILURE: Rescan not done\n" );

        return( Result );
}

//------------------------------------------------------------------------------
// ROUTINE: TestFirmRev
// PURPOSE: 
//------------------------------------------------------------------------------
int TestFirmRev( int OpenClose ) {

        int Result1, Result2, FirmRev, DriverRev;

        // Get driver revision number

        PrintStrDate( "Driver revision running ... \n" );


        if ( OpenClose )  OpenUsbXchHardware( DP.AtodDriverName );

        Result1 = SrUsbXchGetRevDriver( DP.UsbXchHandle, &DriverRev );

        if ( Result1 ) {
                sprintf( DP.Msg, "Driver revision number is %d\n", DriverRev );
                PrintStrRpt( DP.Msg );
                PrintStrDate( "SUCCESS: Driver revision completed\n" );
                }
        else {
                sprintf( DP.Msg, "Driver revision request failed\n" );
                PrintStrRpt( DP.Msg );
                PrintStrDate( "FAILURE: Driver revision completed\n" );
                }


        // Get firmware revision number

        PrintStrRpt( "\n" );
        PrintStrDate( "Firmware revision running ... \n" );

        Result2 = SrUsbXchGetRevFirmware( DP.UsbXchHandle, &FirmRev );

        if ( Result2 ) {
                sprintf( DP.Msg, "Firmware revision number is %d\n", FirmRev );
                PrintStrRpt( DP.Msg );
                PrintStrDate( "SUCCESS: Firmware revision completed\n" );
                }
        else {
                sprintf( DP.Msg, "Firmware revision request failed\n" );
                PrintStrRpt( DP.Msg );
                PrintStrDate( "FAILURE: Firmware revision completed\n" );
                }

        if ( OpenClose )  CloseUsbXch( );


        if ( Result1 && Result2 )
                return( 1 );
        else
                return( 0 );
}


//------------------------------------------------------------------------------
// ROUTINE: TestPacket
// PURPOSE: 
//------------------------------------------------------------------------------
int TestPacket( int OpenClose ) {

        int             Ans, Result, Error, Trys, MaxTry;
        unsigned long   NpacketsRequested, NpacketsRead;
        unsigned long   NpacketsTotal, Nread;
        SRUSBXCH_PACKET PacketArray[MAX_PACKET_REQ];


        PrintStrDate( "Read raw packet data running ... \n" );


        // Initialize variables

        Trys              = 1;
        MaxTry            = 8;
        NpacketsRequested = MAX_PACKET_REQ;
        NpacketsRead      = 0L;
        NpacketsTotal     = 0L;


        if ( OpenClose )  OpenUsbXchFull( DP.AtodDriverName, DP.AtodGpsModel );


        // Set sampling rate (NOTE: Change NTIMEPTS define if change sampling rate!)

        Result = SrUsbXchHelperSpsInit( DP.UsbXchHandle, USBXCH_SPS_INDEX_130, &Error );
        if ( !Result ) {
                PrintStrRpt("ERROR: Failed to set sampling rate\n\n" );
                PrintStrDate( "FAILURE: Read raw packet data complete\n" );
                return( 0 );
                }


        // Start A/D's

        Result = SrUsbXchStart( DP.UsbXchHandle, &Error );
        if ( !Result ) {
                sprintf( DP.Msg, "ERROR: Failed to start.  %s.\n\n", USBXCH_ERROR_MSG[Error] );
                PrintStrRpt( DP.Msg );
                PrintStrDate( "FAILURE: Read raw packet data complete\n" );
                return( 0 );
                }


        // If we're using PCTIME, send two time stamps so we get some
        // NMEA and equipment data (we need two since first one isn't
        // processed until the second arrives).

        if ( DP.AtodGpsModel == SRDAT_GPSMODEL_PCTIME ) {
                SrUsbXchSendTimeStamp( DP.UsbXchHandle, &Error );
                SrUsbXchSendTimeStamp( DP.UsbXchHandle, &Error );
                }

        sprintf( DP.Msg, "Press %s to stop this test if it is taking too long.\n", SR_QENTER );
        PrintStrRpt( DP.Msg );


        // Read Data

        while ( NpacketsRequested > 0L ) {


                // Read raw USB data packets

                Nread = SrUsbXchGetDataAsPackets(
                                                  DP.UsbXchHandle,
                                                 &PacketArray[NpacketsTotal],
                                                  NpacketsRequested,
                                                 &NpacketsRead,
                                                 &Error
                                                  );



                // Check for error during read

                if ( Error != USBXCH_ERROR_NONE ) {
                        sprintf( DP.Msg, "ERROR: SrUsbXchGetDataAsPackets failed with error %d = %s\nStopping acquisition.\n\n",
                                  Error, USBXCH_ERROR_MSG[Error] );
                        PrintStrRpt( DP.Msg );
                        SrUsbXchStop( DP.UsbXchHandle, &Error );
                        PrintStrDate( "FAILURE: Read raw packet data complete\n" );
                        return( 0 );
                        }


                // Check for data taking too long

                if ( Trys >= MaxTry ) {
                        PrintStrDate( "Could not read any data.  Stopping acquisition.\n" );
                        SrUsbXchStop( DP.UsbXchHandle, &Error );
                        PrintStrDate( "FAILURE: Read raw packet data complete\n" );
                        return( 0 );
                        }


                // Check for keypress to exit

                Ans = SrGetKeyCheck();
                if ( Ans != EOF ) {
                        PrintStrDate( "Keypress detected.  Stopping acquisition.\n" );
                        SrUsbXchStop( DP.UsbXchHandle, &Error );
                        PrintStrDate( "FAILURE: Read raw packet data complete\n" );
                        return( 0 );
                        }



                // Prepare to get any remaining data

                NpacketsRequested -= NpacketsRead;
                NpacketsTotal     += NpacketsRead;
                Trys++;


                // If we still need more packets, wait half a second before trying again

                if ( NpacketsRequested > 0L ) {
                        PrintStrRptFlush( "." );
                        SrSleep( 500 );
                        }

                } // end while


        PrintStrRpt( "\n" );


        // Stop A/D's

        Result = SrUsbXchStop( DP.UsbXchHandle, &Error );
        if ( !Result ) {
                sprintf( DP.Msg, "ERROR: Failed to stop A/D's.  Error %s.\n",
                              USBXCH_ERROR_MSG[Error]);
                PrintStrDate( DP.Msg );
                }


        if ( OpenClose )  CloseUsbXch( );



        // Display data

        DisplayPackets( PacketArray, NpacketsTotal, 20, 0 );


        if ( Result )
                PrintStrDate( "SUCCESS: Read raw packet data complete\n" );
        else
                PrintStrDate( "FAILURE: Read raw packet data complete\n" );

        return( Result );
}

//------------------------------------------------------------------------------
// ROUTINE: TestSanity
// PURPOSE: 
//------------------------------------------------------------------------------
int TestSanity( int OpenClose ) {

        int           i, j, Nread, Result;
        unsigned long NanalogReturned, NanalogRequested;

        PrintStrDate( "Read analog/sanity data running ... \n" );


        Result = ReadAll( OpenClose, READALL_REQUEST_ANALOG,
                          &NanalogRequested, &NanalogReturned, NULL, NULL );

        if ( Result ) {

                // Error check that we got all expected analog data

                if ( NanalogReturned != NanalogRequested ) {
                        sprintf( DP.Msg, "ERROR: Analog data NanalogReturned %lu, NanalogRequested %lu\n",
                                 NanalogReturned, NanalogRequested );
                        PrintStrRpt( DP.Msg );
                        }

                // Error check that the data is not all zero

                Nread = (int)NanalogReturned;

                Result = 0;
                for ( i = 0 ; i < Nread ; i++ ) {
                        for ( j = 0 ; j < NCHANNELS ; j++ ) {
                                if ( DP.UsbxchAnalogData[i].AnalogData[j] != 0x00000000 ) {
                                        Result = 1;
                                        break;
                                        }
                                }
                        if ( Result == 1 )
                                break;
                        }

                if ( Result == 0 ) {
                        sprintf( DP.Msg, "ERROR: Analog/Sanity, All analog values are exactly zero.\n" );
                        PrintStrRpt( DP.Msg );
                        }

                // Display the data

                DisplayAnalogData( NanalogReturned, DP.UsbxchAnalogData, NCHANNELS, 20, 0, 1 );

                }


        if ( Result )
                PrintStrDate( "SUCCESS: Read analog/sanity data complete\n" );
        else
                PrintStrDate( "FAILURE: Read analog/sanity data complete\n" );

        return( Result );
}

//------------------------------------------------------------------------------
// ROUTINE: TestSerial
// PURPOSE: 
//------------------------------------------------------------------------------
int TestSerial( int OpenClose ) {

        int           i, Ans, Nbytes, Error, Result, BytesRead, ShowCountChar;
        unsigned long TotalBytes;
        char          Bytes[MAXTESTNMEABYTES];


        PrintStrDate( "Reading serial NMEA strings running ... \n" );


        if ( OpenClose )  OpenUsbXchFull( DP.AtodDriverName, DP.AtodGpsModel );


        // Start A/D's

        PrintStrRpt("\nReading NMEA strings.  Press any key to stop.\n");

        Result = SrUsbXchStart( DP.UsbXchHandle, &Error );
        if ( !Result ) {
                sprintf( DP.Msg, "ERROR: Failed to start.  %s.\n\n", USBXCH_ERROR_MSG[Error] );
                PrintStrRpt( DP.Msg );
                PrintStrDate( "FAILURE: Reading serial NMEA strings complete\n" );
                return( 0 );
                }


        // If we're using PCTIME, send two time stamps so we get some
        // NMEA and equipment data (we need two since first one isn't
        // processed until the second arrives).

        if ( DP.AtodGpsModel == SRDAT_GPSMODEL_PCTIME ) {
                SrUsbXchSendTimeStamp( DP.UsbXchHandle, &Error );
                SrUsbXchSendTimeStamp( DP.UsbXchHandle, &Error );
                }


        Nbytes        = SRDAT_USBPACKET_SIZE; // = MAXTESTNMEABYTES - 1 = 65-1 = 64 = 0x40
        ShowCountChar = 0;
        TotalBytes    = 0L;

        while ( 1 ) {

                // Clear out byte array

                for ( i = 0 ; i < MAXTESTNMEABYTES ; i++ )
                        Bytes[i] = '\0';


                // Read NMEA strings and display

                BytesRead = SrUsbXchSerialRead( DP.UsbXchHandle, Bytes, Nbytes, &Error );


                TotalBytes += (unsigned long)BytesRead;

//                if (BytesRead > 0)
//                        printf( "%s", Bytes );


                for ( i = 0 ; i < BytesRead ; i++ ) {
                        if ( ShowCountChar || Bytes[i] < 0x0020 ) {
                                sprintf( DP.Msg, "0x%X\n", (Bytes[i] & 0x00FF) ); PrintStrRpt( DP.Msg );
                                ShowCountChar = 0;
                                }
                        else
                                sprintf( DP.Msg, "%c", Bytes[i] ); PrintStrRpt( DP.Msg );

                        if (i > 0 && Bytes[i-1] == '\r' && Bytes[i] == '\n')
                                ShowCountChar = 1;
                        }


                // Check for a keypress to quit

                Ans = SrGetKeyCheck( );
                if (Ans != EOF)
                        break;

                // Wait a bit

                SrSleep( 250 );

                } // end while


        PrintStrRpt( "\nStopped reading NMEA strings.\n\n" );

        Result = SrUsbXchStop( DP.UsbXchHandle, &Error );
        if ( !Result ) {
                sprintf( DP.Msg, "ERROR: Failed to stop.  %s.\n\n", USBXCH_ERROR_MSG[Error] );
                PrintStrRpt( DP.Msg );
                PrintStrDate( "FAILURE: Reading serial NMEA strings complete\n" );
                return( 0 );
                }


        if ( OpenClose )  CloseUsbXch( );


        if ( TotalBytes > 0L ) {
                PrintStrDate( "SUCCESS: Reading serial NMEA strings complete\n" );
                return( 1 );
                }
        else {
                PrintStrDate( "FAILURE: Reading serial NMEA strings complete\n" );
                return( 0 );
                }
}

//------------------------------------------------------------------------------
// ROUTINE: TestGetFrame
// PURPOSE: 
//------------------------------------------------------------------------------
int TestGetFrame( int OpenClose ) {

        int    Frame;
        time_t Time1, Time2;


        PrintStrDate( "Get USB frame number running ... \n" );

        if ( OpenClose )  OpenUsbXchHardware( DP.AtodDriverName );

        time( &Time1 );
        SrUsbXchGetFrameNumber( DP.UsbXchHandle, &Frame );
        time( &Time2 );

        if ( OpenClose )  CloseUsbXch( );


        sprintf( DP.Msg, "Frame number %4d occurred between time 0x%lX and 0x%lX\n",
                  Frame, Time1, Time2 );
        PrintStrRpt( DP.Msg );

//FIX - check for failure ?

        PrintStrDate( "SUCCESS: Get USB frame number complete\n" );

        return( 1 );
}


//------------------------------------------------------------------------------
// ROUTINE: TestPipeReset
// PURPOSE: 
//------------------------------------------------------------------------------
int TestPipeReset( int OpenClose ) {

        int Pipe, Result, ResultAll;

        PrintStrDate( "Pipe reset running ... \n" );

        if ( OpenClose )  OpenUsbXchHardware( DP.AtodDriverName );

        ResultAll = 1;

        for ( Pipe = USBXCH_PIPE_MIN ; Pipe < USBXCH_PIPE_MAX+1 ; Pipe++ ) {

                sprintf( DP.Msg, "Resetting pipe %d ...", Pipe );
                PrintStrDateFlush( DP.Msg );

                Result = SrUsbXchPipeReset( DP.UsbXchHandle, Pipe );
                if ( Result )
                        PrintStr( "SUCCESS\n" );
                else {
                        PrintStr( "FAILURE\n" );
                        ResultAll = 0;
                        }
                }

        if ( OpenClose )  CloseUsbXch( );



        if ( ResultAll )
                PrintStrDate( "SUCCESS: Pipe reset complete\n" );
        else
                PrintStrDate( "FAILURE: Pipe reset complete\n" );

        return( ResultAll );
}


//------------------------------------------------------------------------------
// ROUTINE: TestPipeAbort
// PURPOSE: 
//------------------------------------------------------------------------------
int TestPipeAbort( int OpenClose ) {

        int Pipe, Result, ResultAll;

        PrintStrDate( "Pipe abort running ... \n" );

        if ( OpenClose )  OpenUsbXchHardware( DP.AtodDriverName );

        ResultAll = 1;

        for ( Pipe = USBXCH_PIPE_MIN ; Pipe < USBXCH_PIPE_MAX+1 ; Pipe++ ) {

                sprintf( DP.Msg, "Aborting pipe %d ...", Pipe );
                PrintStrDateFlush( DP.Msg );

                Result = SrUsbXchPipeAbort( DP.UsbXchHandle, Pipe );
                if ( Result )
                        PrintStr( "SUCCESS\n" );
                else {
                        PrintStr( "FAILURE\n" );
                        ResultAll = 0;
                        }
                }

        if ( OpenClose )  CloseUsbXch( );


        if ( ResultAll )
                PrintStrDate( "SUCCESS: Pipe abort complete\n" );
        else
                PrintStrDate( "FAILURE: Pipe abort complete\n" );

        return( ResultAll );
}


//------------------------------------------------------------------------------
// ROUTINE: TestPortReset
// PURPOSE: 
//------------------------------------------------------------------------------
int TestPortReset( int OpenClose ) {

        int Result;

        PrintStrDate( "Port reset running ... \n" );

        if ( OpenClose )  OpenUsbXchMinimal( DP.AtodDriverName );

        Result = SrUsbXchPortReset( DP.UsbXchHandle );

        if ( OpenClose )  CloseUsbXch( );


        if ( Result )
                PrintStrDate( "SUCCESS: Port reset complete\n" );
        else
                PrintStrDate( "FAILURE: Port reset complete\n" );

        return( Result );
}

//------------------------------------------------------------------------------
// ROUTINE: TestPortCycle
// PURPOSE: 
//------------------------------------------------------------------------------
int TestPortCycle( int OpenClose ) {

        int Result;

        PrintStrDate( "Port cycle running ... \n" );

        if ( OpenClose )  OpenUsbXchMinimal( DP.AtodDriverName );

        Result = SrUsbXchPortCycle( DP.UsbXchHandle );

        if ( OpenClose )  CloseUsbXch( );


        if ( Result )
                PrintStrDate( "SUCCESS: Port cycle complete\n" );
        else
                PrintStrDate( "FAILURE: Port cycle complete\n" );

        return( Result );
}


// Helper and display functions
//
//
//

//------------------------------------------------------------------------------
// ROUTINE: IsDescriptorGood
// PURPOSE: 
//------------------------------------------------------------------------------
int IsDescriptorGood( USB_DEVICE_DESCRIPTOR *Target ) {

        if ( (Target)                               &&
             (Target->bLength            == 18)     &&
             (Target->bDescriptorType    == 1)      &&
             (Target->bcdUSB             == 512)    &&
             (Target->bDeviceClass       == 0xFF)   &&
             (Target->bDeviceSubClass    == 0xFF)   &&
             (Target->bDeviceProtocol    == 0xFF)   &&
             (Target->bMaxPacketSize0    == 64)     &&
             (Target->idVendor           == 0x15d3) &&
             (Target->idProduct          == 0x5504) &&
//           (Target->bcdDevice          == 0)      && // DID changes for multiple boards
             (Target->iManufacturer      == 0)      &&
             (Target->iProduct           == 0)      &&
             (Target->iSerialNumber      == 0)      &&
             (Target->bNumConfigurations == 1)      )
                return( 1 );
        else
                return( 0 );

}

//------------------------------------------------------------------------------
// ROUTINE: ReadAll
// PURPOSE: 
//------------------------------------------------------------------------------
int ReadAll( int OpenClose, int DataRequested,
             unsigned long *NanalogRequested,
             unsigned long *NanalogReturned,
             unsigned long *NserialReturned,
             unsigned long *NequipReturned ) {

        int             Ans, Result, Error, SrDatError, Trys, MaxTry;
        int             CheckSanity, LastMsg;
        unsigned long   NpacketsRequested, NpacketsRead, NpacketsRemain;
        unsigned long   NpacketsTotal, Nread, NpacketsProcessed;
        unsigned long   NpacketsProcessedData, NpacketsProcessedStatus;
        unsigned long   NanalogRet, NserialRet, NequipRet;
        unsigned long   NanalogRequest, NserialRequest, NequipRequest;
        SRUSBXCH_PACKET PacketArray[MAX_PACKET_REQ];



        // Initialize variables

        Trys              = 1;
        MaxTry            = 8;
        CheckSanity       = 1;
        LastMsg           = -1;
        NpacketsRequested = MAX_PACKET_REQ;
        NpacketsRead      = 0L;
        NpacketsTotal     = 0L;
        NanalogRequest    = NTIMEPTS; // (MAXANALOG/2)
        NserialRequest    = MAXSERIAL;
        NequipRequest     = MAXSERIAL;


        if ( OpenClose )  OpenUsbXchFull( DP.AtodDriverName, DP.AtodGpsModel );


        // Set sampling rate (NOTE: Change NTIMEPTS define if change sampling rate!)

        Result = SrUsbXchHelperSpsInit( DP.UsbXchHandle, USBXCH_SPS_INDEX_130, &Error );
        if ( !Result ) {
                PrintStrRpt("ERROR: Failed to set sampling rate\n" );
                return( 0 );
                }


        // Start A/D's

        Result = SrUsbXchStart( DP.UsbXchHandle, &Error );
        if ( !Result ) {
                sprintf( DP.Msg, "ERROR: Failed to start.  %s.\n\n", USBXCH_ERROR_MSG[Error] );
                PrintStrRpt( DP.Msg );
                return( 0 );
                }


        // If we're using PCTIME, send two time stamps so we get some
        // NMEA and equipment data (we need two since first one isn't
        // processed until the second arrives).

        if ( DP.AtodGpsModel == SRDAT_GPSMODEL_PCTIME ) {
                SrUsbXchSendTimeStamp( DP.UsbXchHandle, &Error );
                SrUsbXchSendTimeStamp( DP.UsbXchHandle, &Error );
                }


        sprintf( DP.Msg, "Press %s to stop this test if it is taking too long.\n", SR_QENTER );
        PrintStrRpt( DP.Msg );


        // Read Data

        while ( NpacketsRequested > 0L ) {


                // Read raw USB data packets

                Nread = SrUsbXchGetDataAsPackets(
                                                  DP.UsbXchHandle,
                                                 &PacketArray[NpacketsTotal],
                                                  NpacketsRequested,
                                                 &NpacketsRead,
                                                 &Error
                                                  );

                // Check for error during read

                if ( Error != USBXCH_ERROR_NONE ) {
                        sprintf( DP.Msg, "ERROR: SrUsbXchGetDataAsPackets failed with error %d = %s\nStopping acquisition.\n",
                                  Error, USBXCH_ERROR_MSG[Error] );
                        PrintStrDate( DP.Msg );
                        SrUsbXchStop( DP.UsbXchHandle, &Error );
                        return( 0 );
                        }


                // Check for data taking too long

                if ( Trys >= MaxTry ) {
                        PrintStrDate( "ERROR: Could not read any data.  Stopping acquisition.\n" );
                        SrUsbXchStop( DP.UsbXchHandle, &Error );
                        return( 0 );
                        }


                // Check for keypress to exit

                Ans = SrGetKeyCheck();
                if ( Ans != EOF ) {
                        PrintStrDate( "Keypress detected.  Stopping acquisition.\n" );
                        SrUsbXchStop( DP.UsbXchHandle, &Error );
                        return( 0 );
                        }

                // Prepare to get any remaining data

                NpacketsRequested -= NpacketsRead;
                NpacketsTotal     += NpacketsRead;
                Trys++;


                // If we got some data, then reset number of trys since
                // that is just to prevent endless waiting when NO data
                // is available

                if ( NpacketsRead > 0 )
                        Trys = 1;



                // If we still need more packets, wait half a second before trying again

                if ( NpacketsRequested > 0L ) {
                        PrintStrRptFlush( "." );
                        SrSleep( 500 );
                        }

                } // end while

        PrintStrRpt( "\n" );


        // Stop A/D's

        Result = SrUsbXchStop( DP.UsbXchHandle, &Error );
        if ( !Result ) {
                sprintf( DP.Msg, "ERROR: Could not stop acquisition.  %s.\n",
                             USBXCH_ERROR_MSG[Error] );
                PrintStrRpt( DP.Msg );
                }

        if ( OpenClose )  CloseUsbXch( );


        // Request processing of all packets, but don't exceed analog array size

        if ( DataRequested == READALL_REQUEST_SERIAL ||
             DataRequested == READALL_REQUEST_EQUIP  )
                NanalogRequest = NpacketsTotal*2;

        if ( NanalogRequest > MAXANALOG )
                NanalogRequest = MAXANALOG;


        // Process the data

        Nread = SrUsbXchConvertPackets(
                                          PacketArray,
                                          NpacketsTotal,
                                          CheckSanity,
                                         &LastMsg,
                                          DP.UsbxchAnalogData,
                                          NanalogRequest,
                                         &NanalogRet,
                                          DP.UsbxchSerialData,
                                          NserialRequest,
                                         &NserialRet,
                                          DP.UsbxchEquipData,
                                          NequipRequest,
                                         &NequipRet,
                                         &NpacketsProcessed,
                                         &NpacketsProcessedData,
                                         &NpacketsProcessedStatus,
                                         &SrDatError
                                          );

        // Check for processing error

        if ( Error != USBXCH_ERROR_NONE ) {
                sprintf( DP.Msg, "ERROR: SrUsbXchConvertPackets failed with error %d = %s\n",
                          Error, USBXCH_ERROR_MSG[Error] );
                PrintStrRpt( DP.Msg );
                if ( NanalogRequested ) *NanalogRequested = NanalogRequest;
                if ( NanalogReturned )  *NanalogReturned  = NanalogRet;
                if ( NserialReturned )  *NserialReturned  = NserialRet;
                if ( NequipReturned )   *NequipReturned   = NequipRet;
                return( 0 );
                }


        // Clean up ALL packets from array after processing
        // (usually only the used packets are removed)

        NpacketsRemain = SrDatUsbPacketCleanUp(
                                                PacketArray,
                                                NpacketsTotal,
                                                NpacketsTotal, // usually NpacketsProcessed
                                               &SrDatError
                                               );

        if ( Error != USBXCH_ERROR_NONE ) {
                sprintf( DP.Msg, "ERROR: SrDatUsbPacketCleanUp failed with error %d = %s\n",
                          Error, USBXCH_ERROR_MSG[Error] );
                PrintStrRpt( DP.Msg );
                if ( NanalogRequested ) *NanalogRequested = NanalogRequest;
                if ( NanalogReturned )  *NanalogReturned  = NanalogRet;
                if ( NserialReturned )  *NserialReturned  = NserialRet;
                if ( NequipReturned )   *NequipReturned   = NequipRet;
                return( 0 );
                }


        if ( NanalogRequested ) *NanalogRequested = NanalogRequest;
        if ( NanalogReturned )  *NanalogReturned  = NanalogRet;
        if ( NserialReturned )  *NserialReturned  = NserialRet;
        if ( NequipReturned )   *NequipReturned   = NequipRet;
        return( 1 );
}

//------------------------------------------------------------------------------
// ROUTINE: ShowTime
// PURPOSE: 
//------------------------------------------------------------------------------
void ShowTime( void ) {

        char *datestr, dbuffer[16];
        char *timestr, tbuffer[16];

        time_t     current_time;
        struct tm *local_time;

        // Get the date and time.

        time( &current_time );
        local_time = localtime( &current_time );

        strftime( dbuffer, 16, "%m/%d/%y", local_time );
        strftime( tbuffer, 16, "%H:%M:%S", local_time );

        datestr = dbuffer;
        timestr = tbuffer;

        sprintf( DP.Msg, "Time is %s\n", timestr );
        PrintStrRpt( DP.Msg );
}

//------------------------------------------------------------------------------
// ROUTINE: ShowDramData
// PURPOSE: 
//------------------------------------------------------------------------------
void ShowDramData( char *WriteData, char *ReadData ) {

        int i;

        PrintStrRpt( " #   Expected  Found\n" );
        for ( i = 0 ; i < DRAM_TEST_SIZE ; i++ ) {
                sprintf( DP.Msg, "%2d:    %2d       %2d\n",
                          i, WriteData[i], ReadData[i] );
                PrintStrRpt( DP.Msg );
                }
}

//------------------------------------------------------------------------------
// ROUTINE: CheckDramResults
// PURPOSE: 
//------------------------------------------------------------------------------
int CheckDramResults( char *WriteData, char *ReadData,
                      int Nvalues, int Nwrote, int Nread ) {
        int i;

        if ( Nvalues != Nwrote ) {
                sprintf( DP.Msg, "Requested %d but wrote %d\n", Nvalues, Nwrote );
                PrintStrRpt( DP.Msg );
                return( 0 );
                }

        if ( Nwrote != Nread ) {
                sprintf( DP.Msg, "Wrote %d but read %d\n", Nwrote, Nread );
                PrintStrRpt( DP.Msg );
                return( 0 );
                }

        for ( i = 0 ; i < Nread ; i++ ) {
                if ( WriteData[i] != ReadData[i] ) {
                        sprintf( DP.Msg, "Data #%d written as 0x%02X but read as 0x%02X\n",
                                 i, (WriteData[i]&0x000000FF), (ReadData[i]&0x000000FF) );
                        PrintStrRpt( DP.Msg );
                        return( 0 );
                        }
                }

        return( 1 );
}

//------------------------------------------------------------------------------
// ROUTINE: ShowDramFlags
// PURPOSE: 
//------------------------------------------------------------------------------
void ShowDramFlags( char *Msg, int *Empty, int *Overflow, int *PartFull ) {

        // Display intro message

        if ( Msg )
                PrintStrRpt( Msg );


        // Display empty status if available

        if ( Empty != NULL ) {
                if ( *Empty == 1 )
                        PrintStrRpt( "Empty is TRUE " );
                else
                        PrintStrRpt( "Empty is FALSE" );
                }



        // Display joining word if needed

        if ( Empty != NULL  && Overflow != NULL )
                PrintStrRpt( " and " );



        // Display overflow status if available

        if ( Overflow != NULL ) {
                if ( *Overflow == 1 )
                        PrintStrRpt( "Overflow is TRUE" );
                else
                        PrintStrRpt( "Overflow is FALSE" );
                }


        // Display joining word if needed

        if ( (Empty != NULL || Overflow != NULL)  &&  PartFull != NULL )
                PrintStrRpt( " and " );



        // Display partfull status if available

        if ( PartFull != NULL ) {
                if ( *PartFull == 1 )
                        PrintStrRpt( "PartFull is TRUE" );
                else
                        PrintStrRpt( "PartFull is FALSE" );
                }


        // Display end of message

        PrintStrRpt( "\n" );
}

//------------------------------------------------------------------------------
// ROUTINE: CheckDramFlags
// PURPOSE: 
//------------------------------------------------------------------------------
int CheckDramFlags( char *Msg, int ExpectedEmpty, int ExpectedOverflow,
                     int ExpectedPartFull ) {

        int Empty, Overflow, PartFull;

        SrUsbXchDramFlags( DP.UsbXchHandle, &Empty, &PartFull, &Overflow );

        ShowDramFlags( Msg, &Empty, &Overflow, &PartFull );

//FIX - Remove once empty flag is working right

//        if (ExpectedEmpty != IGNORE_EMPTY) {
//                if ( Empty != ExpectedEmpty ) {
//                        sprintf( Msg, "ERROR: Empty %s, but expected %s\n",
//                                  TrueFalse[Empty],TrueFalse[ExpectedEmpty] );
//                        PrintStrRpt( Msg );
//                        return( 0 );
//                        }
//                }

        if (ExpectedOverflow != IGNORE_OVERFLOW) {
                if ( Overflow != ExpectedOverflow ) {
                        sprintf( Msg, "ERROR: Overflow %s, but expected %s\n",
                                  TrueFalse[Overflow],TrueFalse[ExpectedOverflow] );
                        PrintStrRpt( Msg );
                        return( 0 );
                        }
                }

        if (ExpectedPartFull != IGNORE_PARTFULL) {
                if ( PartFull != ExpectedPartFull ) {
                        sprintf( Msg, "ERROR: PartFull %s, but expected %s\n",
                                  TrueFalse[PartFull],TrueFalse[ExpectedPartFull] );
                        PrintStrRpt( Msg );
                        return( 0 );
                        }
                }
        return( 1 );
}

//------------------------------------------------------------------------------
// ROUTINE: CheckReadValue
// PURPOSE: 
//------------------------------------------------------------------------------
int CheckReadValue( char ExpectedValue, char ReadValue, int Nread, int Error ) {

        if ( Nread != 1 || Error != USBXCH_ERROR_NONE) {
                sprintf( DP.Msg, "ERROR: Failed to read (nread %d, error %d)\n",
                          Nread, Error );
                PrintStrRpt( DP.Msg );
                return( 0 );
                }

        if ( ReadValue != ExpectedValue ) {
                sprintf( DP.Msg, "ERROR: Expected read value to be %d not %d\n",
                          ExpectedValue, ReadValue );
                PrintStrRpt( DP.Msg );
                return( 0 );
                }

        return( 1 );
}

//------------------------------------------------------------------------------
// ROUTINE: DisplayPackets
// PURPOSE: 
//------------------------------------------------------------------------------
void DisplayPackets( SRUSBXCH_PACKET *PacketArray, unsigned long NpacketsRead,
                     int MaxPrint, int StartCount ) {

        int i, ip, Line, PacketSize, Nprint;

        Nprint = (int) NpacketsRead;
        if ( Nprint > MaxPrint )
                Nprint = MaxPrint;

        sprintf( DP.Msg, "Displaying %d of %lu packets ...\n\n", Nprint, NpacketsRead );
        PrintStrRpt( DP.Msg );

        PacketSize = 64;

        for ( ip = 0 ; ip < Nprint ; ip++ ) {
                Line = 0;
                PrintStrRpt( "  " );
                for ( i = 0 ; i < PacketSize ; i++ ) {
                        sprintf( DP.Msg, "%02X ", (PacketArray[ip].Bytes[i] & 0x00FF) );
                        PrintStrRpt( DP.Msg );
                        Line++;
                        if (Line > 15) {
                                PrintStrRpt( "\n  " );
                                Line = 0;
                                }
                        }
                PrintStrRpt( "\n" );
                }

}

//------------------------------------------------------------------------------
// ROUTINE: DisplayAnalogData
// PURPOSE: 
//------------------------------------------------------------------------------
void DisplayAnalogData( int NanalogRead, SRDAT_USBXCH_ANALOG *UsbxchAnalogData,
                        int Nchan, int MaxPrint, int StartCount, int ShowExtras ) {

        int i, j, Nprint, SampleCount;

        Nprint = NanalogRead;
        if (Nprint > MaxPrint)
                Nprint = MaxPrint;

        SampleCount = StartCount;
        PrintStrRpt( "Analog Data:\n" );
        if ( ShowExtras )
                PrintStrRpt( "Cnt   CH00      CH01      CH02      CH03    PpsEvent San Pps Dig GGA   OBC\n" );
        else
                PrintStrRpt( "Cnt   CH00      CH01      CH02      CH03\n" );

        for ( i = 0 ; i < Nprint ; i++ ) {
                sprintf( DP.Msg, "%3d ", SampleCount );                                PrintStrRpt( DP.Msg );
                SampleCount++;

                // Show Channels
                for ( j = 0 ; j < Nchan ; j++ ) {
                        sprintf( DP.Msg, "0x%06lX  ",
                                  (MASKVAL & UsbxchAnalogData[i].AnalogData[j]) );  PrintStrRpt( DP.Msg );
//SHOWDECIMAL           sprintf( DP.Msg, "%8d  ", UsbxchAnalogData[i].AnalogData[j] ); PrintStrRpt( DP.Msg );
                        }

                if ( ShowExtras ) {

                        sprintf( DP.Msg, "%08X  %02X  %02X  %02X  %02X %08X",
                                  UsbxchAnalogData[i].PpsEventNum,
                                  UsbxchAnalogData[i].Sanity,
                                  UsbxchAnalogData[i].PpsToggle,
                                  UsbxchAnalogData[i].DigitalIn,
                                  UsbxchAnalogData[i].GGACount,
                                  UsbxchAnalogData[i].OnBoardCount
                                  );                                                PrintStrRpt( DP.Msg );
                        if ( UsbxchAnalogData[i].OnBoardCount == 0L )
                                PrintStrRpt( "\n" );
                        else
                                PrintStrRpt( " <<\n" );

                        } // end ShowExtras
                else
                        PrintStrRpt( "\n" );


                } // end for i < Nprint

        PrintStrRpt( "\n" );
}

//------------------------------------------------------------------------------
// ROUTINE: DisplaySerialData
// PURPOSE: 
//------------------------------------------------------------------------------
void DisplaySerialData( int NserialRead, SRDAT_USBXCH_SERIAL *UsbxchSerialData ) {

        int          k;
        unsigned int i, j, NmeaCount, PpsCount;


        // Display serial NMEA data

        if ( NserialRead == 0 )
                PrintStrRpt("No serial data found.  Is GPS attached?\n\n");
        else {

                for ( k = 0 ; k < NserialRead ; k++ ) {
                        PpsCount = (unsigned int)UsbxchSerialData[k].PpsEventNum;
                        sprintf( DP.Msg, "Serial Data with %d NMEA strings and PpsEventNum = %u:\n",
                                  UsbxchSerialData[k].NmeaCount, PpsCount );
                        PrintStrRpt( DP.Msg );

                        NmeaCount = UsbxchSerialData[k].NmeaCount;
                        for ( i = 0 ; i < NmeaCount ; i++ ) {
                                j = i * SRDAT_NMEA_MAX_SIZE;
                                sprintf( DP.Msg, "%02d: %s", i, &UsbxchSerialData[k].NmeaMsg[j] );
                                PrintStrRpt( DP.Msg );
                                }
                        PrintStrRpt( "\n" );
                        }
                } // end else
}

//------------------------------------------------------------------------------
// ROUTINE: DisplayEquipData
// PURPOSE: 
//------------------------------------------------------------------------------
void DisplayEquipData( int NequipRead, SRDAT_USBXCH_EQUIPMENT *UsbxchEquipData ) {

        int          i, PowerHistory, PowerCurrent;
        unsigned int PpsCount;
        double       TempC, TempF;
        char        *PowerString;

        if ( NequipRead == 0 )
                PrintStrRpt("No equipment data found\n");
        else {

                for ( i = 0 ; i < NequipRead ; i++ ) {

                        SrDatUsbTemperatureCompute( UsbxchEquipData[i].Temperature, &TempC, &TempF );
                        PpsCount = (unsigned int)UsbxchEquipData[i].PpsEventNum;


                        if ( SrUsbXchIsPowerBitGood( UsbxchEquipData[i].VoltageGood,
                                USBXCH_POWER_HISTORY ) )
                                PowerHistory = 1;
                        else
                                PowerHistory = 0;

                        if ( SrUsbXchIsPowerBitGood( UsbxchEquipData[i].VoltageGood,
                                USBXCH_POWER_CURRENT ) )
                                PowerCurrent = 1;
                        else
                                PowerCurrent = 0;


                        if ( PowerHistory && PowerCurrent )
                                PowerString = "Power is GOOD";
                        else if ( PowerCurrent )
                                PowerString = "Power currently GOOD, but was bad";
                        else if ( PowerHistory )
                                PowerString = "Power currently BAD, but was good";
                        else
                                PowerString = "Power is BAD";




                        sprintf( DP.Msg, "Equipment Data %d:\n", i );                  PrintStrRpt( DP.Msg );
                        sprintf( DP.Msg, "   PPS Event #  = %d\n", PpsCount );         PrintStrRpt( DP.Msg );
                        sprintf( DP.Msg, "   Voltage Good = %d      (%s)\n",
                                                  UsbxchEquipData[i].VoltageGood,
                                                  PowerString );                    PrintStrRpt( DP.Msg );
                        sprintf( DP.Msg, "   Temperature  = 0x%04X (Degrees %5.1lf C, %3.0lf F)\n",
                                                  UsbxchEquipData[i].Temperature,
                                                  TempC, TempF );                   PrintStrRpt( DP.Msg );
                        PrintStrRpt( "\n" );
                        }
                } // end else
}

// USBXCH device functions
//
//
//

//------------------------------------------------------------------------------
// ROUTINE: OpenUsbXchFull
// PURPOSE: 
//------------------------------------------------------------------------------
void OpenUsbXchFull( char *UsbxchDriverName, int GpsModel ) {

        int Error, UserCfgByte;


        // Prepare default user config byte

        SrUsbXchFillUserCfgByte( USERCFG_NMEA_TX_IDLE_DEFAULT,
                                 USERCFG_NMEA_RX_IDLE_DEFAULT,
                                 USERCFG_PPS_LED_DEFAULT,
                                 USERCFG_PPS_IDLE_DEFAULT,
                                 USERCFG_DB25_ADRDY_DEFAULT,
                                 USERCFG_UNUSED_DEFAULT,
                                 USERCFG_TRIGGER_IDLE_DEFAULT,
                                 &UserCfgByte );


        // Open USBXCH driver

        DP.UsbXchHandle = SrUsbXchOpen( UsbxchDriverName,
                                     SRDAT_ATODMODEL_USB4CH,
                                     GpsModel,
                                     UserCfgByte,
                                     USBXCH_SPS_DEFAULT,
                                     NULL,
                                     &Error );

        if ( DP.UsbXchHandle == BAD_DEVHANDLE ) {
                sprintf( DP.Msg, "ERROR: Failed to open USB driver %s.\n       %s (Err %d)\n\n",
                               UsbxchDriverName, USBXCH_ERROR_MSG[Error], Error );
                PrintStrRpt( "\n" );
                PrintStrDate( DP.Msg );
                CleanExit( 1 );
                }

        if ( Error != USBXCH_ERROR_NONE ) {
                sprintf( DP.Msg, "WARNING: USB driver %s full open return is\n       %s (Err %d)\n\n",
                               UsbxchDriverName, USBXCH_ERROR_MSG[Error], Error );
                PrintStrRpt( "\n" );
                PrintStrDate( DP.Msg );
                }

}

//------------------------------------------------------------------------------
// ROUTINE: OpenUsbXchHardware
// PURPOSE: 
//------------------------------------------------------------------------------
void OpenUsbXchHardware( char *UsbxchDriverName ) {

        int Error;


        // Open USBXCH driver and prepare the hardware

        DP.UsbXchHandle = SrUsbXchHardwareOpen( UsbxchDriverName,
                                                SrUsbXch_HwData,
                                                &Error );

        if ( DP.UsbXchHandle == BAD_DEVHANDLE ) {
                sprintf( DP.Msg, "ERROR: Failed to open USB driver %s.\n       %s (Err %d)\n\n",
                               UsbxchDriverName, USBXCH_ERROR_MSG[Error], Error );
                PrintStrRpt( "\n" );
                PrintStrDate( DP.Msg );
                CleanExit( 1 );
                }
        if ( Error != USBXCH_ERROR_NONE ) {
                sprintf( DP.Msg, "WARNING: USB driver %s hardware open return is\n       %s (Err %d)\n\n",
                               UsbxchDriverName, USBXCH_ERROR_MSG[Error], Error );
                PrintStrRpt( "\n" );
                PrintStrDate( DP.Msg );
                }
}

//------------------------------------------------------------------------------
// ROUTINE: OpenUsbXchMinimal
// PURPOSE: 
//------------------------------------------------------------------------------
void OpenUsbXchMinimal( char *UsbxchDriverName ) {

        int Error;


        // Open USBXCH driver and prepare the hardware

        DP.UsbXchHandle = SrUsbXchMinimalOpen( UsbxchDriverName,
                                            &Error );

        if ( DP.UsbXchHandle == BAD_DEVHANDLE ) {
                sprintf( DP.Msg, "ERROR: Failed to open USB driver %s.\n       %s (Err %d)\n\n",
                               UsbxchDriverName, USBXCH_ERROR_MSG[Error], Error );
                PrintStrRpt( "\n" );
                PrintStrDate( DP.Msg );
                CleanExit( 1 );
                }
        if ( Error != USBXCH_ERROR_NONE ) {
                sprintf( DP.Msg, "WARNING: USB driver %s minimal open return is\n       %s (Err %d)\n\n",
                               UsbxchDriverName, USBXCH_ERROR_MSG[Error], Error );
                PrintStrRpt( "\n" );
                PrintStrDate( DP.Msg );
                }
}

//------------------------------------------------------------------------------
// ROUTINE: OpenUsbXchVerify
// PURPOSE: 
//------------------------------------------------------------------------------
int OpenUsbXchVerify( char *UsbxchDriverName ) {

        int Error;


        // Open USBXCH driver and prepare the hardware

        DP.UsbXchHandle = SrUsbXchMinimalOpen( UsbxchDriverName, &Error );

        if ( DP.UsbXchHandle == BAD_DEVHANDLE )
                return( 0 );

        if ( ! SrUsbXchHelperDownload( DP.UsbXchHandle, SrUsbXch_HwData ) ) {
                CloseUsbXch();
                return( 0 );
                }

        if ( ! SrUsbXchHelper8051RunHold( DP.UsbXchHandle, USB8051_CPU_RUN ) ) {
                CloseUsbXch();
                return( 0 );
                }

        if ( ! SrUsbXchIsPowerGood( DP.UsbXchHandle ) ) {
                CloseUsbXch();
                return( 0 );
                }

        return( 1 );
}

//------------------------------------------------------------------------------
// ROUTINE: CloseUsbXch
// PURPOSE: 
//------------------------------------------------------------------------------
void CloseUsbXch( void ) {

        // Close USBXCH driver

        if ( !SrUsbXchClose( DP.UsbXchHandle ) ) {
                sprintf( DP.Msg, "ERROR: Failed to close USB driver %s\n\n", DP.AtodDriverName );
                PrintStrDate( DP.Msg );
                CleanExit( 1 );
                }

        DP.UsbXchHandle = BAD_DEVHANDLE;
}




// COMMAND LINE PROCESSING FUNCTIONS:
//
// These helper functions set the test options based on the values given
// on the command line.
//


//------------------------------------------------------------------------------
// ROUTINE: ProcessCommandLine
// PURPOSE: 
//------------------------------------------------------------------------------
void ProcessCommandLine( int argc, char *argv[] ) {

        int i, j, DriverNum, DriverNameFound;
        TR *t;
        DO *d;

        char MsgBuffer[MAXSTR];


        // Diag with no arguments at all gives the usage message.

        if ( argc == 1 )

                Usage( );



        // Set defaults

        SrStringCopy( DP.AtodDriverName, USBXCH_DEFAULT_DRIVER_NAME, MAXSTR );
        DriverNameFound  = 0;
        DriverNum        = 0;


        
        // Scan for valid arguments.

        for ( i = 1 ; i < argc ; i++ ) {


                // Check for a "/??" usage message request.

                if ( StrEqu( argv[i], "/??") ) {

                        DP.ShowHidden = 1;
                        Usage( );
                        }



                // Check for a "/?" usage message request.

                if ( StrEqu( argv[i], "/?") )

                        Usage( );



                // Check for a valid driver name

                else if ( SrUsbXchIsValidDriverName( argv[i], &DriverNum ) ) {
                        
                        SrStringCopy( DP.AtodDriverName, argv[i], MAXSTR );

                        if ( !DriverNameFound )
                                DriverNameFound = 1;

                        else
                               ErrorCmd( "multiple driver arguments" );
                        }


                // Check for a "gn" time source.

                else if ( t = TimeTableLookUp( argv[i] ) ) {

                        DP.AtodGpsModel = t->AtodGpsModel;
                        DP.AtodGpsModelName = SrDatGpsModelString( DP.AtodGpsModel );

                        if ( DP.pTimeSourceSelected == NULL )
                                DP.pTimeSourceSelected = t;

                        else
                                ErrorCmd( "multiple \"gn\" time sources" );
                        }


                // Check for a valid test action.

                else if ( d = DiagTestLookUp( argv[i] ) ) {

                        if ( DP.pDiagTestSelected == NULL )
                                DP.pDiagTestSelected = d;

                        else
                                ErrorCmd( "multiple test names" );
                        }

                

                // Unknown argument specified.

                else {
                        sprintf( MsgBuffer, "unknown argument > %s", argv[i] );
                        ErrorCmd( MsgBuffer );
                        }


                } // end for i



        // Collect command line arguments into a single string to save

        j = 0;
        for ( i = 0 ; i < argc ; i++ ) {
                sprintf( &DP.CmdLine[j], "%s ", argv[i] );
                j = strlen( DP.CmdLine );
                }
        DP.CmdLine[j-1] = '\0'; // truncate last space

        return;
}

//------------------------------------------------------------------------------
// ROUTINE: TimeTableLookUp
// PURPOSE: Compare s against the time source table "gn" names.
//------------------------------------------------------------------------------
TR *TimeTableLookUp( char *s ) {

        TR *t;


        // Scan the time source table for a match with s.

        for ( t = TimeTable ; t->GnString ; t++ ) {

                if ( StrEqu( s, t->GnString ) )

                        return( t );  // > success, got a match ...
                }


        return( NULL );  // > failure, hit end of table without a match ...
}

//------------------------------------------------------------------------------
// ROUTINE: DiagTestLookUp
// PURPOSE: 
//------------------------------------------------------------------------------
DO *DiagTestLookUp( char *testname ) {

        DO *d;

        // Search for matching test.

        for ( d = DiagTestTable ; d->name ; d++ ) {

                if ( StrEqu( testname, d->name ) )

                        return( d ) ; // success, got a match ...
                }

        return( NULL ); // failure, hit end of table without a match ...
}

//------------------------------------------------------------------------------
// ROUTINE: Usage
// PURPOSE: Display the usage message.
//------------------------------------------------------------------------------
void Usage( void ) {

        int i;
        TR *t;
        DO *d;


        // Print the top section of the usage message.

        for ( i = 0 ; UsageMsg[i] ; i++ ) {

                sprintf( DP.Msg, "%s\n", UsageMsg[i] );
                if ( i == 0 )
                        PrintStrDate( DP.Msg );
                else
                        PrintStrRpt( DP.Msg );
                }



        // Print allowed test table selections.

        for ( ++i, d = DiagTestTable ; d->name ; d++ ) {

                if ( ((d->hidden == 0) || (DP.ShowHidden)) && (d->text) ) {
                        
                        sprintf( DP.Msg, UsageMsg[i],  // < fixed format string
                                 d->name,
                                 d->text
                                 );
                        PrintStrRpt( DP.Msg );
                        }
                }



        // Print the middle (time table header) section of the usage message.

        for ( ++i ; UsageMsg[i] ; i++ ) {

                sprintf( DP.Msg, "%s\n", UsageMsg[i] );
                PrintStrRpt( DP.Msg );
                }



        // Print the allowed time source table selections.

        for ( ++i, t = TimeTable ; t->GnString ; t++ ) {
        
                sprintf( DP.Msg, UsageMsg[i],  // < fixed format string

                          t->GnString,
                          t->UsageDesc
                          );
                PrintStrRpt( DP.Msg );
                }


        // Print the bottom (examples) section of the usage message.

        for ( ++i ; UsageMsg[i] ; i++ ) {

                sprintf( DP.Msg, "%s\n", UsageMsg[i] );
                PrintStrRpt( DP.Msg );
                }
                

        
        CleanExit( 1 );
}

//------------------------------------------------------------------------------
// ROUTINE: ErrorCmd
// PURPOSE: Print a command line error message and quit.
//------------------------------------------------------------------------------
void ErrorCmd( char *msg ) {

        // Print a command line error message and quit.

        sprintf( DP.Msg, "ERROR: Command line usage : %s \n", msg );
        PrintStrDate( DP.Msg );
        PrintStrRpt( "\n ... Diag /? for help\n" );

        CleanExit( 1 );  // < quit at errorlevel 1.
}

//------------------------------------------------------------------------------
// ROUTINE: PrintString
// PURPOSE: Output privided message to both a log file and the screen,
//          optionally adding a time and date string when writing to the
//          log file.  All output messages should be printed via this function.
//------------------------------------------------------------------------------
void PrintString( char *msg, int mode ) {

        FILE *LogFp;
        char *datestr, dbuffer[16];
        char *timestr, tbuffer[16];
        time_t current_time;
        struct tm *local_time;


        // Get the date and time if requested.

        if ( mode & PS_DATE ) {

                time( &current_time );
                local_time = localtime( &current_time );

                strftime( dbuffer, 16, "%m/%d/%y", local_time );
                strftime( tbuffer, 16, "%H:%M:%S", local_time );

                datestr = dbuffer;
                timestr = tbuffer;

                }



        // Append the message to the log file, optionally including the date.

        if ( mode & PS_LOG ) {

                if ( LogFp = fopen( DP.LogFullPathName, "a" ) ) {
                        if ( mode & PS_DATE )
                                fprintf( LogFp, "[%s] [%s] : ", datestr, timestr );
                        fprintf( LogFp, "%s", msg );
                        fclose( LogFp );
                        }
                }




        // Display the message to the report file

        if ( (mode & PS_REPORT) && DP.RptFp )
                fprintf( DP.RptFp, "%s", msg );



        // Display the message to the screen, optionally flushing
        // the line even if the message did not end with a new line
        // character.

        if ( mode & PS_DISPLAY )
                printf( "%s", msg );

        if ( mode & PS_FLUSH )
                SrFlushPrint();
}

//------------------------------------------------------------------------------
// Routine: PrintRunTimeSummary
// PURPOSE: Display the final DP parameters to run with.
//------------------------------------------------------------------------------
void PrintRunTimeSummary( void ) {

        sprintf( DP.Msg, "CommandLine = {%s}\n", DP.CmdLine );              PrintStrDate( DP.Msg );
        sprintf( DP.Msg, "DiagRev     = {%s}\n", DIAG_REV );                PrintStrRpt(  DP.Msg );
        sprintf( DP.Msg, "LibRev      = %d\n", DP.LibRev );                 PrintStrDate( DP.Msg );
        sprintf( DP.Msg, "DriverName  = \"%s\"\n", DP.AtodDriverName );     PrintStrRpt(  DP.Msg );
        sprintf( DP.Msg, "GpsModel    = \"%s\"\n", DP.AtodGpsModelName );   PrintStrRpt(  DP.Msg );
        sprintf( DP.Msg, "Log File    = \"%s\"\n", DP.LogFullPathName );    PrintStrRpt(  DP.Msg );
        sprintf( DP.Msg, "Report File = \"%s\"\n\n", DP.RptFullPathName );  PrintStrRpt(  DP.Msg );
}

//------------------------------------------------------------------------------
// ROUTINE: StrEqu
// PURPOSE: Compare for exact equality between two strings.
//------------------------------------------------------------------------------
int StrEqu( char *s1, char *s2 ) {

        while ( *s1 && *s2 && (*s1 == *s2) )

                { s1++; s2++; }

        return ( (*s1 == '\0') && (*s2 == '\0') ? 1 : 0 );

}

//------------------------------------------------------------------------------
// ROUTINE: GetDeviceDescriptor
// PURPOSE: This function gets the USBxCH device descriptor.  It provides
//          general info about the device like its Vendor ID (VID) and Product
//          ID (PID).
//------------------------------------------------------------------------------
int GetDeviceDescriptor( DEVHANDLE UsbXchHandle,
                         USB_DEVICE_DESCRIPTOR *pDeviceDescriptor ) {

        BOOL          Ok;
        unsigned long BytesSent;

        if ( !pDeviceDescriptor )
                return( 0 );

        Ok = SrUsbXchOsDriverIoctl( UsbXchHandle,
                                    IOCTL_USBXCH_GET_DEVICE_DESCRIPTOR,
                                    NULL,
                                    0,
                                    pDeviceDescriptor,
                                    sizeof( USB_DEVICE_DESCRIPTOR ),
                                    &BytesSent );

        if ( Ok )
                return( 1 );
        else
                return( 0 );
}

//------------------------------------------------------------------------------
// ROUTINE: DescriptorTypeString
// PURPOSE: This function takes a descriptor type integer and fills in the
//          corresponding string.  Descriptor types include device,
//          configuration, interface, endpoint, and string.
//------------------------------------------------------------------------------
int DescriptorTypeString( int DescType, char *DescTypeString,
                          size_t MaxLen ) {

        if ( !DescTypeString )
                return( 0 );

        if (DescType == USB_DEVICE_DESCRIPTOR_TYPE)
                strcpy( DescTypeString, "Device" );
//WCT           StringCchCopy( DescTypeString, MaxLen, "Device" );

        else if (DescType == USB_CONFIGURATION_DESCRIPTOR_TYPE)
                strcpy( DescTypeString, "Configuration" );
//WCT           StringCchCopy( DescTypeString, MaxLen, "Configuration" );

        else if (DescType == USB_STRING_DESCRIPTOR_TYPE)
                strcpy( DescTypeString, "String" );
//WCT           StringCchCopy( DescTypeString, MaxLen, "String" );

        else if (DescType == USB_INTERFACE_DESCRIPTOR_TYPE)
                strcpy( DescTypeString, "Interface" );
//WCT           StringCchCopy( DescTypeString, MaxLen, "Interface" );

        else if (DescType == USB_ENDPOINT_DESCRIPTOR_TYPE)
                strcpy( DescTypeString, "Endpoint" );
//WCT           StringCchCopy( DescTypeString, MaxLen, "Endpoint" );

        else
                strcpy( DescTypeString, "Unknown" );
//WCT           StringCchCopy( DescTypeString, MaxLen, "Unknown" );

        return( 1 );
}

//------------------------------------------------------------------------------
// ROUTINE: PrepareFileNames
// PURPOSE: Make the full pathnames for the log and report files.  They will
//          be put in the standard \SR\USBXCH\Utilities\Diag directory.
//          (Under Windows, the _get_pgmptr function could be used to get the
//          full path of the executable, but this is not available under
//          Linux.  So we just hardwire the directory instead.)
//------------------------------------------------------------------------------
int PrepareFileNames( void ) {

        // Form the log and report full pathnames.

        sprintf( DP.LogFullPathName, "%s%s", SRDIAGDIR, LogFileBaseExt );
        sprintf( DP.RptFullPathName, "%s%s", SRDIAGDIR, RptFileBaseExt );


        // Return success.

        return( 1 );
}

//------------------------------------------------------------------------------
// ROUTINE: CleanExit
// PURPOSE: Leave the program, while ensuring the keyboard mode is restored.
//------------------------------------------------------------------------------
int CleanExit( int ErrLevel ) {

        PrintStr( "\n" );
        
        SrKeyboardMode( KEYMODE_BLOCK );  // Restore blocking keyboard behavior.

        exit( ErrLevel );
}
