// FILE: Pak2Asc.c
// COPYRIGHT: (c), Symmetric Research, 2009-2010
//
// Pak2Asc converts USBxCH binary PAK files into ASCII text with formatted
// columns for the data and time.
//
//
//


// Include files:

#if defined( SROS_WINXP ) || defined( SROS_WIN2K )
#include <windows.h>
#elif defined( SROS_LINUX )
#endif  // SROS_xxxxx

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <ctype.h>    // for isdigit(), tolower(), etc

#include "SrHelper.h"
#include "SrUsbXch.h"



// Program copyright and screen title:

char *Copyright =

        "\n\nPak2Asc.exe : Copyright (c), Symmetric Research, 2010\n\n";

char *ScreenTitle =

  "USBxCH Pak2Asc Data Conversion Program: (c) SR 2010 : Rev {2010/08/15}\n\n";




// Maximum FILENAME and generic string lengths:
//
// An input filename is comprised of a path, basename, and extension.  The
// path and extension are optional.  In the string:
//
//   c:\directory\foo.pak
//
// the leading c:\directory\ is the path, foo is the basename, and pak
// is the extension.  In Windows, the last '\' denotes the end of the
// path, and the last '.' denotes the beginning of the extension.
//
// A complete filename must be less than MAXFILENAME characters long.
// Even if buffers for the path, basename, and ext are all individually
// dimensioned at MAXFILENAME, the total filename constructed from their
// parts must still be less than MAXFILENAME.
//
//

#define MAXFILENAME             1024
#define MAXSTR                  256   // < generic string length


// Allowed output formats:

#define OUTPUT_FORMAT_NONE     0
#define OUTPUT_FORMAT_FW       1
#define OUTPUT_FORMAT_FW_HEX   2
#define OUTPUT_FORMAT_FW_DBG   3
#define OUTPUT_FORMAT_CSV      4
#define OUTPUT_FORMAT_CSV_HEX  5
#define OUTPUT_FORMAT_CSV_DBG  6
#define OUTPUT_FORMAT_RAW      7
#define OUTPUT_FORMAT_MAX      8

char *OutputFormatName[] = {
        "",
        "fw",
        "fw_hex",
        "fw_debug",
        "csv",
        "csv_hex",
        "csv_debug",
        "raw",
        "",

        NULL
        };




// Pak2Asc RUN TIME PARAMETERS:
//
// Both command line and fixed hardwired parameters are located here.
//    ... access this structure with statements like: PP.Raw ...
//
//

struct Pak2AscParametersLayout {

        int  OutFormat;
        int  CSV;
        int  Hex;
        int  Debug;

        char *FileNameSample;
        char *FileNameStatus;
        char *FileNameTimeSample;
        char *FileNameTimeStatus;
        char *FileNameOutNmea;
        char *FileNameTimeTotal;
        char *FileNameOutData;
        char *FileNameOutCsv;
        char *FileNameRawOut;

        int  AtodGpsModel;

        int  Sequential;
        int  nSequential;

        int  ShowTime;


        // command line input file ...

        int InputFileFlag;

        FILE *fpInput;
        char InputFileName[MAXFILENAME];  // base + ext, no path ...

        char InputFilePath[MAXFILENAME];
        char InputFileBase[MAXFILENAME];
        char InputFileExt[MAXFILENAME];


        // generic full path work area ...

        char FullPathName[MAXFILENAME];

            char *pFullPathNameBase; // -> base+ext part of FullPathName[] ...


        } Pak2AscParameters = {


                // Output format argument values:

                OUTPUT_FORMAT_NONE,     // format, default = NONE
                0,                      // CSV, default = 0
                0,                      // Hex, default = 0
                0,                      // Debug, default = 0


                // Intermediate fixed filenames for each step:

                "Pak2Asc-100-SamplePackets.pak",   // FileNameSample
                "Pak2Asc-101-StatusPackets.pak",   // FileNameStatus
                "Pak2Asc-200-TimeInfoSample.asc",  // FileNameTimeSample
                "Pak2Asc-300-TimeInfoStatus.asc",  // FileNameTimeStatus
                "Pak2Asc-301-Nmea.asc",            // FileNameOutNmea
                "Pak2Asc-400-TimeInfoAll.asc",     // FileNameTimeTotal
                "Pak2Asc-500-Data.asc",            // FileNameOutData
                "Pak2Asc-500-Data.csv",            // FileNameOutCsv
                "Pak2Asc-RAW.asc",                 // FileNameRawOut


                // GPS model:  currently fixed at GARMIN ...

                SRDAT_GPSMODEL_GARMIN,  // AtodGpsModel


                // Sequential processing flags:

                0,      // Sequential on/off, default = off
                0,      // nSequential


                // Display time columns (updated in Step 1)

                0,      // ShowTime on/off, default = off
                

                // Input file parameters:

                0,      // InputFileFlag = 0, input file required

                // all trailing filename areas initialized to 0 by compiler ...
                // all trailing filename areas initialized to 0 by compiler ...

                };


#define PP  Pak2AscParameters  // < PP = shorthand for Pak2AscParameters ...




// USAGE MESSAGE ARRAY:

char *UsageMsg[] = {
                
    "Converts USBxCH binary PAK files from Blast into ASCII text columns        ",
    "                                                                           ",
    "USAGE: Pak2Asc filename format [/?]                                        ",
    "                                                                           ",
    "    filename = PAK file to convert  (required)                             ",
    "      format = output file format  (required)                              ",
    "          /? = show help screen and quit                                   ",
    "                                                                           ",
    "where:                                                                     ",
    "                                                                           ",
    "    filename must be of the form:                                          ",
    "                                                                           ",
    "       nnnnnnnn.pak  for sequential processing                             ",
    "       alpha numeric for single file processing                            ",
    "                                                                           ",
    "                                                                           ",
    "    format must be one of the following:                                   ",
    "      ( use csv for importing into Excel, Matlab, etc ... )                ",
    "                                                                           ",
    "         fw         fixed width ASCII columns                              ",
    "         fw_hex     fw, but with analog and digital in hex                 ",
    "         fw_debug   fw, but includes lower level debug info                ",
    "                                                                           ",
    "         csv        comma separated ASCII values                           ",
    "         csv_hex    csv, but with analog and digital in hex                ",
    "         csv_debug  csv, but includes lower level debug info               ",
    "                                                                           ",
    "         raw        show raw packets for debugging only                    ",
    "                                                                           ",
    "                                                                           ",
    " Examples:                                                                 ",
    "                                                                           ",
    "    Pak2Asc /?                < show help screen and quit                  ",
    "    Pak2Asc test.pak fw       < convert single file to fixed width ASCII   ",
    "    Pak2Asc test.pak csv      < convert to csv format                      ",
    "    Pak2Asc 00000000.pak fw   < sequential processing, starting at 0       ",
    "    Pak2Asc 02.pak fw         < not sequential, must be nnnnnnnn.dat!      ",
    "                                                                           ",

    NULL
    };




// Defines and global variables:

#define NCHANNELS       4
#define MAXDATAFMT     10



// Prototypes:

  void Step1( void );
  void Step2( void );
  void Step3( void );
  void Step4( void );
  void Step5( void );

   int ReadInputFile( SRUSBXCH_PACKET *Packet );
   int FileOpenNextSequential( void );

  FILE *FileOpen( char *FileName, char *ModeString );
   void FileWritePacket(
                        FILE *fp,
                        char *FileName,
                        SRUSBXCH_PACKET *Packet
                        );
   int FileReadPacket(
                        FILE *fp,
                        char *FileName,
                        SRUSBXCH_PACKET *Packet
                        );

   int FileReadTimeSample( FILE *fp, TS *TSpartial );
   int FileReadTimeStatus( FILE *fp, TS *TSpartial, NI *NmeaInfo );

   int FileUpdateTimeSample( FILE *fp, TS *TsPartial, int RequiredPpsCount );
   int FileUpdateTimeStatus( FILE *fp, TS *TsPartial, NI *NmeaInfo,
                             int RequiredPpsCount );
   
   int PpsCountCompare( int ExistingPpsCount, int NewPpsCount );
   
   int IsValidOBC( int PpsToggleLast, int ObcCountLast,
                   int Point, int PpsToggle, int *ObcCount );

   int FillSerialBufferFromFile( FILE *fp, char *FileName );

  void GetTwoAnalogValues( SRUSBXCH_PACKET *Packet,
                           SRDAT_USBXCH_ANALOG *AnalogValues );
   
   int ExtractNmeaInfo(
                        char *NmeaGroup,
                        int NmeaCount,
                        NI *NmeaInfo
                        );

  void CheckPacketSanity(
                        SRUSBXCH_PACKET *CurrentPacket,
                        int nTotalPackets
                        );

   int OutputAnalogData(
                        FILE *fp,
                        int CurrentPt,
                        SRDAT_USBXCH_ANALOG *AnalogValues,
                        TS *TimeStamp0,
                        double SampleTime,
                        int ShowTitle
                        );

  void UpdateTimestamps( FILE *fp, TS *TS0, TS *TS1, int CurrentPt );
   int ComputeTime( TS *TS0, TS *TS1, int CurrentPt, double *SampleTime );
  void ErrorCmd( char *msg );
  void ErrorExec( char *msg );
  void WarningExec( char *msg );
  void ProcessCommandLine( int argc, char *argv[] );
  void Usage( void );
   int MatchFormat( char *argument );
   int StrEqu( char *s1, char *s2 );
   int StrIsEightDecimal( char *s );
   int ParseFileName( char *FileName );
  void PrintRunTimeSummary( void);
  char *ChopLong( char *s);
  void ProcessRawPackets( void );





//------------------------------------------------------------------------------
// ROUTINE: main()
// PURPOSE: Main entry point for Pak2Asc ... convert PAK files to ASCII.
//------------------------------------------------------------------------------

int main( int argc, char *argv[] ) {


        // Announce the program.

        printf( ScreenTitle );


        // Process the command line arguments.

        ProcessCommandLine( argc, argv );  // << fills in the PP structure ...


        // Print a summary of the run time arguments.

        PrintRunTimeSummary( );



        // Check for "raw" Pak2Asc processing:
        //
        //   ... this processing option is primarily for debugging ...
        //
        //

        if ( PP.OutFormat == OUTPUT_FORMAT_RAW )

                ProcessRawPackets(); // << output the raw packets and quit



        // Normal Pak2Asc processing:
        //
        // Make multiple passes to align (analog) sample packets and
        // (GPS/NMEA) status packets to form the final ASCII output.
        //
        //

        // STEP 1: Split the input file(s) into sample and status packets.

        Step1( );


        // STEP 2: Extract (analog) sample packet time information.

        Step2( );


        // STEP 3: Process (GPS NMEA) status packet time information.

        Step3( );


        // STEP 4: Create the timestamp file.

        Step4( );


        // STEP 5: Align analog and GPS data and output it.

        Step5( );



        // Return from main:
        //
        // Linux *requires* main to be of type integer with return 0 for
        // success.  Windows doesn't care ...
        //
        //

        return( 0 );


}

//------------------------------------------------------------------------------
// ROUTINE: Step1
// PURPOSE: Split the input file(s) into sample and status packets.
//------------------------------------------------------------------------------
void Step1( void ) {

    //  FILE                *PP.fpInput;   << see PP struct for this variable
        FILE                *fpSample;
        FILE                *fpStatus;

        SRUSBXCH_PACKET     Packet;
        int                 ptype;

        int                 nTotalPackets;
        int                 nSamplePackets;
        int                 nStatusPackets;


        // Announce STEP 1 is processing.

        printf( "\nSTEP 1 PROCESSING ... \n\n" );


        // Open the input and output files.

        PP.fpInput   = FileOpen( PP.InputFileName,  "rb" );
        fpSample     = FileOpen( PP.FileNameSample, "wb" );
        fpStatus     = FileOpen( PP.FileNameStatus, "wb" );

        nTotalPackets   = 0;
        nSamplePackets  = 0;
        nStatusPackets  = 0;


        // Read from the input and write to the output files.

        while ( ReadInputFile( &Packet ) ) {


                // Check the packet type and split it out to the correct file.

                ptype = SrUsbXchPacketType( &Packet );

                if ( ptype == SRDAT_USBPACKET_TYPE_SAMPLE ) {

                        FileWritePacket( fpSample, PP.FileNameSample, &Packet );
                        nSamplePackets++;

                        }

                else if ( ptype == SRDAT_USBPACKET_TYPE_STATUS ) {

                        FileWritePacket( fpStatus, PP.FileNameStatus, &Packet );
                        nStatusPackets++;

                        }

                else  // unrecognized packet type

                    ErrorExec( "unknown input packet type found, fatal !!!");



                // Also check the packet sanity error codes.

                CheckPacketSanity( &Packet, nTotalPackets );



                // Increment the total packets processed ... and continue on.

                nTotalPackets++;

                }


        // If we found status packets, we have timing info and can gather
        // timestamp info in Steps 2-4 which are used to compute the time
        // columns in Step 5

        if ( nStatusPackets > 0 )
                PP.ShowTime = 1;



        // Close the input and output files.
        //
        // ... close of PP.fpInput is automatically done by ReadInputFile ...
        //

        fclose( fpSample );
        fclose( fpStatus );


        // Print STEP 1 summary info.

        printf( "\n" );

        printf( "step 1 success with:             \n"                    );
        printf( "                                 \n"                    );
        printf( "     Total packets processed: %d \n", nTotalPackets     );
        printf( "     (analog) sample packets: %d \n", nSamplePackets    );
        printf( "   (GPS NMEA) status packets: %d \n", nStatusPackets    );
        printf( "             Last input file: %s \n",

                                            ChopLong( PP.InputFileName ) );
        printf( "\n" );

}

//------------------------------------------------------------------------------
// ROUTINE: Step2
// PURPOSE: Extract (analog) sample packet time information.
//------------------------------------------------------------------------------
void Step2( void ) {

        FILE                *fpSample;
        FILE                *fpTimeSample;

        SRUSBXCH_PACKET     Packet;
        SRDAT_USBXCH_ANALOG AnalogValues[2];

        int                 i;
        int                 nSamplePackets;
        int                 nPoints;
        int                 nObc;
        int                 PpsCount;
        int                 ObcPt;
        int                 ObcCount;
        int                 ObcCountLast;
        int                 PpsToggle;
        int                 PpsToggleLast;
        int                 PpsEvent;
        int                 PpsEventLast;
        int                 PpsEventCorrected;
        int                 FoundObc;

        double              ExtraSec;


#define FMT_TSAMPTITLE " Pps Count     OBC Pt#   OBC Count     ExtraSec\n"
//                      NNNNNNNNNN  NNNNNNNNNN  0xNNNNNNNN FFFFFFFFFFFF
#define FMT_TSAMP      "%10d  %10d  0x%08X %12lf\n"
        


        // Announce STEP 2 is processing.

        printf( "\nSTEP 2 PROCESSING ... \n\n" );


        // Skip this step if we're not showing time

        if ( !PP.ShowTime ) {
                printf( "step 2 skipped: no sample timing data\n\n" );
                return;
                }
        
        

        // Open the input and output files.

        fpSample     = FileOpen( PP.FileNameSample,     "rb" );
        fpTimeSample = FileOpen( PP.FileNameTimeSample, "w" );


        // Output the title line.

        fprintf( fpTimeSample, FMT_TSAMPTITLE );


        // Read from the input and write to the output file.

        nSamplePackets = 0;
        nPoints        = 0;
        nObc           = 0;
        ObcCountLast   = 0;
        PpsToggleLast  = 0;
        PpsEventLast   = 0;


        while ( FileReadPacket( fpSample, PP.FileNameSample, &Packet ) ) {
                
                GetTwoAnalogValues( &Packet, AnalogValues );

                for ( i = 0 ; i < 2 ; i++ ) {

                        PpsEvent = AnalogValues[i].GGACount;
                        if ( PpsEvent == PpsEventLast )
                                PpsEventCorrected = PpsEvent+1;
                        else
                                PpsEventCorrected = PpsEvent;
                        
                        PpsCount  = PpsEventCorrected & 0x0F;
                        PpsToggle = AnalogValues[i].PpsToggle;
                        ObcCount  = AnalogValues[i].OnBoardCount;

                        FoundObc = IsValidOBC( PpsToggleLast, ObcCountLast, 
                                               nPoints, PpsToggle, &ObcCount );
                        if ( FoundObc ) {

                            ObcPt    = nPoints;
                            ExtraSec = ((double)ObcCount) / USBXCH_OBC_FREQ;
                            nObc++;

                            fprintf( fpTimeSample, FMT_TSAMP,
                                                             PpsCount,
                                                             ObcPt,
                                                             ObcCount,
                                                             ExtraSec
                                                   );
                            }
                        
                        ObcCountLast  = ObcCount;
                        PpsToggleLast = PpsToggle;
                        PpsEventLast  = PpsEvent;
                        nPoints++;

                        }

                nSamplePackets++;
                
                }



        // Close the input and output files.

        fclose( fpSample );
        fclose( fpTimeSample );


        // Print STEP 2 summary info.

        printf( "\n" );

        printf( "step 2 success with:             \n"                    );
        printf( "                                 \n"                    );
        printf( "              Sample packets: %d \n", nSamplePackets    );
        printf( "               Analog points: %d \n", nPoints           );
        printf( "         OnBoardCount values: %d \n", nObc              );
        printf( "             Wrote time file: %s \n",

                                        ChopLong( PP.FileNameTimeSample ) );
        printf( "\n" );

}

//------------------------------------------------------------------------------
// ROUTINE: Step3
// PURPOSE: Extract (gps nmea) status packet time information.  Also outputs
//          NMEA strings grouped by second.
//------------------------------------------------------------------------------
void Step3( void ) {

        FILE                  *fpStatus;
        FILE                  *fpTimeStatus;
        FILE                  *fpOutNmea;

        int                    errNmea;
        int                    eofStatus;
        int                    iNmea;
        int                    nNmea;
        int                    nChar;
        int                    nSecond;
        int                    nSecondLast;
        int                    EqFlag;
        int                    PpsCount;
        int                    PowerInfo;
        int                    Temperature;
        int                    DramEmpty;
        int                    DramPartFull;
        int                    DramOverflow;

        double                 TempC;
        double                 TempF;

        char                   NmeaMsgString[SRDAT_NMEA_MAX_SIZE];
        char                   NmeaGroup[SRDAT_NMEA_MAX_BUFF];

        NI                     NmeaInfo;

        SRDAT_USBXCH_EQUIPMENT EquipData;
        
        char                   MsgBuffer[MAXSTR];


#define FMT_TSTATTITLE "  PpsCount    SecSince1970    Sat Ysrc Hsrc Pwr  Temp     PpsEvent      Time as YMD HMS\n"
//                      NNNNNNNNNN FFFFFFFFFFFFFFFFF   NN  NN  NN    N  0xXXXX  NNNNNNNNNN NNNN/NN/NN NN:NN:NN.NNNNNN
#define FMT_TSTAT      "%10d %17lf   %2d  %2d  %2d    %1d  0x%04X  %10d %4d/%02d/%02d %02d:%02d:%02d.%06ld\n"
        


        // Announce STEP 3 is processing.

        printf( "\nSTEP 3 PROCESSING ... \n\n" );


        // Skip this step if we're not showing time

        if ( !PP.ShowTime ) {
                printf( "step 3 skipped: no status timing data\n\n" );
                return;
                }
        

        
        // Open the input and output files.

        fpStatus     = FileOpen( PP.FileNameStatus,     "rb" );
        fpTimeStatus = FileOpen( PP.FileNameTimeStatus, "w" );
        fpOutNmea    = FileOpen( PP.FileNameOutNmea,    "w" );

        
        // Output the title line.

        fprintf( fpTimeStatus, FMT_TSTATTITLE );


        // Read from the input and write to the output files.

        eofStatus   = 0;
        nNmea       = 0;
        nSecondLast = -1;

        SerBufInit();     // Initialize serial buffer parameters

        FillSerialBufferFromFile( fpStatus, PP.FileNameStatus );

        while ( 1 ) {

                errNmea = SrDatSerialReadNmea(
                                        SRDAT_NMEA_MAX_SIZE,
                                        NmeaMsgString,
                                        NULL,                   // don't check MSGID
                                        &nChar,
                                        &nSecond,
                                        &EqFlag,
                                        &EquipData
                                       );


                nChar = SerBufNumChar();

                if ( EqFlag ) {
                        PowerInfo   = EquipData.VoltageGood;
                        Temperature = EquipData.Temperature;
                        SrDatUsbTemperatureCompute( EquipData.Temperature, &TempC, &TempF );
                        SrUsbXchDramFlagsSplit( EquipData.DramFlags,
                                                &DramEmpty, &DramPartFull, &DramOverflow );
                        }
                else {
                        PowerInfo    = USBXCH_POWER_UNKNOWN;
                        Temperature  = USBXCH_TEMP_UNKNOWN;;
                        TempC        = USBXCH_TEMP_DEG_UNKNOWN;
                        TempF        = USBXCH_TEMP_DEG_UNKNOWN;
                        DramEmpty    = -1;
                        DramPartFull = -1;
                        DramOverflow = -1;
                        }

                if ( errNmea == SRDAT_NMEA_ERR_NONE ) {

                        if ( nSecond != -1 ) {              // We found first string in a group

                                if ( (nSecondLast != nSecond )  &&
                                     (nSecondLast != -1)  ) {

                                        ExtractNmeaInfo( NmeaGroup, nNmea, &NmeaInfo );

                                        PpsCount = nSecondLast & 0x0F;

                                        fprintf( fpTimeStatus, FMT_TSTAT,
                                                PpsCount,                 // PpsCount
                                                NmeaInfo.SecSince1970,    // SecSince 1970
                                                NmeaInfo.Nsat,            // Sat
                                                NmeaInfo.YmdSource,       // Ysrc
                                                NmeaInfo.HmsSource,       // Hsrc
                                                PowerInfo,                // Pwr
                                                Temperature,              // Temp
                                                nSecondLast,              // PpsEvent
                                                NmeaInfo.Year,            // Time as YMD HMS ...
                                                NmeaInfo.Month,
                                                NmeaInfo.Day,
                                                NmeaInfo.Hour,
                                                NmeaInfo.Minute,
                                                NmeaInfo.Second,
                                                NmeaInfo.MicroSecond
                                                );
                                        }
                                

                                nNmea       = 1;
                                nSecondLast = nSecond;
                                fprintf( fpOutNmea, "%d\n", nSecond );
                                fprintf( fpOutNmea, "EQUIP: Power %1d, Temp 0x%04X, DegC %5.1f, DegF %5.1f, DramFlags EM/PF/OV %1X/%1X/%1X\n",
                                         PowerInfo, Temperature, TempC, TempF,
                                         DramEmpty, DramPartFull, DramOverflow );
                                }
                        else              // We found trailing string in a group
                                nNmea++;

                        // Save NmeaMsgString into the NmeaGroup for later processing

                        iNmea = (nNmea-1) * SRDAT_NMEA_MAX_SIZE;
                        if ( nNmea <= SRDAT_NMEA_MAX_TYPE )
                                strcpy( &NmeaGroup[iNmea], NmeaMsgString );
                        else {
                                sprintf( MsgBuffer,
                                         "Too many NMEA messages per second, expecting 2, got at least %d", SRDAT_NMEA_MAX_TYPE );
                                ErrorExec( MsgBuffer);
                                }


                        fprintf( fpOutNmea, "%s", NmeaMsgString );
                        }

                else { // ( errNmea != SRDAT_NMEA_ERR_NONE )

                        // No more characters, exit loop and continue

                        if ( nChar == 0 )
                                break;

                        // Error, exit from while loop

                        else {
                                sprintf( MsgBuffer, "Step 3 failed processing NMEA strings (error = %d)", errNmea );
                                printf( "%s\n", MsgBuffer );
                                break;
                                }
                        
                        } // end else errNmea != SRDAT_NMEA_ERR_NONE


                // If the serial buffer is getting low, fill it up again

                if ( nChar < SRDAT_NMEA_MAX_SIZE+5 ) {

                        if ( !eofStatus ) {

                                if ( !FillSerialBufferFromFile( fpStatus, PP.FileNameStatus ) )
                                        eofStatus = 1;

                                } // end if !eofStatus
                        else
                                break; 

                        } // end if nChar

                } // end while 1



        // Close the input and output files.

        fclose( fpStatus );
        fclose( fpTimeStatus );
        fclose( fpOutNmea );


        // Print STEP 3 summary info.

        printf( "\n" );

        printf( "step 3 success with:             \n"                    );
        printf( "                                 \n"                    );
        printf( "             Wrote time file: %s \n",

                                       ChopLong( PP.FileNameTimeStatus ) );

        printf( "      Wrote output NMEA file: %s \n",

                                          ChopLong( PP.FileNameOutNmea ) );
        printf( "\n" );

}


//------------------------------------------------------------------------------
// ROUTINE: Step4
// PURPOSE: Create the timestamp file.
//------------------------------------------------------------------------------
void Step4( void ) {

        FILE   *fpTimeSample;
        FILE   *fpTimeStatus;
        FILE   *fpTimeTotal;

        int     nTime;
        int     TsTitle;

        NI      NmeaInfo;

        TS      TsSample;
        TS      TsStatus;
        TS      TsTotal;

        int     StatusVsSample;


        // Announce STEP 4 is processing.

        printf( "\nSTEP 4 PROCESSING ... \n\n" );


        // Skip this step if we're not showing time

        if ( !PP.ShowTime ) {
                printf( "step 4 skipped: no total timing data\n\n" );
                return;
                }
        

        
        // Open the input and output files.

        fpTimeSample = FileOpen( PP.FileNameTimeSample, "r" );
        fpTimeStatus = FileOpen( PP.FileNameTimeStatus, "r" );
        fpTimeTotal  = FileOpen( PP.FileNameTimeTotal,  "w" );

        
        // Skip over the input title lines.

        SrDatSkipTitleLine( fpTimeSample );
        SrDatSkipTitleLine( fpTimeStatus );


        // Get points from both intermediate files (fpTimeSample + fpTimeStatus) and combine

        nTime   = 0;
        TsTitle = 1;    // Write output title the first time

        while ( FileReadTimeSample( fpTimeSample, &TsSample ) &&
                FileReadTimeStatus( fpTimeStatus, &TsStatus, &NmeaInfo ) ) {


                StatusVsSample = PpsCountCompare( TsSample.PpsCount, TsStatus.PpsCount );
                

                // If the PpsCounts don't match, update the low file

                if ( StatusVsSample > 0 ) { // if ( TsSample.PpsCount < TsStatus.PpsCount ) {
                        if ( !FileUpdateTimeSample( fpTimeSample,
                                                    &TsSample, TsStatus.PpsCount ) )
                                break;
                        }

                if ( StatusVsSample < 0 ) { // if ( TsStatus.PpsCount < TsSample.PpsCount ) {
                        if ( !FileUpdateTimeStatus( fpTimeStatus, &TsStatus,
                                                    &NmeaInfo, TsSample.PpsCount ) )
                                break;
                        }


                // Once the PpsCounts match, combine the time info
                // and write it out

                if ( TsSample.PpsCount == TsStatus.PpsCount ) {

                        TsTotal.PpsCount       = TsSample.PpsCount;
                        TsTotal.Sample         = TsSample.Sample;
                        TsTotal.TotalTime      = NmeaInfo.SecSince1970 + TsSample.ExtraSec;
                        TsTotal.ExtraSec       = TsSample.ExtraSec;
                        TsTotal.ObcCount       = TsSample.ObcCount;
                        TsTotal.NumSat         = NmeaInfo.Nsat;
                        TsTotal.YmdSource      = NmeaInfo.YmdSource;
                        TsTotal.HmsSource      = NmeaInfo.HmsSource;
                        TsTotal.PowerInfo      = TsStatus.PowerInfo;
                        TsTotal.Temperature    = TsStatus.Temperature;


                        SrDatTimestampWriteFile( fpTimeTotal, &TsTotal, TsTitle );
                        TsTitle = 0;

                        nTime++;
                        }
                else {
                        printf( "Trouble with alignment\n" );
                        exit( 1 );
                        }

                }


        // Close all files.

        fclose( fpTimeStatus );
        fclose( fpTimeSample );
        fclose( fpTimeTotal );



        // Print STEP 4 summary info.

        printf( "\n" );

        printf( "step 4 success with:             \n"                    );
        printf( "                                 \n"                    );
        printf( "       Time stamps processed: %d \n", nTime             );
        printf( "             Wrote time file: %s \n",

                                        ChopLong( PP.FileNameTimeTotal ) );

        printf( "\n" );

}

//------------------------------------------------------------------------------
// ROUTINE: Step5
// PURPOSE: Align analog and GPS data and output it.
//------------------------------------------------------------------------------
void Step5( void ) {

        FILE               *fpSample;
        FILE               *fpTimeTotal;
        FILE               *fpOutData;

        char               *FileNameOut;

        int                 i;
        int                 nPoints;

        double              SampleTime;

        SRUSBXCH_PACKET     Packet;
        SRDAT_USBXCH_ANALOG AnalogValues[2];
        
        TS                  TimeStamp0;
        TS                  TimeStamp1;



        // Announce STEP 5 is processing.

        printf( "\nSTEP 5 PROCESSING ... \n\n");


        if ( PP.CSV )
                FileNameOut = PP.FileNameOutCsv;
        else
                FileNameOut = PP.FileNameOutData;


        // Open the input and output files.

        fpSample    = FileOpen( PP.FileNameSample,    "rb" );

        if ( PP.ShowTime )  // Only open this one if we're showing time
        fpTimeTotal = FileOpen( PP.FileNameTimeTotal, "r" );

        fpOutData   = FileOpen( FileNameOut,          "w" );


        
        // Fill first two time stamps from timetmp file

        if ( PP.ShowTime ) {
                SrDatSkipTitleLine( fpTimeTotal );
                if ( !SrDatTimestampReadFile( fpTimeTotal, &TimeStamp0, 0 ) ||
                     !SrDatTimestampReadFile( fpTimeTotal, &TimeStamp1, 0 ) ) {
                        WarningExec( "Not enough timestamps to compute time.  Showing data only." );
                        fclose( fpTimeTotal );
                        PP.ShowTime = 0;
                        }
                }


        // Show title line (no data)

        OutputAnalogData( fpOutData, 0, NULL, NULL, 0.0L, 1 );




        // Read and process data one packet at a time

        nPoints = 0;
        
        while ( FileReadPacket( fpSample, PP.FileNameSample, &Packet ) ) {

                GetTwoAnalogValues( &Packet, AnalogValues );

                for ( i = 0 ; i < 2 ; i++ ) {

                        if ( PP.ShowTime ) {
                                UpdateTimestamps( fpTimeTotal, &TimeStamp0, &TimeStamp1, nPoints );
                                ComputeTime( &TimeStamp0, &TimeStamp1, nPoints, &SampleTime );
                                }

                        OutputAnalogData( fpOutData, nPoints, &AnalogValues[i],
                                                     &TimeStamp0, SampleTime, 0 );

                        nPoints++;

                        }

                }


        // Close the input and output files.

        fclose( fpSample );
        fclose( fpOutData );

        if ( PP.ShowTime )
                fclose( fpTimeTotal );



        // Print STEP 5 summary info.

        printf( "\n" );

        printf( "step 5 success with:             \n"                    );
        printf( "                                 \n"                    );
        printf( "               Analog points: %d \n", nPoints           );
        printf( "             Final data file: %s \n", ChopLong( FileNameOut ) );
        printf( "\n\n" );

        if ( PP.ShowTime )
                printf( "SUCCESS: final ASCII data >> %s\n\n", FileNameOut );
        else {
                printf( "WARNING: Not enough time stamps to show time and temperature.\n\n" );
                printf( "         Final ASCII data >> %s\n\n", FileNameOut );
                }
}

//------------------------------------------------------------------------------
// ROUTINE: FileOpen
// PURPOSE: Open the specified filename.
//------------------------------------------------------------------------------
FILE *FileOpen( char *filename, char *modestring ) {

        FILE *fp;
        char MsgBuffer[MAXSTR], *CreateOpenStr;


        // Form the full path name.
        //
        // Here filename is the base + ext.  Append it to the path for the
        // full pathname.
        //
        //

        sprintf( PP.pFullPathNameBase, "%s", filename );


        // Check mode string, can be w, wb, r, or rb

        if ( StrEqu( modestring, "w" ) || StrEqu( modestring, "wb" ) )
                CreateOpenStr = "creating";
        
        else // StrEqu( modestring, "r" ) || StrEqu( modestring, "rb" ) )
                CreateOpenStr = " opening";


        // Try to open it.

        fp = fopen( PP.FullPathName, modestring );

        if ( fp ) { // success, return the file pointer ...

                printf( "%s ... %s\n", CreateOpenStr, ChopLong( filename ) );
                return( fp );

                }

        else { // fp == NULL, fatal error ...

                sprintf( MsgBuffer, "failed to open %s", ChopLong( filename ) );

                    ErrorExec( MsgBuffer );

                }

        return( NULL ); // quiet the compiler, should never get here
}

//------------------------------------------------------------------------------
// ROUTINE: FileOpenNextSequential
// PURPOSE: Try to open the next sequential input file.
//------------------------------------------------------------------------------
int FileOpenNextSequential( void ) {


        // Form the next sequential input filename.

        PP.nSequential++;

            sprintf( PP.InputFileName, "%08d.%s", PP.nSequential, PP.InputFileExt );


        // Append the new filename to the full pathname.

        sprintf( PP.pFullPathNameBase, "%s", PP.InputFileName);


        // Try to open it.

        PP.fpInput = fopen( PP.FullPathName, "rb" );

        if ( PP.fpInput ) { // success, next file opened ...

                printf( " opening ... %s\n", ChopLong( PP.InputFileName ) );
                return( 1 );

                }

        else { // failed, but not fatal, just got to end of file sequence ...

            printf( "sequence ended, no %s\n", ChopLong( PP.InputFileName ) );

                return( 0 );

                }

}

//------------------------------------------------------------------------------
// ROUTINE: ReadInputFile
// PURPOSE: Read a USB packet from the input file.
//------------------------------------------------------------------------------
int ReadInputFile( SRUSBXCH_PACKET *packet ) {

        int nbytes;


        // Read a packet from the input file.
        //
        // ... for sequential processing, increment to next file if necessary
        //

ReadPacket:

        nbytes = fread( packet, 1, SRDAT_USBPACKET_SIZE, PP.fpInput );


        if ( nbytes == SRDAT_USBPACKET_SIZE ) // success

                return( 1 );


        else if ( nbytes == 0 ) { // EOF

                fclose( PP.fpInput ); // close the current input file ...


                // if sequential processing, try to increment onward ...

                if ( PP.Sequential && FileOpenNextSequential() )

                        goto ReadPacket;

                else
                        return( 0 );

                }


        else // nbytes != SRDAT_USBPACKET_SIZE, fatal error ...

                ErrorExec( "fatal ReadInputFile, nbytes != packet size ... !!!" );


        return( 0 ); // quiet the compiler, should never get here
}

//------------------------------------------------------------------------------
// ROUTINE: FileWritePacket
// PURPOSE: Write a USBxCH packet to fp.
//------------------------------------------------------------------------------
void FileWritePacket( FILE *fp, char *FileName, SRUSBXCH_PACKET *Packet ) {

        int  nbytes;
        char MsgBuffer[MAXSTR];


        // Write one USBxCH packet to fp.

        nbytes = fwrite( Packet, 1, SRDAT_USBPACKET_SIZE, fp );


        // Check the return.

        if ( nbytes != SRDAT_USBPACKET_SIZE ) {

            sprintf( MsgBuffer, "failed to write %s", ChopLong( FileName ) );

                      ErrorExec( MsgBuffer );  // < fatal ...
                }


        else  // nbytes == SRDAT_USBPACKET_SIZE ... success

                return;

}

//------------------------------------------------------------------------------
// ROUTINE: FileReadPacket
// PURPOSE: Read a packet from fp.
//------------------------------------------------------------------------------
int FileReadPacket( FILE *fp, char *FileName, SRUSBXCH_PACKET *Packet ) {

        int  nbytes;
        char MsgBuffer[MAXSTR];


        // Read a packet from the specified file.

        nbytes = fread( Packet, 1, SRDAT_USBPACKET_SIZE, fp );


        if ( nbytes == SRDAT_USBPACKET_SIZE ) // success

                return( 1 );


        else if ( nbytes == 0 )  // EOF, tell caller all done ...

               return( 0 );


        else { // nbytes != SRDAT_USBPACKET_SIZE, fatal error ...

                sprintf( MsgBuffer, "failed to read from %s", FileName );

                    ErrorExec( MsgBuffer );

                }

        return( 0 ); // quiet the compiler, should never get here
}

//------------------------------------------------------------------------------
// ROUTINE: FileReadTimeSample
// PURPOSE: Read the time sample file to get partial timestamp info.  Returns
//          1 on success, 0 otherwise.
//------------------------------------------------------------------------------
int FileReadTimeSample( FILE *fp, TS *TsPartial ) {

        int    Nscan, PpsCount, ObcPt, ObcCount;
        double ExtraSec;
        char   Line[MAXSTR];

        // Error check arguments

        if ( !fp || !TsPartial )
                return( 0 );

        
        // Read one line

        if ( !fgets( Line, MAXSTR-1, fp ) )
                return( 0 );


        // Read sample time info

        Nscan = sscanf( Line, FMT_TSAMP, // format defined in Step 2
                        &PpsCount,
                        &ObcPt,
                        &ObcCount,
                        &ExtraSec );
        
        if ( Nscan == 4 ) {
                TsPartial->PpsCount = PpsCount;
                TsPartial->Sample   = ObcPt;
                TsPartial->ObcCount = ObcCount;
                TsPartial->ExtraSec = ExtraSec;
                return( 1 );
                }
        else
                return( 0 );
}

//------------------------------------------------------------------------------
// ROUTINE: FileReadTimeStatus
// PURPOSE: Read the time status file to get partial timestamp info.  Returns
//          1 on success, 0 otherwise.
//------------------------------------------------------------------------------
int FileReadTimeStatus( FILE *fp, TS *TsPartial, NI *NmeaInfo ) {

        int    Nscan, GpsCount, FullGpsCount, PowerInfo, Temperature;
        char   Line[MAXSTR];
        NI     NmeaTmp;

        // Error check arguments

        if ( !fp || !TsPartial || !NmeaInfo )
                return( 0 );

        
        // Read one line

        if ( !fgets( Line, MAXSTR-1, fp ) )
                return( 0 );


        // Read status time info

        Nscan = sscanf( Line, FMT_TSTAT,          // format defined in Step 3
                        &GpsCount,                // PpsCount
                        &NmeaTmp.SecSince1970,    // SecSince 1970
                        &NmeaTmp.Nsat,            // Sat
                        &NmeaTmp.YmdSource,       // Ysrc
                        &NmeaTmp.HmsSource,       // Hsrc
                        &PowerInfo,               // Pwr
                        &Temperature,             // Temp
                        &FullGpsCount,            // PpsEvent
                        &NmeaTmp.Year,            // Time as YMD HMS ...
                        &NmeaTmp.Month,
                        &NmeaTmp.Day,
                        &NmeaTmp.Hour,
                        &NmeaTmp.Minute,
                        &NmeaTmp.Second,
                        &NmeaTmp.MicroSecond
                      );
        
        
        if ( Nscan == 15 ) {
                TsPartial->PpsCount    = GpsCount;
                TsPartial->PowerInfo   = PowerInfo;
                TsPartial->Temperature = Temperature;
                *NmeaInfo              = NmeaTmp;
                return( 1 );
                }
        else
                return( 0 );
}


//------------------------------------------------------------------------------
// ROUTINE: FileUpdateTimeSample
// PURPOSE: Read the time sample file repeatedly until it returns partial
//          timestamp info that matches the RequiredPpsCount.  Returns
//          1 on success, 0 otherwise.
//------------------------------------------------------------------------------
int FileUpdateTimeSample( FILE *fp, TS *TsPartial, int RequiredPpsCount ) {

        int RequiredVsPartial;

        if ( !fp || !TsPartial )
                return( 0 );

        RequiredVsPartial = PpsCountCompare( TsPartial->PpsCount, RequiredPpsCount );

        while ( RequiredVsPartial > 0 ) { // was ( TsPartial->PpsCount < RequiredPpsCount ) {

                if ( !FileReadTimeSample( fp, TsPartial ) ) {
                        printf( "Can't update PpsCount from TimeSample file\n" );
                        return( 0 );
                        }
                RequiredVsPartial = PpsCountCompare( TsPartial->PpsCount, RequiredPpsCount );
                }

        return( 1 );
}

//------------------------------------------------------------------------------
// ROUTINE: FileUpdateTimeStatus
// PURPOSE: Read the time Status file repeatedly until it returns partial
//          timestamp info that matches the RequiredPpsCount.  Returns
//          1 on success, 0 otherwise.
//------------------------------------------------------------------------------
int FileUpdateTimeStatus( FILE *fp, TS *TsPartial, NI *NmeaInfo,
                          int RequiredPpsCount ) {

        int RequiredVsPartial;

        if ( !fp || !TsPartial || !NmeaInfo )
                return( 0 );

        RequiredVsPartial = PpsCountCompare( TsPartial->PpsCount, RequiredPpsCount );

        while ( RequiredVsPartial > 0 ) { // was ( TsPartial->PpsCount < RequiredPpsCount ) {

                if ( !FileReadTimeStatus( fp, TsPartial, NmeaInfo ) ) {
                        printf( "Can't update PpsCount from TimeStatus file\n" );
                        return( 0 );
                        }
                RequiredVsPartial = PpsCountCompare( TsPartial->PpsCount, RequiredPpsCount );
                }

        return( 1 );
}

//------------------------------------------------------------------------------
// ROUTINE: PpsCountCompare
// PURPOSE: Compare new PpsCount from current packet with existing PpsCount in TS.
//          Return 0 if counts are equal, -1 or -2 if new is less than existing, 
//          and 1 or 2 if new count is more than existing.  If the counts are
//          close to each other It should be trivial, 
//          but PpsCounts wrap around to 0 after they reach 15.  So, a new count of
//          0 will be considered more than an existing count of 15.  If the
//          two counts differ by more than 4, we don't really know how to 
//          interpret things properly, so we fall back to treating them like
//          regular numbers.  Returns 3 if existing is unset (-1) and -3 if
//          new is unset (-1).
//------------------------------------------------------------------------------
int PpsCountCompare( int ExistingPpsCount, int NewPpsCount ) {
#define DIFFLIMIT        4
#define HALFSHIFT        8
#define FULLSHIFT       16

        int Diff, DiffPlain, DiffShift;
        int ExistingShift, NewShift;

        // Deal with the easy case where they are identical

        if ( ExistingPpsCount == NewPpsCount ) {
                return( 0 );
                }

        // Check for counts being unset

        if ( ExistingPpsCount == -1 )
                return( 3 );

        if ( NewPpsCount == -1 )
                return( -3 );


        // Shift halfway around

        ExistingShift = ExistingPpsCount + HALFSHIFT;
        if ( ExistingShift >= FULLSHIFT )
                ExistingShift -= FULLSHIFT;

        NewShift = NewPpsCount + HALFSHIFT;
        if ( NewShift >= FULLSHIFT )
                NewShift -= FULLSHIFT;



        // Compute difference between counts plain and shifted

        DiffPlain = ExistingPpsCount - NewPpsCount;
        if ( DiffPlain < 0 )
                DiffPlain = -DiffPlain;

        DiffShift = ExistingShift - NewShift;
        if ( DiffShift < 0 )
                DiffShift = -DiffShift;


        // Work with smaller difference

        if ( DiffShift < DiffPlain )
                Diff = DiffShift;
        else
                Diff = DiffPlain;


        // Difference is large, so treat counts simply

        if ( Diff > DIFFLIMIT ) {

                if ( ExistingPpsCount > NewPpsCount ) {
                        return( -2 );
                        }
                else {
                        return( 2 );
                        }
                }


        // Else compensate for wrap

        else {
                if ( DiffShift < DiffPlain ) {
                        if ( ExistingShift > NewShift ) {
                                return( -1 );
                                }
                        else {
                                return( 1 );
                                }
                        }

                else { // DiffShift >= DiffPlain
                        if ( ExistingPpsCount > NewPpsCount ) {
                                return( -1 );
                                }
                        else {
                                return( 1 );
                                }
                        }
                }
}

//------------------------------------------------------------------------------
// ROUTINE: IsValidOBC
// PURPOSE: Check the specified analog data value to see if the OnBoardCount
//          is valid.  If so, it returns a 1, otherwise 0.
//          Typically, OBC is valid if it is non-zero.
//------------------------------------------------------------------------------
int IsValidOBC( int PpsToggleLast, int ObcCountLast,
                int Point, int PpsToggle, int *ObcCount ) {

        // Ignore very first point

        if ( Point == 0 )
                return( 0 );


        // Look for PpsToggle changed

        else if ( PpsToggleLast != PpsToggle ) {

                // Correct OBC if it occurred 1 sample back

                if ( *ObcCount == 0 )
                        *ObcCount = ObcCountLast;

                return( 1 );
                }


        // Otherwise OBC is not valid

        else
                return( 0 );
}

//------------------------------------------------------------------------------
// ROUTINE: FillSerialBufferFromFile
// PURPOSE: Read status packets from Fptr with name FileName and use them
//          to fill the serial buffer which will be used to parse out the
//          NMEA strings.  Returns 1 on success, 0 if any of the arguments are
//          missing or the end of file was reached.  It exits if there was a
//          problem filling the serial buffer.
//------------------------------------------------------------------------------
int FillSerialBufferFromFile( FILE *fp, char *FileName ) {

        SRUSBXCH_PACKET CurrentPacket;
        int             Ok, Error, EarlyExit;
        char            MsgBuffer[MAXSTR];

        if (!fp || !FileName )
                return( 0 );

        EarlyExit = 0;

        while ( FileReadPacket( fp, FileName, &CurrentPacket ) ) {

                Ok = SrDatUsbSerialFillBuffer( &CurrentPacket, 0, &Error );

                if (!Ok || Error != USBXCH_ERROR_NONE ) {
                        sprintf( MsgBuffer, "Pak2Asc SrDatUsbSerialFillBuffer error %d, ok %d\n", Error, Ok );
                        ErrorExec( MsgBuffer );
                        }

                // Serial buffer is too full for one more serial packet

                if ( SerBufNumFree() < SRDAT_USBPACKET_STATUS_BYTES ) {
                        EarlyExit = 1;
                        break;
                        }

                }



        return( EarlyExit ? 1 : 0 );


}

//------------------------------------------------------------------------------
// ROUTINE: ExtractNmeaInfo
// PURPOSE: 
//------------------------------------------------------------------------------
int ExtractNmeaInfo( char *NmeaGroup, int NmeaCount, NI *NmeaInfo ) {

        int Ok;

        if ( !NmeaGroup || NmeaCount < 0 || !NmeaInfo )
                return( 0 );


        // Get the best NMEA info possible from the provided group of NMEA strings

        Ok = SrDatRequestNmeaInfo2( NmeaGroup, NmeaCount,
                                    &NmeaInfo->Year,
                                    &NmeaInfo->Month,
                                    &NmeaInfo->Day,
                                    &NmeaInfo->YmdSource,
                                    &NmeaInfo->YmdIsValid,
                                    &NmeaInfo->Hour,
                                    &NmeaInfo->Minute,
                                    &NmeaInfo->Second,
                                    &NmeaInfo->MicroSecond,
                                    &NmeaInfo->HmsSource,
                                    &NmeaInfo->HmsIsValid,
                                    &NmeaInfo->Latitude,
                                    &NmeaInfo->Longitude,
                                    &NmeaInfo->LocSource,
                                    &NmeaInfo->LocIsValid,
                                    &NmeaInfo->Altitude,
                                    &NmeaInfo->PosSource,
                                    &NmeaInfo->PosIsValid,
                                    &NmeaInfo->Nsat,
                                    &NmeaInfo->SatSource,
                                    &NmeaInfo->SatIsValid
                                  );

        // If all went well, output the data to the specified file
        
        if ( Ok ) {
                Ok = SrDatSecTimeCombine( NmeaInfo->Year, NmeaInfo->Month, NmeaInfo->Day,
                                          NmeaInfo->Hour, NmeaInfo->Minute, NmeaInfo->Second, 
                                          NmeaInfo->MicroSecond, &NmeaInfo->SecSince1970 );

                NmeaInfo->SecIsValid = Ok;
                
                if ( Ok && NmeaInfo->YmdIsValid && NmeaInfo->HmsIsValid ) {

                        return( 1 );
                        }
                else {
                        printf( "Time/date is not valid and not printed to file\n" );
                        }
                }
        else
                printf("SrDatRequestNmeaInfo2 failed\n" );

        return( 0 );
}

//------------------------------------------------------------------------------
// ROUTINE: CheckPacketSanity
// PURPOSE: Check the packet sanity to make sure it is the expected value.
//------------------------------------------------------------------------------
void CheckPacketSanity( SRUSBXCH_PACKET *Packet, int nTotalPackets ) {

        int Error;
        int ptype;
        int sanity;

        char MsgBuffer[MAXSTR];

#define STARTUP  -1

        static int ExpectedSampleSanity = STARTUP;
        static int ExpectedStatusSanity = STARTUP;

#define Increment( s )     s = ( (s+1) & SRDAT_USBPACKET_SANITY_MASK )




        // Initalize the error flag.

        Error = 0;


        // Check the sanity for the respective packet type.
        //
        // Note (analog) sample packets have two samples per 64 byte packet.
        // Each of the two samples has its own sanity.
        //
        //

         ptype = SrUsbXchPacketType( Packet );
        sanity = SrUsbXchPacketSanity( Packet );


        if ( ptype == SRDAT_USBPACKET_TYPE_SAMPLE ) { // (analog) sample packet


                // If starting up, init the expected sanity.

                if ( ExpectedSampleSanity == STARTUP )

                        ExpectedSampleSanity = sanity;


                // Check the actual sanity against the expected.

                if ( sanity == ExpectedSampleSanity ) {

                        Increment( ExpectedSampleSanity );


                        // check the second packet sub sample too ...

                        if ( SrUsbXchPacketSanity2( Packet) == ExpectedSampleSanity )

                                Increment( ExpectedSampleSanity );

                        else
                                Error = 2;

                        }

                else
                        Error = 1;

                }


        
        else { // ptype == SRDAT_USBPACKET_TYPE_STATUS, (GPS NMEA) status packet


                // If starting up, init the expected sanity.

                if ( ExpectedStatusSanity == STARTUP )

                        ExpectedStatusSanity = sanity;


                // Check the actual sanity against the expected value.

                if ( sanity == ExpectedStatusSanity  )

                        Increment( ExpectedStatusSanity );

                else
                        Error = 3;
                
                }
        


        // Fatal error if any acutal sanity did not match expected.

        if ( Error ) {

            sprintf( MsgBuffer, "sanity errror for packet %d", nTotalPackets );

                ErrorExec( MsgBuffer );

                }

                
}

//------------------------------------------------------------------------------
// ROUTINE: GetTwoAnalogValues
// PURPOSE: Convert the packet data into analog structures.  Each sample packet
//          contains info for two analog structures.  If an error occurs, this
//          function exits.
//------------------------------------------------------------------------------
void GetTwoAnalogValues( SRUSBXCH_PACKET *Packet,
                         SRDAT_USBXCH_ANALOG *AnalogValues ) {

                  int Error;
        unsigned long Nsamples;
                 char MsgBuffer[MAXSTR];        

        if ( !Packet || !AnalogValues )
                ErrorExec( "Invalid parameters passed to GetTwoAnalogValues\n" );
        
        Nsamples = SrDatUsbAnalogFill( Packet, 0, AnalogValues, &Error );

        if ( ( Error != USBXCH_ERROR_NONE ) || ( Nsamples != 2 ) ) {
                sprintf( MsgBuffer,
                         "Processing sample packet failed, error %d, nsamples %lu\n",
                         Error, Nsamples );
                ErrorExec( MsgBuffer );
                }
}

//------------------------------------------------------------------------------
// ROUTINE: OutputAnalogData
// PURPOSE: Write the analog data out to the specified file in pretty format.
//          Standard out can be used to make the data come to the screen instead.
//          If Debug is one, the specialized info such as the sanity byte,
//          the digital input, and the onboard count are included in the output.
//          Otherwise, only the 4 analog values themselves (plus the index and
//          time columns are included).  Returns 1 on success and 0 if any of
//          the arguments are invalid.
//------------------------------------------------------------------------------
int OutputAnalogData( FILE *fp, int CurrentPt,
                      SRDAT_USBXCH_ANALOG *AnalogValues,
                      TS *TimeStamp0, double SampleTime, int ShowTitle ) {

        int    i, pos, Yr, Mo, Dy, Hh, Mm, Ss, PowerGoodC, PowerGoodH;
        long   Usec;
        double TempC, TempF;
        char   *Fmt, *ObcSetStr, ChannelString[MAXSTR], DigitalString[MAXSTR];

        // Error check - can pass null for analog data if only showing title
        
        if ( !fp || CurrentPt < 0 )
                return( 0 );
        if ( !ShowTitle && ( !AnalogValues || !TimeStamp0 ) )
                return( 0 );

        
        if ( ShowTitle ) {

                // Prepare string with channel names

                if ( PP.CSV )
                        Fmt = ",CH%02d";
                else
                        Fmt = "        CH%02d";

                pos = 0;
                for ( i = 0 ; i < NCHANNELS ; i++ ) {
                        sprintf( &ChannelString[pos], Fmt, i );
                        pos = strlen( ChannelString );
                        }

                // Write header/title info according to selected output format

                if ( !PP.CSV )  // fixed width
                        fprintf( fp, "# Pak2Asc\n# format=%s\n#\n", OutputFormatName[PP.OutFormat] );
                
                if ( PP.OutFormat == OUTPUT_FORMAT_CSV_DBG )
                        fprintf( fp, "Pt#%s,D0,D1,D2,D3,PwrH,DegC,DegF,Seconds,Date(YMD),Time(HMS),FracSec,PwrC,San,Pps,Nmea,OBC,ObcSet\n", ChannelString );
                else if ( PP.OutFormat == OUTPUT_FORMAT_CSV_HEX )
                        fprintf( fp, "Pt#%s,Dig,PwrH,DegC,DegF,Seconds,Date(YMD),Time(HMS),FracSec\n", ChannelString );
                else if ( PP.OutFormat == OUTPUT_FORMAT_CSV )
                        fprintf( fp, "Pt#%s,D0,D1,D2,D3,PwrH,DegC,DegF,Seconds,Date(YMD),Time(HMS),FracSec\n", ChannelString );
                else if ( PP.OutFormat == OUTPUT_FORMAT_FW_DBG )
                        fprintf( fp, "# Data:\n#   Pt#  %s  Dig PwrH DegC  DegF   Time in Seconds     Time as YMD HMS    Frac Sec PwrC San  Pps Nmea    OBC\n", ChannelString );
                else if ( PP.OutFormat == OUTPUT_FORMAT_FW_HEX )
                        fprintf( fp, "#   Pt#  %s  Dig PwrH DegC  DegF   Time in Seconds     Time as YMD HMS    Frac Sec\n", ChannelString );
                else // if ( PP.OutFormat == OUTPUT_FORMAT_FW )
                        fprintf( fp, "#   Pt#  %s D0 D1 D2 D3 PwrH DegC  DegF   Time in Seconds     Time as YMD HMS    Frac Sec\n", ChannelString );

                return( 1 );

                } // end if ShowTitle

        
        
        // Compute values for output below

        if ( PP.ShowTime ) {
                PowerGoodC = SrUsbXchIsPowerBitGood( TimeStamp0->PowerInfo,
                                                     USBXCH_POWER_CURRENT );

                PowerGoodH = SrUsbXchIsPowerBitGood( TimeStamp0->PowerInfo,
                                                     USBXCH_POWER_HISTORY );

                SrDatUsbTemperatureCompute( TimeStamp0->Temperature, &TempC, &TempF );

                SrDatSecTimeSplit( SampleTime, &Yr, &Mo, &Dy, &Hh, &Mm, &Ss, &Usec );
                }
        
        else { // not enough timestamps so set up dummy defaults
                PowerGoodC = 0;
                PowerGoodH = 0;
                TempC      = -99.9;
                TempF      = -99.9;
                SampleTime = 0.0L;
                Yr         = 1970;
                Mo         = 1;
                Dy         = 1;
                Hh         = 0;
                Mm         = 0;
                Ss         = 0;
                Usec       = 0L;
                }



        // Prepare analog channel data

        if ( PP.OutFormat == OUTPUT_FORMAT_FW_HEX ||
             PP.OutFormat == OUTPUT_FORMAT_FW_DBG )     
                Fmt = "0x%08lX  ";
        else if ( PP.OutFormat == OUTPUT_FORMAT_CSV_HEX )
                Fmt = ",0x%08lX";
        else if ( PP.OutFormat == OUTPUT_FORMAT_CSV ||
                  PP.OutFormat == OUTPUT_FORMAT_CSV_DBG )
                Fmt = ",%+ld";
        else // if ( PP.OutFormat == OUTPUT_FORMAT_FW || unknown
                Fmt = " %+9ld  ";
        pos = 0;
        for ( i = 0 ; i < NCHANNELS ; i++ ) {
                sprintf( &ChannelString[pos], Fmt, AnalogValues->AnalogData[i] );
                pos = strlen( ChannelString );
                }

        // Prepare digital data

        if ( PP.Hex ) {
                if ( PP.CSV )
                        Fmt = ",0x%02X";           // CSV_HEX
                else
                        Fmt = "0x%02X ";           // FW_HEX, FW_DBG
                sprintf( DigitalString, Fmt, AnalogValues->DigitalIn );
                }
        else {
                if ( PP.CSV )
                        Fmt = ",%d,%d,%d,%d";      // CSV, CSV_DBG
                else
                        Fmt = "%d  %d  %d  %d  ";  // FW
                sprintf( DigitalString, Fmt, (AnalogValues->DigitalIn & 0x01),
                                            ((AnalogValues->DigitalIn & 0x02)>>1),
                                            ((AnalogValues->DigitalIn & 0x04)>>2),
                                            ((AnalogValues->DigitalIn & 0x08)>>3) );
                }

        
        // Prepare ObcSet string

        if ( PP.CSV ) {
                if ( PP.Debug ) {
                        if ( AnalogValues->OnBoardCount != 0x0 )
                                ObcSetStr = "<<"; // DBGCSV(#)
                        else
                                ObcSetStr = "--"; // DBGCSV(0)
                        }
                // CSV, HEXCSV
                }
        else {
                if ( PP.Debug  &&  AnalogValues->OnBoardCount != 0x0 )
                        ObcSetStr = "<<"; // DBGFIX(#)
                else
                        ObcSetStr = "  "; // FIXED, HEXFIX, DBGFIX(0)
                }


        
        // Output data as requested by selected format 

        if ( PP.OutFormat == OUTPUT_FORMAT_CSV_DBG )
                fprintf( fp, "%ld%s%s,%1d,%.1lf,%.1lf,%.6lf,%4d/%02d/%02d,%02d:%02d:%02d,0.%06ld,%1d,%d,%d,%d,%d,%s\n",
                         CurrentPt, ChannelString, DigitalString,
                         PowerGoodH, TempC, TempF,
                         SampleTime, Yr, Mo, Dy, Hh, Mm, Ss, Usec,
                         PowerGoodC,
                         AnalogValues->Sanity,
                         AnalogValues->PpsToggle,
                         AnalogValues->GGACount,
                         AnalogValues->OnBoardCount,
                         ObcSetStr
                       );

        else if ( PP.OutFormat == OUTPUT_FORMAT_CSV_HEX  ||
                  PP.OutFormat == OUTPUT_FORMAT_CSV )
                fprintf( fp, "%ld%s%s,%1d,%.1lf,%.1lf,%.6lf,%4d/%02d/%02d,%02d:%02d:%02d,0.%06ld\n",
                         CurrentPt, ChannelString, DigitalString,
                         PowerGoodH, TempC, TempF,
                         SampleTime, Yr, Mo, Dy, Hh, Mm, Ss, Usec
                       );
        
        else if ( PP.OutFormat == OUTPUT_FORMAT_FW_DBG )
                fprintf( fp, "%010ld %s%s %1d %5.1lf %5.1lf  %17.6lf  %4d/%02d/%02d %02d:%02d:%02d  0.%06ld  %1d  0x%02X 0x%02X 0x%02X 0x%08X %s\n",
                         CurrentPt, ChannelString, DigitalString,
                         PowerGoodH, TempC, TempF,
                         SampleTime, Yr, Mo, Dy, Hh, Mm, Ss, Usec,
                         PowerGoodC,
                         AnalogValues->Sanity,
                         AnalogValues->PpsToggle,
                         AnalogValues->GGACount,
                         AnalogValues->OnBoardCount,
                         ObcSetStr
                       );

        else // if ( PP.OutFormat == OUTPUT_FORMAT_FW_HEX ) ||
             // if ( PP.OutFormat == OUTPUT_FORMAT_FW )
                fprintf( fp, "%010ld %s%s %1d %5.1lf %5.1lf  %17.6lf  %4d/%02d/%02d %02d:%02d:%02d  0.%06ld\n",
                         CurrentPt, ChannelString, DigitalString,
                         PowerGoodH, TempC, TempF,
                         SampleTime, Yr, Mo, Dy, Hh, Mm, Ss, Usec
                       );

        return( 1 );
}

//------------------------------------------------------------------------------
// ROUTINE: UpdateTimestamps
// PURPOSE: If current point is beyond time stamp #1, then update the time stamps
//          by reading a new one in from the timestamp file.
//------------------------------------------------------------------------------
void UpdateTimestamps( FILE *fp, TS *TS0, TS *TS1, int CurrentPt ) {
        
        TS TsTmp;
        
        static int eofTimeTmp = 0;

        // Error check

        if ( !TS0 || !TS1 )
                return;


        // Update time stamp if needed

        if ( CurrentPt >= TS1->Sample ) {

                if ( !eofTimeTmp ) {

                        // New time stamp needed, so read one and update

                        if ( SrDatTimestampReadFile( fp, &TsTmp, 0 ) ) {
                                *TS0 = *TS1;
                                *TS1 = TsTmp;
                                }

                        // Reached EOF, so use old timestamps without updating

                        else
                                eofTimeTmp = 1;

                        }

                }
}


//------------------------------------------------------------------------------
// ROUTINE: ComputeTime
// PURPOSE: Compute the time of the specified sample based on the information
//          contained in the timestamp structures.
//------------------------------------------------------------------------------
int ComputeTime( TS *TS0, TS *TS1, int CurrentPt, double *SampleTime ) {

        int    Ok;
        double x1, x2, y1, y2, xk, yk;

        if ( !TS0 || !TS1 || !SampleTime )
                return( 0 );

        x1 = (double)TS0->Sample;
        x2 = (double)TS1->Sample;
        y1 = TS0->TotalTime;
        y2 = TS1->TotalTime;
        xk = (double)CurrentPt;

        Ok = SrDatEquation2pt( x1, y1, x2, y2, xk, &yk );

        *SampleTime = yk;

        return( Ok );
}

//------------------------------------------------------------------------------
// ROUTINE: ErrorCmd
// PURPOSE: Print a command line error message and quit.
//------------------------------------------------------------------------------
void ErrorCmd( char *msg ) {

        // Print a command line error message and quit.

        printf(

               "ERROR: Command line usage : %s \n\n ... Pak2Asc /? for help\n",
               msg

               );

        exit( 1 );  // < quit at errorlevel 1.

}

//------------------------------------------------------------------------------
// ROUTINE: ErrorExec
// PURPOSE: An execution error occurred.
//------------------------------------------------------------------------------
void ErrorExec( char *msg ) {

        // A fatal execution error occurred.

        printf( "\nERROR: %s \n\n", msg );

        exit( 2 );  // < quit at errorlevel 2.

}

//------------------------------------------------------------------------------
// ROUTINE: WarningExec
// PURPOSE: A non-fatal execution error occurred.
//------------------------------------------------------------------------------
void WarningExec( char *msg ) {

        // A non-fatal execution error occurred.

        printf( "\nWARNING: %s \n\n", msg );
}

//------------------------------------------------------------------------------
// ROUTINE: ProcessCommandLine
// PURPOSE: Gather the command line arguments, filling in PP.
//          See the UsageMsg structure defined above for allowed options.
//------------------------------------------------------------------------------
void ProcessCommandLine( int argc, char *argv[] ) {

        int i, ifmt;
        char MsgBuffer[MAXSTR];


        // No command line arguments gives the usage message.

        if ( argc == 1 )

                Usage( );


        // Check for too many arguments.

        if ( argc > 3 )

                ErrorCmd( "too many command line arguments" );


        // Scan for valid arguments.

        for ( i = 1 ; i < argc ; i++ ) {


                // Check for usage message request.

                if ( StrEqu( argv[i], "/?") )

                        Usage( );


                // Check for any allowed format.

                else if ( ifmt = MatchFormat( argv[i] ) ) {
                        
                        if ( PP.OutFormat == OUTPUT_FORMAT_NONE )
                                PP.OutFormat = ifmt;

                        else
                                ErrorCmd( "multiple format arguments" );
                        }

                

                // If not one of the above, it must be a filename ...

                else {

                        if ( ParseFileName( argv[i] ) ) {

                                if ( PP.InputFileFlag == 0 )

                                        PP.InputFileFlag = 1;

                                else

                                    ErrorCmd( "multiple filename arguments" );
                                }

                        else {

                                sprintf( MsgBuffer, "filename %s is too long",

                                                         ChopLong( argv[i] )

                                                         );

                                    ErrorCmd( MsgBuffer );

                                }

                        }

                }



        // An input filename is *required* on the command line.

        if ( !PP.InputFileFlag )

                ErrorCmd( "an input filename is required ..." );


        // If the input file base name is eight digits, turn on sequential.

        if ( StrIsEightDecimal( PP.InputFileBase ) ) {

                PP.Sequential = 1;  // < set the Sequential processing flag on


                // Get the starting filename as a binary integer.

                sscanf( PP.InputFileBase, "%d", &PP.nSequential );

                }



        // An output format is *required* on the command line.

        if ( PP.OutFormat == OUTPUT_FORMAT_NONE )

                ErrorCmd( "an output format is required ..." );


        // Set csv, hex, and debug helper variables based on selected format

        if ( PP.OutFormat == OUTPUT_FORMAT_CSV    ||
             PP.OutFormat == OUTPUT_FORMAT_CSV_HEX ||
             PP.OutFormat == OUTPUT_FORMAT_CSV_DBG )
                PP.CSV = 1;

        if ( PP.OutFormat == OUTPUT_FORMAT_FW_HEX ||
             PP.OutFormat == OUTPUT_FORMAT_CSV_HEX ||
             PP.OutFormat == OUTPUT_FORMAT_FW_DBG )  // fixed debug mode uses hex too
                PP.Hex = 1;
        
        if ( PP.OutFormat == OUTPUT_FORMAT_FW_DBG ||
             PP.OutFormat == OUTPUT_FORMAT_CSV_DBG )
                PP.Debug = 1;

}

//------------------------------------------------------------------------------
// ROUTINE: Usage
// PURPOSE: Display the usage message.
//------------------------------------------------------------------------------
void Usage( void ) {

        int i;


        // Print the usage message.

        for ( i = 0 ; UsageMsg[i] ; i++ )

                printf( "%s\n", UsageMsg[i] );

        exit( 0 );  // < quit at error level 0.
}

//------------------------------------------------------------------------------
// ROUTINE: MatchFormat
// PURPOSE: Check if argument is one of the allowed format names.  If not, 
//          return 0.  Otherwise, return the index of that matched.
//------------------------------------------------------------------------------
int MatchFormat( char *argument ) {

        int i;

        if ( !argument )
                return( 0 );

        for ( i = OUTPUT_FORMAT_NONE+1 ; i < OUTPUT_FORMAT_MAX ; i++ ) {

                if ( StrEqu( argument, OutputFormatName[i] ) )
                        return( i );

                } // end for i

        return( 0 );
}

//------------------------------------------------------------------------------
// ROUTINE: StrEqu
// PURPOSE: Compare for exact equality between two strings.
//------------------------------------------------------------------------------
int StrEqu( char *s1, char *s2 ) {

        while ( *s1 && *s2 && (*s1 == *s2) )

                { s1++; s2++; }

        return ( (*s1 == '\0') && (*s2 == '\0') ? 1 : 0 );

}

//------------------------------------------------------------------------------
// ROUTINE: StrIsEightDecimal
// PURPOSE: Check a string for being eight decimal digits.
//------------------------------------------------------------------------------
int StrIsEightDecimal( char *s ) {

        int i;

        i = 0;

        while ( isdigit( *s ) )

                { s++; i++; }

        return ( (*s == '\0') && (i == 8) ? 1 : 0 );

}

//------------------------------------------------------------------------------
// ROUTINE: ParseFileName
// PURPOSE: Parse an input full pathname, determining any path and extension.
//------------------------------------------------------------------------------
int ParseFileName( char *fullpath ) {

        char *s, *path, *ext;
        int i;


        // Copy the command line input file to PP.FullPathName[].
        //
        //  ... and check to make sure not too long
        //
        //

        for ( s = fullpath , i = 0 ; *s && (i < MAXFILENAME) ; s++ , i++)

                PP.FullPathName[i] = *s;

        PP.FullPathName[i] = '\0';

        if ( i > (MAXFILENAME - 80) )

                return( 0); // don't even bother if too long ...




        // Scan for the last '\' to mark the filename path.
        //
        //  ... allow both '\' and '/' for Windows/Linux compatibility
        //

        path = NULL;

        for ( s = PP.FullPathName ; *s ; s++ ) {

                if ( (*s == '\\') || (*s == '/') )

                        path = s;

                } // path now == NULL, or points to last '\' ...



        // Scan from the path to the last '.' to mark the extension.

        ext = NULL;

        for ( s = ( path ? path + 1 : PP.FullPathName ) ; *s ; s++ ) {

                if ( *s == '.' )

                        ext = s;

                } // ext now == NULL, or points to last '.' ...





        // Copy out the input path into PP.InputFilePath[].
        //
        // ... the trailing '\' is included in PP.InputFilePath[]
        //

        if ( path ) {

                for ( s = PP.FullPathName , i = 0 ; s <= path ; s++ , i++ )

                        PP.InputFilePath[i] = *s;

                PP.InputFilePath[i] = '\0';

                }
        else

                PP.InputFilePath[0] = '\0';




        // Copy out the base filename into InputFileBase[].

        s = ( path ? path + 1 : PP.FullPathName );

        if ( ext ) { // gather up to the extension ...

                for ( i = 0 ; s < ext ; s++ , i++ )

                        PP.InputFileBase[i] = *s;

                }

        else { // ext == NULL

                for ( i = 0 ; *s ; s++ , i++ )

                        PP.InputFileBase[i] = *s;

                }

        PP.InputFileBase[i] = '\0';




        // Copy out the filename extension into InputFileExt[].

        if ( ext ) { // gather from the extension onward ...

                for ( s = ext + 1 , i = 0 ; *s ; s++ , i++ )

                        PP.InputFileExt[i] = *s;

                PP.InputFileExt[i] = '\0';

                }
        else

                PP.InputFileExt[0] = '\0';





        // Convert PP.InputFileExt[] to lowercase.

        for ( s = PP.InputFileExt ; *s ; s++ )

                *s = tolower( *s);



        // Point pFullPathNameBase to the base part of FullPathName[].

        PP.pFullPathNameBase = ( path ? path + 1 : PP.FullPathName );



        // Form the InputFileName[].

        sprintf( PP.InputFileName, "%s.%s", PP.InputFileBase, PP.InputFileExt);



        // Return success.
        //
        // Almost anything will be accepted for a filename if it is not too
        // long ... fopen will further qualify later ...
        //
        //

        return( 1);


}

//------------------------------------------------------------------------------
// ROUTINE: SkipTitleLine
// PURPOSE: Read in an discard one line from the specified file.
//------------------------------------------------------------------------------
void SkipTitleLine( FILE *fp ) {

        int c;

        c = fgetc( fp );
        while ( c != EOF && c != '\n' )
                c = fgetc( fp );
}

//------------------------------------------------------------------------------
// ROUTINE: PrintRunTimeSummary
// PURPOSE: Display the final PP parameters to run with.
//------------------------------------------------------------------------------
void PrintRunTimeSummary( void ) {


        // Display the Pak2Asc parameters they are running with:

        printf( "Beginning Pak2Asc with:\n"                                );
        printf( "\n"                                                       );
        printf( " Input FilePath: %s\n",   ChopLong( PP.InputFilePath )    );
        printf( " Input FileName: %s\n",   ChopLong( PP.InputFileBase )    );
        printf( "  Input FileExt: %s\n",   ChopLong( PP.InputFileExt  )    );
        printf( "  Output Format: %s\n",   OutputFormatName[PP.OutFormat]  );
        printf( "     Sequential: %s\n",   PP.Sequential ? "ON" : "OFF"    );

    if ( PP.Sequential )

        printf( "         nStart: %08d\n", PP.nSequential                  );

    else

        printf( "         nStart: not applicable ... \n"                   );
        printf( "\n"                                                       );
}

//------------------------------------------------------------------------------
// ROUTINE: ChopLong
// PURPOSE: If a string is too long for display, chop it ...
//------------------------------------------------------------------------------
char *ChopLong( char *s ) {

        int i; static char MsgBuffer[MAXSTR];


        // Copy the string up to CHOPLENGTH.

#define CHOPLENGTH  40

        for ( i = 0 ; (i < CHOPLENGTH) && *s ; i++ , s++ )

                MsgBuffer[i] = *s;


        // Add '~' if more than CHOPLENGTH.

        if ( i == CHOPLENGTH )

                MsgBuffer[i++] = '~';


        // NULL terminate.

        MsgBuffer[i] = '\0';


        // Return the chopped string.

        return( MsgBuffer );

}

//------------------------------------------------------------------------------
// ROUTINE: ProcessRawPackets
// PURPOSE: Write raw packet bytes to the ASCII output file.
//------------------------------------------------------------------------------
void ProcessRawPackets( void ) {

    //  FILE            *PP.fpInput;   << see PP struct for this variable
        FILE            *fpRawOut;

        SRUSBXCH_PACKET Packet;

        int             nTotalPackets;
        int             Error;



        // Open the input and output files.

        PP.fpInput = FileOpen( PP.InputFileName,  "rb" );
        fpRawOut   = FileOpen( PP.FileNameRawOut, "w"  );


        // Read from the input and write to the output file.

        nTotalPackets = 0;
        
        while ( ReadInputFile( &Packet ) ) {

                SrUsbXchPacketShow( fpRawOut, &Packet, nTotalPackets, 1, &Error );
                
                nTotalPackets++;

                }


        // Close the input and output files.
        //
        // ... close of PP.fpInput is automatically done by ReadInputFile ...
        //

        fclose( fpRawOut );



        // Print RAW summary info.

        printf( "\n" );

        printf( "FINISHED Raw processing with:     \n"                    );
        printf( "                                  \n"                    );
        printf( "     Total packets processed: %d  \n", nTotalPackets     );
        printf( "             Last input file: %s  \n",
                                             ChopLong( PP.InputFileName ) );
        printf( "\n" );
        
        printf( "\nSUCCESS: final ASCII data >> %s\n\n", PP.FileNameRawOut );


        // Quit the program.

        exit( 0 );
}
